/* ================================== */
/* LLREK LICENSE */
/* ================================== */
/* Copyright (c) 2013 Carnegie Mellon University. */
/* All Rights Reserved. */

/* Redistribution and use in source and binary forms, with or without */
/* modification, are permitted provided that the following conditions are */
/* met: */

/* 1. LLREK includes and/or makes use of certain third party software */
/* ("Third Party Software"). You agree to comply with any and all the */
/* Third Party Software terms and conditions listed below and/or */
/* contained in any separate license file distributed with LLREK. The */
/* parties who own the Third Party Software ("Third Party Licensors") are */
/* intended third party beneficiaries to this License with respect to the */
/* terms applicable to their Third Party Software. */

/* 2. Redistributions of source code must retain the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers. */

/* 3. Redistributions in binary form must reproduce the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers in the documentation and/or other materials provided with */
/* the distribution. */

/* 4. Products derived from this software may not include Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" in the name of such derived product, nor shall Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" be used to endorse or promote products derived from this */
/* software without prior written permission. For written permission, */
/* please contact permission@sei.cmu.edu. */

/* This material is based upon work funded and supported by the */
/* Department of Defense under Contract No. FA8721-05-C-0003 with */
/* Carnegie Mellon University for the operation of the Software */
/* Engineering Institute, a federally funded research and development */
/* center. */

/* Any opinions, findings and conclusions or recommendations expressed in */
/* this material are those of the author(s) and do not necessarily */
/* reflect the views of the United States Department of Defense. */

/* NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING */
/* INSTITUTE MATERIAL IS FURNISHEDON AN AS-IS BASIS. CARNEGIE */
/* MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR */
/* IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF */
/* FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS */
/* OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT */
/* MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, */
/* TRADEMARK, OR COPYRIGHT INFRINGEMENT. */

/* This material has been approved for public release and unlimited */
/* distribution. */

/* DM-0000665 */

#ifndef __BVC_BVENV_HPP__
#define __BVC_BVENV_HPP__

#include "./BvcUtil.hpp"

namespace ufo
{
  //number of PCP lock levels. note that the highest level is reserved
  //for the CPU lock.
  static const unsigned PCPLEVELS = 8;

  //names of prefixes of names of LLREK-specific functions
  static const char *LLINIT = "llrek_init";
  static const char *LLFINAL = "llrek_final";
  static const char *LLREAD = "llrek_shared_read_";
  static const char *LLWRITE = "llrek_shared_write_";
  static const char *LLPCP_LOCK = "llrek_pcp_lock";
  static const char *LLPCP_UNLOCK = "llrek_pcp_unlock";
  static const char *LLCPU_LOCK = "llrek_cpu_lock";
  static const char *LLCPU_UNLOCK = "llrek_cpu_unlock";
  static const char *LLJOB_START = "llrek_job_start_";
  static const char *LLJOB_END = "llrek_job_end_";
  static const char *LLSNAP = "llrek_snapshot_";

  class BVEnv;
  
  namespace bitvc
  {
    struct EvalVisitor : public std::unary_function<Expr, VisitAction>
    {
      BVEnv &env;

      EvalVisitor (BVEnv &e) : env (e) {}
      inline VisitAction operator () (Expr exp) const;
    };

    //-- the control boolean variable for a CFG edge
    inline Expr mkEdgVar (Expr src, Expr dst)
    {
      return bvop::bvConst(mk<TUPLE> (src, dst),1);
    }
  } //namespace bitvc

  /** An environment for interpreting formulas. 
      Maps between llvm::Value and expr::Expr 
  */
  class BVEnv : public std::unary_function<Expr,Expr>
  {
  public:
    typedef std::map<Expr, Expr> ExprExprMap;

  private:

    /** environment map */
    ExprExprMap env;
    
    /** global map for fresh names */
    typedef boost::shared_ptr<ExprExprMap> PtrExprExprMap;
    PtrExprExprMap genv;

    ExprFactory *efac;

    bitvc::EvalVisitor evalVisitor;
  
  public:
    
    typedef ExprExprMap::iterator iterator;
    typedef ExprExprMap::const_iterator const_iterator;
	
    //-- for a value, create a BV varible of appropriate width
    Expr toExpr (const Value &v) const 
    { 
      if(const CallInst *ci = dyn_cast<const CallInst>(&v))
        return toExprCall(*ci);
 
      return bvop::bvConst(mkTerm (&v, *efac), bitvc::getBW(&v));
    }
	
    //-- for a value, create a BV varible of appropriate width
    Expr toExprValue (const Value &v) const 
    { 
      return bvop::bvConst(mkTerm (&v, *efac), bitvc::getBW(&v));
    }

    //-- for a call instruction
    Expr toExprCall(const CallInst &ci) const 
    {
      //-- for a call to access a global variable, create BV of width
      //-- corresponding to its first argument, i.e., the global var
      Function *Callee = ci.getCalledFunction ();
      if (Callee->getName ().startswith (LLREAD) || Callee->getName ().startswith (LLWRITE))
        return bvop::bvConst(mkTerm (static_cast<const Value*>(&ci), *efac), 
                             bitvc::getBW(ci.getArgOperand(0))); 
      
      //-- for a call to get or release a PCP lock, create BV of width
      //-- equal to the maximum number of PCP levels. note that CPU
      //-- locks are treated a special types of PCP locks.
      if (Callee->getName ().equals (LLPCP_LOCK) || Callee->getName ().equals (LLPCP_UNLOCK) ||
          Callee->getName ().equals (LLCPU_LOCK) || Callee->getName ().equals (LLCPU_UNLOCK))
        return bvop::bvConst(mkTerm (static_cast<const Value*>(&ci), *efac),PCPLEVELS); 

      //-- for a job start or job end event, create BV of width equal
      //-- to the maximum number of PCP levels
      if (Callee->getName ().startswith (LLJOB_START) || Callee->getName ().startswith (LLJOB_END))
        return bvop::bvConst(mkTerm (static_cast<const Value*>(&ci), *efac),PCPLEVELS); 

      //-- otherwise, treat it like a value
      return toExprValue(static_cast<const Value&>(ci));
    }

    //-- for a basic block return a boolean variable, i.e., a BV
    //-- variable of width 1
    Expr toExpr (const BasicBlock &v) const { 
      return bvop::bvConst(mkTerm (&v ,*efac),1); 
    }

    Expr toExpr (const Expr &v) const { return v; }
    Expr toExpr (const std::string &v) { return mkTerm (v, *efac); }
    
    BVEnv(ExprFactory& e): genv(new ExprExprMap ()), efac(&e), 
                           evalVisitor (*this) {}
    
    BVEnv (const BVEnv &other) 
      : env (other.env), 
        genv(other.genv), 
        efac (other.efac), 
        evalVisitor(*this) 
    {}

    /** noarg constructor for containers */
    BVEnv() : efac(NULL), evalVisitor (*this) {}
    
    BVEnv& operator=(const BVEnv& other)
    {
      BVEnv tmp(other);
      std::swap (tmp.env, env);
      std::swap (tmp.genv, genv);
      efac = tmp.efac;

      return *this;
    }
	
    ExprFactory &getExprFactory () { return *efac; }

    /** Evaluates the expression in the given environment */
    Expr eval (Expr exp) { return dagVisit (evalVisitor, exp); }
    Expr operator() (Expr exp) { return eval (exp); }
    
    template <typename T> bool isBound (const T& v) const
    {
      return (env.find (toExpr (v)) != env.end());
    }

    /// assign bindings from the other environment. Global environment
    /// is unchanged. All local bindings are dropped.
    void assign (BVEnv &other)
    { 
      env.clear ();
      env.insert (other.begin (), other.end ()); 
    }
    
    iterator begin () { return env.begin (); }
    iterator end () { return env.end (); }

    const_iterator begin () const { return env.begin (); }
    const_iterator end  () const { return env.end (); }

    void erase (iterator i) { return env.erase (i); }
    size_t erase (const Expr &key) { return env.erase (key); }
     

    /** remove all bindings */
    void clear () { env.clear (); }
    
    
    bool has_binding (const Expr v, const Expr u) const
    {
      ExprExprMap::const_iterator it = env.find (v);
      return it != env.end () && it->second == u;
    }

    // -- lookup and refresh if needed
    template <typename T>
    Expr lookup(const T& v)
    {
      Expr expr = toExpr (v);
      ExprExprMap::const_iterator it = env.find (expr);
      if (it == env.end ()) 
	return refresh (expr);
      
      return it->second;
    }
    
    void printEnv() const 
    {
      errs() << "Printing Env\n\n";
      forall (const ExprExprMap::value_type &v, env)
        errs () << *v.first << " ===> " << *v.second << "\n";
    }

    template <typename T>
    void bind (const T& v, const Expr &u)
    {
      //check if first bound expression is a variant
      assert (isOpX<VARIANT> (u));
      env [toExpr(v)] = u;
    }

    template <typename T>
    Expr refresh (const T& _v)
    {
      Expr v = toExpr (_v);
            
      ExprExprMap::iterator geit = genv->find (v);
        
      int idx = 0;
      if (geit != genv->end ())
	{
	  idx = variant::variantNum (geit->second) + 1;
	  // -- genv always maps v to 'variant of v'
	  assert (variant::mainVariant (geit->second) == v);
	}
      
      Expr u = variant::variant (idx, v);
     
      (*genv)[v] = u;
      bind (v, u);
      
      return u;
    }
    
    /**
     * Unifies (intersects) this environment with a collection of
     * environments. Refreshed keys are returned in refreshed set.
     */
    template <typename Iterator> 
    void unify (Iterator begin, Iterator end, ExprSet &refreshed)
    {
      for (Iterator oit = begin; oit != end; ++oit)
	{
	  BVEnv* oEnv = *oit;
	  
	  ExprExprMap::iterator me = env.begin ();
	  while (me != env.end ())
	    {
	      Expr key = me->first;
	      
	      ExprExprMap::iterator other = oEnv->env.find (key);

	      // -- not found, not in the intersection
	      if (other == oEnv->env.end ())
		{
		  env.erase (me++);
		  refreshed.erase (key);
		  continue;
		}

	      // -- key is in both me and other, check value
	      if (me->second != other->second)
		// -- needs refresh
		refreshed.insert (key);
	      ++me;
	    }	  
	}
    
      // -- refresh anything that needs refresh
      for (ExprSet::const_iterator it = refreshed.begin (), 
	     end = refreshed.end (); it != end; ++it)
	refresh (*it);
      
    }    
  };

  namespace bitvc
  {
    VisitAction EvalVisitor::operator () (Expr exp) const
    {
      // -- if bound, return value
      if (env.isBound (exp))
	return VisitAction::changeTo (env.lookup (exp));

      // -- if unbound, but is a terminal that environment knows about
      // -- implicitly
      if (isOpX<BB> (exp) || isOpX<VALUE> (exp) || 
	  isOpX<NODE> (exp) || isOpX<VARIANT> (exp)){
	return VisitAction::changeTo (env.lookup (exp));
      }
      return VisitAction::doKids ();
    }
  }

  /** returns a path from entry to exit, but not including exit */
  template <typename SmtModel, typename OutputIterator>
  void doCfgCex (const BasicBlock *en, const BasicBlock *ex, 
		 BVEnv &edgEnv,
		 SmtModel &model, OutputIterator out)
  {
    const BasicBlock *v;
    v = en;
    *(out++) = en;
    
    do
      {
	const BasicBlock* newV = NULL;
	bool modelEdge = false;
	
	assert (edgEnv.isBound (*v));
	Expr src = edgEnv.lookup (*v);

	// for all successors
	for (succ_const_iterator it = succ_begin (v), end = succ_end (v); 
	     it != end; ++it)
	  {
	    const BasicBlock *u = *it;

	    // -- the edge variable to the exit variable is missing
	    // -- so if u is exit node, we guess that edge, unless some 
	    // -- other edge variable will reset it
	    if (u == ex) 
              {
                newV = u;
                continue;
              }

	  
	    // -- skip all successors that are not on the right cpg-edge
	    if (!edgEnv.isBound (*u)) continue;

	    if (!newV) newV = u;

	    Expr dst = edgEnv.lookup (*u);
	    Expr edgTerm = bitvc::mkEdgVar (src, dst);
	    if (isOpX<TRUE> (model (edgTerm)))
	      {
		modelEdge = true;
                newV = u;
		break;
	      }
	  }
	assert (newV && "Could not guess a counterexample!!!");
	
	if (newV != ex) *(out++) = newV;
	if (!modelEdge && newV != ex)
	  errs () << "WARNING: NO MODEL: guessed cex step: " 
		  << v->getName () << " --> "
		  << newV->getName () << "\n";
	v = newV;
      } 
    while (v != ex);
  }
} //namespace ufo

#endif //__BVC_BVENV_HPP__
