/* ================================== */
/* LLREK LICENSE */
/* ================================== */
/* Copyright (c) 2013 Carnegie Mellon University. */
/* All Rights Reserved. */

/* Redistribution and use in source and binary forms, with or without */
/* modification, are permitted provided that the following conditions are */
/* met: */

/* 1. LLREK includes and/or makes use of certain third party software */
/* ("Third Party Software"). You agree to comply with any and all the */
/* Third Party Software terms and conditions listed below and/or */
/* contained in any separate license file distributed with LLREK. The */
/* parties who own the Third Party Software ("Third Party Licensors") are */
/* intended third party beneficiaries to this License with respect to the */
/* terms applicable to their Third Party Software. */

/* 2. Redistributions of source code must retain the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers. */

/* 3. Redistributions in binary form must reproduce the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers in the documentation and/or other materials provided with */
/* the distribution. */

/* 4. Products derived from this software may not include Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" in the name of such derived product, nor shall Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" be used to endorse or promote products derived from this */
/* software without prior written permission. For written permission, */
/* please contact permission@sei.cmu.edu. */

/* This material is based upon work funded and supported by the */
/* Department of Defense under Contract No. FA8721-05-C-0003 with */
/* Carnegie Mellon University for the operation of the Software */
/* Engineering Institute, a federally funded research and development */
/* center. */

/* Any opinions, findings and conclusions or recommendations expressed in */
/* this material are those of the author(s) and do not necessarily */
/* reflect the views of the United States Department of Defense. */

/* NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING */
/* INSTITUTE MATERIAL IS FURNISHEDON AN AS-IS BASIS. CARNEGIE */
/* MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR */
/* IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF */
/* FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS */
/* OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT */
/* MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, */
/* TRADEMARK, OR COPYRIGHT INFRINGEMENT. */

/* This material has been approved for public release and unlimited */
/* distribution. */

/* DM-0000665 */

#include "ufo/InitializePasses.h"

#include "llvm/Support/CommandLine.h"
#include "llvm/Module.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Target/TargetData.h"

#include "ufo/Cpg/Topo.hpp"
#include "ufo/Util/NameValues.hpp"

#include "ufo/Smt/UfoZ3.hpp"
#include "ufo/CptLiveValues.hpp"

using namespace llvm;
using namespace ufo;

//-- options
static cl::opt<bool>
BvcDebug ("bvc-debug", cl::desc ("Generate debug output"),
          cl::init (false));
static cl::opt<bool>
BvcStp ("bvc-stp", cl::desc ("Use STP as backend solver"),
        cl::init (false));
static cl::opt<bool>
BvcBoolector ("bvc-boolector", cl::desc ("Use Boolector as backend solver"),
              cl::init (false));

#include "./Z3Conv.hpp"
#include "./BVEdgeCondComp.hpp"

namespace ufo
{
  struct BitVC : public ModulePass 
  {
    static char ID;
    BitVC () : ModulePass(ID) { errs () << "== Creating BitVC Pass\n";}    

    static bool exit;
    virtual bool runOnModule(Module &M) 
    { 
      if (exit) return false;
      else exit = true;

      //run unit test and return
      //unitTest();
      //return false;        

      TargetData theTD(&M);
      bitvc::TD = &theTD;
      Function& F = *(M.getFunction("main")); 
      if (F.getName().compare("main") != 0) return false;

      //sort function topologically and check for cycles
      Topo &topo = getAnalysis<Topo>(F);
      assert(!topo.hasBackEdges(&F) && "ERROR: main is not acyclic");

      //compute the VC
      errs () << "Start generating bitvector VC \n";
      ExprFactory m_efac;
      BVEnv env (m_efac);
      BVEdgeCondComp condComp (m_efac, env, F, false);
      condComp.compute ();
      Expr econd = condComp.getCond ();

      if(BvcDebug) errs() << "VC : " << *econd << '\n';
      errs () << "Done generating bitvector VC \n";
              
      //check SAT of VC
      verify(condComp,m_efac);

      return false;
    }

    void verify(const BVEdgeCondComp &condComp,ExprFactory &efac)
    {
      errs() << "starting verification ...\n";
      if(condComp.getAsserts().empty()) return;

      Z3_config cfg = Z3_mk_config();
      Z3_context ctx = Z3_mk_context(cfg);
      Z3_solver slvr = Z3_mk_solver_for_logic(ctx,Z3_mk_string_symbol(ctx,"QF_BV"));
      bitvc::Z3Conv z3conv(ctx);
      Z3_ast trans = z3conv.marshal(condComp.getCond());
      Z3_ast spec = z3conv.marshal(mknary<AND>(mk<TRUE>(efac),condComp.getAsserts()));

      //if using STP or Boolector
      if(BvcStp || BvcBoolector) {
        errs() << "exporting to SMT2 ...\n";

        //create temporary file
        std::string errInfo;
        char tmpFileName[32];
        strncpy(tmpFileName,"/tmp/mmc.XXXXXX",32);
        raw_fd_ostream outf(mkstemp(tmpFileName),true);

        //create smt string and dump to file
        Z3_set_ast_print_mode(ctx,Z3_PRINT_SMTLIB2_COMPLIANT);
        Z3_string bnch = Z3_benchmark_to_smtlib_string(ctx,"foo","QF_BV","unknown","",1,&trans,Z3_mk_not(ctx,spec));
        outf << bnch;
        outf.close();

        //solve using STP
        if(BvcStp) {
          std::string cmd = "/home/chaki/stuff/ufo/stp/stp-r1673/install/bin/stp --simplifying-minisat --SMTLIB2 " + std::string(tmpFileName);
          errs() << cmd << '\n';
          int stpStatus = ::system(cmd.c_str());
          errs() << "status = " << stpStatus << '\n';
          errs() << "VERIFICATION " << 
            (stpStatus == 0 ? "SUCCESSFUL" : 
             (stpStatus == 256 ? "FAILED" : "UNKNOWN")) << '\n';
        }
        //solve using Boolector
        else {
          std::string cmd = "/home/chaki/stuff/ufo/boolector/boolector-1.5.118-6b56be4-121013/boolector --smt2 " + std::string(tmpFileName);
          errs() << cmd << '\n';
          int btrStatus = ::system(cmd.c_str());
          errs() << "status = " << btrStatus << '\n';
          errs() << "VERIFICATION " << 
            (btrStatus == 5120 ? "SUCCESSFUL" : 
             (btrStatus == 2560 ? "FAILED" : "UNKNOWN")) << '\n';
        }
      }
      //else if using Z3
      else {
        Z3_solver_assert(ctx,slvr,trans);
        Z3_solver_assert(ctx,slvr,Z3_mk_not(ctx,spec));
        errs() << "done marshalling .. now solving ...\n";
        Z3_lbool res = Z3_solver_check(ctx,slvr);

        //-- print stats
        //Z3_stats stats = Z3_solver_get_statistics(ctx,slvr);
        //Z3_stats_inc_ref(ctx,stats);
        //errs() << Z3_stats_to_string(ctx,stats) << '\n';
        //Z3_stats_dec_ref(ctx,stats);

        //-- print result
        errs() << "VERIFICATION " << 
          (res == Z3_L_FALSE ? "SUCCESSFUL" : 
           (res == Z3_L_TRUE ? "FAILED" : "UNKNOWN")) << '\n';
      }

      //-- cleanup
      Z3_del_context(ctx);
      Z3_del_config(cfg);
    }

    void unitTest()
    {
      ExprFactory efac;

      errs() << "starting BV unit test\n";
      Expr bvs8 = bvop::bvsort(8,efac);
      errs() << *bvs8 << '\n';

      Expr bvs300 = bvop::bvsort(300,efac);
      errs() << *bvs300 << '\n';

      Expr decl18 = bvop::bvConstDecl(mkTerm<string>("var18",efac),18);
      errs() << *decl18 << '\n';

      Expr var28 = bvop::bvConst(mkTerm<string>("var28",efac),28);
      errs() << *var28 << '\n';

      Expr andE = mk<BVAND>(decl18,var28);
      errs() << *andE << '\n';

      Expr orE = mk<BVOR>(decl18,var28);
      errs() << *orE << '\n';
    }

    virtual void getAnalysisUsage(AnalysisUsage &AU) const
    {
      AU.addRequired<Topo> ();
      AU.addRequired<NameValues> ();
      AU.setPreservesAll ();
    }
  };

  char BitVC::ID = 0;
  bool BitVC::exit = false;
}

INITIALIZE_PASS_BEGIN(BitVC, "bitvc", "BitVector VC generator", false, false)
INITIALIZE_PASS_DEPENDENCY(Topo)
INITIALIZE_PASS_DEPENDENCY(NameValues)
INITIALIZE_PASS_END(BitVC, "bitvc", "BitVector VC generator", false, false)
