/* ================================== */
/* LLREK LICENSE */
/* ================================== */
/* Copyright (c) 2013 Carnegie Mellon University. */
/* All Rights Reserved. */

/* Redistribution and use in source and binary forms, with or without */
/* modification, are permitted provided that the following conditions are */
/* met: */

/* 1. LLREK includes and/or makes use of certain third party software */
/* ("Third Party Software"). You agree to comply with any and all the */
/* Third Party Software terms and conditions listed below and/or */
/* contained in any separate license file distributed with LLREK. The */
/* parties who own the Third Party Software ("Third Party Licensors") are */
/* intended third party beneficiaries to this License with respect to the */
/* terms applicable to their Third Party Software. */

/* 2. Redistributions of source code must retain the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers. */

/* 3. Redistributions in binary form must reproduce the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers in the documentation and/or other materials provided with */
/* the distribution. */

/* 4. Products derived from this software may not include Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" in the name of such derived product, nor shall Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" be used to endorse or promote products derived from this */
/* software without prior written permission. For written permission, */
/* please contact permission@sei.cmu.edu. */

/* This material is based upon work funded and supported by the */
/* Department of Defense under Contract No. FA8721-05-C-0003 with */
/* Carnegie Mellon University for the operation of the Software */
/* Engineering Institute, a federally funded research and development */
/* center. */

/* Any opinions, findings and conclusions or recommendations expressed in */
/* this material are those of the author(s) and do not necessarily */
/* reflect the views of the United States Department of Defense. */

/* NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING */
/* INSTITUTE MATERIAL IS FURNISHEDON AN AS-IS BASIS. CARNEGIE */
/* MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR */
/* IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF */
/* FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS */
/* OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT */
/* MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, */
/* TRADEMARK, OR COPYRIGHT INFRINGEMENT. */

/* This material has been approved for public release and unlimited */
/* distribution. */

/* DM-0000665 */

#include "ufo/InitializePasses.h"

#include "llvm/Support/CommandLine.h"
#include "llvm/Module.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Support/IRBuilder.h"
#include "llvm/Support/InstVisitor.h"
#include "llvm/Bitcode/ReaderWriter.h"
#include "llvm/Transforms/Utils/Cloning.h"

#include "ufo/Smt/UfoZ3.hpp"
#include "ufo/CptLiveValues.hpp"

using namespace llvm;
using namespace ufo;

//-- options
static cl::opt<bool>
G2lDebug ("g2l-debug", cl::desc ("Generate Glob2Loc debug output"),
          cl::init (false));

namespace ufo
{
  struct G2LVisitor : public InstVisitor<G2LVisitor> 
  {
    ///dead instructions that should be removed
    std::set<Instruction*> deadInsts;

    ///map from global variables to alloca instructions
    std::map<GlobalVariable*,AllocaInst*> gvmap;

    ///the assume function
    Function *assumeFunc;

    //map from globals to arguments and allocas
    std::map<GlobalVariable*,AllocaInst*> &glob2alloca;

    ///constructor with assume function
    G2LVisitor(Function *af,std::map<GlobalVariable*,AllocaInst*> &g2a) : 
      assumeFunc(af),glob2alloca(g2a) {}

    AllocaInst *getAlloca(GlobalVariable *gv,Instruction *ins)
    {
      return glob2alloca[gv];
    }

    ///return a vector of indices. indices start from the second
    ///operand onward.
    std::vector<Value*> getIdxs(User *ce)
    {
      std::vector<Value*> idxs;
      User::op_iterator oit = ce->op_begin();
      ++oit;          
      for(User::op_iterator oend = ce->op_end();oit != oend;++oit)
        idxs.push_back(oit->get());
      return idxs;
    }

    void visitLoadInst(LoadInst &li)
    {
      if(G2lDebug) errs() << "processing : " << li << '\n';
      Value *ptr = li.getPointerOperand();
      if(GlobalVariable *gv = dyn_cast<GlobalVariable>(ptr)) {
        AllocaInst *ai = getAlloca(gv,&li);
        LoadInst *newli = new LoadInst(ai,0,false,&li);
        if (!li.use_empty()) li.replaceAllUsesWith(newli);
        deadInsts.insert(&li);
        return;
      }

      // -- constant expression, this is usuaally a GEP
      if(ConstantExpr *ce = dyn_cast<ConstantExpr>(ptr)) {
        if(ce->getOpcode() != Instruction::GetElementPtr) return;
        if(GlobalVariable *gv = dyn_cast<GlobalVariable>(ce->getOperand(0))) {
          AllocaInst *ai = getAlloca(gv,&li);
          std::vector<Value*> idxs = getIdxs(ce);
          GetElementPtrInst *gepi = GetElementPtrInst::Create(ai,idxs.begin(),idxs.end(),"",&li);
          LoadInst *newli = new LoadInst(gepi,0,false,&li);
          if (!li.use_empty()) li.replaceAllUsesWith(newli);
          deadInsts.insert(&li);
        }
        return;
      }
    }

    void visitStoreInst(StoreInst &si)
    {
      if(G2lDebug) errs() << "processing : " << si << '\n';
      Value *ptr = si.getPointerOperand();
      if(GlobalVariable *gv = dyn_cast<GlobalVariable>(ptr)) {
        AllocaInst *ai = getAlloca(gv,&si);
        StoreInst *newsi = new StoreInst(si.getValueOperand(),ai,false,&si);
        if (!si.use_empty()) si.replaceAllUsesWith(newsi);
        deadInsts.insert(&si);
        return;
      }

      // -- constant expression, this is usuaally a GEP
      if(ConstantExpr *ce = dyn_cast<ConstantExpr>(ptr)) {
        if(ce->getOpcode() != Instruction::GetElementPtr) return;
        if(GlobalVariable *gv = dyn_cast<GlobalVariable>(ce->getOperand(0))) {
          AllocaInst *ai = getAlloca(gv,&si);
          std::vector<Value*> idxs = getIdxs(ce);
          GetElementPtrInst *gepi = GetElementPtrInst::Create(ai,idxs.begin(),idxs.end(),"",&si);
          StoreInst *newsi = new StoreInst(si.getValueOperand(),gepi,false,&si);
          if (!si.use_empty()) si.replaceAllUsesWith(newsi);
          deadInsts.insert(&si);
        }
        return;
      }
    }

    ///set of GEP instructions with variable index
    std::set<GetElementPtrInst*> dynGEPs;

    ///convert GEPs with variable indices to a switch between GEPIs
    ///with constant indices
    void switchGEPIs()
    {
      for(std::set<GetElementPtrInst*>::iterator git = dynGEPs.begin(),ge = dynGEPs.end();git != ge;++git) {
        GetElementPtrInst *gepi = *git;
        Value *arg0 = gepi->idx_begin()->get();
        Value *arg1 = (gepi->idx_begin() + 1)->get();
        const PointerType *pt = gepi->getPointerOperandType();
        const ArrayType *at = dyn_cast<const ArrayType>(pt->getElementType());
        uint64_t arrSize = at->getNumElements();

        std::set<Value*> gepiUses;
        for (Value::use_iterator uit = gepi->use_begin(), ue = gepi->use_end();uit != ue; ++uit)
          gepiUses.insert(*uit);

        for (std::set<Value*>::iterator uit = gepiUses.begin(), ue = gepiUses.end();uit != ue; ++uit) {
          Instruction *ii = dyn_cast<Instruction>(*uit);
          if(!ii) continue;

          std::vector<BasicBlock*> swBBs(arrSize+1);
          BasicBlock *oldBB = ii->getParent();
          BasicBlock *newBB = oldBB->splitBasicBlock(ii);
          oldBB->getTerminator()->eraseFromParent();

          //create the default block
          swBBs[0] = BasicBlock::Create(oldBB->getContext(),"",newBB->getParent(),newBB);        
          
          SwitchInst *swi = SwitchInst::Create(arg1,swBBs[0],arrSize,oldBB);
          for(uint64_t i = 0;i < arrSize;++i) {
            swBBs[i+1] = BasicBlock::Create(oldBB->getContext(),"",newBB->getParent(),newBB);        
            swi->addCase(ConstantInt::get(dyn_cast<IntegerType>(arg1->getType()),i),swBBs[i+1]);
          }
        
          PHINode *phi = ii->use_empty() ? NULL : PHINode::Create(ii->getType(),"",ii);

          for(uint64_t i = 0;i <= arrSize;++i) {
            //default case
            if(i == 0) {
              Value *zero = ConstantInt::getSigned(assumeFunc->arg_begin()->getType(),0);
              InvokeInst::Create(assumeFunc,newBB,newBB,&zero,&zero+1,"",swBBs[0]);
              if(phi) phi->addIncoming(Constant::getNullValue(ii->getType()),swBBs[i]);
              if(phi) phi->addIncoming(Constant::getNullValue(ii->getType()),swBBs[i]);
              continue;
            }

            //other cases
            Constant *cind = ConstantInt::get(arg1->getType(),i-1,false);
            Value *newidx[] = {arg0,cind};
            GetElementPtrInst *newgepi = 
              GetElementPtrInst::CreateInBounds(gepi->getPointerOperand(),newidx,newidx+2,"",swBBs[i]);
            Instruction *newii = ii->clone();
            newii->insertAfter(newgepi);
            newii->replaceUsesOfWith(gepi,newgepi);
            if(phi) phi->addIncoming(newii,swBBs[i]);
            BranchInst::Create(newBB,swBBs[i]);
          }

          if(phi) ii->replaceAllUsesWith(phi);
          if(ii->use_empty()) deadInsts.insert(ii);
        }

        //remove instruction if no longer used
        if(gepi->use_empty()) deadInsts.insert(gepi);          
      }
    }

    void visitGetElementPtrInst (GetElementPtrInst &gepi)
    {      
      if(G2lDebug) errs() << "processing : " << gepi << '\n';
      Value *ptr = gepi.getPointerOperand();
      if(GlobalVariable *gv = dyn_cast<GlobalVariable>(ptr)) {
        AllocaInst *ai = getAlloca(gv,&gepi);
        std::vector<Value*> idxs = getIdxs(&gepi);
        GetElementPtrInst *newgepi = GetElementPtrInst::Create(ai,idxs.begin(),idxs.end(),"",&gepi);
        if (!gepi.use_empty()) gepi.replaceAllUsesWith(newgepi);
        deadInsts.insert(&gepi);

        //check if new GEP has variable index
        Value *arg = (newgepi->idx_begin() + 1)->get();
        if(isa<ConstantInt>(*arg)) return;
        const PointerType *pt = newgepi->getPointerOperandType();
        const ArrayType *at = dyn_cast<const ArrayType>(pt->getElementType());
        if(at) dynGEPs.insert(newgepi);
        return;
      }

      //check if GEP has variable index
      Value *arg = (gepi.idx_begin() + 1)->get();
      if(isa<ConstantInt>(*arg)) return;
      const PointerType *pt = gepi.getPointerOperandType();
      const ArrayType *at = dyn_cast<const ArrayType>(pt->getElementType());
      if(at) dynGEPs.insert(&gepi);
    }
  };

  struct Glob2Loc : public ModulePass 
  {
    static char ID;
    
    //the assume function
    Function *assumeFunc;

    Glob2Loc () : ModulePass(ID) { if(G2lDebug) errs () << "== Creating Glob2Loc Pass\n"; }

    ///at beginning
    void doInitialization(Module &M)
    {
      assumeFunc = M.getFunction("__CPROVER_assume");
      if(!assumeFunc) {
        LLVMContext &ctxt = M.getContext();
        std::vector<const Type*> argTypes(1);
        argTypes[0] = Type::getInt32Ty(ctxt);
        FunctionType *ft = FunctionType::get(Type::getVoidTy(ctxt),argTypes,false);
        assumeFunc = Function::Create(ft,GlobalValue::ExternalLinkage,"__CPROVER_assume",&M);
      }
    }    

    ///at end
    void doFinalization(Module &M)
    {
      WriteBitcodeToFile(&M,outs());
    }

    ///map from functions to globals they use
    std::map< Function*,std::set<GlobalVariable*> > func2Globs;

    ///recursively compute the set of functions that use a global
    void globUsage(GlobalVariable *gv,User *user)
    {
      if(Instruction *i = dyn_cast<Instruction>(user)) {
        func2Globs[i->getParent()->getParent()].insert(gv);
        return;
      }

      for(Value::use_iterator uit = user->use_begin(),ue = user->use_end();uit != ue;++uit)
        globUsage(gv,*uit);      
    }

    ///per function
    void runOnFunction(Function &F,std::map<GlobalVariable*,AllocaInst*> glob2alloca) 
    { 
      if(G2lDebug) {
        errs () << "Start converting globals to locals on function " << F.getName() << '\n';
        errs() << "===== before\n" << F << '\n';
      }

      G2LVisitor g2lv(assumeFunc,glob2alloca);
      g2lv.visit(F);
      g2lv.switchGEPIs();
      for(std::set<Instruction*>::iterator it = g2lv.deadInsts.begin(),e = g2lv.deadInsts.end();it != e;++it)
        (*it)->eraseFromParent();

      if(G2lDebug) {
        errs() << "===== after\n" << F << '\n';
        errs () << "Done converting globals to locals on function " << F.getName() << '\n';
      }
    }
    
    ///per function
    virtual bool runOnModule(Module &M) 
    {
      doInitialization(M);

      //compute map from functions to globals they use
      for(Module::iterator fit = M.begin(),fe = M.end();fit != fe;++fit)
        if(!fit->isDeclaration()) func2Globs[fit] = std::set<GlobalVariable*>();

      for(Module::global_iterator git = M.global_begin(),ge = M.global_end();git != ge;++git)
        for(Value::use_iterator uit = git->use_begin(),ue = git->use_end();uit != ue;++uit)
          globUsage(git,*uit);

      //clone each function, replacing global variables with new paramteres
      for(std::map< Function*,std::set<GlobalVariable*> >::const_iterator it1 = func2Globs.begin(),e1 = func2Globs.end();it1 != e1;++it1) {
        Function *F = it1->first;        
        bool isMain = F->getName().equals("main");
        
        //if this is the main function, make a clone and delete main
        if(isMain) {
          ValueToValueMapTy emptyMap;
          Function *newMain = llvm::CloneFunction(F,emptyMap,false,0);
          F->eraseFromParent();
          F = newMain;
        }
        
        const FunctionType *ft = F->getFunctionType();
        const Type *retType = F->getReturnType();

        //create parameters for cloned function. first create fresh
        //parameters for initial values of global variables.
        std::vector<const Type*> params;
        for(std::set<GlobalVariable*>::const_iterator it2 = it1->second.begin(),e2 = it1->second.end();it2 != e2;++it2)
          params.push_back((*it2)->getType()->getElementType());

        //next, create fresh parameters for final values of global
        //variables.
        for(std::set<GlobalVariable*>::const_iterator it2 = it1->second.begin(),e2 = it1->second.end();it2 != e2;++it2)
          params.push_back((*it2)->getType());

        //finally, copy the parameters from original function
        for(unsigned i = 0;i < ft->getNumParams();++i)
          params.push_back(ft->getParamType(i));

        //create new function
        FunctionType *newft = FunctionType::get(retType,params,false);
        Function *newFunc = Function::Create(newft,GlobalValue::ExternalLinkage,
                                             isMain ? "main" : F->getName() + "_renamed",
                                             &M);

        //create map from input values of global variables to
        //parameters
        Function::arg_iterator arg1 = newFunc->arg_begin();
        std::map<GlobalVariable*,Argument*> iglob2arg;
        for(std::set<GlobalVariable*>::const_iterator it2 = it1->second.begin(),e2 = it1->second.end();it2 != e2;++it2,++arg1) {
          iglob2arg[*it2] = arg1;
          arg1->setName((*it2)->getName() + "_i");
        }

        //create map from output values of global variables to
        //parameters
        std::map<GlobalVariable*,Argument*> oglob2arg;
        for(std::set<GlobalVariable*>::const_iterator it2 = it1->second.begin(),e2 = it1->second.end();it2 != e2;++it2,++arg1) {
          oglob2arg[*it2] = arg1;
          arg1->setName((*it2)->getName() + "_o");
        }

        //create the old parameter to new parameter map
        ValueToValueMapTy oldp2newp;
        for(Function::arg_iterator arg2 = F->arg_begin(),e2 = F->arg_end();arg2 != e2;++arg2,++arg1) {
          oldp2newp[arg2] = arg1;
          arg1->setName(arg2->getName());
        }
        
        //clone old function into new
        SmallVectorImpl<ReturnInst*> rets(4);
        llvm::CloneFunctionInto(newFunc,F,oldp2newp,!it1->second.empty(),rets,"",0);

        //create allocas corresponding to globals
        std::map<GlobalVariable*,AllocaInst*> glob2alloca;
        for(std::set<GlobalVariable*>::const_iterator it2 = it1->second.begin(),e2 = it1->second.end();it2 != e2;++it2,++arg1) {
          //create alloca
          BasicBlock &entryBB = newFunc->getEntryBlock();
          IRBuilder<> bldr(&entryBB,entryBB.begin());
          Twine lvname = (*it2)->getName() + "_l";
          AllocaInst *ai = bldr.CreateAlloca(iglob2arg[*it2]->getType(),0,lvname);
          glob2alloca[*it2] = ai;

          //store input value of global to alloca
          StoreInst *si = bldr.CreateStore(iglob2arg[*it2],ai);

          //copy alloca back to output value of global
          for(Function::iterator bit = newFunc->begin(),be = newFunc->end();bit != be;++bit) {
            TerminatorInst *ti = bit->getTerminator();
            if(!isa<ReturnInst>(ti)) continue;
            LoadInst *li2 = new LoadInst(ai,0,false,ti);
            StoreInst *si2 = new StoreInst(li2,oglob2arg[*it2],false,ti);
          }
        }

        runOnFunction(*newFunc,glob2alloca);
      }
      
      doFinalization(M);

      return true;
    }

    virtual void getAnalysisUsage(AnalysisUsage &AU) const
    {
      AU.setPreservesAll ();
    }
  };

  char Glob2Loc::ID = 0;
}

INITIALIZE_PASS_BEGIN(Glob2Loc, "glob2loc", "Convertor of globals to locals", false, false)
INITIALIZE_PASS_END(Glob2Loc, "glob2loc", "Convertor of globals to locals", false, false)
