/* ================================== */
/* LLREK LICENSE */
/* ================================== */
/* Copyright (c) 2013 Carnegie Mellon University. */
/* All Rights Reserved. */

/* Redistribution and use in source and binary forms, with or without */
/* modification, are permitted provided that the following conditions are */
/* met: */

/* 1. LLREK includes and/or makes use of certain third party software */
/* ("Third Party Software"). You agree to comply with any and all the */
/* Third Party Software terms and conditions listed below and/or */
/* contained in any separate license file distributed with LLREK. The */
/* parties who own the Third Party Software ("Third Party Licensors") are */
/* intended third party beneficiaries to this License with respect to the */
/* terms applicable to their Third Party Software. */

/* 2. Redistributions of source code must retain the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers. */

/* 3. Redistributions in binary form must reproduce the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers in the documentation and/or other materials provided with */
/* the distribution. */

/* 4. Products derived from this software may not include Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" in the name of such derived product, nor shall Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" be used to endorse or promote products derived from this */
/* software without prior written permission. For written permission, */
/* please contact permission@sei.cmu.edu. */

/* This material is based upon work funded and supported by the */
/* Department of Defense under Contract No. FA8721-05-C-0003 with */
/* Carnegie Mellon University for the operation of the Software */
/* Engineering Institute, a federally funded research and development */
/* center. */

/* Any opinions, findings and conclusions or recommendations expressed in */
/* this material are those of the author(s) and do not necessarily */
/* reflect the views of the United States Department of Defense. */

/* NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING */
/* INSTITUTE MATERIAL IS FURNISHEDON AN AS-IS BASIS. CARNEGIE */
/* MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR */
/* IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF */
/* FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS */
/* OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT */
/* MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, */
/* TRADEMARK, OR COPYRIGHT INFRINGEMENT. */

/* This material has been approved for public release and unlimited */
/* distribution. */

/* DM-0000665 */

#include "ufo/InitializePasses.h"

#include "llvm/Support/CommandLine.h"
#include "llvm/Module.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Target/TargetData.h"
#include "llvm/Transforms/Utils/Cloning.h"
#include "llvm/ADT/StringExtras.h"

#include "ufo/Cpg/Topo.hpp"
#include "ufo/Util/NameValues.hpp"

#include "ufo/Smt/UfoZ3.hpp"
#include "ufo/CptLiveValues.hpp"

using namespace llvm;
using namespace ufo;

//-- options
static cl::opt<bool>
RekscDebug ("reksc-debug", cl::desc ("Generate debug output"),
            cl::init (false));
static cl::opt<bool>
RekscStp ("reksc-stp", cl::desc ("Use STP as backend solver"),
          cl::init (false));
static cl::opt<bool>
RekscBoolector ("reksc-boolector", cl::desc ("Use Boolector as backend solver"),
                cl::init (false));
static cl::opt<bool>
RekscModel ("reksc-model", cl::desc ("Generate model if SAT"),
            cl::init (false));
static cl::opt<std::string>
RekscSnap ("reksc-snap", cl::desc ("Level of snapshotting shared variables (NONE, WRITE, ALL)"),
            cl::init ("NONE"));
static cl::opt<bool>
RekscCnf ("reksc-cnf", cl::desc ("Generate CNF and stop"),
          cl::init (false));

#include "./Z3Conv.hpp"
#include "./RekSCVCGen.hpp"

namespace ufo
{
  ///number of bits used for round variables, priorities, and local
  ///clocks
  static const unsigned RNDBITS = 10;
  static const unsigned PRIOBITS = 4;
  static const unsigned LCLKBITS = 18;

  ///prefixes used to identify task parameters
  static const char *LLTBOUND = "llrek_time_bound";
  static const char *LLPRIO = "llrek_base_priority_";
  static const char *LLPERIOD = "llrek_period_";
  static const char *LLWCET = "llrek_wcet_";
  static const char *LLARRIVAL = "llrek_arrival_";

  ///number of bits used for lamport clock variables
  static const unsigned CLKBITS = RNDBITS + PRIOBITS + LCLKBITS;

  //forward declaration of Task
  struct Task;

  ///info about a job
  struct Job
  {
    ///the task
    Task *task;

    ///the job id
    unsigned jobId;

    ///the call instruction marking the start of the job
    const CallInst *jobStart;

    ///the call instruction marking the end of the job
    const CallInst *jobEnd;

    ///the list of reads, writes, snapshots and priority operations in
    ///topologically sorted order
    std::list<const CallInst*> accesses;

    ///map from global variables to their writes
    std::map< GlobalVariable*,std::list<const CallInst*> > writes;

    ///map from global variables to their snapshots
    std::map<GlobalVariable*,const CallInst*> snapshots;

    ///list of priority operations of this job
    std::list<const CallInst*> priOps;

    ///the starting and ending round variables for the job
    Expr startRnd, endRnd;

    ///default constructor
    Job() : task(0),jobId(-1),jobEnd(0) {}

    ///constructor
    Job(Task *t,unsigned jid,const CallInst *js,const CallInst *je,ExprFactory &efac);

    //-- NOTE: the default copy constructor and assignment operators
    //-- should work fine

    ///equality of jobs
    bool operator == (const Job &rhs) const;

    ///return the priority of the job
    unsigned prio() const;

    ///return the arrival time of this job
    unsigned arr() const;

    ///return the maximum departure time of this job
    unsigned dep() const;

    ///a string representation
    std::string dump() const;

    ///return true iff j1 always finishes before j2 starts
    friend bool operator << (const Job &j1,const Job &j2);

    ///return true iff j1 could be preempted by j2
    friend bool operator ^ (const Job &j1,const Job &j2);

    ///return true iff (j1 ^ j2) or (j2 ^ j1)
    friend bool operator * (const Job &j1,const Job &j2);

    ///return true iff j1 << j2 || j1 ^ j2. NOTE: this is a strict job
    ///order defined in our VMCAI'13 paper
    friend bool operator < (const Job &j1,const Job &j2);
  };

  ///a task descriptor
  struct Task
  {
    ///the task id
    unsigned taskId;

    ////name of the job function
    std::string funcName; 

    ///the job function for the task
    Function *jobFunc;

    ///task parameters
    unsigned prio; //priority
    unsigned period; //period
    unsigned wcet; //wcet
    unsigned arrival; //arrival time
    unsigned resp; //response time 

    ///the task function -- created during sequentialization
    Function *taskFunc;

    ///the number of jobs and vector of info for each job. these are
    ///created along with taskFunc
    unsigned numJobs;
    std::vector<Job> jobs;

    ///default constructor
    Task() : taskFunc(0) {}

    ///constructor
    Task(unsigned tid,std::string _fn,unsigned _prio,unsigned _period,
         unsigned _wcet,unsigned _arrival) :
      taskId(tid),funcName(_fn),jobFunc(0),prio(_prio),period(_period),
      wcet(_wcet),arrival(_arrival),resp(0),taskFunc(0) {}

    //-- NOTE: the default copy constructor and assignment operators
    //-- should work fine

    //print to errs()
    void dump() const
    {
      errs() << "task " << taskId << "{\n" 
             << "  funcName = " << funcName << "\n"
             << "  prio = " << prio << "\n"
             << "  period = " << period << "\n"
             << "  wcet = " << wcet << "\n"
             << "  arrival = " << arrival << "\n"
             << "  response time = " << resp << "\n"
             << "}\n";
    }
  };

  //-- we can now define various things now that all declarations are
  //-- in place

  ///job constructor
  Job::Job(Task *t,unsigned jid,const CallInst *js,const CallInst *je,ExprFactory &efac) :
    task(t),jobId(jid),jobStart(js),jobEnd(je)
  {
    //create the job start and end variables
    std::string buf;
    raw_string_ostream os(buf);
    os << "start_rnd_" << task->taskId << "_" << jobId;
    startRnd = bvop::bvConst(mkTerm<std::string>(os.str(),efac),RNDBITS);
    buf.clear();
    os << "end_rnd_" << task->taskId << "_" << jobId;
    endRnd = bvop::bvConst(mkTerm<std::string>(os.str(),efac),RNDBITS);      
    if(RekscDebug) errs() << "start round = " << *startRnd
                          << " end round = " << *endRnd << '\n';
  }

  ///equality of jobs
  bool Job::operator == (const Job &rhs) const
  {
    return task->taskId == rhs.task->taskId && jobId == rhs.jobId;
  }

  ///return the priority of the job
  unsigned Job::prio() const { return task->prio; }

  ///return the arrival time of this job
  unsigned Job::arr() const { return task->arrival + jobId * task->period; }

  ///return the maximum departure time of this job
  unsigned Job::dep() const { return arr() + task->resp; }

  ///a string representation
  std::string Job::dump() const
  {    
    std::string buf;
    raw_string_ostream os(buf);
    os << task->funcName << "_" << jobId;
    return buf;
  }

  ///return true iff j1 always finishes before j2 starts
  bool operator << (const Job &j1,const Job &j2)
  {
    return (j1.prio() <= j2.prio() && j1.dep() <= j2.arr()) ||
      (j1.prio() > j2.prio() && j1.arr() <= j2.arr());
  }

  ///return true iff j2 could be preempted by j2
  bool operator ^ (const Job &j1,const Job &j2)
  {
    return j1.prio() < j2.prio() && j1.arr() < j2.arr() && j2.arr() < j1.dep();
  }

  ///return true iff (j1 ^ j2) or (j2 ^ j1)
  bool operator * (const Job &j1,const Job &j2)
  {
    return (j1 ^ j2) || (j2 ^ j1);
  }

  ///return true iff j1 << j2 || j1 ^ j2. NOTE: this is a strict job
  ///order defined in our VMCAI'13 paper
  bool operator < (const Job &j1,const Job &j2)
  {
    return (j1 << j2) || (j1 ^ j2);
  }

  ///information for an event. has a Lamport clock, and a function
  ///call marking the event
  struct Event
  {
    ///the event itself
    const CallInst *event;

    ///the job that this event belongs to
    Job *job;

    ///the lamport clock variable
    Expr lampClk;

    ///default constructor
    Event() : event(0),job(0) {}

    ///constructor
    Event(const CallInst *ci,Job *j,ExprFactory &efac) : event(ci),job(j)
    {
      //create the lamport clock
      std::string buf;
      raw_string_ostream os(buf);
      os << "lamp_clock_" << ci;
      lampClk = bvop::bvConst(mkTerm<std::string>(os.str(),efac),CLKBITS);
      if(RekscDebug) errs() << "lamp_clock = " << *lampClk << '\n';
    }

    //-- NOTE: the default copy constructor and assignment operators
    //-- should work just fine.
  };

  ///a generic observer event, i.e., an event with an additional
  ///lamport clock for the event it observes. subsequently we will
  ///specialize this to read, and the four priority operations
  struct Observer : public Event
  {
    ///the lamport clock of the observed event
    Expr obsClk;

    ///default constructor
    Observer() {}

    ///constructor
    Observer(const CallInst *ci,Job *j,ExprFactory &efac) : Event(ci,j,efac)
    {
      //create the observer clock
      std::string buf;
      raw_string_ostream os(buf);
      os << "obs_clock_" << ci;
      obsClk = bvop::bvConst(mkTerm<std::string>(os.str(),efac),CLKBITS);
      if(RekscDebug) errs() << " obs clock = " << *obsClk << '\n';
    }

    //-- NOTE: the default copy constructor and assignment operators
    //-- should work just fine.
  };

  ///information for each read access. a read is an observer event.
  struct Read : public Observer
  {
    ///default constructor
    Read() : Observer() {}

    ///constructor
    Read(const CallInst *ci,Job *j,ExprFactory &efac) : Observer(ci,j,efac) {}
  };

  ///information for each snapshot. a snapshot is an observer event.
  struct Snapshot : public Observer
  {
    ///default constructor
    Snapshot() : Observer() {}

    ///constructor
    Snapshot(const CallInst *ci,Job *j,ExprFactory &efac) : Observer(ci,j,efac) {}
  };

  ///types of priority operations
  enum PriOpType { PRIOP_NONE, JOB_START, JOB_END, PCP_LOCK, PCP_UNLOCK, CPU_LOCK, CPU_UNLOCK };

  ///information for each priority operation. note that a priority
  ///operation is an observer event.
  struct PriOp : public Observer
  {
    ///the type of the operation
    PriOpType type;

    ///default constructor
    PriOp() : Observer(), type(PRIOP_NONE) {}

    ///constructor
    PriOp(PriOpType t,const CallInst *ci,Job *j,ExprFactory &efac) : Observer(ci,j,efac),type(t) {}
  };

  ///information for each write access, a write is just an event
  struct Write : public Event
  {
    ///default constructor
    Write() {}

    ///constructor
    Write(const CallInst *ci,Job *j,ExprFactory &efac) : Event(ci,j,efac) {}
  };

  ///the main class for doing SC based sequentialization
  struct RekSC : public ModulePass 
  {
    static char ID;
    ExprFactory efac;
    Expr trueE,falseE;
    BVEnv env;
    
    //-- the periodic program being analyzed
    std::vector<Task> tasks; //the array of tasks
    unsigned timeBound; //the time-bound for verification

    //-- the sorted list of all jobs according to the strict order
    //-- defined in our VMCAI'13 paper
    std::vector<Job*> sortedJobs;

    //-- map from global variables to list of task ids that write to
    //-- them
    std::map< GlobalVariable*,std::set<unsigned> > glob2Writers;

    ///map from global variables to their initial values
    std::map<GlobalVariable*,Expr> glob2Inits;

    //-- maps from global variables to its reads and writes
    std::map< GlobalVariable*,std::list<const CallInst*> > glob2Reads;
    std::map< GlobalVariable*,std::list<const CallInst*> > glob2Writes;

    //-- maps from reads, writes, snapshots, and priority operations
    //-- to their info
    std::map<const CallInst*,Read> reads;
    std::map<const CallInst*,Write> writes;
    std::map<const CallInst*,Snapshot> snapshots;
    std::map<const CallInst*,PriOp> priOps;

    ///the INIT function and its exit block
    Function *initFunc;
    BasicBlock *initExit;

    ///the FINAL function and its exit block
    Function *finalFunc;
    BasicBlock *finalExit;

    ///map from functions to their VCGens
    std::map<Function*,RekSCVCGen> vcgens;

    ///the expression representing the job scheduling constraints
    Expr schedCons;

    ///the expression representing the lamport clock constraints
    Expr clockCons;

    ///the constraint on observed clocks
    Expr obsCons;

    ///the constraints generated by processing final
    Expr finalCons;

    RekSC () : ModulePass(ID),env(efac),timeBound(UINT_MAX),
               initFunc(0),initExit(0)
    { 
      errs () << "== Creating RekSC Pass\n";
      trueE = mk<TRUE> (efac);
      falseE =  mk<FALSE> (efac);
      schedCons = trueE;
      clockCons = trueE;
      obsCons = trueE;
      finalCons = trueE;
    }

    ///compute VC for a given function
    void computeVC(Function *F)
    {
      //get function, sort topologically and check for cycles
      Topo &topo = getAnalysis<Topo>(*F);
      assert(!topo.hasBackEdges(F) && "ERROR: function is not acyclic");

      //compute the VC
      errs () << "Start generating bitvector VC \n";
      std::map<Function*,RekSCVCGen>::iterator vcgenit = 
        vcgens.insert(std::pair<Function*,RekSCVCGen>(F,RekSCVCGen(efac, env, *F, topo, false))).first;
      RekSCVCGen &vcgen = vcgenit->second;
      vcgen.compute ();
      if(RekscDebug) errs() << "VC : " << *vcgen.getCond() << '\n';
      errs () << "Done generating bitvector VC \n";
    }

    ///split the return instruction of INIT into separate block. then
    ///compute VC of INIT.
    void processInit(Module &M)
    {
      initFunc = M.getFunction(LLINIT);
      if(!initFunc || initFunc->isDeclaration()) return;

      //split the return instruction
      BasicBlock &bb = initFunc->getEntryBlock();
      initExit = bb.splitBasicBlock(bb.getTerminator(),"");

      //compute VC
      computeVC(initFunc);
    }

    ///split the return instruction of FINAL into separate block. then
    ///compute VC of FINAL.
    void processFinal(Module &M)
    {
      finalFunc = M.getFunction(LLFINAL);
      if(!finalFunc || finalFunc->isDeclaration()) return;

      //split the return instruction
      BasicBlock &bb = finalFunc->getEntryBlock();
      finalExit = bb.splitBasicBlock(bb.getTerminator(),"");

      //compute VC
      computeVC(finalFunc);

      //add clock constraints and observation constraints. we want the
      //clocks to be bigger than all job clocks. so we just set all
      //clocks with round equal to total number of jobs. for
      //observation, we only deal with reads, and treat them like
      //ordinary reads.

      //list of conjuncts of the final constraints
      std::list<Expr> res;

      //for each basic block
      for(Function::iterator fit = finalFunc->begin(),fe = finalFunc->end();fit != fe;++fit) {
        //for each instruction
        for(BasicBlock::iterator bit = fit->begin(),be = fit->end();bit != be;++bit) {
          if(CallInst *ci = dyn_cast<CallInst>(bit)) {
            Function *Callee = ci->getCalledFunction ();

            //ignore non-reads
            if(!Callee->getName ().startswith (LLREAD)) continue;

            //create observed clock for the read
            std::string buf;
            raw_string_ostream os(buf);
            os << "obs_clock_" << ci;
            Expr robs = bvop::bvConst(mkTerm<std::string>(os.str(),efac),CLKBITS);

            //collect constraints that equal read value to value wriiten
            //by the observed write. initialize it with the case that no
            //write is observed, i.e., with the initial value of the
            //global variable read. note that in this case, the observed
            //clock is assumed to be 0, since this is only possible if no
            //write is observed.
            std::list<Expr> readCons;
            readCons.push_back(mk<AND>(mk<EQ>(numExpr(0,CLKBITS),robs),
                                       mk<EQ>(env.lookup(*ci),defaultObserved(ci))));

            //ensure that the max clock is as big as every write to the
            //same global variable that has occurred
            BOOST_FOREACH(const Event &write,canObserve(ci)) {
              //dump may-observe relation between read and write
              errs() << "<<OBS>> <<" << RekscSnap << ">> " << ci << " " << write.event << '\n';
              Expr pre = executed(write.event);
              Expr wclk = write.lampClk;
              Expr post = mk<BVULE>(wclk,robs);          
              res.push_back(mk<IMPL>(pre,post));

              //add constraint equating value 
              readCons.push_back(mk<AND>(mk<AND>(pre,mk<EQ>(wclk,robs)),
                                         mk<EQ>(env.lookup(*ci),env.lookup(*write.event))));
            }

            //add disjunction of readCons
            res.push_back(mknary<OR>(falseE,readCons));

          }
        }
      }

      finalCons = mknary<AND> (trueE, res.begin (), res.end ());
      if(RekscDebug) errs() << "final cons\n" << *finalCons << '\n';
    }

    ///add function calls snapshotting shared variables
    void addSnapshots(BasicBlock *bb,const Task &td,Module &M)
    {
      //if no snapshotting
      if(RekscSnap == "NONE") return;

      //depending on the type of snapshotting, either snapshot all
      //variables, or variables written by this function
      typedef std::map< GlobalVariable*,std::set<unsigned> > G2W;
      BOOST_FOREACH(const G2W::value_type &vt,glob2Writers) {
        if(RekscSnap == "ALL" ||
           (RekscSnap == "WRITE" && vt.second.count(td.taskId))) {
          Function *snapFunc = M.getFunction(LLSNAP + vt.first->getName().str());
          CallInst::Create(snapFunc,vt.first,"",bb);
        }
      }
    }

    ///create the function for each task by creating a function that
    ///calls the job function the correct number of times determined
    ///by the period and the time bound, and then inlining it.
    void createTaskFunc(Task &td,Module &M)
    {
      //create task function declaration
      LLVMContext &ctxt = M.getContext();
      std::vector<const Type*> argTypes;
      FunctionType *ft = FunctionType::get(Type::getVoidTy(ctxt),argTypes,false);
      td.taskFunc = Function::Create(ft,GlobalValue::ExternalLinkage,"task_func_" + td.funcName,&M);

      //allocate space for job end instructions
      td.numJobs = timeBound / td.period;
      td.jobs.resize(td.numJobs);

      //create the task body -- call the job function correct no. of times
      BasicBlock *bb = BasicBlock::Create(ctxt,"",td.taskFunc,0);
      for(unsigned i = 0;i < td.numJobs;++i) {
        //create and call function to indicate start of the job
        Function *jobStartFunc = Function::Create(ft,GlobalValue::ExternalLinkage,
                                                  LLJOB_START + td.funcName + "_" + utostr(i),&M);
        CallInst *jstart = CallInst::Create(jobStartFunc,"",bb);

        //call the job function
        CallInst::Create(td.jobFunc,"",bb);

        //snapshot shared variables
        addSnapshots(bb,td,M);

        //create and call function to indicate end of the job
        Function *jobEndFunc = Function::Create(ft,GlobalValue::ExternalLinkage,
                                                LLJOB_END + td.funcName + "_" + utostr(i),&M);
        CallInst *jend = CallInst::Create(jobEndFunc,"",bb);

        //create the job
        td.jobs[i] = Job(&td,i,jstart,jend,efac);
      }

      //create the return instruction at the end
      ReturnInst::Create(ctxt,bb);
    }

    ///return the maximal subset of a set of jobs. assume the given
    ///set is sorted.
    std::list<Job*> maximal(const std::list<Job*> &jobs)
    {
      std::list<Job*> res;
      for(std::list<Job*>::const_iterator jit = jobs.begin(),je = jobs.end();jit != je;++jit) {
        bool isMax = true;
        std::list<Job*>::const_iterator jit2 = jit;          
        for(++jit2;isMax && jit2 != je;++jit2) {
          if(**jit << **jit2) isMax = false;
        }
        if(isMax) res.push_back(*jit);
      }
      return res;
    }

    ///return list of jobs that interfere with the argument
    std::list<Job*> interfering(Job &job)
    {
      std::list<Job*> res;
      BOOST_FOREACH(Job *j,sortedJobs)
        if(*j * job) res.push_back(j);
      return res;
    }

    ///return list of jobs that can preempt the argument
    std::list<Job*> preempting(Job &job)
    {
      std::list<Job*> res;
      BOOST_FOREACH(Job *j,sortedJobs)
        if(job ^ *j) res.push_back(j);
      return res;
    }

    ///return list of jobs that can be preempted by the argument
    std::list<Job*> preempted(Job &job)
    {
      std::list<Job*> res;
      BOOST_FOREACH(Job *j,sortedJobs)
        if(*j ^ job) res.push_back(j);
      return res;
    }

    ///sort jobs according to the VMCAI'13 order. we do it now because
    ///the Job objects are still pretty small in size. their accessors
    ///have not been set. so we can simply sort them as objects using
    ///their defined "<" operator, and then copy their pointers into
    ///the sortedJobs array.
    void sortJobs()
    {
      //sort the jobs
      std::set<Job> jobSet;
      BOOST_FOREACH(Task &td,tasks) {
        BOOST_FOREACH(Job &j,td.jobs)
          jobSet.insert(j);
      }

      //insert pointers into sortedJobs
      if(RekscDebug) errs() << "++++++++++++++ begin sorted jobs\n";
      sortedJobs.resize(jobSet.size());
      unsigned id = 0;
      BOOST_FOREACH(const Job &j,jobSet) {
        sortedJobs[id++] = &(j.task->jobs[j.jobId]);
        if(RekscDebug) errs() << j.dump() << '\n';
      }
      if(RekscDebug) errs() << "++++++++++++++ end sorted jobs\n";
    }

    ///inline a function. when this method is done, func should only
    ///call functions that are declaration and that are LLREK-specific
    void inlineFunc(Function *func)
    {
      for(;;) {
        //collect callsites to inline
        std::list<CallInst*> callSites;
        for(Function::iterator fit = func->begin(),fe = func->end();fit != fe;++fit) {
          for(BasicBlock::iterator bit = fit->begin(),be = fit->end();bit != be;++bit) {
            if(CallInst *ci = dyn_cast<CallInst>(bit)) {
              Function *Callee = ci->getCalledFunction ();
              if(!Callee || Callee->isDeclaration() || 
                 Callee->getName ().startswith (LLREAD) || Callee->getName ().startswith (LLWRITE))
                continue;
              callSites.push_back(ci);
            }
          }
        }

        //if no call-sites to inline
        if(callSites.empty()) break;

        //inline all call-sites
        BOOST_FOREACH(CallInst *ci,callSites) {
          InlineFunctionInfo ifi(0,bitvc::TD);
          assert(InlineFunction(ci,ifi));
        }
      }

      if(RekscDebug) func->dump();
    }

    ///collect info about the task function and populate various data
    ///structures
    void collectTaskInfo(Task &td)
    {
      //create glob2Reads, glob2Writes, reads, and writes. also
      //associate reads and writes in topological order with the
      //appropriate jobs
      Topo &topo = getAnalysis<Topo>(*td.taskFunc);
      unsigned jobId = 0; //start with the first job
      BOOST_FOREACH(const BasicBlock *bb,topo.topo[td.taskFunc]) {
        for(BasicBlock::const_iterator bit = bb->begin(),be = bb->end();bit != be;++bit) {
          if(const CallInst *ci = dyn_cast<CallInst>(bit)) {
            if(RekscDebug) errs() << "+++ handling " << *ci << '\n';

            //process job end
            if(ci == td.jobs[jobId].jobEnd) {
              priOps[ci] = PriOp(JOB_END,ci,&(td.jobs[jobId]),efac);
              if(RekscDebug) errs() << "=== adding to job " << jobId << '\n';
              td.jobs[jobId].accesses.push_back(ci);
              td.jobs[jobId].priOps.push_back(ci);
              if(RekscDebug)
                errs() << "EVENT: " << *lampClock(ci) << " : " << td.jobs[jobId].dump()  << " : " << *ci << '\n';
              //move to next job
              ++jobId;
              continue;
            }

            //process job start
            if(ci == td.jobs[jobId].jobStart) {
              priOps[ci] = PriOp(JOB_START,ci,&(td.jobs[jobId]),efac);
              if(RekscDebug) errs() << "=== adding to job " << jobId << '\n';
              td.jobs[jobId].accesses.push_back(ci);
              td.jobs[jobId].priOps.push_back(ci);
              if(RekscDebug)
                errs() << "EVENT: " << *lampClock(ci) << " : " << td.jobs[jobId].dump()  << " : " << *ci << '\n';
              continue;
            }

            //get called function
            Function *Callee = ci->getCalledFunction ();

            //process PCP lock
            if(Callee->getName ().equals (LLPCP_LOCK)) {
              priOps[ci] = PriOp(PCP_LOCK,ci,&(td.jobs[jobId]),efac);
              if(RekscDebug) errs() << "=== adding to job " << jobId << '\n';
              td.jobs[jobId].accesses.push_back(ci);
              td.jobs[jobId].priOps.push_back(ci);
              if(RekscDebug)
                errs() << "EVENT: " << *lampClock(ci) << " : " << td.jobs[jobId].dump()  << " : " << *ci << '\n';
              continue;
            }

            //process CPU lock
            if(Callee->getName ().equals (LLCPU_LOCK)) {
              priOps[ci] = PriOp(CPU_LOCK,ci,&(td.jobs[jobId]),efac);
              if(RekscDebug) errs() << "=== adding to job " << jobId << '\n';
              td.jobs[jobId].accesses.push_back(ci);
              td.jobs[jobId].priOps.push_back(ci);
              if(RekscDebug)
                errs() << "EVENT: " << *lampClock(ci) << " : " << td.jobs[jobId].dump()  << " : " << *ci << '\n';
              continue;
            }

            //process PCP unlock
            if(Callee->getName ().equals (LLPCP_UNLOCK)) {
              priOps[ci] = PriOp(PCP_UNLOCK,ci,&(td.jobs[jobId]),efac);
              if(RekscDebug) errs() << "=== adding to job " << jobId << '\n';
              td.jobs[jobId].accesses.push_back(ci);
              td.jobs[jobId].priOps.push_back(ci);
              if(RekscDebug)
                errs() << "EVENT: " << *lampClock(ci) << " : " << td.jobs[jobId].dump()  << " : " << *ci << '\n';
              continue;
            }

            //process CPU unlock
            if(Callee->getName ().equals (LLCPU_UNLOCK)) {
              priOps[ci] = PriOp(CPU_UNLOCK,ci,&(td.jobs[jobId]),efac);
              if(RekscDebug) errs() << "=== adding to job " << jobId << '\n';
              td.jobs[jobId].accesses.push_back(ci);
              td.jobs[jobId].priOps.push_back(ci);
              if(RekscDebug)
                errs() << "EVENT: " << *lampClock(ci) << " : " << td.jobs[jobId].dump()  << " : " << *ci << '\n';
              continue;
            }

            //process snapshot
            if(Callee->getName ().startswith (LLSNAP)) {
              snapshots[ci] = Snapshot(ci,&(td.jobs[jobId]),efac);
              if(RekscDebug) errs() << "=== adding to job " << jobId << '\n';
              td.jobs[jobId].accesses.push_back(ci);
              GlobalVariable *gv = dyn_cast<GlobalVariable>(ci->getArgOperand(0));
              td.jobs[jobId].snapshots[gv] = ci;
              if(RekscDebug)
                errs() << "EVENT: " << *lampClock(ci) << " : " << td.jobs[jobId].dump()  << " : " << *ci << '\n';
              continue;
            }

            //process read
            if(Callee->getName ().startswith (LLREAD)) {
              if(GlobalVariable *gv = dyn_cast<GlobalVariable>(ci->getArgOperand(0))) {
                glob2Reads[gv].push_back(ci);
                if(RekscDebug) errs() << "glob2Reads[" << *gv << "] <=== " << *ci << '\n';
                reads[ci] = Read(ci,&(td.jobs[jobId]),efac);
                if(RekscDebug) errs() << "=== adding to job " << jobId << '\n';
                td.jobs[jobId].accesses.push_back(ci);
                if(RekscDebug)
                  errs() << "EVENT: " << *lampClock(ci) << " : " << td.jobs[jobId].dump()  << " : " << *ci << '\n';
              }
              continue;
            }

            //process write
            if(Callee->getName ().startswith (LLWRITE)) {
              if(GlobalVariable *gv = dyn_cast<GlobalVariable>(ci->getArgOperand(0))) {
                glob2Writes[gv].push_back(ci);
                if(RekscDebug) errs() << "glob2Writes[" << *gv << "] <=== " << *ci << '\n';
                writes[ci] = Write(ci,&(td.jobs[jobId]),efac);
                if(RekscDebug) errs() << "=== adding to job " << jobId << '\n';
                td.jobs[jobId].accesses.push_back(ci);
                td.jobs[jobId].writes[gv].push_back(ci);
                if(RekscDebug)
                  errs() << "EVENT: " << *lampClock(ci) << " : " << td.jobs[jobId].dump()  << " : " << *ci << '\n';
              }
              continue;
            }
          }
        }
      }
    }

    ///create the periodic program
    void createPP(Module &M)
    {
      //-- create set of tasks by going through global variables
      typedef std::map<StringRef,Task> N2T;
      N2T name2Tasks;
      BOOST_FOREACH(const GlobalVariable &gv,M.getGlobalList()) {
        StringRef gvName = gv.getName();
        if(gvName.equals(LLTBOUND))
          timeBound = dyn_cast<ConstantInt>(gv.getInitializer())->getZExtValue();
        else if(gvName.startswith(LLPRIO)) {
          StringRef taskName = gvName.substr(strlen(LLPRIO));
          name2Tasks[taskName].prio = dyn_cast<ConstantInt>(gv.getInitializer())->getZExtValue();
        } else if(gvName.startswith(LLPERIOD)) {
          StringRef taskName = gvName.substr(strlen(LLPERIOD));
          name2Tasks[taskName].period = dyn_cast<ConstantInt>(gv.getInitializer())->getZExtValue();
        } else if(gvName.startswith(LLWCET)) {
          StringRef taskName = gvName.substr(strlen(LLWCET));
          name2Tasks[taskName].wcet = dyn_cast<ConstantInt>(gv.getInitializer())->getZExtValue();
        } else if(gvName.startswith(LLARRIVAL)) {
          StringRef taskName = gvName.substr(strlen(LLARRIVAL));
          name2Tasks[taskName].arrival = dyn_cast<ConstantInt>(gv.getInitializer())->getZExtValue();
        }
      }

      //-- must set time bound
      assert(timeBound != UINT_MAX && "ERROR: no time bound set!!");

      //-- sort tasks by priority, assign job function name and job
      //-- function, set response time to 0
      typedef std::map<unsigned,Task> P2T;
      P2T prio2Tasks;
      BOOST_FOREACH(const N2T::value_type &vt,name2Tasks) {
        Task t = vt.second;
        t.funcName = vt.first.str();
        t.jobFunc = M.getFunction(t.funcName);
        t.resp = 0;
        prio2Tasks[t.prio] = t;
      }

      //-- add sorted tasks to global tasks vector, and assign taskId
      tasks.resize(prio2Tasks.size());
      unsigned taskId = 0;
      BOOST_FOREACH(const P2T::value_type &vt,prio2Tasks) {
        Task t = vt.second;
        t.taskId = taskId;
        tasks[taskId] = t;
        ++taskId;
      }

      //collect all job functions and inline them
      std::set<Function*> jobFuncs;
      BOOST_FOREACH(const Task &td,tasks) jobFuncs.insert(td.jobFunc);
      BOOST_FOREACH(Function *func,jobFuncs) inlineFunc(func);

      //-- update glob2Writers, and create necessary snapshot
      //-- functions
      BOOST_FOREACH(const Task &td,tasks) {
        for(Function::iterator fit = td.jobFunc->begin(),fe = td.jobFunc->end();fit != fe;++fit) {
          for(BasicBlock::iterator bit = fit->begin(),be = fit->end();bit != be;++bit) {
            if(CallInst *ci = dyn_cast<CallInst>(bit)) {
              Function *Callee = ci->getCalledFunction ();
              if(!Callee->getName ().startswith (LLWRITE)) continue;
              GlobalVariable *gv = dyn_cast<GlobalVariable>(ci->getArgOperand(0));
              glob2Writers[gv].insert(td.taskId);

              //create snapshot function
              std::string fnName = LLSNAP + gv->getName().str();
              if(!M.getFunction(fnName)) {
                std::vector<const Type*> argTypes(1,gv->getType());
                FunctionType *ft = FunctionType::get(gv->getType()->getElementType(),argTypes,false);
                Function::Create(ft,GlobalValue::ExternalLinkage,fnName,&M);
              }
            }
          }
        }
      }

      //-- compute response times of tasks via RMS
      unsigned numTasks = tasks.size();
      for(int i = numTasks - 1;i >= 0;--i) {
        tasks[i].resp = tasks[i].wcet;
        unsigned curr = 0;
        while(tasks[i].resp != curr) {
          curr = tasks[i].resp;
          tasks[i].resp = tasks[i].wcet;
          for(int j = numTasks - 1;j > i;--j) {
            unsigned prBound = (unsigned)ceil(curr * 1.0 / tasks[j].period);
            tasks[i].resp += prBound * tasks[j].wcet; 
          }
        }

        if(RekscDebug) errs() << "resp[" << i << "] = " << tasks[i].resp << '\n';        
      }

      if(RekscDebug) {
        errs() << "time bound  = " << timeBound << '\n';
        BOOST_FOREACH(const Task &t,tasks) t.dump();
      }
    }

    ///given a number N and a bit-width W, return the bit-vector
    ///constant for N with width W
    Expr numExpr(unsigned N,unsigned W)
    {
      return bvop::bvConst(mkTerm<llui_t>(N,efac),W);
    }

    ///compute job scheduling constraints, i.e., constraints on start
    ///and end rounds of jobs
    void computeSchedConstraints()
    {
      //list of conjuncts of the constraints
      std::list<Expr> res;
      
      //create the expression for number of rounds. number of rounds
      //equals total number of jobs.
      unsigned numRounds = 0;
      BOOST_FOREACH(Task &td,tasks) numRounds += td.numJobs;
      Expr numRnds = numExpr(numRounds,RNDBITS);
      
      //jobs are sequential
      BOOST_FOREACH(Task &td,tasks) {
        BOOST_FOREACH(Job &j,td.jobs) {
          res.push_back(mk<BVULE>(j.startRnd,j.endRnd));
          //XXX: this is an optimization
          res.push_back(mk<BVULT>(j.endRnd,numRnds));
        }
      }
      
      //jobs are well-separated
      BOOST_FOREACH(Task &td1,tasks) {
        BOOST_FOREACH(Job &j1,td1.jobs) {
          BOOST_FOREACH(Task &td2,tasks) {
            BOOST_FOREACH(Job &j2,td2.jobs) {
              if(j1 << j2) 
                res.push_back(mk<BVULT>(j1.endRnd,j2.startRnd));
              //XXX: this next "else if" is not needed any more
              else if(j1 ^ j2) {
                //XXX: the following constraint is a bug
                //res.push_back(mk<BVULE>(j1.startRnd,j2.startRnd));

                //XXX: jobs are well nested. this is an optimization
                Expr e1 = mk<BVULE>(j1.startRnd,j2.startRnd);
                Expr e2 = mk<BVULT>(j2.startRnd,j1.endRnd);
                Expr e3 = mk<BVULT>(j2.endRnd,j1.endRnd);
                res.push_back(mk<IMPL>(mk<AND>(e1,e2),e3));
              }
            }
          }
        }
      }

      schedCons = mknary<AND> (trueE, res.begin (), res.end ());
      if(RekscDebug) errs() << "sched cons\n" << *schedCons << '\n';
    }

    ///given a lamport clock variable, return the round component of
    ///it, i.e., the topmost RNDBITS bits
    Expr clockRound(const Expr &clk)
    {
      unsigned high = CLKBITS - 1;
      unsigned low = high - RNDBITS + 1;
      return mk<BVEXTRACT>(clk,mkTerm<unsigned>(high,efac),mkTerm<unsigned>(low,efac));
    }

    ///given a lamport clock variable, return the priority component
    ///of it, i.e., the PRIOBITS after the top CLKBITS
    Expr clockPrio(const Expr &clk)
    {
      unsigned high = CLKBITS - RNDBITS - 1;
      unsigned low = high - PRIOBITS + 1;
      return mk<BVEXTRACT>(clk,mkTerm<unsigned>(high,efac),mkTerm<unsigned>(low,efac));
    }

    ///given a lamport clock variable, return the local clock of it,
    ///i.e., the lowest LCLKBITS
    Expr clockLocal(const Expr &clk)
    {
      unsigned high = LCLKBITS - 1;
      unsigned low = 0;
      return mk<BVEXTRACT>(clk,mkTerm<unsigned>(high,efac),mkTerm<unsigned>(low,efac));
    }

    ///given a read or write, return its lamport clock
    Expr lampClock(const CallInst *ci)
    {
      //try each possible event. return first one that matches.
      std::map<const CallInst*,Read>::const_iterator it1 = reads.find(ci);
      if(it1 != reads.end()) return it1->second.lampClk;
      std::map<const CallInst*,Write>::const_iterator it2 = writes.find(ci);
      if(it2 != writes.end()) return it2->second.lampClk;
      std::map<const CallInst*,Snapshot>::const_iterator it3 = snapshots.find(ci);
      if(it3 != snapshots.end()) return it3->second.lampClk;
      std::map<const CallInst*,PriOp>::const_iterator it4 = priOps.find(ci);
      return it4->second.lampClk;
    }

    ///generate constraints on lamport clocks
    void computeClockConstraints()
    {
      //list of conjuncts of the constraints
      std::list<Expr> res;

      //within each job, round components of lamport clocks must be
      //topologically sorted, but don't have to increase
      //strictly. however, the local clock components have to increase
      //strictly. in fact we fix the local clocks to be 1, 2,
      //.. etc. NOTE that we start numbering local clocks from 1
      //because we want legal lamport clocks to be positive.
      BOOST_FOREACH(Task &td,tasks) {
        BOOST_FOREACH(Job &j,td.jobs) {
          unsigned locClk = 1;
          const CallInst *prevci = 0;

          BOOST_FOREACH(const CallInst *ci,j.accesses) {
            Expr lampClk = lampClock(ci);
            Expr rnd = clockRound(lampClk);

            //round must not be inside a higher priority job
            BOOST_FOREACH(Task &td2,tasks) {
              BOOST_FOREACH(Job &j2,td2.jobs) {
                if(j ^ j2) {
                  Expr e1 = mk<BVULE>(rnd,j2.startRnd);
                  Expr e2 = mk<BVULT>(j2.endRnd,rnd);
                  res.push_back(mk<OR>(e1,e2));
                }
              }
            }

            //rounds must respect start and end rounds of job
            res.push_back(mk<BVULE>(j.startRnd,rnd));
            res.push_back(mk<BVULE>(rnd,j.endRnd));

            //XXX: round of job start/end event must equal the job
            //start/end round. this is an optimization
            if(ci == j.jobStart) res.push_back(mk<EQ>(rnd,j.startRnd));
            else if(ci == j.jobEnd) res.push_back(mk<EQ>(rnd,j.endRnd));

            //round must not be less than the round of the previous access
            if(prevci) {
              Expr prevLampClk = lampClock(prevci);
              Expr prevRnd = clockRound(prevLampClk);
              res.push_back(mk<BVULE>(prevRnd,rnd));
            } else 
              assert(ci == j.jobStart && "ERROR: first event must be a the job start!");
            
            //update previous access
            prevci = ci;

            //priority of clock equals job priority
            Expr prio = clockPrio(lampClk);
            res.push_back(mk<EQ>(prio,numExpr(j.prio(),PRIOBITS)));
            //local clock of clock equals current local clock
            Expr loc = clockLocal(lampClk);
            res.push_back(mk<EQ>(loc,numExpr(locClk,LCLKBITS)));

            ++locClk;
          }
        }
      }      

      clockCons = mknary<AND> (trueE, res.begin (), res.end ());
      if(RekscDebug) errs() << "clock cons\n" << *clockCons << '\n';
    }

    ///return an expression to indicate if the argument instruction
    ///has been executed. this is essentially the "control" variable
    ///for the basic block (created during VCGEN) to which the
    ///instruction belongs.
    Expr executed(const CallInst *ci)
    {
      return eqOne(env.lookup(*(ci->getParent())));
    }

    ///the generic canObserve() for both reads and snapshots
    std::list<Event> canObserve(const Observer &read)
    {
      Job &rjob = *(read.job);
      GlobalVariable *gv = dyn_cast<GlobalVariable>(read.event->getArgOperand(0));
      std::list<Event> res;

      //if no snapshotting
      if(RekscSnap == "NONE") {
        BOOST_FOREACH(const CallInst *wi,glob2Writes[gv]) {
          Write &write = writes[wi];
          Job &wjob = *(write.job);
          if(!(rjob << wjob)) res.push_back(write);
        }
        return res;
      }

      //some snapshotting
      if(RekscSnap == "WRITE" || RekscSnap == "ALL") {
        //find all jobs that are less than rjob and snapshot gv
        std::list<Job*> lessSnap;
        BOOST_FOREACH(Job *j,sortedJobs) {
          if(*j << rjob && j->snapshots.count(gv))
            lessSnap.push_back(j);
        }

        //add snapshots of maximal prior jobs
        BOOST_FOREACH(Job *j,maximal(lessSnap))
          res.push_back(snapshots[j->snapshots[gv]]);

        //add writes of jobs that can be preempted by reader
        BOOST_FOREACH(Job *j,preempted(rjob)) {
          if(j->writes.count(gv)) {
            BOOST_FOREACH(const CallInst *ci,j->writes[gv])
              res.push_back(writes[ci]);
          }
        }

        //add snapshots of jobs that can preempt the reader
        BOOST_FOREACH(Job *j,preempting(rjob)) {
          if(j->snapshots.count(gv))
            res.push_back(snapshots[j->snapshots[gv]]);
        }

        //add writes of the reader job
        if(rjob.writes.count(gv)) {
          BOOST_FOREACH(const CallInst *ci,rjob.writes[gv])
            res.push_back(writes[ci]);
        }

        return res;
      }

      assert(0 && "ERROR: unknown snapshot type!");
    }

    ///given a read, return a set of writes that it can observe
    std::list<Event> canObserve(const Read &read)
    {
      return canObserve(static_cast<const Observer&>(read));
    }

    ///given a snapshot, return a set of writes that it can observe
    std::list<Event> canObserve(const Snapshot &read)
    {      
      return canObserve(static_cast<const Observer&>(read));
    }

    ///given a priority operation, return the set of priority
    ///operations that it can observe. these are priority operations
    ///of its own job or other interfering jobs
    std::list<PriOp> canObserve(const PriOp &read)
    {
      Job &rjob = *(read.job);
      std::list<PriOp> res;

      //no snapshotting
      if(RekscSnap == "NONE") {
        typedef std::map<const CallInst*,PriOp> PriOpMap;
        BOOST_FOREACH(const PriOpMap::value_type &y,priOps) {
          const PriOp &write = y.second;
          Job &wjob = *(write.job);
          if(&read != &write && !(rjob << wjob)) res.push_back(write);
        }
        return res;
      }

      //some snapshotting
      if(RekscSnap == "WRITE" || RekscSnap == "ALL") { 
        //find all jobs that are less than rjob
        std::list<Job*> lessSnap;
        BOOST_FOREACH(Job *j,sortedJobs)
          if(*j << rjob) lessSnap.push_back(j);

        //add ends of maximal prior jobs
        BOOST_FOREACH(Job *j,maximal(lessSnap))
          res.push_back(priOps[j->jobEnd]);

        //add priops of interfering jobs
        BOOST_FOREACH(Job *j,interfering(rjob)) {
          BOOST_FOREACH(const CallInst *ci,j->priOps)
            res.push_back(priOps[ci]);
        }

        //add writes of the reader job
        BOOST_FOREACH(const CallInst *ci,rjob.priOps)
          if(ci != read.event) res.push_back(priOps[ci]);

        return res;
      }

      assert(0 && "ERROR: unknown snapshot type!");
    }

    ///given a read in FINAL, return a set of writes that it can
    ///observe
    std::list<Event> canObserve(const CallInst *ci)
    {      
      std::list<Event> res;
      GlobalVariable *gv = dyn_cast<GlobalVariable>(ci->getArgOperand(0));

      //if no snapshotting
      if(RekscSnap == "NONE") {
        BOOST_FOREACH(const CallInst *wi,glob2Writes[gv]) res.push_back(writes[wi]);
        return res;
      }

      //some snapshotting.
      if(RekscSnap == "WRITE" || RekscSnap == "ALL") { 
        //find all jobs that are less than rjob and snapshot gv
        std::list<Job*> lessSnap;
        BOOST_FOREACH(Job *j,sortedJobs)
          if(j->snapshots.count(gv)) lessSnap.push_back(j);

        //add snapshots of maximal prior jobs
        BOOST_FOREACH(Job *j,maximal(lessSnap))
          res.push_back(snapshots[j->snapshots[gv]]);

        return res;        
      }

      assert(0 && "ERROR: unknown snapshot type!");
    }

    ///the defaultObserved() for a base Observer
    Expr defaultObserved(const Observer &read)
    {
      GlobalVariable *gv = dyn_cast<GlobalVariable>(read.event->getArgOperand(0));
      //if no snapshotting
      if(RekscSnap == "NONE") return glob2Inits[gv];

      //some snapshotting. 
      if(RekscSnap == "WRITE" || RekscSnap == "ALL")
        return glob2Inits[gv];

      assert(0 && "ERROR: unknown snapshot type!");
    }

    ///the default value observed by a read event, i.e., if it does
    ///not observe another write
    Expr defaultObserved(const Read &read)
    {
      return defaultObserved(static_cast<const Observer&>(read));
    }

    ///the default value observed by a snapshot event, i.e., if it
    ///does not observe another write
    Expr defaultObserved(const Snapshot &read)
    {
      return defaultObserved(static_cast<const Observer&>(read));
    }

    ///the default value observed by a priority operation, i.e., if it
    ///does not observe another priority operation
    Expr defaultObserved(const PriOp &read)
    {
      //no snapshotting
      if(RekscSnap == "NONE") return numExpr(0,PCPLEVELS);

      //some snapshotting. 
      if(RekscSnap == "WRITE" || RekscSnap == "ALL") return numExpr(0,PCPLEVELS);

      assert(0 && "ERROR: unknown snapshot type!");
    }

    ///the default value observed by a read in FINAL, i.e., if it does
    ///not observe another write
    Expr defaultObserved(CallInst *ci)
    {
      GlobalVariable *gv = dyn_cast<GlobalVariable>(ci->getArgOperand(0));

      //if no snapshotting
      if(RekscSnap == "NONE") return glob2Inits[gv];

      //some snapshotting. 
      if(RekscSnap == "WRITE" || RekscSnap == "ALL") return glob2Inits[gv];

      assert(0 && "ERROR: unknown snapshot type!");
    }

    ///generate constraints about observed clocks -- in essence they
    ///say that a read must observe the most recent write
    void computeObsConstraints()
    {
      //list of conjuncts of the constraints
      std::list<Expr> res;

      //for every read
      typedef std::map<const CallInst*,Read> ReadMap;
      BOOST_FOREACH(const ReadMap::value_type &x,reads) {
        const Read &read = x.second;
        Job &rjob = *(read.job);
        Expr rclk = read.lampClk;
        Expr robs = read.obsClk;
        GlobalVariable *gv = dyn_cast<GlobalVariable>(read.event->getArgOperand(0));

        //collect constraints that equal read value to value wriiten
        //by the observed write. initialize it with the case that no
        //write is observed, i.e., with the initial value of the
        //global variable read. note that in this case, the observed
        //clock is assumed to be 0, since this is only possible if no
        //write is observed.
        if(RekscDebug)
          errs() << "++++++ read = " << *read.event 
                 << "   lookup1 = " << *(env.lookup(*read.event)) 
                 << "   lookup2 = " << (env.lookup(*read.event)) 
                 << "   z3 = " << bitvc::toZ3Str(env.lookup(*read.event)) << '\n';

        std::list<Expr> readCons;
        readCons.push_back(mk<AND>(mk<EQ>(numExpr(0,CLKBITS),robs),
                                   mk<EQ>(env.lookup(*read.event),defaultObserved(read))));

        if(RekscDebug)
          errs() << "+++ read cons\n" << bitvc::toZ3Str(readCons.back()) << '\n';

        //ensure that the max clock is as big as every write to the
        //same global variable that has occurred
        BOOST_FOREACH(const Event &write,canObserve(read)) {
          //dump may-observe relation between read and write
          errs() << "<<OBS>> <<" << RekscSnap << ">> " << read.event << " " << write.event << '\n';

          Expr exec = executed(write.event);
          Expr wclk = write.lampClk;
          Expr pre = mk<AND>(exec,mk<BVULT>(wclk,rclk));
          Expr post = mk<BVULE>(wclk,robs);          
          res.push_back(mk<IMPL>(pre,post));

          //add constraint equating value 
          readCons.push_back(mk<AND>(mk<AND>(pre,mk<EQ>(wclk,robs)),
                                     mk<EQ>(env.lookup(*read.event),env.lookup(*write.event))));
        }

        //add disjunction of readCons
        res.push_back(mknary<OR>(falseE,readCons));
      }

      //for every snapshot
      typedef std::map<const CallInst*,Snapshot> SnapMap;
      BOOST_FOREACH(const SnapMap::value_type &x,snapshots) {
        const Snapshot &read = x.second;
        Job &rjob = *(read.job);
        Expr rclk = read.lampClk;
        Expr robs = read.obsClk;
        GlobalVariable *gv = dyn_cast<GlobalVariable>(read.event->getArgOperand(0));

        //collect constraints that equal snapshot value to value
        //written by the observed write.
        std::list<Expr> readCons;
        readCons.push_back(mk<AND>(mk<EQ>(numExpr(0,CLKBITS),robs),
                                   mk<EQ>(env.lookup(*read.event),defaultObserved(read))));

        BOOST_FOREACH(const Event &write,canObserve(read)) {
          //dump may-observe relation between read and write
          errs() << "<<OBS>> <<" << RekscSnap << ">> " << read.event << " " << write.event << '\n';

          Expr exec = executed(write.event);
          Expr wclk = write.lampClk;
          Expr pre = mk<AND>(exec,mk<BVULT>(wclk,rclk));
          Expr post = mk<BVULE>(wclk,robs);          
          res.push_back(mk<IMPL>(pre,post));

          //add constraint equating value 
          readCons.push_back(mk<AND>(mk<AND>(pre,mk<EQ>(wclk,robs)),
                                     mk<EQ>(env.lookup(*read.event),env.lookup(*write.event))));
        }

        //add disjunction of readCons
        res.push_back(mknary<OR>(falseE,readCons));
      }

      //for every priority operation
      typedef std::map<const CallInst*,PriOp> PriOpMap;
      BOOST_FOREACH(const PriOpMap::value_type &x,priOps) {
        const PriOp &read = x.second;
        Job &rjob = *(read.job);
        Expr rclk = read.lampClk;
        Expr robs = read.obsClk;

        //collect constraints that equate op value to value wriiten by
        //the observed op.
        std::list<Expr> readCons;
        readCons.push_back(mk<AND>(mk<EQ>(numExpr(0,CLKBITS),robs),
                                   mk<EQ>(env.lookup(*read.event),defaultObserved(read))));

        //ensure that the max clock is as big as every other priority
        //operation that has occured earlier
        BOOST_FOREACH(const PriOp &write,canObserve(read)) {
          //dump may-observe relation between read and write
          errs() << "<<OBS>> <<" << RekscSnap << ">> " << read.event << " " << write.event << '\n';

          Expr exec = executed(write.event);
          Expr wclk = write.lampClk;
          Expr pre = mk<AND>(exec,mk<BVULT>(wclk,rclk));
          Expr post = mk<BVULE>(wclk,robs);          
          res.push_back(mk<IMPL>(pre,post));

          //add constraint depending on the type of the operation that
          //relates the value of this operation with that of the
          //observed operation
          std::list<Expr> readConj;
          readConj.push_back(pre);
          readConj.push_back(mk<EQ>(wclk,robs));
          
          //the lock priority, default to CPU lock
          unsigned lockPrio = PCPLEVELS - 1;

          switch(read.type) {
          case JOB_START:
            //acquired PCP locks carry over
            readConj.push_back(mk<EQ>(env.lookup(*read.event),env.lookup(*write.event)));
            {
              //all locks with priority greater than or equal to the
              //current job's base priority must be free.
              unsigned lockMask = 0;
              for(unsigned p = rjob.prio();p < PCPLEVELS;++p)
                lockMask |= 0x1 << p;

              readConj.push_back(mk<EQ>(numExpr(0,PCPLEVELS),
                                        (mk<BVAND>(numExpr(lockMask,PCPLEVELS),
                                                   env.lookup(*read.event)))));              
            }
            break;
          case JOB_END:
            //acquired PCP locks carry over
            readConj.push_back(mk<EQ>(env.lookup(*read.event),env.lookup(*write.event)));
            break;
          case PCP_LOCK:
            //get the lock priority and acquire it
            lockPrio = dyn_cast<ConstantInt>(read.event->getArgOperand(0))->getZExtValue();
            readConj.push_back(mk<EQ>(env.lookup(*read.event),
                                      mk<BVOR>(env.lookup(*write.event),
                                               numExpr(0x1 << lockPrio,PCPLEVELS))));
            break;
          case CPU_LOCK: 
            //acquire the lock -- note this is highest priority
            readConj.push_back(mk<EQ>(env.lookup(*read.event),
                                      mk<BVOR>(env.lookup(*write.event),
                                               numExpr(0x1 << lockPrio,PCPLEVELS))));
            break;
          case PCP_UNLOCK:
            //get the lock priority and release it
            lockPrio = dyn_cast<ConstantInt>(read.event->getArgOperand(0))->getZExtValue();
            readConj.push_back(mk<EQ>(env.lookup(*read.event),
                                      mk<BVAND>(env.lookup(*write.event),
                                                numExpr(~(0x1 << lockPrio),PCPLEVELS))));
            break;
          case CPU_UNLOCK:
            //release the lock -- note this is highest priority
            readConj.push_back(mk<EQ>(env.lookup(*read.event),
                                      mk<BVAND>(env.lookup(*write.event),
                                                numExpr(~(0x1 << lockPrio),PCPLEVELS))));
            break;
          default:
            assert(0);
          }

          readCons.push_back(mknary<AND>(trueE,readConj));
        }

        //add disjunction of readCons
        res.push_back(mknary<OR>(falseE,readCons));
      }

      if(RekscDebug) {
        BOOST_FOREACH(Expr e,res)
          errs() << "+++ obs cons : " << bitvc::toZ3Str(e) << '\n';
      }

      obsCons = mknary<AND> (trueE, res.begin (), res.end ());
      if(RekscDebug) errs() << "obs cons\n" << *obsCons << '\n';
      //errs() << "obs cons\n" << bitvc::toZ3Str(obsCons) << '\n';
    }

    ///compute initial values of globals
    void computeGlobInits()
    {
      //collect all shared globals
      typedef std::map< GlobalVariable*,std::list<const CallInst*> > GMap;
      std::set<GlobalVariable*> allGlobs;
      BOOST_FOREACH(GMap::value_type &vt,glob2Reads) allGlobs.insert(vt.first);
      BOOST_FOREACH(GMap::value_type &vt,glob2Writes) allGlobs.insert(vt.first);

      //store initial values of all shared globals. if no initializer,
      //assume a non-deterministic value
      BOOST_FOREACH(GlobalVariable *gv,allGlobs) {
        if(gv->hasInitializer()) {
          if(ConstantInt *ci = dyn_cast<ConstantInt>(gv->getInitializer()))
            glob2Inits[gv] = numExpr(ci->getZExtValue(),bitvc::getBW(gv));
          else
            glob2Inits[gv] = env.lookup(*(gv->getInitializer()));
        }
        else glob2Inits[gv] = env.lookup(*gv);

        if(RekscDebug) 
          errs() << "++++ INIT : " << *gv << " === " << *glob2Inits[gv] << '\n';
      }

      //process statements of the INIT function to update initial
      //values of global variables to user specified values
      std::map<Function*,RekSCVCGen>::iterator vcgit = vcgens.find(initFunc);
      if(vcgit == vcgens.end()) return;

      //we assume that the init function only writes values to shared
      //variables, and that it has linear control flow. so we traverse
      //the instructions in topogical order and collect values written
      //to shared globals.
      
      Topo &topo = getAnalysis<Topo>(*initFunc);
      BOOST_FOREACH(const BasicBlock *bb,topo.topo[initFunc]) {
        for(BasicBlock::const_iterator bit = bb->begin(),be = bb->end();bit != be;++bit) {
          if(const CallInst *ci = dyn_cast<const CallInst>(bit)) {
            Function *Callee = ci->getCalledFunction ();
            if(Callee && Callee->getName ().startswith (LLWRITE))
              if(GlobalVariable *gv = dyn_cast<GlobalVariable>(ci->getArgOperand(0))) {
                glob2Inits[gv] = vcgit->second.lookup(ci->getArgOperand(1));
                if(RekscDebug) errs() << "++++++ init [" << *gv << "] = " << *(glob2Inits[gv]) << '\n';
              }            
          }
        }
      }
    }

    ///compute various constraints
    void computeConstraints()
    {
      //compute initial values of globals
      computeGlobInits();

      //generate job scheduling constraints
      computeSchedConstraints();

      //generate constraints on lamport clocks
      computeClockConstraints();

      //generate constraints on observed clocks
      computeObsConstraints();
    }

    ///solve the constraints
    void solve()
    {
      errs() << "starting verification ...\n";

      typedef std::map<Function*,RekSCVCGen> VCGMap;

      //collect asserts from tasks
      std::list<Expr> asserts;
      BOOST_FOREACH(Task &td,tasks) {
        const ExprVector &ev = vcgens.find(td.taskFunc)->second.getAsserts();
        asserts.insert(asserts.end(),ev.begin(),ev.end());
      }
      
      //collect asserts from FINAL
      if(vcgens.count(finalFunc)) {
        const ExprVector &ev = vcgens.find(finalFunc)->second.getAsserts();
        asserts.insert(asserts.end(),ev.begin(),ev.end());
      }

      if(asserts.empty()) {
        errs() << "ERROR: No Assertions to Verify!!\n";
        return;
      }

      //construct the problem
      std::list<Expr> probCons;

      //-- for INIT, assert its VC and that it ends
      VCGMap::iterator initIt = vcgens.find(initFunc);
      if(initIt != vcgens.end()) {
        probCons.push_back(initIt->second.getCond());
        probCons.push_back(eqOne(env.lookup(*initExit)));
      }

      //-- for FINAL, assert its VC and that it ends
      VCGMap::iterator finalIt = vcgens.find(finalFunc);
      if(finalIt != vcgens.end()) {
        probCons.push_back(finalIt->second.getCond());
        probCons.push_back(eqOne(env.lookup(*finalExit)));
      }

      //-- for each task, assert its VC and that it ends
      BOOST_FOREACH(Task &td,tasks) {
        VCGMap::iterator it = vcgens.find(td.taskFunc);
        probCons.push_back(it->second.getCond());
        const Job &lastJob = td.jobs.back();
        probCons.push_back(executed(lastJob.jobEnd));
      }

      //-- assert other constraints
      probCons.push_back(schedCons);
      probCons.push_back(clockCons);
      probCons.push_back(obsCons);
      probCons.push_back(finalCons);

      Z3_config cfg = Z3_mk_config();
      if(RekscModel) Z3_set_param_value(cfg, "model", "true");
      Z3_context ctx = Z3_mk_context(cfg);
      Z3_solver slvr = Z3_mk_solver_for_logic(ctx,Z3_mk_string_symbol(ctx,"QF_BV"));
      bitvc::Z3Conv z3conv(ctx);

      Z3_ast problem = z3conv.marshal(mknary<AND>(trueE,probCons));
      Z3_ast spec = z3conv.marshal(mknary<AND>(trueE,asserts));

      //if using STP or Boolector
      if(RekscStp || RekscBoolector) {
        errs() << "exporting to SMT2 ...\n";
        
        //create temporary file
        std::string errInfo;
        char tmpFileName[32];
        strncpy(tmpFileName,"/tmp/mmc.XXXXXX",32);
        raw_fd_ostream outf(mkstemp(tmpFileName),true);
        
        //create smt string and dump to file
        Z3_set_ast_print_mode(ctx,Z3_PRINT_SMTLIB2_COMPLIANT);
        Z3_string bnch = Z3_benchmark_to_smtlib_string(ctx,"foo","QF_BV","unknown","",1,&problem,Z3_mk_not(ctx,spec));
        outf << bnch;
        outf.close();
        
        //solve using STP
        if(RekscStp) {
          //if generating CNF and stopping
          if(RekscCnf) {
            std::string cmd = "stp --simplifying-minisat " 
              + (RekscModel ? std::string("-p ") : std::string("")) 
              + "--output-CNF --exit-after-CNF "
              + "--SMTLIB2 " + std::string(tmpFileName);
            errs() << cmd << '\n';
            int stpStatus = ::system(cmd.c_str());
          } else {
            std::string cmd = "stp --simplifying-minisat " 
              + (RekscModel ? std::string("-p ") : std::string("")) 
              + "--SMTLIB2 " + std::string(tmpFileName);
            errs() << cmd << '\n';
            int stpStatus = ::system(cmd.c_str());
            errs() << "status = " << stpStatus << '\n';
            errs() << "VERIFICATION " << 
              (stpStatus == 0 ? "SUCCESSFUL" : 
               (stpStatus == 256 ? "FAILED" : "UNKNOWN")) << '\n';
          }
        }
        //solve using Boolector
        else {
          std::string cmd = "boolector --smt2 " + std::string(tmpFileName);
          errs() << cmd << '\n';
          int btrStatus = ::system(cmd.c_str());
          errs() << "status = " << btrStatus << '\n';
          errs() << "VERIFICATION " << 
            (btrStatus == 5120 ? "SUCCESSFUL" : 
             (btrStatus == 2560 ? "FAILED" : "UNKNOWN")) << '\n';
        }
      }
      //else if using Z3
      else {
        //-- assert problem and spec
        Z3_solver_assert(ctx,slvr,problem);
        Z3_solver_assert(ctx,slvr,Z3_mk_not(ctx,spec));

        errs() << "done marshalling .. now solving ...\n";
        Z3_lbool res = Z3_solver_check(ctx,slvr);

        //-- print stats
        //Z3_stats stats = Z3_solver_get_statistics(ctx,slvr);
        //Z3_stats_inc_ref(ctx,stats);
        //errs() << Z3_stats_to_string(ctx,stats) << '\n';
        //Z3_stats_dec_ref(ctx,stats);

        //-- print model
        if(RekscModel && res == Z3_L_TRUE)
          printf("model:\n%s\n", 
                 Z3_model_to_string(ctx, Z3_solver_get_model(ctx,slvr)));

        //-- print result
        errs() << "VERIFICATION " << 
          (res == Z3_L_FALSE ? "SUCCESSFUL" : 
           (res == Z3_L_TRUE ? "FAILED" : "UNKNOWN")) << '\n';
      }

      //-- cleanup
      Z3_del_context(ctx);
      Z3_del_config(cfg);
    }

    ///top-level method called on the module
    virtual bool runOnModule(Module &M) 
    { 
      TargetData theTD(&M);
      bitvc::TD = &theTD;
      
      //-- create the periodic program
      createPP(M);
      //-- create functions for each task
      BOOST_FOREACH(Task &td,tasks) createTaskFunc(td,M);
      //-- sort jobs according to VMCAI'13 order
      sortJobs();
      //-- inline bodies of task functions
      BOOST_FOREACH(Task &td,tasks) inlineFunc(td.taskFunc);
      //-- collect useful info about tasks
      BOOST_FOREACH(Task &td,tasks) collectTaskInfo(td);
      //-- process the init function
      processInit(M);
      //-- compute VCs for each task
      BOOST_FOREACH(const Task &td,tasks) computeVC(td.taskFunc);
      //-- compute constraints
      computeConstraints();
      //-- process the final function
      processFinal(M);
      //-- solve the constraints to get the result
      solve();

      return true;
    }

    virtual void getAnalysisUsage(AnalysisUsage &AU) const
    {
      AU.addRequired<Topo> ();
      AU.addRequired<NameValues> ();
      AU.setPreservesAll ();
    }
  };

  char RekSC::ID = 0;
}

INITIALIZE_PASS_BEGIN(RekSC, "reksc", "RekSC at logical level", false, false)
INITIALIZE_PASS_DEPENDENCY(Topo)
INITIALIZE_PASS_DEPENDENCY(NameValues)
INITIALIZE_PASS_END(RekSC, "reksc", "RekSC at logical level", false, false)
