/* ================================== */
/* LLREK LICENSE */
/* ================================== */
/* Copyright (c) 2013 Carnegie Mellon University. */
/* All Rights Reserved. */

/* Redistribution and use in source and binary forms, with or without */
/* modification, are permitted provided that the following conditions are */
/* met: */

/* 1. LLREK includes and/or makes use of certain third party software */
/* ("Third Party Software"). You agree to comply with any and all the */
/* Third Party Software terms and conditions listed below and/or */
/* contained in any separate license file distributed with LLREK. The */
/* parties who own the Third Party Software ("Third Party Licensors") are */
/* intended third party beneficiaries to this License with respect to the */
/* terms applicable to their Third Party Software. */

/* 2. Redistributions of source code must retain the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers. */

/* 3. Redistributions in binary form must reproduce the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers in the documentation and/or other materials provided with */
/* the distribution. */

/* 4. Products derived from this software may not include Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" in the name of such derived product, nor shall Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" be used to endorse or promote products derived from this */
/* software without prior written permission. For written permission, */
/* please contact permission@sei.cmu.edu. */

/* This material is based upon work funded and supported by the */
/* Department of Defense under Contract No. FA8721-05-C-0003 with */
/* Carnegie Mellon University for the operation of the Software */
/* Engineering Institute, a federally funded research and development */
/* center. */

/* Any opinions, findings and conclusions or recommendations expressed in */
/* this material are those of the author(s) and do not necessarily */
/* reflect the views of the United States Department of Defense. */

/* NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING */
/* INSTITUTE MATERIAL IS FURNISHEDON AN AS-IS BASIS. CARNEGIE */
/* MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR */
/* IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF */
/* FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS */
/* OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT */
/* MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, */
/* TRADEMARK, OR COPYRIGHT INFRINGEMENT. */

/* This material has been approved for public release and unlimited */
/* distribution. */

/* DM-0000665 */

#ifndef __BVC_REKSCVCGEN_HPP__
#define __BVC_REKSCVCGEN_HPP__

#include "./BvcUtil.hpp"
#include "./BVEnv.hpp"
#include "./Z3Conv.hpp"

namespace ufo
{
  class RekSCVCGen
  {
  public:
    typedef std::map< GlobalVariable*,std::list<const CallInst*> > GlobAccess;

  protected:
    ExprFactory& efac;
    BVEnv& env;
    Function &F;
    Topo &topo;
    Expr trueE;
    Expr falseE;
    
    /** true if pointers are tracked in VC. Controlled by
	--ufo-track-ptr command line option. */
    bool trackPtr;

    // -- true if compute() was called and condition is available
    bool computed;
    
    // -- the computed condition
    Expr cond;    

    // -- negated assertions
    ExprVector asserts;   

  public:
    RekSCVCGen (ExprFactory &f, BVEnv& e, Function &_F,Topo &_t,bool auto_update) :
      efac(f), env(e), F(_F), topo(_t), trueE (mk<TRUE> (efac)), falseE(mk<FALSE>(efac)),
      computed(false), trackPtr (ufocl::UFO_TRACK_PTR)
    { 
      cond = trueE;
      if (auto_update) updateEnvToFirstInst (env,F); 
    }

    RekSCVCGen (const RekSCVCGen &other) 
      : efac(other.efac),env(other.env),F(other.F),topo(other.topo),
        trueE(other.trueE),falseE(other.falseE),
        trackPtr(other.trackPtr),computed(other.computed),
        cond(other.cond),asserts(other.asserts)
    {}
    
    /**
     * Updated BVEnv from the begining of the basic block, to
     * the environment at the first non-PHI instruction of the block.
     * Idempotent.
     */
    static BVEnv &updateEnvToFirstInst (BVEnv &env,Function &_F) 
    {
      ExprFactory &efac = env.getExprFactory ();
      
      for(Function::const_iterator fit = _F.begin(),fe = _F.end();fit != fe;++fit) {
        for (BasicBlock::const_iterator it = fit->begin(), end = fit->end (); it != end; ++it) {
          // -- stop at first non-phi node
          if (!isa<PHINode> (*it)) break;
            
          // map phi to what phi' is pointing to
          const Value *phi = it;
          Expr u = env.lookup (variant::prime (mkTerm (phi, efac)));
          env.bind (*phi, u);
        }
      }
      return env;
    }
    
    Expr getPhiArg (const BasicBlock* bb, const PHINode& phi) const
    {
      return lookup (phi.getIncomingValueForBlock (bb));
    }

    Expr getPhiRes(const Instruction& i) { return env.lookup (i); }

    //-- compute the condition under which an edge is executed from sr
    //-- to ds.
    Expr edgeCondition (const BasicBlock* sr, const BasicBlock* ds)
    {
      ExprVector res;
      for (BasicBlock::const_iterator it = sr->begin (), end = sr->end (); it != end; ++it) {
        // -- skip PHI nodes
        if (isa<PHINode> (*it)) continue;

        //terminator instruction
        if (const TerminatorInst *ti = dyn_cast<const TerminatorInst> (it)) {
          const BranchInst* bi = dyn_cast<const BranchInst> (ti);

          // -- last instruction in the bb, nothing to do
          if (bi == NULL) break;

          // -- uncoditional branch, nothing to do
          if (!bi->isConditional ()) break;
	      
          // -- if the branching condition is a constant
          if (const ConstantInt* cint = dyn_cast<const ConstantInt>(bi->getCondition())) {            
            if ((cint->isZero () && bi->getSuccessor (1) != ds) ||
                (!cint->isZero () && bi->getSuccessor (0) != ds))
              return falseE;
            break;
          }

          // -- branch condition not a constant
          const Instruction* c = dyn_cast<Instruction>(bi->getCondition ());
          assert (c != NULL);
          res.push_back (bi->getSuccessor (0) == ds ? 
                         eqOne(env.lookup (*c)) : eqZero(env.lookup (*c)));

          // -- last instruction
          break;
        }

        // -- call instruction -- we handle assume()'s here
        if (const CallInst *ci = dyn_cast<const CallInst> (it)) {
          Function *Callee = ci->getCalledFunction ();
          if (Callee == 0) continue;
	      
          // -- skip if not the assume function
          if (!Callee->getName ().equals ("__VERIFIER_assume") &&
              !Callee->getName ().equals ("__CPROVER_assume")) continue;

          // -- the condition
          Value *c = ci->getArgOperand (0);
	      
          // -- if the assume() condition is a constant
          if (isa<ConstantInt> (c)) {
            const ConstantInt* cint = dyn_cast<const ConstantInt>(c);
            if (cint->isZero ()) return falseE;
            continue;		  
          }

          res.push_back (mk<NEQ>(env.lookup (*c),
                                 bvop::bvConst(mkTerm<uint64_t>(0,efac),
                                               bitvc::getBW(c))));
        }
      }

      return mknary<AND> (trueE, res.begin (), res.end ());
    }

    //-- compute control flow constraints
    Expr doControl (const BasicBlock* bb)
    {
      // -- refresh variable for this block
      Expr bbVar = env.lookup (*bb);

      // skip entry block and blocks with no predecessors
      if (bb == &(F.getEntryBlock()) || pred_begin(bb) == pred_end(bb))
	return trueE;
	
      ExprVector preCond,evars;

      // -- for all predecessors of bb
      for (const_pred_iterator PI = pred_begin (bb),E = pred_end(bb); PI != E; ++PI) {
        const BasicBlock *pred = *PI;
	  
        Expr predVar = env.lookup(*pred);
        Expr condc = eqOne(predVar);
	  
        // -- create an edge variable
        Expr edgeV = eqOne(bitvc::mkEdgVar (predVar, bbVar));
        condc = boolop::land (condc, edgeV);
        evars.push_back (edgeV);
	  
        condc = boolop::land (condc, (edgeCondition(pred, bb)));
	  
        // -- for every PHINode
        for (BasicBlock::const_iterator it = bb->begin(), 
               end = bb->end (); it != end; ++it) {
          if (!isa<PHINode>(*it)) break;
	      
          // -- only attempt to approximate integer registers and
          // -- pointers (which we treat as integer addresses,
          // -- i.e., never dereferenced
          if (!it->getType ()->isIntegerTy () && 
              (!trackPtr || !it->getType ()->isPointerTy ()))
            continue;

          const PHINode *phi = dyn_cast<const PHINode> (&*it);
          assert (phi != NULL);
		  
          Expr phiArg = getPhiArg(pred, *phi);
          Expr phiRes = getPhiRes (*phi);
          //add condition from phinode with respect to pred
          Expr phiE = mk<EQ>(phiRes, phiArg);
          condc = boolop::land (condc, phiE);
        }

        if (condc != trueE) preCond.push_back (condc);
      }
        
      Expr res = boolop::limp (eqOne(bbVar), 
			       mknary<OR> (trueE,
					   preCond.begin (), preCond.end ()));
      // -- if have predecessors, say that exactly one must be taken
      if (!evars.empty ())
      	res = boolop::land (res,
                            boolop::limp (eqOne(bbVar),
      					  exactlyOne (evars, efac)));
      return res;      
    }

    //-- lookup a value from the environment
    Expr lookup (const Value *v) const
    {
      assert (v != NULL);
      
      //constant integer
      if (isa<ConstantInt>(v)) {
        const ConstantInt *ci = cast<const ConstantInt>(v);          
        uint64_t cival = ci->getType()->getSignBit() ?
          ci->getSExtValue() : ci->getZExtValue();
        return bvop::bvConst(mkTerm(cival,efac),ci->getBitWidth());
      }
      
      //null pointer -- return zero bitvector of appropriate width
      if (isa<ConstantPointerNull>(v)) {
        const ConstantPointerNull *cpn = cast<const ConstantPointerNull>(v);
        return bvop::bvConst(mkTerm<uint64_t>(0,efac), bitvc::getBW(cpn));
      }

      // -- any other constant, is represented by symbolic value
      if (isa<Constant> (v)) return mkTerm (v, efac);
      
      return env.lookup (*v);
    }

    //-- generate VC for a logical operation
    Expr doLogic(const Expr &lhs, const Instruction &i)
    {
      Value* v1 = i.getOperand(0);
      Value* v2 = i.getOperand(1);
      
      const IntegerType* t1 = dyn_cast<const IntegerType>(v1->getType());
      const IntegerType* t2 = dyn_cast<const IntegerType>(v2->getType());

      // -- operation not over Integer type -- shouldn't happen
      if (t1 == NULL || t2 == NULL) return trueE;

      //if (t1->getBitWidth() != 1 || t2->getBitWidth() != 1)
      // -- cannot handle such operators on integers
      // XXX add workaround for shifts with constant values
      //return trueE;

      Expr op1;
      Expr op2;
      
      ConstantInt* c1;
      ConstantInt* c2;

      // -- if operands are constants, get their values
      // -- otherwise, resolve in the environment
      if ((c1 = dyn_cast<ConstantInt>(v1)) != NULL) {
        uint64_t c1val = c1->getType()->getSignBit() ?
          c1->getSExtValue() : c1->getZExtValue();
	op1 = bvop::bvConst(mkTerm(c1val,efac),c1->getBitWidth());
      } else
	op1 = env.lookup (*v1);
      
      if(RekscDebug)
        errs() << "++++++ v1 = " << *v1 
               << "   lookup1 = " << *op1 
               << "   lookup2 = " << op1
               << "   BW = " << bitvc::getBW(op1)
               << "   z3 = " << bitvc::toZ3Str(op1) << '\n';

      if ((c2 = dyn_cast<ConstantInt>(v2)) != NULL) {
        uint64_t c2val = c2->getType()->getSignBit() ?
          c2->getSExtValue() : c2->getZExtValue();
	op2 = bvop::bvConst(mkTerm(c2val,efac),c2->getBitWidth());
      } else
	op2 = env.lookup (*v2);

      switch(i.getOpcode())
	{
	case BinaryOperator::And:
	  return mk<EQ>(lhs, mk<BVAND>(op1,op2));
	case BinaryOperator::Or:
	  return mk<EQ>(lhs, mk<BVOR>(op1,op2));
	case BinaryOperator::Xor:
	  return mk<EQ>(lhs, mk<BVXOR>(op1,op2));	
        case BinaryOperator::Shl:
	  return mk<EQ>(lhs, mk<BVSHL>(op1,op2));
        case BinaryOperator::AShr:
	  return mk<EQ>(lhs, mk<BVASHR>(op1,op2));
        case BinaryOperator::LShr:  
	  return mk<EQ>(lhs, mk<BVLSHR>(op1,op2));

	default:
          errs() << "ERROR: unhandled opcode " << i.getOpcode() <<
            " in doLogic()\n";
          assert(0);
	  return trueE;
	}
    }
    
    //-- handle arithmetic operations
    Expr doArith(const Expr &lhs, const Instruction &i)
    {
      Value* v1 = i.getOperand(0);
      Value* v2 = i.getOperand(1);


      Expr op1 = lookup (v1);
      Expr op2 = lookup (v2);
      
      switch(i.getOpcode())
	{
	case BinaryOperator::Add:
	  return mk<EQ>(lhs ,mk<BVADD>(op1, op2));
	case BinaryOperator::Sub:
	  return mk<EQ>(lhs ,mk<BVSUB>(op1, op2));
	case BinaryOperator::Mul:
	  return mk<EQ>(lhs ,mk<BVMUL>(op1, op2));
	case BinaryOperator::UDiv:
	  return mk<EQ>(lhs ,mk<BVUDIV>(op1, op2));
	case BinaryOperator::SDiv:
	  return mk<EQ>(lhs ,mk<BVSDIV>(op1, op2));
	case BinaryOperator::URem:
	  return mk<EQ>(lhs ,mk<BVUREM>(op1, op2));
	case BinaryOperator::SRem:
	  return mk<EQ>(lhs ,mk<BVSREM>(op1, op2));
        }

      assert (0 && "ERROR: Unhandled case in doArith()");
    }

    //-- Over approximates a comparison instruction
    Expr doCompare (Expr lhs, const ICmpInst &icmp)
    {
      const Value* v1 = icmp.getOperand(0);
      const Value* v2 = icmp.getOperand(1);

      Expr op1 = lookup (v1);
      Expr op2 = lookup (v2);
        
      //equality between lhs and one/zero bit
      Expr lhsEqOne = eqOne(lhs);
      Expr lhsEqZero = eqZero(lhs);

      switch(icmp.getPredicate()) {
      case CmpInst::ICMP_EQ:
        return mk<IFF>(lhsEqOne, mk<EQ>(op1,op2));
      case CmpInst::ICMP_NE:
        return mk<IFF>(lhsEqZero, mk<EQ>(op1,op2));
      case CmpInst::ICMP_UGT:
        return mk<IFF>(lhsEqOne,mk<BVUGT>(op1,op2));
      case CmpInst::ICMP_SGT:
        return mk<IFF>(lhsEqOne,mk<BVSGT>(op1,op2));
      case CmpInst::ICMP_UGE:
        return mk<IFF>(lhsEqOne,mk<BVUGE>(op1,op2));
      case CmpInst::ICMP_SGE:
        return mk<IFF>(lhsEqOne,mk<BVSGE>(op1,op2));
      case CmpInst::ICMP_ULT:
        return mk<IFF>(lhsEqOne,mk<BVULT>(op1,op2));
      case CmpInst::ICMP_SLT:
        return mk<IFF>(lhsEqOne,mk<BVSLT>(op1,op2));
      case CmpInst::ICMP_ULE:
        return mk<IFF>(lhsEqOne,mk<BVULE>(op1,op2));
      case CmpInst::ICMP_SLE:
        return mk<IFF>(lhsEqOne,mk<BVSLE>(op1,op2));
      }

      assert (0 && "ERROR: Unhandled case in doCompare()");
    }

    ///process call to assert
    Expr doAssert(const CallInst *ci,const BasicBlock *bb)
    {
      Expr bbVar = eqOne(env.lookup(*bb));

      Value *c = ci->getArgOperand (0);
      // -- if the assert() condition is a constant
      if (isa<ConstantInt> (c)) {
        const ConstantInt* cint = dyn_cast<const ConstantInt>(c);
        asserts.push_back(mk<IMPL>(bbVar,cint->isZero () ? falseE : trueE));
        if(RekscDebug) {
          errs() << "assert VC: " << *(asserts.back()) << '\n';
          errs() << "assert Z3: " << bitvc::toZ3Str(asserts.back()) << '\n';
        }
        return trueE;
      }
      asserts.push_back
        (mk<IMPL>(bbVar,
                  mk<NEQ>(env.lookup (*c),bvop::bvConst(mkTerm<uint64_t>(0,efac),
                                                        bitvc::getBW(c)))));
      if(RekscDebug) {
        errs() << "assert VC: " << *(asserts.back()) << '\n';
        errs() << "assert Z3: " << bitvc::toZ3Str(asserts.back()) << '\n';
      }
      return trueE;
    }

    ///process call to read
    Expr doRead(const CallInst *ci,const BasicBlock *bb)
    {
      //sanity check
      assert(dyn_cast<GlobalVariable>(ci->getArgOperand(0)));
      //over-approximate the function with non-determinism
      return trueE;
    }

    ///process call to write
    Expr doWrite(const CallInst *ci,const BasicBlock *bb)
    {
      //sanity check
      assert(dyn_cast<GlobalVariable>(ci->getArgOperand(0)));
      //equate the value written with the second argument to
      //call. take care of differences in bit-widths.
      Value *rhs = ci->getArgOperand(1);
      unsigned lbw = bitvc::getBW(ci->getArgOperand(0));
      unsigned rbw = bitvc::getBW(rhs);

      if(lbw == rbw)
        return mk<EQ>(env.lookup(*ci),lookup(rhs));
      if(lbw < rbw)
        return mk<EQ>(env.lookup(*ci),mk<BVEXTRACT>(lookup(rhs),
                                                    mkTerm<unsigned>(lbw - 1, efac),
                                                    mkTerm<unsigned>(0, efac)));
      return mk<EQ>(lookup(rhs),mk<BVEXTRACT>(env.lookup(*ci),
                                              mkTerm<unsigned>(rbw - 1, efac),
                                              mkTerm<unsigned>(0, efac)));
    }

    // -- generate VC for a statement
    Expr doStatement(const Instruction* i, const BasicBlock* bb) 
    {
      //errs () << "\nDoing Statement " << *i << '\n';
      int opc = i->getOpcode();
       
      // -- instructions that do not produce a value
      if (opc == BranchInst::Br || opc == ReturnInst::Ret)
	return trueE;

      // -- call statement (collect assert and over-approximate by
      // -- nondet)
      if (opc == CallInst::Call) {
        const CallInst *ci = dyn_cast<const CallInst>(i);
        Function *Callee = ci->getCalledFunction ();
        if (Callee == 0) return trueE;

        //-- handle assert
        if (Callee->getName ().equals ("assert")) return doAssert(ci,bb);

        //-- handle read
        if (Callee->getName ().startswith (LLREAD)) return doRead(ci,bb);

        //-- handle write
        if (Callee->getName ().startswith (LLWRITE)) return doWrite(ci,bb);

        //-- by default over approximate with non-determinism
        return trueE;
      }

      // -- load instruction
      if(const LoadInst *li = dyn_cast<const LoadInst>(i)) {
        const Value *ptr = li->getPointerOperand();
        if(!isa<GlobalVariable>(ptr)) return trueE;
        return mk<EQ>(env.lookup(*i),env.lookup(*ptr));
      }

      // -- store instruction
      if(const StoreInst *si = dyn_cast<const StoreInst>(i)) {
        const Value *ptr = si->getPointerOperand();
        if(!isa<GlobalVariable>(ptr)) return trueE;
        return mk<EQ>(env.lookup(*ptr),env.lookup(*(si->getValueOperand())));
      }

      // -- extract value
      if(const ExtractValueInst *evi = dyn_cast<const ExtractValueInst>(i)) {
        unsigned extractSize = bitvc::TD->getTypeSizeInBits(evi->getType());
        const Value *ao = evi->getAggregateOperand();
        const Type *aot = ao->getType();
        ExtractValueInst::idx_iterator idxs = evi->idx_begin(); 
        //compute offset
        unsigned offset = 0;
        for(unsigned ii = 0,ei = evi->getNumIndices();ii < ei;++ii) {
          const Type *it = ExtractValueInst::getIndexedType(aot,ii);
          unsigned itbw = bitvc::TD->getTypeSizeInBits(it);
          offset += itbw * idxs[ii];
        }
        if(RekscDebug) errs() << "++++++ offset = " << offset << " size = " << extractSize << '\n';
        unsigned low = offset,high = offset + extractSize - 1;
        return mk<EQ>(env.lookup(*i), mk<BVEXTRACT>(env.lookup(*ao),
                                                    mkTerm<unsigned>(high,efac),
                                                    mkTerm<unsigned>(low,efac)));
      }

      // -- floating point compare (over-approximate by nondet)
      if (opc == FCmpInst::FCmp)
	return trueE;

      // -- if not tracking pointers, and i is not an integer type,
      // -- replace by non-determinism
      if (!trackPtr && !i->getType ()->isIntegerTy ()) return trueE;

      // -- create new non-deterministic value
      Expr lhs = env.lookup (*i);

      // -- handle ZExt
      if (opc == ZExtInst::ZExt)
        return mk<EQ>(lhs,
                      mk<BVZEXT>(lookup(i->getOperand(0)),
                                 mkTerm<unsigned>(bitvc::getBW(i),efac)));

      // -- handle SExt
      if (opc == SExtInst::SExt)
        return mk<EQ>(lhs,
                      mk<BVSEXT>(lookup(i->getOperand(0)),
                                 mkTerm<unsigned>(bitvc::getBW(i),efac)));

      // -- truncate instruction
      if (opc == TruncInst::Trunc)
        return mk<EQ>(lhs,mk<BVEXTRACT>(lookup(i->getOperand(0)),
                                        mkTerm<unsigned>(bitvc::getBW(lhs) - 1, efac),
                                        mkTerm<unsigned>(0, efac)));
      
      // -- treat all other casts as non-determinism
      if (isa<CastInst> (i)) {
        return trueE;
        //assert(0 && "ERROR: can't handle cast instruction");
	//return mk<EQ> (lhs, env.lookup (*(i->getOperand (0))));
      }

      //assert (opc != TruncInst::Trunc);

      // -- handle select statement
      if (opc == SelectInst::Select) {
        const SelectInst *si = cast<SelectInst>(i);
        
        const Value *c = si->getCondition();
        const Value *v1 = si->getTrueValue();
        const Value *v2 = si->getFalseValue();        

        //errs () << "SELECT: " << c << " " << v1 << " " << v2 << "\n" ;
        
        Expr cond;
        Expr op1;
        Expr op2;
        
        const ConstantInt* ci;
        if ((ci = dyn_cast<ConstantInt>(c)) != NULL)
          cond = ci->isOne () ? trueE : falseE;	
        else cond = eqOne(env.lookup(*c));

        op1 = lookup (v1);
        op2 = lookup (v2);
        
        return mk<EQ>(lhs, mk<ITE>(cond , op1, op2));
      }

      switch(opc) {	
      case BinaryOperator::And:
      case BinaryOperator::Or:
      case BinaryOperator::Xor:
      case BinaryOperator::Shl:
      case BinaryOperator::AShr:
      case BinaryOperator::LShr:
        return doLogic(lhs, *i);
        
      case ICmpInst::ICmp:
        return doCompare(lhs, cast<ICmpInst>(*i));

      case BinaryOperator::Add:
      case BinaryOperator::Sub:
      case BinaryOperator::Mul:
      case BinaryOperator::UDiv:
      case BinaryOperator::SDiv:
      case BinaryOperator::URem:
      case BinaryOperator::SRem:
        return doArith(lhs, *i);
	
      default:
        assert(0 && "ERROR: Unhandled case in doStatement()");
        // -- do nothing -- non-deterministic assignment
        //DEBUG(errs () << "default: " <<  *i << "\n");
        return trueE;
      }
    }
     
    //-- compute constraints for actions
    Expr doAction(const BasicBlock *bb)
    {
      Expr actions = trueE;
      
      //iterate over the instructions over control-flow order. this is
      //important to maintain topological ordrer.
      for (BasicBlock::const_iterator it = bb->begin(), 
             end = bb->end (); it != end; ++it) {
        // -- skip phi node
        if (isa<PHINode> (*it)) continue;
        
        if(RekscDebug) errs() << "statement: " << *it << '\n';
        Expr vc = doStatement(it, bb);
        if(RekscDebug) {
          errs() << "statement VC: " << *vc << '\n';
          errs() << "statement Z3: " << bitvc::toZ3Str(vc) << '\n';
        }
        actions = boolop::land(actions,vc);
      }

      return actions;
    }

    // -- returns the condition of the edge
    Expr getCond() const { assert (computed); return cond; }

    // -- returns the asserts of the edge
    const ExprVector &getAsserts() const { assert (computed); return asserts; }

    // -- compute the VC for an LBE edge
    void compute ()
    {
      //must not have called this function earlier
      assert (!computed);
      
      std::vector<Expr> actions;
      std::vector<Expr> control;
      
      // -- all blocks are dealt with similarly
      Expr r;
        
      // -- for every BB in the function in topological order
      const std::vector<const BasicBlock*> &bbs = topo.topo[&F];
      for(size_t i = 0,e = bbs.size();i < e;++i) {
        const BasicBlock *bb = bbs[i];
        if(RekscDebug) errs() << "computing VC for BB " << *bb << "\n";
        
        r = doControl (bb);
        if(RekscDebug) {
          errs() << "control VC: " << *r << '\n';
          errs() << "control Z3: " << bitvc::toZ3Str(r) << '\n';
        }
        if (r != trueE) control.push_back (r);

        r = doAction (bb);
        if (r != trueE) actions.push_back (r);
      }
      
      //emptiness check
      if (actions.empty ()) actions.push_back (trueE);
      if (control.empty ()) control.push_back (trueE);
      
      //combine control and actions constraints
      cond = boolop::land (boolop::land (control), boolop::land (actions));

      //update computed
      computed = true;
    }
  };

} //namespace ufo

#endif //__BVC_REKSCVCGEN_HPP__
