/* ================================== */
/* LLREK LICENSE */
/* ================================== */
/* Copyright (c) 2013 Carnegie Mellon University. */
/* All Rights Reserved. */

/* Redistribution and use in source and binary forms, with or without */
/* modification, are permitted provided that the following conditions are */
/* met: */

/* 1. LLREK includes and/or makes use of certain third party software */
/* ("Third Party Software"). You agree to comply with any and all the */
/* Third Party Software terms and conditions listed below and/or */
/* contained in any separate license file distributed with LLREK. The */
/* parties who own the Third Party Software ("Third Party Licensors") are */
/* intended third party beneficiaries to this License with respect to the */
/* terms applicable to their Third Party Software. */

/* 2. Redistributions of source code must retain the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers. */

/* 3. Redistributions in binary form must reproduce the above copyright */
/* notice, this list of conditions and the following acknowledgments and */
/* disclaimers in the documentation and/or other materials provided with */
/* the distribution. */

/* 4. Products derived from this software may not include Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" in the name of such derived product, nor shall Carnegie */
/* Mellon University, "SEI and/or Software Engineering */
/* Institute" be used to endorse or promote products derived from this */
/* software without prior written permission. For written permission, */
/* please contact permission@sei.cmu.edu. */

/* This material is based upon work funded and supported by the */
/* Department of Defense under Contract No. FA8721-05-C-0003 with */
/* Carnegie Mellon University for the operation of the Software */
/* Engineering Institute, a federally funded research and development */
/* center. */

/* Any opinions, findings and conclusions or recommendations expressed in */
/* this material are those of the author(s) and do not necessarily */
/* reflect the views of the United States Department of Defense. */

/* NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING */
/* INSTITUTE MATERIAL IS FURNISHEDON AN AS-IS BASIS. CARNEGIE */
/* MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR */
/* IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF */
/* FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS */
/* OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT */
/* MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, */
/* TRADEMARK, OR COPYRIGHT INFRINGEMENT. */

/* This material has been approved for public release and unlimited */
/* distribution. */

/* DM-0000665 */

#ifndef __BVC_Z3CONV_HPP__
#define __BVC_Z3CONV_HPP__

#include "./BvcUtil.hpp"

namespace ufo
{  
  namespace bitvc
  {
    //-- a convertor from Expr to Z3_ast
    struct Z3Conv
    {
      //-- the Z3 context to be used
      Z3_context ctx;

      //-- constructor
      Z3Conv(Z3_context c) : ctx(c) {}

      //-- convert decl to string
      std::string declToStr(const Expr &decl)
      {
        string buf; raw_string_ostream ostr(buf);

        //constant -- unsigned long 
        if(isOpX< Terminal<unsigned long> >(decl->left())) {
          unsigned long val = static_cast<unsigned long>(getTerm<unsigned long>(decl->left()));
          ostr << val; ostr.flush();
          return buf;
        }
 
        //constant -- unsigned long long 
        if(isOpX<LLUIT>(decl->left())) {
          llui_t val = static_cast<llui_t>(getTerm<llui_t>(decl->left()));
          ostr << val; ostr.flush();
          return buf;
        }
          
        //variable -- tuple (e.g., for edges)
        if(isOpX<TUPLE>(decl->left())) {
          ostr << "tup_" << decl->left(); ostr.flush();          
          return buf;
        }

        //variable -- terminal value
        if(isOpX< Terminal<Value*> >(decl->left())) {
          Value *val = getTerm<Value*>(decl->left());
          ostr << "vbl_" << val; ostr.flush();          
          return buf;
        }

        //variable -- terminal const value
        if(isOpX< Terminal<const Value*> >(decl->left())) {
          const Value *val = getTerm<const Value*>(decl->left());
          ostr << "vbl_" << val; ostr.flush();          
          return buf;
        }

        //variable -- terminal basic block
        if(isOpX< Terminal<BasicBlock*> >(decl->left())) {
          BasicBlock *val = getTerm<BasicBlock*>(decl->left());
          ostr << "bb_" << val; ostr.flush();          
          return buf;
        }

        //variable -- terminal const basic block
        if(isOpX< Terminal<const BasicBlock*> >(decl->left())) {
          const BasicBlock *bb = getTerm<const BasicBlock*>(decl->left());
          ostr << "bb_" << bb; ostr.flush();          
          return buf;
        }

        //variable -- terminal string
        if(isOpX< Terminal<std::string> >(decl->left())) {
          return getTerm<std::string>(decl->left());
        }

        errs() << "+++ decl = " << *(decl->left()) << '\n';
        errs() << "+++ operator = " << typeid(decl->left()->op()).name() << '\n';
        assert(0 && "ERROR: cannot convert decl to string!");
      }

      //-- convert
      Z3_ast marshal(const Expr &e)
      {
        //errs() << "converting " << *e << " to Z3\n";
        //errs() << "operator = " << typeid(e->op()).name() << '\n';

        //shortcut for true and false
        if (isOpX<TRUE>(e)) return Z3_mk_true (ctx);
        if (isOpX<FALSE>(e)) return Z3_mk_false (ctx);

        //variable
        if(isOpX<VARIANT>(e)) {
          Expr decl = variant::mainVariant(e)->left();
          assert(bind::isFdecl(decl));
          std::string declStr = declToStr(decl);
          int vnum = variant::variantNum(e);
          Z3_sort bvsort = Z3_mk_bv_sort(ctx,bvop::bvwidth(decl->right())); 
          string buf; raw_string_ostream ostr(buf);
          ostr << "vnt_" << declStr << "_" << vnum; ostr.flush();          
          return Z3_mk_const(ctx,Z3_mk_string_symbol(ctx,buf.c_str()),bvsort);
        }

        //constant or variable
        if(isOpX<FAPP>(e)) {
          Expr decl = e->left();
          assert(bind::isFdecl(decl));
          std::string declStr = declToStr(decl);
          Z3_sort bvsort = Z3_mk_bv_sort(ctx,bvop::bvwidth(decl->right())); 

          //constant -- unsigned long or unsigned long long
          if(isOpX< Terminal<unsigned long> >(decl->left()) || isOpX<LLUIT>(decl->left()))
            return Z3_mk_numeral(ctx,declStr.c_str(),bvsort);

          //non-constant
          return Z3_mk_const(ctx,Z3_mk_string_symbol(ctx,declStr.c_str()),bvsort);
        }
        
        //-- sign extension
        if(isOpX<BVZEXT>(e) || isOpX<BVSEXT>(e)) {
	  Z3_ast arg = marshal(e->left());
          unsigned oldbw = getBW(e->left());
          unsigned newbw = static_cast<unsigned>(getTerm<unsigned>(e->right())); 
          return isOpX<BVZEXT>(e) ?
            Z3_mk_zero_ext(ctx,newbw-oldbw,arg) :
            Z3_mk_sign_ext(ctx,newbw-oldbw,arg);
        }

        //ite
        if(isOpX<ITE>(e)) {
	  Z3_ast t1 = marshal(e->arg(0));
	  Z3_ast t2 = marshal(e->arg(1));
	  Z3_ast t3 = marshal(e->arg(2));
          return Z3_mk_ite(ctx,t1,t2,t3);
        }

        //extract
        if(isOpX<BVEXTRACT>(e)) {
          Z3_ast t = marshal(e->arg(0));
          unsigned high = getTerm<unsigned>(e->arg(1));
          unsigned low = getTerm<unsigned>(e->arg(2));
          return Z3_mk_extract(ctx,high,low,t);
        }

        int arity = e->arity ();

        if(arity == 1) {
	  Z3_ast arg = marshal(e->left());
          if(isOpX<NEG>(e)) return Z3_mk_not(ctx,arg);
          if(isOpX<BVNOT>(e)) return Z3_mk_bvnot(ctx,arg);
        }

        if(arity == 2) {
	  Z3_ast t1 = marshal(e->left());
	  Z3_ast t2 = marshal(e->right());
	  Z3_ast args [2] = {t1, t2};

          //-- logical operations
          if(isOpX<AND>(e)) return Z3_mk_and(ctx, 2, args);
          if(isOpX<OR>(e)) return Z3_mk_or(ctx, 2, args);
          if(isOpX<IMPL>(e)) return Z3_mk_implies(ctx, t1, t2);
          if(isOpX<IFF>(e)) return Z3_mk_iff(ctx, t1, t2);
          if(isOpX<BVAND>(e)) return Z3_mk_bvand(ctx, t1, t2);
          if(isOpX<BVOR>(e)) return Z3_mk_bvor(ctx, t1, t2);
          if(isOpX<BVXOR>(e)) return Z3_mk_bvxor(ctx, t1, t2);
          if(isOpX<BVSHL>(e)) return Z3_mk_bvshl(ctx, t1, t2);
          if(isOpX<BVASHR>(e)) return Z3_mk_bvashr(ctx, t1, t2);
          if(isOpX<BVLSHR>(e)) return Z3_mk_bvlshr(ctx, t1, t2);

          //-- arithmetic operations
          if(isOpX<BVADD>(e)) return Z3_mk_bvadd(ctx, t1, t2);
          if(isOpX<BVSUB>(e)) return Z3_mk_bvsub(ctx, t1, t2);
          if(isOpX<BVMUL>(e)) return Z3_mk_bvmul(ctx, t1, t2);
          if(isOpX<BVUDIV>(e)) return Z3_mk_bvudiv(ctx, t1, t2);
          if(isOpX<BVSDIV>(e)) return Z3_mk_bvsdiv(ctx, t1, t2);
          if(isOpX<BVUREM>(e)) return Z3_mk_bvurem(ctx, t1, t2);
          if(isOpX<BVSREM>(e)) return Z3_mk_bvsrem(ctx, t1, t2);

          //-- comparison operations
          if(isOpX<EQ>(e)) return Z3_mk_eq(ctx,t1,t2);
          if(isOpX<NEQ>(e)) return Z3_mk_not(ctx, Z3_mk_eq(ctx,t1,t2));
          if(isOpX<BVULT>(e)) return Z3_mk_bvult(ctx,t1,t2);
          if(isOpX<BVSLT>(e)) return Z3_mk_bvslt(ctx,t1,t2);
          if(isOpX<BVULE>(e)) return Z3_mk_bvule(ctx,t1,t2);
          if(isOpX<BVSLE>(e)) return Z3_mk_bvsle(ctx,t1,t2);
          if(isOpX<BVUGE>(e)) return Z3_mk_bvuge(ctx,t1,t2);
          if(isOpX<BVSGE>(e)) return Z3_mk_bvsge(ctx,t1,t2);
          if(isOpX<BVUGT>(e)) return Z3_mk_bvugt(ctx,t1,t2);
          if(isOpX<BVSGT>(e)) return Z3_mk_bvsgt(ctx,t1,t2);
        }
       
        //-- handle n-ary AND and OR
        if(isOpX<AND>(e) || isOpX<OR>(e)) {
          std::vector<Z3_ast> args;
          for (ENode::args_iterator it = e->args_begin(), end = e->args_end();
               it != end; ++it) {
            args.push_back (marshal(*it));
          }
          
          return isOpX<AND>(e) ?
            Z3_mk_and (ctx, args.size (), &args[0]) :
            Z3_mk_or (ctx, args.size (), &args[0]);
        }

        errs() << "ERROR : cannot convert " << *e << " to Z3\n";
        assert(0);
      }
    };

    //convert an expression to Z3 and return a string
    std::string toZ3Str(const Expr &e);

  } //namespace bitvc

} //namespace ufo

#endif //__BVC_Z3CONV_HPP__
