open Cil
open Rtcbautil
open Rtcbainput
open Rekoptions

(* open Graph.Pack *)

(* string map *)
module SM = Map.Make(struct type t = string let compare = compare end)
  
(** Varinfo map *)
module VM = ViMap

(* set of strings *)
module SS = Set.Make(String) 

(** Balanced Tree to keep track of jobs *)
module BT = BalancedTree

(** type for the tree of jobs *)
type jobtree  = BT.t

(** type of a job *)
type jobt = BT.n

(** switch for experimental optimized version -- still to be proven
    correct *)
let exper = ref true

(** whether to do recursive or non-recursive sequentialization --
    default is recursive *)
let recursiveSeq = ref true

(** whether to do unbounded sequentialization -- default is false *)
let unboundedSeq = ref false

(*********************************************************************)
(* Task Guessing code *)
(*********************************************************************)

(** number of hyper-periods to analyze *)
let hyperPeriods : int ref = ref 1

(** the tree of jobs *)
let jobTree : jobtree ref = ref (BT.make [])

let numOfJobs (unit) : int = BT.size !jobTree (BT.root !jobTree)
let jobLowRound (job:jobt) = (BT.id !jobTree job) - (BT.level_of job)
let jobHighRound (job:jobt) = (BT.max_id !jobTree job)
  
(** first job of a task *)
let firstJobOfTask (t:task) : jobt = 
  BT.firstInGeneration !jobTree t.tid
    
type taskInfo =
    {
      taskName : string;
      
      (** priority of the task. Priorities are unique. Priority of a
	  task is also its id *)
      mutable priority : int option;
      
      (** entry point *)
      mutable entryPt : fundec option;
      
      (** worst case execution time *)
      mutable wcet : int option;

      (** arival time *)
      mutable arrival : range;
      
      (** period *)
      mutable period : int option;
      
      mutable latency : int;
      
      mutable pbounds : int array
    }
      
(** array of taskInfo sorted by priorities *)
let taskArray : taskInfo array ref = ref [||]

let arr_leq x y =
  if x.rmax = -1 || y.rmin = -1 then false
  else x.rmax <= y.rmin

let arr_lt x y = 
  if x.rmax = -1 || y.rmin = -1 then false
  else x.rmax < y.rmin

let arr_add x y = 
  if x.rmax = -1 || x.rmin = -1 then {rmin = -1; rmax = -1;}
  else {rmin = x.rmin + y; rmax = x.rmax + y}


let jobPriority = BT.level_of 
let jobTask = BT.level_of

let jobArrival (job:jobt) =
  let tid = jobTask job in
  let arrival = !taskArray.(tid).arrival in
  let period = oget (!taskArray.(tid).period) in
  arr_add arrival ((BT.generationId !jobTree job) * period)

let jobRT (job:jobt) = 
  let tid = BT.level_of job in  !taskArray.(tid).latency


let jobPreceeds high low = 
  let jh = (jobArrival high) in
  let jl = (jobArrival low) in
  (*int_string ("ja_high " ^ (string_of_int jh.rmax) ^ "\n");
  print_string ("ja_low " ^ (string_of_int jl.rmin) ^ "\n"); *)
  let r = arr_leq  jh jl
  in 
  (*print_string 
    ((string_of_int (BT.id !jobTree high)) ^ 
	"<=" ^ 
	(string_of_int (BT.id !jobTree low)) ^ 
	" is " ^ 
	(string_of_bool r) ^ 
	"\n")
  ;*)
  r

let jobFollows high low = 
  let low_completion = arr_add (jobArrival low) (jobRT low) in
  arr_leq low_completion (jobArrival high)
    
let jobMayPreempt high low = 
  let x = not ( (jobPreceeds high low) || jobFollows high low)   
  in 
  (*let jid1 = BT.id !jobTree high in
  let jid2 = BT.id !jobTree low in
  print_string 
    ((string_of_int jid1) ^
	" may preempt " ^
	(string_of_int jid2) ^
	": " ^
	(string_of_bool x) ^ "\n");*)
  x

let jobMayNotPreceed high low = 
  arr_lt (jobArrival low) (jobArrival high)

let jobMayNotFollow high low =
  let low_completion = arr_add (jobArrival low) (jobRT low) in  
  arr_lt (jobArrival high) low_completion



(** compute a hyper-period for a task array*)
let hyperPeriod ta = 
  Array.fold_left (fun res ti -> lcm res (oget ti.period) ) 1 ta

(** convert task array to a list of tasks for rtcbautil *)
let taskArrayToList ta : task list = 
  (* number of jobs in previous iteration *)
  let res =  Array.mapi 
    (fun i a -> 
      { 
	name = a.taskName; 
	tid = i;
        tperiod = oget a.period;
	tslices = 1;
	tjobs = intceil !timeBound (oget a.period)
      }
    ) 
    ta 
  in Array.to_list res
	   
	   
let guessTasks file : taskInfo array  = 
  print_string "Guessing tasks\n";

  (** map from task names to task information *)
  let taskMap : taskInfo SM.t ref = ref SM.empty in

  let taskName vn p =
    let vnLen = String.length vn and
	pLen = String.length p in
    String.sub vn pLen (vnLen - pLen) 
  in
  
  let findOrCreateTask n = 
    try 
      SM.find n !taskMap
    with 
      | Not_found ->
	let ti = {taskName = n; 
		  priority = None; 
		  entryPt = None;
		  wcet = None;
		  arrival = { rmin = -1; rmax = -1 };
		  period = None;
		  latency = 0;
		  pbounds = [||]
		 } in
	taskMap := SM.add n ti !taskMap; 
	ti
  in
  
  let intFromInit init = 
    match init.init with
	Some (SingleInit e) when isConstant e ->
	  (match isInteger e with
	      Some int64 -> i64_to_int int64
	    | None -> invalid_arg "expecting an int")
      | _ -> invalid_arg "missing init"
  in
  let 
      taskEntryPt = "__startrek_entry_pt_" and
      basePriority = "__startrek_base_priority_" and
      taskWcet = "__startrek_wcet_" and
      taskPeriod = "__startrek_period_" and
      sTimeBound = "__startrek_time_bound" and
      taskLatency = "__startrek_latency_" and
      taskArrivalMin = "__startrek_arrival_min_" and
      taskArrivalMax = "__startrek_arrival_max_"
  in
  iterGlobals 
    file
    (fun g -> 
      match g with
	| GFun (fd, loc) when 
	    Util.hasPrefix taskEntryPt fd.svar.vname  -> 
	  let name = taskName fd.svar.vname taskEntryPt in
	  (findOrCreateTask name).entryPt <- Some fd
	| GVar (v, init, loc) when 
	    Util.hasPrefix basePriority v.vname  -> 
	  let name = taskName v.vname basePriority and
	      priority = intFromInit init in
	  (findOrCreateTask name).priority <- Some (priority) 
	| GVar (v, init, _) when
	    Util.hasPrefix taskWcet v.vname ->
	  let name = taskName v.vname taskWcet and
	      wcet = intFromInit init in
	  (findOrCreateTask name).wcet <- Some (wcet)
	| GVar (v, init, _) when
	    Util.hasPrefix taskPeriod v.vname ->
	  let name = taskName v.vname taskPeriod and
	      period = intFromInit init in
	  (findOrCreateTask name).period <- Some (period)
	| GVar (v, init, _) when
	    Util.hasPrefix taskLatency v.vname ->
	  let name = taskName v.vname taskLatency and
	      latency = intFromInit init in
	  (findOrCreateTask name).latency <- latency
	| GVar (v, init, _) when v.vname = sTimeBound ->
	  setTimeBound (intFromInit init)
	| GVar (v, init, _) when
	    Util.hasPrefix taskArrivalMin v.vname ->
	  let name = taskName v.vname taskArrivalMin and
	      minArrival = intFromInit init in
	  let task = findOrCreateTask name in
	  task.arrival <- { rmin = minArrival; rmax = task.arrival.rmax}
	| GVar (v, init, _) when
	    Util.hasPrefix taskArrivalMax v.vname ->
	  let name = taskName v.vname taskArrivalMax and
	      maxArrival = intFromInit init in
	  let task = findOrCreateTask name in
	  task.arrival <- { rmin = task.arrival.rmin; rmax = maxArrival}
	    
	| _ -> () 
    ) ;
  let taskArray = 
    Array.of_list (SM.fold (fun k d r -> d :: r ) !taskMap [])  in
  Array.sort (fun t1 t2 -> compare t1.priority t2.priority) taskArray; 
  taskArray
	  
let sanityCheck ta : unit = 
  for i = 0 to pred (Array.length ta) do
    begin
      (* All priorities are specified *)
      (match ta.(i).priority with 
	  None -> invalid_arg "No priority!"
	| _ -> ()
      );
      
      (* Rate Monotonic Priority Assignment *)
      if i < pred (Array.length ta) then
	begin
	  let x = oget ta.(i).period and
	      y = oget ta.(succ i).period in
	  if x < y then invalid_arg "Not RMS!" ;
	  if x mod y <> 0 then invalid_arg "Not harmonic"
	end
    end
  done

(*********************************************************************)
(* Globals *)
(*********************************************************************)
(** number of rounds *)
let rounds : int ref = ref 0

let rangesInArrival : bool ref = ref false
  
(*********************************************************************)
(*** Guessing task arguments **)
(*********************************************************************)      
let initTaskList file : unit =

  let checkArrivals ta =
    let x = Array.fold_left  
      (fun r t -> 
	(* compute arrival range if not set *)
	if (t.arrival.rmax == -1 || t.arrival.rmin = -1) then
	  begin
	    let m = (oget t.period) - t.latency in
	    let rng = { rmin = max 0 t.arrival.rmin; 
			rmax = if t.arrival.rmax == -1 then m 
			  else t.arrival.rmax; }
	    in t.arrival <- rng
	  end;
	if (t.arrival.rmax != t.arrival.rmin) then true
	else r
      )
      false
      ta
    in
    if x then
      begin
	print_string "RANGES found\n";
	rangesInArrival := true
      end
    else
      rangesInArrival := false
  in
  let computeJobTree ta = 
    let r = ref [] in
    print_string "JOB_TREE: start\n" ;
    for i = (Array.length ta) - 2 downto 0 do
      let d = (oget ta.(i).period) / (oget ta.(succ i).period) in
      r :=  d :: !r  ;
      print_string ("\tLevel " ^ (string_of_int i) ^ " has degree " ^ 
		       (string_of_int d) ^ "\n")
    done ; 
    print_string "JOB_TREE: done\n";
    BT.make !r 
  in


  taskArray := guessTasks file;
  sanityCheck !taskArray;
  
  (* normalize timebound to the closest hyper period *)
  let hp = hyperPeriod !taskArray in
  hyperPeriods := intceil !timeBound hp;
  setTimeBound (hp);
  print_string ("Using time bound: " ^ (string_of_int !timeBound) ^ "\n");
  print_string ("      with hyper-period: " ^ (string_of_int hp) ^ "\n");  

  begin 
    try 
      (* compute PB *)
      
      (* compute the rtInput from the arguments *)
      let rti = new rtInput in
      let ti2rt (ti:taskInfo) : unit = 
	let rt : rt_task = {
	rtname = "";
	rtwcet = oget ti.wcet;
	rtperiod = oget ti.period;
      } in rti#addTask (rt)
      in
      Array.iter ti2rt !taskArray;

      (* compute pbInput from rtInput *)
      let pbi = new pbInput in
      pbi#populate(rti);
      (* set latency and pbounds *)
      for i = 0 to pred (Array.length !taskArray) do
	let pbt = List.nth (pbi#getTasks) i in
	!taskArray.(i).pbounds <- Array.of_list pbt.pbpbounds;
	!taskArray.(i).latency <- pbt.pblatency;
      done;
    with invalid_arg ->nopbounds := true
  end
  ;
  tasks := taskArrayToList !taskArray ; 
  jobTree := computeJobTree !taskArray;
  checkArrivals !taskArray


(*********************************************************************)
(* Signatures *)
(*********************************************************************)
    
    
(** Type to pass common global values *)
module type ARGS = sig
  (* current file *)
  val file : file

  (* the "started" variable *)
  val startedVar : varinfo

  (* the "finished" variable *)
  val finishedVar : varinfo

  (* the current job *)
  val jobVar : varinfo

  (* lock variable *)
  val lockVar : varinfo

  (* The current hyper period *)
  val hyperPeriodVar : varinfo
end

(*********************************************************************)
(* Modules *)
(*********************************************************************)

(** the basic KJV module -- this contains all the common functions *)
module KJV (Args : ARGS) = 
struct

  (*********************************************************************)
  (* Common functions and types  *)
  (*********************************************************************)

  let vvtype = TFun (voidType, Some [], false, [])
  let csType = TFun (boolType,Some [],false,[])
  let assumeFn = findOrCreateFunc Args.file "__CPROVER_assume" assumeType
  let assertFn = findOrCreateFunc Args.file "assert" assertType 

  (*********************************************************************)
  (* Names  *)
  (*********************************************************************)
  let runJobFnName = "__startrek_run_job"
  let initSharedFnName = "__startrek_init_shared"
  let initGlobalsFnName = "__startrek_init_globals"
  let scheduleJobsFnName = "__startrek_schedule_jobs"
  let userFinalFnName = "__startrek_user_final"
  let userInitialFnName = "__startrek_user_init"
  let hyperFnName = "__startrek_hyperperiod"
  let preemptFnNamePrefix = "__startrek_preempt"
  let mainFnName = "main"

  let sharedReadPrefix = "__startrek_shared_read_"
  let sharedWritePrefix = "__startrek_shared_write_"

  (*********************************************************************)
  (* Non-deterministic values functions  *)
  (*********************************************************************)

  let ndbFn = findOrCreateFunc Args.file "nondet_bool" ndbType

  (** Returns a varinfo of a function that returns a non-deterministic value 
      of a given type *)
  let nondetFn (typ:typ) : varinfo =
    let tyname = 
      match typ with
	  TInt (IChar, _) -> "char"
	| TInt (ISChar, _) -> "schar"
	| TInt (IUChar, _) -> "uchar"
	| TInt (IBool, _) -> "_Bool"
	| TInt (IInt, _) -> "int"
	| TInt (IUInt, _) -> "uint"
	| TInt (IShort, _) -> "short"
	| TInt (IUShort, _) -> "ushort"
	| TInt (ILong, _) -> "long"
	| TInt (IULong, _) -> "ulong"
	| TInt (ILongLong, _) -> "long_long"
	| TInt (IULongLong, _) -> "ulong_long"
	| TFloat (FFloat, _) -> "float"
	| TFloat (FDouble, _) -> "double"
	| TFloat (FLongDouble, _) -> "long_double"
	| TNamed (tyinfo, _) -> tyinfo.tname
	| _ -> invalid_arg "Unsupported type"
    in
    findOrCreateFunc 
      Args.file 
      ("nondet_" ^ tyname) 
      (TFun (typ, Some [], false, []))
		
  (*********************************************************************)
  (* create the array of all jobs in DFS order in job tree *)
  (*********************************************************************)
  let jobsArray =
    Array.of_list (BT.fold_dfs (fun res t n -> res@[n]) [] !jobTree (BT.root !jobTree))    

  (*********************************************************************)
  (* given an array of jobs ja, return an expression that represents
     the fact that all of them have started *)
  (*********************************************************************)
  let allStarted (ja:jobt array) =
    let e = ref (integer 1) in
      for j = 0 to ((Array.length ja) - 1) do
        let ne = Formatcil.cExp
          "%e:e %b:land %v:startedVar[%d:jid]"
          [
            ("e", Fe !e);
            ("land", Fb LAnd);
            ("jid", Fd (BT.id !jobTree ja.(j)));
            ("startedVar", Fv Args.startedVar);
          ]
        in
          e := ne
      done;
      !e

  (*********************************************************************)
  (* given an array of jobs ja, return an expression that represents
     the fact that all of them have finished *)
  (*********************************************************************)
  let allFinished (ja:jobt array) =
    let e = ref (integer 1) in
      for j = 0 to ((Array.length ja) - 1) do
        let ne = Formatcil.cExp
          "%e:e %b:land %v:finishedVar[%d:jid]"
          [
            ("e", Fe !e);
            ("land", Fb LAnd);
            ("jid", Fd (BT.id !jobTree ja.(j)));
            ("finishedVar", Fv Args.finishedVar);
          ]
        in
          e := ne
      done;
      !e
		
  (*********************************************************************)
  (* given a job j return an array of all jobs that must finish
     before j starts *)
  (*********************************************************************)
  let mustFinishBefore (job:jobt) : jobt array =
    let subtree (job:jobt) : jobt list =
      BT.fold_dfs (fun res t n -> res@[n]) [] !jobTree job
    in
    let rec prevSubtrees (job:jobt) : jobt list =
      if (BT.hasPrev job) then (prevSubtrees (BT.prev job)) @ (subtree job)
      else subtree job
    in
    let rec mustFinishBeforeRec (job:jobt) : jobt list =
      if (BT.isRoot job) then []
      else if (BT.hasPrev job) then 
        (prevSubtrees (BT.prev job)) @ (mustFinishBeforeRec (BT.parent_of job))
      else (mustFinishBeforeRec (BT.parent_of job))
    in
      Array.of_list (mustFinishBeforeRec job)

  (*********************************************************************)
  (*** rewritting user-supplied init and final functions **)
  (*********************************************************************)
  let createUserInitial (f:file) =
    let fd = findOrCreateFundec f userInitialFnName vvtype in
    fd.svar.vinline <- true

  (* create a final function *)
  let createUserFinal (f:file) =
    let fd = findOrCreateFundec f userFinalFnName vvtype in
    fd.svar.vinline <- true

  (*********************************************************************)
  (*** creating the function that runs a sleeping job *)
  (*********************************************************************)
  let createRunJob allInterleavings (f:file) =
    let fd = findOrCreateFundec f runJobFnName vvtype in
      fd.sbody <- mkBlock 
        (allInterleavings f fd (if !exper then (Array.sub jobsArray 0 1) else jobsArray))

  (*********************************************************************)
  (* create the hyper period function *)
  (*********************************************************************)
  let createHyperPeriod (f:file) =
    let fd = findOrCreateFundec f hyperFnName vvtype in
    let runJobFn = var (findOrCreateFunc f runJobFnName vvtype) 
    in
    let inst = ref [] in
    let stmts = ref [] in
      (* initialize startedVar and finishedVar *)
      for i = 0 to (!rounds - 1) do
        stmts := !stmts @ 
          (Formatcil.cStmts
             "%v:startedVar[%d:i] = 0; %v:finishedVar[%d:i] = 0;"
             (fun n t -> makeTempVar fd ~name:n t)
             rekloc
             [
               ("i", Fd i);
               ("startedVar", Fv Args.startedVar);
               ("finishedVar", Fv Args.finishedVar);
             ])
      done;
      (* call runJob *)
      for i = 0 to (if !exper then 0 else (!rounds - 1)) do
        inst := !inst@[Call (None, Lval  runJobFn, [], rekloc)]
      done;
      stmts := !stmts @ [mkStmt (Instr !inst)] @
        (Formatcil.cStmts "assume(%e:e);"
           (fun n t -> makeTempVar fd ~name:n t) rekloc
           [ ("assume", Fv assumeFn); ("e", Fe (allFinished jobsArray)) ])
      ;
      fd.sbody <- mkBlock !stmts
      
  (*********************************************************************)
  (*** creating the main function **)
  (*********************************************************************)	
  let createMain (f:file) = 
    let fd = findOrCreateFundec f mainFnName vvtype in
    let userInitFn = findOrCreateFunc Args.file userInitialFnName vvtype in
    let stmts = ref (Formatcil.cStmts "init();"
           (fun n t -> makeTempVar fd ~name:n t) rekloc
           [ ("init", Fv userInitFn); ])
    in
    let hyperFn = findOrCreateFunc Args.file hyperFnName vvtype in
      if (!unboundedSeq) then (
        stmts := !stmts @ Formatcil.cStmts "while(1) hyperFn();"
          (fun n t -> makeTempVar fd ~name:n t) rekloc
          [ ("hyperFn", Fv hyperFn); ]
      ) else (
        for i = 1 to !hyperPeriods do
          stmts := !stmts @ Formatcil.cStmts "hyperFn();"
           (fun n t -> makeTempVar fd ~name:n t) rekloc
           [ ("hyperFn", Fv hyperFn); ]
        done ;
      );
      let finalFn = findOrCreateFunc Args.file userFinalFnName vvtype in
        stmts := !stmts @ Formatcil.cStmts "finalFn();"
           (fun n t -> makeTempVar fd ~name:n t) rekloc
           [ ("finalFn", Fv finalFn); ];
        fd.sbody <- mkBlock !stmts

  (*********************************************************************)
  (** replaces access to shared variables in special functions, i.e.,
      do not preempt *)
  (*********************************************************************)
  class sharedAccessReplacerSpecial = 
  object(self) inherit nopCilVisitor as super

    (* visit an instruction -- we are only interested in calls to
       renamed functions *)
    method vinst (ins:instr) = begin
      match ins with
          (* read a global variable -- must be of the form x = R(&v) *)
          Call(Some(Var x, NoOffset),
               Lval(Var {vname = fn}, NoOffset),
               [AddrOf(Var v, NoOffset)], loc)
            when Util.hasPrefix sharedReadPrefix fn ->
              x.vtype <- v.vtype;
              ChangeTo [Set (var x,Lval(var v),loc)]
	        
        (* write a global variable -- must be of the form W(&v,e) *)
        | Call(None,Lval(Var {vname = fn}, NoOffset),
               [AddrOf(Var v, NoOffset); e],loc)
            when Util.hasPrefix sharedWritePrefix fn ->            
            ChangeTo [Set(var v,e,loc)]
              
        (* otherwise, do nothing *)
        | _ -> SkipChildren
    end
  end
end;;

(** Time Bounded Analysis for Harmonic Tasks -- generates a non
    recursive sequentilal file *)
module KJVNonRec (Args : ARGS) = 
struct

  module K = KJV (Args)

  (*********************************************************************)
  (* Visitors  *)
  (*********************************************************************)   
    
  (** changes calls to foo to calls to foo_i *)
  class sharedAccessReplacer (jid:int) (renamedFuncs:SS.t) = 
  object(self) inherit nopCilVisitor as super
    (* the preemption function for this job *)
    val prfunc = findOrCreateFunc Args.file
      (K.preemptFnNamePrefix ^ "_" ^ (string_of_int jid))
      (TFun (boolType, Some [], false, []))

    (* visit an instruction -- we are only interested in calls to
       renamed functions *)
    method vinst (ins:instr) = begin
      match ins with
          (* read a global variable -- must be of the form x = R(&v) *)
          Call(Some(Var x, NoOffset),
           Lval(Var {vname = fn}, NoOffset),
           [AddrOf(Var v, NoOffset)], loc)
            when Util.hasPrefix K.sharedReadPrefix fn ->
              x.vtype <- v.vtype;
              ChangeTo
                [Call(None,Lval (var prfunc),[],loc); Set (var x,Lval(var v),loc)]
	        
        (* write a global variable -- must be of the form W(&v,e) *)
        | Call(None,Lval(Var {vname = fn}, NoOffset),
               [AddrOf(Var v, NoOffset); e],loc)
            when Util.hasPrefix K.sharedWritePrefix fn ->            
            ChangeTo
              [Call(None,Lval (var prfunc),[],loc); Set(var v,e,loc)]

        | Call (None, Lval (Var {vname = fn}, NoOffset), [], loc)
            when fn = cpuLockFnName ->
            ChangeTo ([Set (var Args.lockVar, one, builtinLoc)])
        | Call (None, Lval (Var {vname = fn}, NoOffset), [], loc)
            when fn = cpuUnlockFnName ->
            ChangeTo ([Set (var Args.lockVar, zero, builtinLoc)])
              (* call to a renamed function *)
        | Call (lhs, Lval (Var vi, off), args, loc)
            when SS.mem vi.vname renamedFuncs ->
            let vit = copyVarinfo vi (vi.vname ^ "_" ^ (string_of_int jid)) in
	      ChangeTo [Call(lhs,Lval(Var vit, off), args, loc)]
        (* otherwise, do nothing *)
        | _ -> SkipChildren
    end
  end

  (*********************************************************************)
  (*** creating the body of a function that simulates all possible
       interleavings of the set of jobs passed as argument *)
  (*********************************************************************)
  let allInterleavings (f:file) (fd:fundec) (prjobs:jobt array) =
    let jobStartFunction (j:jobt) =
      let jid = BT.id !jobTree j in
      let tid = jobTask j in
      let epvar = (oget (!taskArray.(tid).entryPt)).svar in
        findOrCreateFunc f (epvar.vname ^ "_" ^ (string_of_int jid)) K.vvtype
    in
    let stmts = ref [] in
      (* if all schedulable jobs have been executed already, then
         return *)
    let e = K.allStarted prjobs in
      stmts := !stmts @ 
        (Formatcil.cStmts
           "
             if(%e:e) return;
           "
           (fun n t -> makeTempVar fd ~name:n t)
           rekloc
           [
             ("e", Fe e);
           ])
      ;
      (* for each schedulable job *)
      for j = 0 to ((Array.length prjobs) - 1) do
        let job = prjobs.(j) in
        let jid = BT.id !jobTree job in
        let prfunc = findOrCreateFunc f (K.preemptFnNamePrefix ^ "_" ^ (string_of_int jid)) K.vvtype in
          (* only schedule this job if all jobs that must finish
             before it have finished *)
        let allfin = K.allFinished (K.mustFinishBefore job) in
          stmts := !stmts @ 
            (Formatcil.cStmts
               ("
                 %t:boolt cv = ndb(); 
                 if(cv %b:land %u:lnot %v:startedVar[%d:jid] %b:land %u:lnot %v:finishedVar[%d:jid] %b:land %e:allfin) {
                   %v:startedVar[%d:jid] = 1;
                   jobFn ();
                   %v:finishedVar[%d:jid] = 1;
               " ^ (if !exper then "prfunc ();" else "") ^
               "
                   return;
                 }
               ")
               (fun n t -> makeTempVar fd ~name:n t)
               rekloc
               [
                 ("boolt", Ft boolType);
                 ("ndb", Fv K.ndbFn);
                 ("land", Fb LAnd);
                 ("lnot", Fu LNot);
                 ("jid", Fd jid);
                 ("startedVar", Fv Args.startedVar);
                 ("finishedVar", Fv Args.finishedVar);
                 ("jobFn", Fv (jobStartFunction job));
                 ("prfunc", Fv prfunc);
                 ("allfin", Fe allfin);
               ])
      done;
      !stmts

  (*********************************************************************)
  (*** creating the tasks **)
  (*********************************************************************)
  let createPreemptFn (f:file) =
    (* for each job in DFS order *)
    for i = 0 to (Array.length K.jobsArray - 1) do                   
      let job = K.jobsArray.(i) in
      let jid = BT.id !jobTree job in
      let fdh = findOrCreateFundec f (K.preemptFnNamePrefix ^ "_helper_" ^ (string_of_int jid)) K.vvtype in
      let fd = findOrCreateFundec f (K.preemptFnNamePrefix ^ "_" ^ (string_of_int jid)) K.vvtype in
        (* compute the set of jobs that can preempt this job -- this
           is essentially the jobs in the sub-tree rooted at this job *)
          if (not (BT.isLeaf !jobTree job)) then (
            let prjobs = 
              Array.of_list (List.tl (BT.fold_dfs (fun res t n -> res@[n]) [] !jobTree job))
            in              
              fdh.sbody <- mkBlock (allInterleavings f fdh prjobs);
              let stmts = ref (Formatcil.cStmts
                                 "if (%v:lockVar) return;"
                                 (fun n t -> makeTempVar fd ~name:n t) rekloc
                                 [ ("lockVar", Fv Args.lockVar); ])
              in
                for k = 0 to (if !exper then 0 else (Array.length prjobs - 1)) do
                  stmts := !stmts@[mkStmtOneInstr (Call (None, Lval (var fdh.svar), [], rekloc))]
                done;
                fd.sbody <- mkBlock !stmts
          );
      done
      
  (*********************************************************************)
  (*** creating the tasks **)
  (*********************************************************************)
  let createTasks (f:file) =
    (* the new globals *)
    let newGlobs = ref [] in
      (* the names of renamed functions *)
    let renamedFuncs = ref SS.empty in
    iterGlobals 
      f 
      (fun g -> 
	 match g with 
	   | GFun(fd, loc) ->                
	       if ((fd.svar.vname = K.userFinalFnName) ||
		     (fd.svar.vname = K.userInitialFnName) ||
                     (Util.hasPrefix K.preemptFnNamePrefix fd.svar.vname)) then (                 
                 fd.sbody <- visitCilBlock (new K.sharedAccessReplacerSpecial) fd.sbody;
                 newGlobs := !newGlobs@[g]
               ) else (
                 renamedFuncs := SS.add fd.svar.vname !renamedFuncs;
                 (* for each task, copy all functions, appending the task id to the function name *)
                 for i = 0 to (!rounds - 1) do                   
                   let prfd = findOrCreateFundec Args.file
                     (K.preemptFnNamePrefix ^ "_" ^ (string_of_int i))
                     (TFun (boolType, Some [], false, []))
                   in
                     newGlobs := !newGlobs@[GVarDecl(prfd.svar,loc)];
                     let fdi = copyFunction fd (fd.svar.vname ^ "_" ^ (string_of_int i)) in
                       fdi.sbody <- visitCilBlock (new sharedAccessReplacer i !renamedFuncs) fdi.sbody;
                       newGlobs := !newGlobs@[GFun(fdi,loc)]
                 done
               )
	   | _ -> newGlobs := !newGlobs@[g] 
      );
      f.globals <- !newGlobs

  (*********************************************************************)	
  (** run everything *)
  (*********************************************************************)	
  let run unit : unit = begin
    (* create the preemption functions *)
    createPreemptFn Args.file;
    (* create the tasks *)
    createTasks Args.file;
    (* create the runJob function *)
    K.createRunJob allInterleavings Args.file;
    (* create user initial function *)
    K.createUserInitial Args.file;
    (* create user final function *)
    K.createUserFinal Args.file;
    (* create hyper-period function *)
    K.createHyperPeriod Args.file; 
    (* create the main function *)
    K.createMain Args.file;
  end
    
end;;

(** Time Bounded Analysis for Harmonic Tasks -- generates a recursive
    sequentilal file *)
module KJVRec (Args : ARGS) = 
struct

  module K = KJV (Args)

  (*********************************************************************)
  (* Visitors  *)
  (*********************************************************************)   
    
  (** call copier for the recursive version *)
  class sharedAccessReplacer = 
  object(self) inherit nopCilVisitor as super
    (* the preemption function for this job *)
    val prfunc = findOrCreateFunc Args.file K.preemptFnNamePrefix
      (TFun (boolType, Some [], false, []))

    (* visit an instruction *)
    method vinst (ins:instr) = begin
      match ins with
          (* read a global variable -- must be of the form x = R(&v) *)
          Call(Some(Var x, NoOffset),
           Lval(Var {vname = fn}, NoOffset),
           [AddrOf(Var v, NoOffset)], loc)
            when Util.hasPrefix K.sharedReadPrefix fn ->
              x.vtype <- v.vtype;
              ChangeTo
                [Call(None,Lval (var prfunc),[],loc); Set (var x,Lval(var v),loc)]
	   
          (* write a global variable -- must be of the form W(&v,e) *)
        | Call(None,Lval(Var {vname = fn}, NoOffset),
               [AddrOf(Var v, NoOffset); e],loc)
            when Util.hasPrefix K.sharedWritePrefix fn ->            
            ChangeTo
              [Call(None,Lval (var prfunc),[],loc); Set(var v,e,loc)]

        | Call (None, Lval (Var {vname = fn}, NoOffset), [], loc)
            when fn = cpuLockFnName ->
            ChangeTo ([Set (var Args.lockVar, one, builtinLoc)])
        | Call (None, Lval (Var {vname = fn}, NoOffset), [], loc)
            when fn = cpuUnlockFnName ->
            ChangeTo ([Set (var Args.lockVar, zero, builtinLoc)])
        (* otherwise, do nothing *)
        | _ -> SkipChildren
    end
  end

  (*********************************************************************)
  (*** creating the body of a function that simulates all possible
       interleavings of the set of jobs passed as argument *)
  (*********************************************************************)
  let allInterleavings (f:file) (fd:fundec) (prjobs:jobt array) =
    let jobStartFunction (j:jobt) =
      let tid = jobTask j in
      let epvar = (oget (!taskArray.(tid).entryPt)).svar in
        findOrCreateFunc f epvar.vname K.vvtype
    in
    let stmts = ref [] in
      (* if all schedulable jobs have been executed already, then
         return *)
    let e = K.allStarted prjobs in
      stmts := !stmts @ 
        (Formatcil.cStmts
           "
             if(%e:e) return;
           "
           (fun n t -> makeTempVar fd ~name:n t)
           rekloc
           [
             ("e", Fe e);
           ])
      ;
      (* for each schedulable job *)
      for j = 0 to ((Array.length prjobs) - 1) do
        let job = prjobs.(j) in
        let jid = BT.id !jobTree job in
        let prfunc = findOrCreateFunc f (K.preemptFnNamePrefix ^ "_helper_" ^ (string_of_int jid)) K.vvtype in
          (* only schedule this job if all jobs that must finish
             before it have finished *)
        let allfin = K.allFinished (K.mustFinishBefore job) in
          stmts := !stmts @ 
            (Formatcil.cStmts
               ("
                 %t:boolt cv = ndb(); 
                 if(cv %b:land %u:lnot %v:startedVar[%d:jid] %b:land %u:lnot %v:finishedVar[%d:jid] %b:land %e:allfin) {
                   %v:startedVar[%d:jid] = 1;
                   %v:jobVar = %d:jid;
                   jobFn ();
                   %v:finishedVar[%d:jid] = 1;
               " ^ (if !exper then "prfunc ();" else "") ^
               "
                   return;
                 }
               ")
               (fun n t -> makeTempVar fd ~name:n t)
               rekloc
               [
                 ("boolt", Ft boolType);
                 ("ndb", Fv K.ndbFn);
                 ("land", Fb LAnd);
                 ("lnot", Fu LNot);
                 ("jid", Fd jid);
                 ("startedVar", Fv Args.startedVar);
                 ("jobVar", Fv Args.jobVar);
                 ("finishedVar", Fv Args.finishedVar);
                 ("jobFn", Fv (jobStartFunction job));
                 ("prfunc", Fv prfunc);
                 ("allfin", Fe allfin);
               ])
      done;
      !stmts

  (*********************************************************************)
  (*** creating the tasks **)
  (*********************************************************************)
  let createPreemptFn (f:file) =
      (* for each job in DFS order -- create the helper function *)
    for i = 0 to (Array.length K.jobsArray - 1) do                   
      let job = K.jobsArray.(i) in
      let jid = BT.id !jobTree job in
      let fdh = findOrCreateFundec f (K.preemptFnNamePrefix ^ "_helper_" ^ (string_of_int jid)) K.vvtype in
        (* compute the set of jobs that can preempt this job -- this
           is essentially the jobs in the sub-tree rooted at this job *)
        if (not (BT.isLeaf !jobTree job)) then (
          let prjobs = 
            Array.of_list (List.tl (BT.fold_dfs (fun res t n -> res@[n]) [] !jobTree job))
          in              
            fdh.sbody <- mkBlock (allInterleavings f fdh prjobs);
        );
    done;
      let fd = findOrCreateFundec f K.preemptFnNamePrefix K.vvtype in
      let oldjv = makeLocalVar fd "oldjob" ucharType in
      let stmts = ref (Formatcil.cStmts "if (%v:lockVar) return;"
                         (fun n t -> makeTempVar fd ~name:n t) rekloc
                         [ ("lockVar", Fv Args.lockVar); ])
      in
        (* now create the main prempt function *)
        for i = 0 to (Array.length K.jobsArray - 1) do                   
          let job = K.jobsArray.(i) in
          let jid = BT.id !jobTree job in
          let prhelp = findOrCreateFunc f (K.preemptFnNamePrefix ^ "_helper_" ^ (string_of_int jid)) K.vvtype in
            (* compute the set of jobs that can preempt this job -- this
               is essentially the jobs in the sub-tree rooted at this job *)
            if (not (BT.isLeaf !jobTree job)) then (
              stmts := !stmts@
                (Formatcil.cStmts 
                   "if (%v:jobVar == %d:i) { %v:oldjv = %v:jobVar; prhelp(); %v:jobVar = %v:oldjv; }"
                   (fun n t -> makeTempVar fd ~name:n t) rekloc
                   [ 
                     ("oldjv", Fv oldjv);
                     ("jobVar", Fv Args.jobVar);
                     ("i", Fd i);
                     ("prhelp", Fv prhelp); 
                   ])
            );
        done;
        fd.sbody <- mkBlock !stmts
      
  (*********************************************************************)
  (*** creating the tasks **)
  (*********************************************************************)
  let createTasks (f:file) =
    (* the new globals *)
    let newGlobs = ref [] in
    iterGlobals 
      f 
      (fun g -> 
	 match g with 
	   | GFun(fd, loc) ->                
	       if ((fd.svar.vname = K.userFinalFnName) ||
		     (fd.svar.vname = K.userInitialFnName) ||
                     (Util.hasPrefix K.preemptFnNamePrefix fd.svar.vname)) then (                 
                 fd.sbody <- visitCilBlock (new K.sharedAccessReplacerSpecial) fd.sbody;
                 newGlobs := !newGlobs@[g]
               ) else (                   
                 let prfd = findOrCreateFundec Args.file K.preemptFnNamePrefix
                   (TFun (boolType, Some [], false, []))
                 in
                   newGlobs := !newGlobs@[GVarDecl(prfd.svar,loc)];
                   fd.sbody <- visitCilBlock (new sharedAccessReplacer) fd.sbody;
                   newGlobs := !newGlobs@[GFun(fd,loc)]
               )
	   | _ -> newGlobs := !newGlobs@[g] 
      );
      f.globals <- !newGlobs

  (*********************************************************************)	
  (** run everything *)
  (*********************************************************************)	
  let run unit : unit = begin
    (* create the preemption functions *)
    createPreemptFn Args.file;
    (* create the tasks *)
    createTasks Args.file;
    (* create the runJob function *)
    K.createRunJob allInterleavings Args.file;
    (* create user initial function *)
    K.createUserInitial Args.file;
    (* create user final function *)
    K.createUserFinal Args.file;
    (* create hyper-period function *)
    K.createHyperPeriod Args.file; 
    (* create the main function *)
    K.createMain Args.file;
  end
    
end;;

(*********************************************************************)
(*** the Cil feature, where everything begins *)
(*********************************************************************)

(* type of the started and finished array - the argument is the size
   of the array *)
let startedVarType (size:int) = 
  Formatcil.cType "%t:t [%e:R]"
    [
      ("t", Ft boolType); 
      ("R", Fe (integer size))
    ] 

(* command line option for first pass *)
let feature : featureDescr = { 
  fd_name = "kjv";              
  fd_enabled = ref false;
  fd_description = "Kidd et al. SPIN'10";
  fd_extraopt = 
    [
      "--kjvnonrec", Arg.Clear recursiveSeq, " do non-recursive sequentialization";
      "--kjvunbounded", Arg.Set unboundedSeq, " do unbounded sequentialization";
    ];
  fd_doit = (
    function (f: file) -> 
      print_string ("Analyzing " ^ f.fileName ^ "\n");

      (* compute task specific info *)
      initTaskList f ;
      
      (* #rounds == total # of jobs *)
      rounds := List.fold_left (fun res t -> res + t.tjobs) 0 !tasks ;
        
      (* package arguments into a module *)
      let module A = struct 
	let file = f 
	let startedVar = makeGlobalVar "__startrek_started" (startedVarType !rounds)
	let finishedVar = makeGlobalVar "__startrek_finished" (startedVarType !rounds)
        let jobVar = makeGlobalVar "__startrek_job" ucharType
        let lockVar = makeGlobalVar "__startrek_lock" boolType
	let hyperPeriodVar = makeGlobalVar "__startrek_hyper_period" ucharType
      end 
      in        
	(* register global variables created so far with the file *)
	f.globals <- 
          (GVar (A.startedVar, {init = Some(makeZeroInit (startedVarType !rounds))}, builtinLoc))::
          (GVar (A.finishedVar, {init = Some(makeZeroInit (startedVarType !rounds))}, builtinLoc))::
          (GVar (A.jobVar, {init = None}, builtinLoc))::
          (GVar (A.lockVar, {init = Some (makeZeroInit boolType)},builtinLoc))::
          (GVar (A.hyperPeriodVar, {init = None}, builtinLoc))::
          f.globals ;
        
        (* generate KJV sequentialization *)
        if (!recursiveSeq) then 
          let module M = KJVRec (A) in
            M.run ()
        else
          let module M = KJVNonRec (A) in
            M.run ()
  );
  fd_post_check = false
} 
