open Cil
open Rtcbautil
open Rtcbainput
open Rekoptions
open Inliner

(* string map *)
module SM = Map.Make(struct type t = string let compare = compare end)
  
(* Varinfo map *)
module VM = ViMap

let isTopTask t = (t.tid = pred (List.length !tasks))

let oget a = 
  match a with 
    | Some x -> x 
    | None -> invalid_arg "Expecting None, got Some"

let rec gcd a b =
  if      a = 0 then b
  else if b = 0 then a
  else if a > b then gcd b (a mod b)
  else gcd a (b mod a)

let lcm m n =  (abs m * n) / (gcd m n)

let intceil a b = (a / b) + if (a mod b = 0) then 0 else 1
    
(*********************************************************************)
(* Task Guessing code *)
(*********************************************************************)


(** number of hyper-periods to analyze *)
let hyperPeriods : int ref = ref 1

type taskInfo =
    {
      taskName : string;
      
      (** priority of the task. Priorities are unique. Priority of a
	  task is also its id *)
      mutable priority : int option;
      
      (** entry point *)
      mutable entryPt : fundec option;
      
      (** worst case execution time *)
      mutable wcet : int option;
      
      (** period *)
      mutable period : int option;
      
      mutable latency : int;
      
      mutable pbounds : int array
    }
      

(** array of taskInfo sorted by priorities *)
let taskArray : taskInfo array ref = ref [||]

(** compute a hyper-period for a task array*)
let hyperPeriod ta = 
  Array.fold_left (fun res ti -> lcm res (oget ti.period) ) 1 ta

(** convert task array to a list of tasks for rtcbautil *)
let taskArrayToList ta : task list = 
  (* number of jobs in previous iteration *)
  let res =  Array.mapi 
    (fun i a -> 
      { 
	name = a.taskName; 
	tid = i;
        tperiod = oget a.period;
	tslices = 1;
	tjobs = intceil !timeBound (oget a.period)
      }
    ) 
    ta 
  in Array.to_list res
	   
	   
let guessTasks file : taskInfo array  = 
  print_string "Guessing tasks\n";


  (** map from task names to task information *)
  let taskMap : taskInfo SM.t ref = ref SM.empty in


  let taskName vn p =
    let vnLen = String.length vn and
	pLen = String.length p in
    String.sub vn pLen (vnLen - pLen) 
  in
  
  let findOrCreateTask n = 
    try 
      SM.find n !taskMap
    with 
      | Not_found ->
	let ti = {taskName = n; 
		  priority = None; 
		  entryPt = None;
		  wcet = None;
		  period = None;
		  latency = 0;
		  pbounds = [||]
		 } in
	taskMap := SM.add n ti !taskMap; 
	ti
  in
  
  let intFromInit init = 
    match init.init with
	Some (SingleInit e) when isConstant e ->
	  (match isInteger e with
	      Some int64 -> i64_to_int int64
	    | None -> invalid_arg "expecting an int")
      | _ -> invalid_arg "missing init"
  in
  let 
      taskEntryPt = "__startrek_entry_pt_" and
      basePriority = "__startrek_base_priority_" and
      taskWcet = "__startrek_wcet_" and
      taskPeriod = "__startrek_period_" and
      sTimeBound = "__startrek_time_bound" and
      taskLatency = "__startrek_latency" 
  in
  iterGlobals 
    file
    (fun g -> 
      match g with
	| GFun (fd, loc) when 
	    Util.hasPrefix taskEntryPt fd.svar.vname  -> 
	  let name = taskName fd.svar.vname taskEntryPt in
	  (findOrCreateTask name).entryPt <- Some fd
	| GVar (v, init, loc) when 
	    Util.hasPrefix basePriority v.vname  -> 
	  let name = taskName v.vname basePriority and
	      priority = intFromInit init in
	  (findOrCreateTask name).priority <- Some (priority) 
	| GVar (v, init, _) when
	    Util.hasPrefix taskWcet v.vname ->
	  let name = taskName v.vname taskWcet and
	      wcet = intFromInit init in
	  (findOrCreateTask name).wcet <- Some (wcet)
	| GVar (v, init, _) when
	    Util.hasPrefix taskPeriod v.vname ->
	  let name = taskName v.vname taskPeriod and
	      period = intFromInit init in
	  (findOrCreateTask name).period <- Some (period)
	| GVar (v, init, _) when
	    Util.hasPrefix taskLatency v.vname ->
	  let name = taskName v.vname taskLatency and
	      latency = intFromInit init in
	  (findOrCreateTask name).latency <- latency
	| GVar (v, init, _) when v.vname = sTimeBound ->
	  setTimeBound (intFromInit init)
	| _ -> () 
    ) ;
  let taskArray = 
    Array.of_list (SM.fold (fun k d r -> d :: r ) !taskMap [])  in
  Array.sort (fun t1 t2 -> compare t1.priority t2.priority) taskArray; 
  taskArray
	  
let sanityCheck ta : unit = 
  for i = 0 to pred (Array.length ta) do
    begin
      (* All priorities are specified *)
      (match ta.(i).priority with 
	  None -> invalid_arg "No priority!"
	| _ -> ()
      );
      
      (* Rate Monotonic Priority Assignment *)
      if (i < pred (Array.length ta)) then
	(match (ta.(i).period, ta.(i+1).period) with
	    (Some(x), Some(y)) when y > x -> 
	      invalid_arg "Not RMS!"
	  | _ -> ()
	);
      
      
    end
  done



(*********************************************************************)
(* Globals *)
(*********************************************************************)

(** number of rounds *)
let rounds : int ref = ref 0



(*********************************************************************)
(*** Guessing task arguments **)
(*********************************************************************)      
let initTaskList file : unit =

  taskArray := guessTasks file;
  sanityCheck !taskArray;
  
  (* normalize timebound to the closest hyper period *)
  let hp = hyperPeriod !taskArray in
  hyperPeriods := intceil !timeBound hp;
  setTimeBound (hp);
  print_string ("Using time bound: " ^ (string_of_int !timeBound) ^ "\n");
  print_string ("      with hyper-period: " ^ (string_of_int hp) ^ "\n");  

  if (!nohyperunroll) then 
    begin 
      setTimeBound (hp * !hyperPeriods);
      hyperPeriods := 1;
      print_string "Disabled hyper-period unrolling!!!\n";
    end;
  
  begin 
    try 
      (* compute PB *)
      
      (* compute the rtInput from the arguments *)
      let rti = new rtInput in
      let ti2rt (ti:taskInfo) : unit = 
	let rt : rt_task = {
	rtname = "";
	rtwcet = oget ti.wcet;
	rtperiod = oget ti.period;
      } in rti#addTask (rt)
      in
      Array.iter ti2rt !taskArray;

      (* compute pbInput from rtInput *)
      let pbi = new pbInput in
      pbi#populate(rti);
      (* set latency and pbounds *)
      for i = 0 to pred (Array.length !taskArray) do
	let pbt = List.nth (pbi#getTasks) i in
	!taskArray.(i).pbounds <- Array.of_list pbt.pbpbounds;
	!taskArray.(i).latency <- pbt.pblatency;
      done;
    with invalid_arg -> nopbounds := true
  end;
    
  tasks := taskArrayToList !taskArray 



(*********************************************************************)
(* Signatures *)
(*********************************************************************)


(** Type to pass common global values *)
module type ARGS = sig
  (* current file *)
  val file : file

  (* current job *)
  val jobVar : varinfo

  (* round in which current job ends *)
  val jobEndVar : varinfo

  (* round in which an error was found *)
  val errorRoundVar : varinfo

  (* current task *)
  val taskVar : varinfo

  (* current round *)
  val roundVar : varinfo

  (* lock variable *)
  val lockVar : varinfo

  (* boolean variable. True when first context switch has happened *)
  val isfirstcsVar : varinfo 

  (* The current hyper period *)
  val hyperPeriodVar : varinfo

end

(** Signature of a manger of shared variables. *)
module type SHARED_VAR_MANAGER = sig
  (** opaque information about a variable *)
  type sv
  
  (** returns true if a given variable is managed by this module*)
  val mem : varinfo -> bool

  (** promote varinfo to a shared variable *)
  val makeSharedVar : varinfo -> unit

  (** iterate over all shared variables *)
  val iterShared : (varinfo -> sv -> unit) -> unit 

  val foldShared : (varinfo -> sv -> 'b -> 'b) -> 'b -> 'b

  (** get shared variable information for a given task *)
  val getSharedVar : varinfo  -> sv

  (** lval for a shared variable given a round expression *)
  val getSharedLval : sv -> int  -> lval

  (** lval for a symbolic constant that initialized a shared vairable *)
  val getSharedInitLval : sv -> int -> lval
    
  (** mkSharedReadStmt l s e returns a stmt that loads s[e] into l *)
  val mkSharedReadStmt : lval -> sv -> exp -> stmt

  (** mkSharedWriteStmt s e a retuns a stmt that writes a into s[e] *)
  val mkSharedWriteStmt : sv -> exp -> exp -> stmt

  (** register any global created as side-effects with with the
     current file. Must be called at the end. *)
  val registerGlobals : file -> unit

end

module type ASSERT_VAR_MANAGER = sig

  val getAssertLval : int -> int -> exp -> lval

  val mkAsssertWriteStmt : int -> task -> exp -> exp -> stmt 

  val newAssertFun : file -> varinfo

  val registerGlobals : file -> unit
    
  val map : (Cil.lval -> 'a) -> 'a list

end

module type SCHED_VAR_MANAGER = sig
  type schedvars
    
  val taskSchedVars : int -> schedvars
  val jobStartLval : int -> int -> lval
  val jobEndLval : int -> int -> lval
  val mkReadJobEndStmt : lval -> task -> exp -> stmt
  val registerGlobals : file -> unit

end

(*********************************************************************)
(* Modules *)
(*********************************************************************)

(** an implementation of SHARED_VAR_MANAGER *)
module ArraySharedVarManager (Args : ARGS) =
struct
  
  type sv = 
      {
	(** array for round values *)
	xs : varinfo;
	(** array for initial values *)
	sc : varinfo;
      }

  (** map from variables to shared info. The state. *)
  let sharedVars :  sv VM.t ref = ref VM.empty

  (** returns shared variable info for a variable *)
  let svi (v) = (VM.find v !sharedVars)

  (** returns true if a variable is managed *)
  let mem (v) = (VM.mem v !sharedVars)

  (** Creates a new shared variable from a varinfo *)
  let makeSharedVar (v:varinfo) : unit =

    let makeSharedVarName (v:varinfo) : string = 
      "_" ^ v.vname ^ "_" in

    let makeSharedInitName (v:varinfo) : string = 
      "_i_" ^ v.vname ^ "_" in
      
    let atype = 
      Formatcil.cType "%t:t [%e:R]"
	[
	  ("t", Ft v.vtype); 
	  ("R", Fe (integer !rounds))
	] in 
    
    let xs = makeGlobalVar (makeSharedVarName v) atype and
	sc = makeGlobalVar (makeSharedInitName v) atype; 
    in
    sharedVars := 
      VM.add v {xs = xs; sc = sc} !sharedVars

  (** Returns a shared varinfo for varinfo v*)
  let getSharedVar = svi

  (** Returns an Lval for accessing a shared variable v in a given round *)
  let getSharedLval (sv:sv) (r:int) : lval =  
    Formatcil.cLval "xs[%e:r]" [("xs", Fv sv.xs); ("r", Fe (integer r))]

  (** Returns an Lval for the symbolic constant representing an
      intial value of a shared variable in a given round *)
  let getSharedInitLval (sv:sv) (r:int) : lval =
    Formatcil.cLval "sc[%e:r]" [("sc", Fv sv.sc); ("r", Fe (integer r))]

  (** Returns a statement that sets out to the value of shared
      variable sv in round re *)
  let mkSharedReadStmt (out:lval) (sv:sv) (re:exp) : stmt = 
    Formatcil.cStmt 
      "%l:out = xs[%e:re];"
      (fun n t -> assert false)
      rekloc
      [
	("out", Fl out);
	("xs", Fv sv.xs);
	("re", Fe re)
      ]
      
  (** Returns a statement that writes arg to the shared variable sv
      in round re *)
  let mkSharedWriteStmt (sv:sv) (re:exp) (arg:exp) : stmt = 
    Formatcil.cStmt
      "xs[%e:re] = %e:arg;"
      (fun n t -> assert false)
      rekloc
      [
	("xs", Fv sv.xs);
	("re", Fe re);
	("arg", Fe arg)
      ]
      
  (** iterate over shared variables*)
  let iterShared (f : varinfo -> sv -> unit) : unit =
    VM.iter f !sharedVars

  let foldShared (f : varinfo -> sv -> 'b -> 'b) (r:'b) : 'b = 
    VM.fold f !sharedVars r
      
  let registerGlobals f : unit = 
    (* register global variables *)
    let fn res g = 
      let out = g :: res in 
      match g with 
	| GVar (v, _, _) when mem v -> 
	  let sv = svi v in
	  GVar (sv.xs, {init = None}, rekloc) :: 
	    GVar (sv.sc, {init = None}, rekloc) :: out
	| _ -> out
    in
    f.globals <- List.fold_left fn [] (List.rev f.globals)
    (* iterShared  *)
    (*   (fun v sv -> *)
    (* 	 f.globals <-  *)
    (* 	   GVar(sv.xs, { init = None }, rekloc) ::  *)
    (* 	   GVar(sv.sc, { init = None }, rekloc) :: *)
    (* 	   f.globals) *)
end


(** an implementation of SHARED_VAR_MANAGER *)
module NoArraySharedVarManager (Args : ARGS) =
struct
  
  type sv = 
      {
	(** array for round values *)
	xs : varinfo array;
	(** array for initial values *)
	sc : varinfo array;
      }

  (** map from variables to shared info. The state. *)
  let sharedVars :  sv VM.t ref = ref VM.empty

  (** returns shared variable info for a variable *)
  let svi (v) = (VM.find v !sharedVars)

  (** returns true if a variable is managed *)
  let mem (v) = (VM.mem v !sharedVars)

  (** Creates a new shared variable from a varinfo *)
  let makeSharedVar (v:varinfo) : unit =

    let makeSharedVarName (v:varinfo) (r:int) : string = 
      "_" ^ v.vname ^ "_r" ^ (string_of_int r) in

    let makeSharedInitName (v:varinfo) (r:int) : string = 
      "_i_" ^ v.vname ^ "_r" ^ (string_of_int r) in
      
    let atype = v.vtype in
    
    let xs = Array.init !rounds 
      (fun i -> makeGlobalVar (makeSharedVarName v i) atype) in
    
    let sc = Array.init !rounds 
      (fun i -> makeGlobalVar (makeSharedInitName v i) atype)  in

    sharedVars := 
      VM.add v {xs = xs; sc = sc} !sharedVars

  (** Returns a shared varinfo for varinfo v*)
  let getSharedVar = svi

  (** Returns an Lval for accessing a shared variable v in a given round *)
  let getSharedLval (sv:sv) (r:int) : lval =  var sv.xs.(r)


  (** Returns an Lval for the symbolic constant representing an
      intial value of a shared variable in a given round *)
  let getSharedInitLval (sv:sv) (r:int) : lval = var sv.sc.(r)

  (** Returns a statement that sets out to the value of shared
      variable sv in round re *)
  let mkSharedReadStmt (out:lval) (sv:sv) (re:exp) : stmt = 
    mkSwitchStmt re 0 !rounds 
      (fun i -> 
	 [mkStmtOneInstr 
	    (Set (out, Lval (getSharedLval sv i), rekloc))]
      )

    (** Returns a statement that writes arg to the shared variable sv
	in round re *)
  let mkSharedWriteStmt (sv:sv) (re:exp) (arg:exp) : stmt = 
    mkSwitchStmt re 0 !rounds
      (fun i -> 
	 [mkStmtOneInstr 
	    (Set (getSharedLval sv i, arg, rekloc))]
      )
      
  (** iterate over shared variables*)
  let iterShared (f : varinfo -> sv -> unit) : unit =
    VM.iter f !sharedVars
      
  let foldShared (f : varinfo -> sv -> 'b -> 'b) (r:'b) : 'b = 
    VM.fold f !sharedVars r
      
      
  let registerGlobals f : unit = 
    (* register global variables *)
    iterShared 
      (fun v sv ->
	 Array.iter 
	   (fun e -> 
	      f.globals <- (GVar (e, { init = None }, rekloc)) :: f.globals)
	   sv.xs ;
	 Array.iter 
	   (fun e -> 
	      f.globals <- (GVar (e, { init = None }, rekloc)) :: f.globals)
	   sv.sc )

end

module ArrayAssertVarManager (Args : ARGS) =
struct

  (*********************************************************************)
  (* Managing assert variables  *)
  (*********************************************************************)

  let taskVar = Args.taskVar
  let jobVar = Args.jobVar
  let errorRoundVar = Args.errorRoundVar
  let roundVar = Args.roundVar

  let assertVars : varinfo array list ref = ref []

  let assertVarName (aid:int) (tid:int) : string = 
    "__startrek_Assert_t" ^ (string_of_int tid) ^ "_i" ^ (string_of_int aid)

  let getAssertVar (aid:int) (tid:int) = (List.nth !assertVars aid).(tid) 

  let getAssertLval (aid : int) (tid : int) (job:exp) : lval =

    Formatcil.cLval "a[%e:j]" 
      [("a", Fv (getAssertVar aid tid)); ("j", Fe job)]
      
  let mkAsssertWriteStmt (aid:int) (t:task) (job:exp) (arg:exp) : stmt = 
    (* mkSwitchStmt job 0 t.tjobs 
      (fun j -> 
	 [mkStmtOneInstr
	    (Set (getAssertLval aid t.tid (integer j), arg, rekloc))]) *)
    Formatcil.cStmt "a[%e:j] = %e:arg;"
      (fun n t -> assert false)
      rekloc
      [
	("a", Fv (getAssertVar aid t.tid)); 
	("j", Fe job);
	("arg", Fe arg);
      ]
		   
   
  (** Creates an assert variable. Creates an assert function. Returns
      varinfo of the created function*)
  let newAssertFun file : varinfo = 

    let createAssertFunc (id:int) : varinfo =

      let fd = findOrCreateFundec file
	("__startrek_assert_i" ^ (string_of_int id)) assertType in

      let switchStmt = 
	mkSwitchStmt (Lval (var taskVar)) 0 (List.length !tasks) 
	  (fun i -> 
	     [mkAsssertWriteStmt id (taskByTid i) (Lval (var jobVar)) zero]) in

      let b = Formatcil.cStmts 
	" if (hp != %e:last_hp) return;
          if (arg) return ; if (round < eround) eround = round; %s:switch"
	(fun n t -> assert false)
	rekloc
	[
	  ("arg", Fv (List.hd fd.sformals));
	  ("switch", Fs switchStmt);
	  ("eround", Fv errorRoundVar);
	  ("round", Fv roundVar);
	  ("hp", Fv Args.hyperPeriodVar);
	  ("last_hp", Fe (integer (pred !hyperPeriods)));
	]
      in	
      fd.svar.vinline <- true;
      fd.svar.vstorage <- Static;
      fd.sbody <- mkBlock b;
      (* XXX A hack to get a prototype *)
      Args.file.globals <- GVarDecl (fd.svar,rekloc) :: Args.file.globals;
      fd.svar
	
    in

    let id = List.length !assertVars in

    let entry = 
      List.map 
	(fun t -> 
	   (* create assert variables for task t *)
	   let atype = Formatcil.cType "%t:t [%e:j]"
	     [("t", Ft boolType); ("j", Fe (integer t.tjobs))] in
	     makeGlobalVar (assertVarName id t.tid) atype
	)
	!tasks 
    in

    assertVars := !assertVars @ [Array.of_list entry] ; 
    createAssertFunc id
      
  let registerGlobals file : unit = 
    let mkInit (size:int) = 
      let onelist = ref [] in 
      for i = 0 to size - 1 do 
	onelist := (Index(integer i, NoOffset), SingleInit one) :: !onelist
      done ;
      CompoundInit (boolType, !onelist)
    in

    List.iter 
      (fun entry -> 
	Array.iteri 
	  (fun (tid:int) (av : varinfo) -> 
	     let task = taskByTid tid in
	      file.globals <- 
		GVar (av, {init = Some (mkInit task.tjobs)}, rekloc) :: 
		file.globals
	  )
	  entry
      ) 
      !assertVars
  

  let map  (fn : lval -> 'a) : 'a list =
    (* fold over all jobs *)
    let rec fold_job v j max res = 
      if (j = max) then res
      else
	let lv = 
	  Formatcil.cLval "v[%e:j]" [("v", Fv v); ("j", Fe (integer j))] in
	  fold_job v (j + 1) max ((fn lv) :: res)
    in
      
    (* fold over all assert variables *)
    let rec fold_entry b i res =
      if (i == Array.length b) then res 
      else
	let x = fold_job b.(i) 0 (taskByTid i).tjobs res in
	  fold_entry b (i + 1) x
    in
      
      (* fold over all assert vars entries *)
      List.fold_left (fun r e -> fold_entry e 0 r) [] !assertVars
end


module ArraySchedVarManager (Args : ARGS) =
struct
  (*********************************************************************)
  (* Managing job scheduling variables  *)
  (*********************************************************************)
  type schedvars = { sv: varinfo; ev: varinfo }

  let jvtype = Args.roundVar.vtype

  let jobStartVarName (tid : int)  : string = 
    "__startrek_start_t" ^ (string_of_int tid) 
  let jobEndVarName (tid : int) : string = 
    "__startrek_end_t" ^ (string_of_int tid) 

  let schedVars : schedvars array =  
    Array.init (numOfTasks ())
      (fun tid -> 
	 let atype = Formatcil.cType "%t:T[%e:j]"
	   [
	     ("T", Ft jvtype);
	     ("j", Fe (integer (taskByTid tid).tjobs))
	   ] in
	   {
	     sv =  makeGlobalVar (jobStartVarName tid) atype;
	     ev = makeGlobalVar (jobEndVarName tid) atype;
	   }
      )

    
  let taskSchedVars (tid:int) : schedvars = schedVars.(tid)

  let jobStartLval (tid:int) (job:int) : lval = 
    Formatcil.cLval "sv[%e:job]" 
      [("sv", Fv (taskSchedVars tid).sv);
       ("job", Fe (integer job))
      ]

  let jobEndLval (tid:int) (job:int) : lval = 
    Formatcil.cLval "ev[%e:job]" 
      [("ev", Fv (taskSchedVars tid).ev);
       ("job", Fe (integer job))
      ]

    
  let mkReadJobEndStmt (out:lval) (t:task) (se:exp) : stmt = 
    Formatcil.cStmt "%l:out = ev[%e:se];"
      (fun n t -> assert false)
      rekloc
      [
	("out", Fl out);
	("ev", Fv (taskSchedVars t.tid).ev);
	("se", Fe se);
      ]

  let registerGlobals f : unit = 
    Array.iter 
      (fun e -> 
	 f.globals <- 
	   GVar (e.sv, {init = None}, rekloc) ::
	   GVar (e.ev, {init = None}, rekloc) ::
	   f.globals)
      schedVars
end


module NoArraySchedVarManager (Args : ARGS) =
struct
  (*********************************************************************)
  (* Managing job scheduling variables  *)
  (*********************************************************************)
  type schedvars = { sv: varinfo array; ev: varinfo array }

  let jobVarType = Args.roundVar.vtype
  let jobStartVarName (tid : int) (job : int) : string = 
    "__startrek_start_t" ^ (string_of_int tid) ^ "_j" ^ (string_of_int job)
  let jobEndVarName (tid : int) (job:int) : string = 
    "__startrek_end_t" ^ (string_of_int tid) ^ "_j" ^ (string_of_int job)

  let schedVars : schedvars array =  
    Array.init (numOfTasks ())
      (fun tid -> 
	 let jobs = (taskByTid tid).tjobs in 
	   {
	     sv = Array.init jobs 
	       (fun j -> makeGlobalVar (jobStartVarName tid j) jobVarType);
	     ev = Array.init jobs 
	       (fun j -> makeGlobalVar (jobEndVarName tid j) jobVarType)
	   }
      )

    
  let taskSchedVars (tid:int) : schedvars = schedVars.(tid)

  let jobStartLval (tid:int) (job:int) : lval = var (taskSchedVars tid).sv.(job)

  let jobEndLval (tid:int) (job:int) : lval = var (taskSchedVars tid).ev.(job)
    
  let mkReadJobEndStmt (out:lval) (t:task) (se:exp) : stmt = 
    mkSwitchStmt se 0 t.tjobs
      (fun i -> 
	 [mkStmtOneInstr 
	    (Set (out, Lval (jobEndLval t.tid i), rekloc))]
      )


  let registerGlobals f : unit = 
    Array.iter 
      (fun e-> 
	 Array.iter 
	   (fun s -> f.globals <- 
	      GVar (s, {init = None}, rekloc) :: f.globals)
	   e.sv ;
	 
	 Array.iter 
	   (fun s -> f.globals <- 
	      GVar (s, {init = None}, rekloc) :: f.globals)
	   e.ev
      )
      schedVars
end

(** Real-Time CBA Slice *)
module RekCba 
  (Args : ARGS) 
  (Svm :  SHARED_VAR_MANAGER) 
  (Sched : SCHED_VAR_MANAGER)
  (Asrt : ASSERT_VAR_MANAGER) 
  
  = 
struct


  (** map from varinfo to corresponding read function *)
  let rfnMap : fundec VM.t ref = ref VM.empty

  (** map from varinfo to corresponding write function *)
  let wfnMap : fundec VM.t ref = ref VM.empty


  let getReadFn v = VM.find v !rfnMap 
  let setReadFn v fd : unit =  rfnMap := VM.add v fd !rfnMap

  let setWriteFn v fd = wfnMap := VM.add v fd !wfnMap
  let getWriteFn v = VM.find v !wfnMap
    

  (*********************************************************************)
  (* module-global constants *)
  (*********************************************************************)

  (*********************************************************************)
  (* Arguments *)
  (*********************************************************************)
  let taskVar = Args.taskVar
  let jobVar = Args.jobVar
  let jobEndVar = Args.jobEndVar
  let errorRoundVar = Args.errorRoundVar
  let roundVar = Args.roundVar
  let lockVar = Args.lockVar
  let isfirstcsVar = Args.isfirstcsVar
  let hyperPeriodVar = Args.hyperPeriodVar

  (*********************************************************************)
  (* Common functions and types  *)
  (*********************************************************************)

  let csType = TFun (boolType,Some [],false,[])
  let assumeFn = findOrCreateFunc Args.file "__CPROVER_assume" assumeType
  let assertFn = findOrCreateFunc Args.file "assert" assertType 

  (*********************************************************************)
  (* Names  *)
  (*********************************************************************)
  let initSharedFnName = "__startrek_init_shared"
  let initGlobalsFnName = "__startrek_init_globals"
  let scheduleJobsFnName = "__startrek_schedule_jobs"
  let checkAssumptionsFnName = "__startrek_check_assumptions"
  let userFinalFnName = "__startrek_user_final"
  let userInitialFnName = "__startrek_user_init"
  let checkAssertFnName = "__startrek_check_assertions"
  let mainFnName = "__main"
  let realMainFnName = "main"

  let sharedReadPrefix = "__startrek_shared_read_"
  let sharedWritePrefix = "__startrek_shared_write_"


(** returns a statement that assumes x < y ===> y > z **)
  let notBetweenStmt (x:lval) (y:lval) (z:lval) : stmt =
    (* x < y ==> y > z *)
    Formatcil.cStmt 
      "if (%l:x < %l:y)  assume (%l:y > %l:z); "
      (fun n t -> assert false)
      rekloc
      [
	("x", Fl x);
	("y", Fl y);
	("z", Fl z);
	("assume", Fv assumeFn);
      ]

      
  (** returns a statement that assumes x < y ===> y > z **)
      
  let notBetweenStmt2 (x:lval) (y:lval) (z:lval) : stmt =
    (* x < y ==> y > z *)
    (* __CPROVER_assume (x >= y || (z < y)); *)

    let xGey = BinOp (Ge, Lval x, Lval y, boolType) and
	zLty = BinOp (Lt, Lval z, Lval y, boolType) in
    let disj = BinOp (LOr, xGey, zLty, boolType) in 
    mkStmtOneInstr (Call (None, Lval (var assumeFn), [disj], rekloc))
      
    
      

  (*********************************************************************)
  (* Read/Write flags for Partial-Order Reduction  *)
  (*********************************************************************)

  (** map from shared variables to read flag variables *)
  let readFlags : varinfo VM.t ref = ref VM.empty

  (** map from shared variables to write flag variables *)
  let writeFlags : varinfo VM.t ref = ref VM.empty


  (** create read/write flags for every shared variable *)
  let createRWFlags unit : unit = 
    
    let readFlagName v = "__startrek_rflag_" ^ v.vname in
    let writeFlagName v = "__startrek_wflag_" ^ v.vname in

    (* creaete flag variables for shared variables *)
    let flags = List.fold_left 
      (fun res g -> 
	match g with 
	  | GVar (v, _, _) when Svm.mem v -> 
	    (* found a shared variable *)
	    let readf = makeGlobalVar (readFlagName v) boolType and
		writef = makeGlobalVar (writeFlagName v) boolType in
	    readFlags := VM.add v readf !readFlags;
	    Svm.makeSharedVar readf;

	    writeFlags := VM.add v writef !writeFlags;
	    Svm.makeSharedVar writef; 

	    GVar (readf, {init = None}, rekloc) ::
	      GVar (writef, {init = None}, rekloc) :: res
	  | _ -> res
      )
      []
      Args.file.globals 
    in
    
    (* add flag variables to globals *)
    Args.file.globals <- flags @ Args.file.globals


  let globalTypes = ref []
  let globalVars = ref []

  let pushGlobal g : unit = 
    pushGlobal g ~types:globalTypes ~variables:globalVars

  let pushGlobals gl = List.iter pushGlobal gl

  (*********************************************************************)
  (* Non-deterministic values functions  *)
  (*********************************************************************)

  let ndbFn = findOrCreateFunc Args.file "nondet_bool" ndbType

  (** Returns a varinfo of a function that returns a non-deterministic value 
      of a given type *)
  let nondetFn (typ:typ) : varinfo =
    let tyname = 
      match typ with
	  TInt (IChar, _) -> "char"
	| TInt (ISChar, _) -> "schar"
	| TInt (IUChar, _) -> "uchar"
	| TInt (IBool, _) -> "_Bool"
	| TInt (IInt, _) -> "int"
	| TInt (IUInt, _) -> "uint"
	| TInt (IShort, _) -> "short"
	| TInt (IUShort, _) -> "ushort"
	| TInt (ILong, _) -> "long"
	| TInt (IULong, _) -> "ulong"
	| TInt (ILongLong, _) -> "long_long"
	| TInt (IULongLong, _) -> "ulong_long"
	| TFloat (FFloat, _) -> "float"
	| TFloat (FDouble, _) -> "double"
	| TFloat (FLongDouble, _) -> "long_double"
	| TNamed (tyinfo, _) -> tyinfo.tname
	| _ -> invalid_arg "Unsupported type"
    in
    findOrCreateFunc 
      Args.file 
      ("nondet_" ^ tyname) 
      (TFun (typ, Some [], false, []))





  (*********************************************************************)
  (* Context Switch Functions  *)
  (*********************************************************************)
      
  let csFnName (t:task) = "__startrek_cs_t" ^ (string_of_int t.tid)

  let getCsFn (t:task) : varinfo = 
    findOrCreateFunc Args.file (csFnName t) csType

  (*********************************************************************)
  (*** creating  context switch functions **)
  (*********************************************************************)
  let createCS (f:file) =


    (*create CS for a task *)
    let createCSTask (t:task) : unit =
      let fd = findOrCreateFundec f (csFnName t) csType in
	
      (* the body of context switch *)
      let body = 
	if isTopTask t then 
	  [mkStmt (Return (Some zero, rekloc))]
	else
	  begin 
            let stmts = ref [] in
	      (* for each higher priority task *)
              for tid2 = t.tid + 1 to numOfTasks() - 2 do 
		let t2 = taskByTid tid2 in
		  (* for each job of this higher task *)
		  for j2 = 0 to pred t2.tjobs do 
		    let 
			s2 = Sched.jobStartLval t2.tid j2 and
			e2 = Sched.jobEndLval t2.tid j2 and
			rv = var roundVar in
		    stmts := !stmts @ [notBetweenStmt s2 rv e2]
		  done;
              done;
	    Formatcil.cStmts 
	      " if (first) { first = 0; }

                if (lock) return 0;
                %t:Bool c = ndb (); if (c) return 0;
                
                %t:rty o = round; round = nd(); 
                assume (round > o); assume (round <= ej); 

                if (round != ej) { %S:w } return 1;
              "
	      (fun n t -> makeTempVar fd ~name:n t)
	      rekloc
	      [
		("rty", Ft roundVar.vtype);
		("round", Fv roundVar);
		("nd", Fv (nondetFn roundVar.vtype));
		("Bool", Ft boolType);
		("ndb", Fv ndbFn);
		("assume", Fv assumeFn);
		("lock", Fv lockVar);
		("first", Fv isfirstcsVar);
		("ej", Fv jobEndVar);
		("w", FS (if !nopriorcs then [] else !stmts))
	      ] 
	  end
      in
      fd.sbody.bstmts <- body @ fd.sbody.bstmts ; 
      fd.svar.vinline <- true;
      fd.svar.vstorage <- Static;
      Args.file.globals <- 
	GVarDecl (fd.svar, rekloc) :: Args.file.globals 
    in
    (* create context-switch for each task *)
    List.iter createCSTask !tasks


  (*********************************************************************)
  (*** shared variable read/write functions **)
  (*********************************************************************)

  let getSharedReadFn (v:varinfo) = 

    let readFnName (v:varinfo) = 
      "__startrek_read_" ^ v.vname in
    
    let makeSwitchRead readFn = 
      let cvar = makeTempVar readFn ~name:"c" boolType in
      let oldRound = makeTempVar readFn ~name:"or" roundVar.vtype in
      (* list of statements *)
      let slist = 
	
	let fn (t:task) (res:stmt list) : stmt list = 
	  let poblock = 
	    if !dopo then
	      begin
		let wf = VM.find v !writeFlags and
		    rf = VM.find v !readFlags in
		let svwf = Svm.getSharedVar wf and
		    svrf = Svm.getSharedVar rf in

		let wfVar = makeTempVar readFn ~name:"wf" boolType in
		
		let orndE = (Lval (var oldRound)) and 
		    rndE = (Lval (var roundVar)) in 
		
		Formatcil.cStmts
		  "if (c) { %s:owf %s:wfread assume (wf); } %s:rf"
		  (fun n t -> assert false)
		  rekloc
		  [
		    ("c", Fv cvar);
		    ("owf", Fs (Svm.mkSharedWriteStmt svwf orndE zero));
		    ("wfread", Fs (Svm.mkSharedReadStmt (var wfVar) svwf rndE));
		    ("wf", Fv wfVar);
		    ("rf", Fs (Svm.mkSharedWriteStmt svrf rndE one));
		    ("assume", Fv assumeFn)
		  ]
	      end
	    else [] 
	  in
	  let caseBody = 
	    mkStmt 
	      (Block 
		 (mkBlock
		    (Formatcil.cStmts 
		       "or = r; c = cs(); { %S:po }"
		       (fun n t -> assert false)
		       rekloc
		       [
			 ("cs", Fv (getCsFn t));
			 ("or", Fv oldRound);
			 ("r", Fv roundVar);
			 ("c", Fv cvar);
			 ("po", FS poblock)
		       ]
		    ))) 
	  in
	  
	  (* add a case label *)
	  caseBody.labels <- 
	    (Case (integer t.tid, rekloc)) :: caseBody.labels ;
	  caseBody :: mkStmt (Break rekloc) :: res
	    
	in
	List.fold_right fn !tasks [] 
      in
      mkStmt (Switch (Lval (var Args.taskVar), 
		      mkBlock slist, slist, rekloc))
    in
    try (getReadFn v).svar with 
      (* no read function found, create one *)
      | Not_found -> 
	begin
	  let readFnType = TFun (v.vtype, Some [], false, []) in
	  let readFn = emptyFunction (readFnName v) in 
	  
	  setFunctionTypeMakeFormals readFn readFnType ; 
	  readFn.svar.vinline <- true ;
	  readFn.svar.vstorage <- Static ;
	  
	  let rvar = makeTempVar readFn ~name:"r" v.vtype in
	  let sv = Svm.getSharedVar v in
	  let rstmt = 
	    Svm.mkSharedReadStmt (var rvar) sv (Lval (var roundVar)) in
	  let retStmt = mkStmt (Return (Some (Lval (var rvar)), rekloc)) in
	  readFn.sbody <- mkBlock ([makeSwitchRead readFn ; rstmt ; retStmt]);
	  setReadFn v readFn ; 
	  readFn.svar
	end
	  
  let getSharedWriteFn (v:varinfo)  = 

    let writeFnName (v:varinfo) = 
      "__startrek_write_" ^ v.vname in

    let makeSwitchWrite writeFn = 
      
      let cvar = makeTempVar writeFn ~name:"c" boolType in
      let oldRound = makeTempVar writeFn ~name:"or" roundVar.vtype in

      let slist =
	let fn (t:task) (l:stmt list) : stmt list = 

	  let poblock = 
	    if !dopo then 
	      let wf = VM.find v !writeFlags and
		  rf = VM.find v !readFlags in
	      let svwf = Svm.getSharedVar wf  and
		  svrf = Svm.getSharedVar rf  in
	      
	      let wfVar = makeTempVar writeFn ~name:"wf" boolType and
		  rfVar = makeTempVar writeFn ~name:"rf" boolType in
	      
	      let orndE = (Lval (var oldRound)) and 
		  rndE = (Lval (var roundVar)) in 

	      Formatcil.cStmts
		"if (c)  
                    { %s:orf %s:owf %s:readrf %s:readwf assume (wf | rf); } 
                  %s:writewf" 
		(fun n t -> assert false)
		rekloc
		[
		  ("c", Fv cvar);
		  ("orf", Fs (Svm.mkSharedWriteStmt svrf orndE zero));
		  ("owf", Fs (Svm.mkSharedWriteStmt svwf orndE zero));
		  ("readrf", Fs (Svm.mkSharedReadStmt (var rfVar) svrf rndE));
		  ("readwf", Fs (Svm.mkSharedReadStmt (var wfVar) svwf rndE));
		  ("wf", Fv wfVar);
		  ("rf", Fv rfVar);
		  ("writewf", Fs (Svm.mkSharedWriteStmt svwf rndE one));
		  ("assume", Fv assumeFn);
		]	      
	    else
	      []
	  in
	  let caseBody = 
	    mkStmt 
	      (Block (
		mkBlock 
		  (Formatcil.cStmts
                     "or = r; c = cs(); { %S:po } "
		     (fun n t -> makeTempVar writeFn ~name:n t)
		     rekloc
		     [
		       ("or", Fv oldRound);
		       ("r", Fv roundVar);
		       ("c", Fv cvar);
		       ("cs", Fv (getCsFn t));
		       ("po", FS poblock)
		     ])))
	  in
	  caseBody.labels <- (Case (integer t.tid, rekloc)) :: 
	    caseBody.labels;
	  caseBody :: mkStmt (Break rekloc) :: l 
	in
	List.fold_right fn !tasks []
      in
      mkStmt (Switch (Lval (var Args.taskVar), 
		      mkBlock slist, slist, rekloc)) 
    in

    try (getWriteFn v).svar with
      (* no write function found, create one *)
      | Not_found -> 
	begin
	  let writeFnType = 
	    TFun (voidType,Some [("arg",v.vtype,[])],false,[]) in
	  let writeFn = emptyFunction (writeFnName v) in 
	  
	  setFunctionTypeMakeFormals writeFn writeFnType ; 
	  writeFn.svar.vinline <- true ;
	  writeFn.svar.vstorage <- Static ;


	  let arg = Lval (var (List.hd writeFn.sformals)) in
	  let sv = Svm.getSharedVar v in
	  let wstmt = Svm.mkSharedWriteStmt sv (Lval (var roundVar)) arg in
	  writeFn.sbody <- mkBlock ([makeSwitchWrite writeFn; wstmt]) ;
	  
	  setWriteFn v writeFn ; 
	  writeFn.svar
	end

  (*********************************************************************)
  (* Visitors  *)
  (*********************************************************************)
    
  (** Finds and registers shared variables *)
  class sharedVariableFinder = object (self)
    inherit nopCilVisitor as super
      
    method vinst (ins : instr) = begin
      match ins with
          (* read a global variable -- must be of the form x = R(&v) *)
          Call(Some(x),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when 
	      Util.hasPrefix sharedReadPrefix fn && 
		(not (Svm.mem v)) -> 
		  Svm.makeSharedVar v ; SkipChildren

		    
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when 
	      Util.hasPrefix sharedWritePrefix fn && 
		(not (Svm.mem v)) -> 
	  Svm.makeSharedVar v ; SkipChildren
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren

    end
  end
	  
  (** rewrites accesses to shared variables and lock functions *)
  class globalReplacer = 
  object(self) 
    inherit nopCilVisitor as super
      
    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
	  (* read a global variable -- must be of the form x = R(&v) *)
	  Call(Some(Var x, NoOffset),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when Util.hasPrefix sharedReadPrefix fn -> 
	      x.vtype <- v.vtype;
	      ChangeTo 
		[Call (Some(Var x, NoOffset), 
		       Lval (var (getSharedReadFn v)), [], loc)]
		
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when Util.hasPrefix sharedWritePrefix fn ->
	  ChangeTo 
	    ([Call (None, Lval (var (getSharedWriteFn v)), [e], loc)])

	| Call (None, Lval (Var {vname = fn}, NoOffset), [], loc) 
	    when fn = cpuLockFnName ->
	  ChangeTo ([Set (var lockVar, one, loc)])
	| Call (None, Lval (Var {vname = fn}, NoOffset), [], loc) 
	    when fn = cpuUnlockFnName ->
	  ChangeTo ([Set (var lockVar, zero, loc)])
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end



  (**  pulls out assertions from task bodies *)
  class assertReplacer = object(self) inherit nopCilVisitor as super
            
    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
          (* read a global variable -- must be of the form x = R(&v) *)
          Call(None,
	       Lval(Var {vname = fn}, NoOffset), [e], loc)  
	    when fn = "assert" -> 
	      ChangeTo 
		[Call (None, 
		       Lval (var (Asrt.newAssertFun Args.file)), 
		       [e], 
		       loc)]
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end



  (** replaces accesses to shared variables in user initializer *)
  class initialReplacer = 
  object(self) 
    inherit nopCilVisitor as super
      
    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
	  (* read a global variable -- must be of the form x = R(&v) *)
	  Call(Some(Var x, NoOffset),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when Util.hasPrefix sharedReadPrefix fn -> 
	      let sv = Svm.getSharedVar v in
	      x.vtype <- v.vtype;
	      ChangeTo [Set (var x, 
			     Lval (Svm.getSharedLval sv 0), rekloc)]

		
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when Util.hasPrefix sharedWritePrefix fn ->
	  let sv = Svm.getSharedVar v in
	  ChangeTo [Set (Svm.getSharedLval sv  0, e, rekloc)]
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end

  (** replaces accesses to shared variables in user finalizer *)
  class finalReplacer = 
  object(self) 
    inherit nopCilVisitor as super
      
    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
          (* read a global variable -- must be of the form x = R(&v) *)
          Call(Some(x),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when Util.hasPrefix sharedReadPrefix fn -> 
	      let sv = Svm.getSharedVar v in
	      ChangeTo 
		[Set (x, Lval (Svm.getSharedLval sv (pred !rounds)), 
			       rekloc)]
		
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when Util.hasPrefix sharedWritePrefix fn ->
	  let sv = Svm.getSharedVar v in
	  ChangeTo 
	    [Set (Svm.getSharedLval sv (pred !rounds), e, loc)]
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end


  (*********************************************************************)
  (* Creation of shared variables  *)
  (*********************************************************************)
 
  (* create variables for initial values *)
  let createInitVars (f:file) = 
    let svf = new sharedVariableFinder in
    iterGlobals 
      f 
      (fun g -> 
	match g with
	  | GFun(fd, _) -> fd.sbody <- visitCilBlock svf fd.sbody
	  | _ -> ()
      )
      
  (* create all global variables *)
  let createGlobalVars = createInitVars



  (*********************************************************************)
  (*** creating job scheduling function **)
  (*********************************************************************)



  (** create schedule_jobs  function **)
  let createScheduleJobs unit : unit =
    let ftype =  TFun (voidType,Some [],false,[]) in
    let fd = findOrCreateFundec Args.file scheduleJobsFnName ftype in
    fd.svar.vinline <- true;

    (** Returns nesting constraints for job j1 of a low priority task
	t and job j2 of a higher priority task j2 **)
    (* XXX This is an older version of nesting constraints. Does not
       easily support preemption bounds *)
    (* let nestingConstraints2 t j1 t2 j2 = *)
    (*   let *)
    (* 	  s1 = Sched.jobStartLval t.tid j1 and *)
    (* 	  s2 = Sched.jobStartLval t2.tid j2 and *)
    (* 	  e1 = Sched.jobEndLval t.tid j1 and *)
    (* 	  e2 = Sched.jobEndLval t2.tid j2 in *)
    (*   [notBetweenStmt s1 s2 e1 ; notBetweenStmt s1 e2 e1] *)
    (* in *)

    (** Returns nesting constraints for job j2 of a low priority task
	t2 and job j1 of a higher priority task t **)
    let nestingConstraints t j1 t2 j2 =
      (* ensures that j1 occurs either before, after, or inside j2 *)
      (* easily supports preemption bounds *)
      let
    	  s1 = Sched.jobStartLval t.tid j1 and
    	  s2 = Sched.jobStartLval t2.tid j2 and
    	  e1 = Sched.jobEndLval t.tid j1 and
    	  e2 = Sched.jobEndLval t2.tid j2 in
      
      let pb = 
	if !nopbounds then 0 else 
	  !taskArray.(t2.tid).pbounds.(t.tid - t2.tid - 1) in
      let pbound = 
	if !nopbounds || j1 < pb  then [] 
	else 
	  Formatcil.cStmts 
	    "assume (%l:je < %l:s2);"
	    (fun n t -> assert false)
	    rekloc
	    [
	      ("assume", Fv assumeFn);
	      ("s2", Fl s2);
	      ("je", Fl (Sched.jobEndLval t.tid (j1 - pb)));
	    ]	  
      in 

      Formatcil.cStmts
    	"if (%l:s2 <= %l:e1)
           if (%l:s1 <= %l:e2)
             {
               assume (%l:s2 <= %l:s1);
               assume (%l:e1 < %l:e2);
               %S:pbound

             }
        "
    	(fun n t -> assert false)
    	rekloc
    	[
    	  ("s1", Fl s1);
    	  ("s2", Fl s2);
    	  ("e1", Fl e1);
    	  ("e2", Fl e2);
	  ("pbound", FS pbound);
    	  ("assume", Fv assumeFn);
    	]
      in
      
    (* create the body of the function for a specific task *)
    let perTask (t:task) : unit =
      (* initialize start and end rounds *)
      for i = 0 to pred t.tjobs do 
        (* the top priority task starts and ends in the same round *)
        if isTopTask t then
          fd.sbody.bstmts <- fd.sbody.bstmts @
	    Formatcil.cStmts
	    "%l:sr = ndc (); %l:er = %l:sr;"
	    (fun n t -> assert false)
	    rekloc
	    [
	      ("sr", Fl (Sched.jobStartLval t.tid i));
	      ("er", Fl (Sched.jobEndLval t.tid i));
	      ("ndc", Fv (nondetFn ucharType))
	    ]

        (* other tasks can start end anywhere *)
        else 
          fd.sbody.bstmts <- fd.sbody.bstmts @
	    Formatcil.cStmts
	    "%l:sr = ndc (); %l:er = ndc();"
	    (fun n t -> makeTempVar fd ~name:n t)
	    rekloc
	    [
	      ("sr", Fl (Sched.jobStartLval t.tid i));
	      ("er", Fl (Sched.jobEndLval t.tid i));
	      ("ndc", Fv (nondetFn ucharType))
	    ]
      done;

      (* first jobs starts after round 0, last job finishes by the last round *)
      fd.sbody.bstmts <- fd.sbody.bstmts @
	Formatcil.cStmts
	"assume (0 <= %l:sr); assume (%l:er <= %e:lim );"
	(fun n t -> assert false)
	rekloc
	[
	  ("sr", Fl (Sched.jobStartLval t.tid 0));
	  ("er", Fl (Sched.jobEndLval t.tid (pred t.tjobs)));
	  ("lim", Fe (integer (pred !rounds)));
          ("assume", Fv assumeFn)
	];

      (* 
	 Each job starts no later than it ends.  This is asserted for
	 all but top task since it always starts and ends in the same
	 round 
      *)
      if not (isTopTask t) then
        for i = 0 to pred t.tjobs do 
          fd.sbody.bstmts <- fd.sbody.bstmts @
	    Formatcil.cStmts
	    "assume ( %l:sr <= %l:er );"
	    (fun n t -> assert false)
	    rekloc
	    [
	      ("sr", Fl (Sched.jobStartLval t.tid i));
	      ("er", Fl (Sched.jobEndLval t.tid i));
              ("assume", Fv assumeFn)
	    ]
        done;
      
      (* each job ends (strictly) before the next job starts *)
      for i = 0 to t.tjobs - 2 do 
        fd.sbody.bstmts <- fd.sbody.bstmts @
	  Formatcil.cStmts
	  "assume ( %l:er <= %l:sr - 1 );"
	  (fun n t -> assert false)
	  rekloc
	  [
	    ("sr", Fl (Sched.jobStartLval t.tid (i + 1)));
	    ("er", Fl (Sched.jobEndLval t.tid i));
            ("assume", Fv assumeFn)
	  ]
      done;

      (* constraints to ensure proper nesting *)
      (* no constraints if
	 (a) nesting constraints are turned off OR
	 (b) t is a topTask AND pbounds are disabled
      *)
      if  !nesting && not ( (isTopTask t) && !nopbounds) then 
        (* for every job of this task *)
        for j1 = 0 to pred t.tjobs  do 
          (* for each task of lower priority *)
          for tid2 = 0 to pred t.tid  do 
            let t2 = taskByTid tid2 in
            (* for each job of this higher task *)
            for j2 = 0 to pred t2.tjobs do 
	      fd.sbody.bstmts <- fd.sbody.bstmts @
		(nestingConstraints t j1 t2 j2)
            done;
          done;
        done;
      
    in
    List.iter perTask !tasks


  (*********************************************************************)
  (*** creating the init function **)
  (*********************************************************************)
    
  (*
    vi3[j][s] = ndVi (); vi2[j][s] = vi3[j][s]
  *)

  (** create a init function *)
  let createInit (f:file) =
    let initType = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f initGlobalsFnName initType in
    fd.svar.vinline <- true;

    (* for every shared variable *)
    Svm.iterShared
      (fun v sv ->
	(* for every round, except the first one *)
	for r = pred !rounds downto 1 do
	  fd.sbody.bstmts <-
	    Formatcil.cStmts
	    "%l:sc = nd () ; %l:xs = %l:sc;"
	    (fun n t -> makeTempVar fd ~name:n t)
	    rekloc
	    [("sc", Fl (Svm.getSharedInitLval sv r)); 
	     ("xs", Fl (Svm.getSharedLval sv r));
	     ("nd", Fv (nondetFn v.vtype))
	    ] @
	    fd.sbody.bstmts
	done)

	
	
  (*********************************************************************)
  (*** creating the tasks **)
  (*********************************************************************)

  (** rewrites all accesses to shared variables, lock functions, and
      asserts *)
  let createTasks (f:file) = 
    let glblr = new globalReplacer and
	asrtr = new assertReplacer in
      iterGlobals 
	f 
	(fun g -> 
	  match g with 
	    | GFun(fd, _) -> 
	      if (not (fd.svar.vname = userFinalFnName) &&
		    not (fd.svar.vname = userInitialFnName)) then
		fd.sbody <- visitCilBlock asrtr (visitCilBlock glblr fd.sbody)
	    | _ -> ()
	)
	

  (*********************************************************************)
  (*** creating the check assuptions function **)
  (*********************************************************************)
  let createCheckAssumption (f:file) = 
    let checkerType = TFun (voidType,Some [],false,[]) in
    let fd = findOrCreateFundec f checkAssumptionsFnName checkerType in
    fd.svar.vinline <- true;

    (* for every shared variable *)
    Svm.iterShared
      (fun v sv ->
	(* for every round, except the first one *)
	for r = 0 to !rounds - 2 do
	  fd.sbody.bstmts <-
	    Formatcil.cStmts
	    "assume (%l:sc == %l:xs);"
	    (fun n t -> makeTempVar fd ~name:n t)
	    rekloc
	    [("sc", Fl (Svm.getSharedInitLval sv (succ r))); 
	     ("xs", Fl (Svm.getSharedLval sv r));
	     ("assume", Fv assumeFn)
	    ] @
	    fd.sbody.bstmts
	done)
    
	
  (*********************************************************************)
  (*** rewritting user-supplied init and final functions **)
  (*********************************************************************)

  let createUserInitial (f:file) =
    let finalType = TFun (voidType, Some [], false,[]) in
    let fd = findOrCreateFundec f userInitialFnName finalType in
    fd.svar.vinline <- true;

    
    fd.sbody <-  visitCilBlock (new initialReplacer) fd.sbody	

  (* create a final function *)
  let createUserFinal (f:file) =
    let finalType = TFun (voidType, Some [], false,[]) in
    let fd = findOrCreateFundec f userFinalFnName finalType in
    fd.svar.vinline <- true;
    fd.sbody <-  visitCilBlock (new finalReplacer) fd.sbody


  (*********************************************************************)
  (*** creating check assertions function **)
  (*********************************************************************)
  let createCheckAssertions (f:file) = 
    let atype = TFun (voidType, Some [], false,[]) in
      
    let body = Asrt.map (fun lv -> Call (None, Lval (var assertFn), [Lval lv], 
					 rekloc)) in
    let fd = findOrCreateFundec f checkAssertFnName atype in
    fd.svar.vinline <- true;
    fd.sbody <- mkBlock ([mkStmt (Instr body)])



  (*********************************************************************)
  (*** creating the main function **)
  (*********************************************************************)
  (** create a main function *)	
  let createMain (f: file) = 
    
    let ftype = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f mainFnName ftype in
    let loc = rekloc in

    (* a choice variable *)
    let cvar = makeTempVar fd ~name:"c" (TInt (IBool, [])) in

    let taskCalls = 
      List.map 
	(fun t -> 
	  let taskFn = (oget (!taskArray.(t.tid).entryPt)).svar in 
	  let body = ref [] in
	  
	  for j = 0 to t.tjobs - 1 do
	    body := !body @
	      Formatcil.cStmts
	      "
               job = %e:j; isfst = 1; round = %l:sr; 
                if (round < eround) {
                  jobend = %l:er; 
                  cvar = task (); 
                  lock = 0; 
                  assume (round == jobend);
                }
              "
	      (fun n t -> assert false)
	      rekloc
	      [
		("eround", Fv errorRoundVar);
		("job", Fv jobVar);
		("j", Fe (integer j));
		("lock", Fv lockVar);
		("cvar", Fv cvar);
		("task", Fv taskFn);
		("isfst", Fv isfirstcsVar);
		("round", Fv roundVar);
		("sr", Fl (Sched.jobStartLval t.tid j));
		("jobend", Fv jobEndVar);
		("er", Fl (Sched.jobEndLval t.tid j));
		("assume", Fv assumeFn);
	      ]
	  done;
	  Formatcil.cStmt 
	    "{ tid = %e:id; %S:body }"
	    (fun n t -> assert false)
	    rekloc
	    [
	      ("tid", Fv taskVar);
	      ("id", Fe (integer t.tid));
	      ("body", FS !body)
	    ]
	)
	!tasks 
    in
    
    fd.sbody <- mkBlock 
      (Formatcil.cStmts
	 " eround = %e:max;
           sjobs(); init() ; 
           { %S:tasks } 
           round = %e:max;
           checker(); checkassert (); if (hp == %e:last_hp) { final (); }"
	 (fun n t -> makeTempVar fd ~name:n t)
	 loc
	 [
	   ("eround", Fv errorRoundVar);
	   ("round", Fv roundVar);
	   ("max", Fe (integer !rounds));
	   ("sjobs", Fv (findOrCreateFunc f scheduleJobsFnName ftype));
	   ("init", Fv (findOrCreateFunc f initGlobalsFnName ftype));
	   ("checker", Fv (findOrCreateFunc f checkAssumptionsFnName ftype)); 
	   ("final", Fv (findOrCreateFunc f userFinalFnName ftype));
	   ("checkassert", Fv (findOrCreateFunc f checkAssertFnName ftype));
	   ("tasks", FS taskCalls);
	   ("hp", Fv hyperPeriodVar);
	   ("last_hp", Fe (integer (pred !hyperPeriods)));
	 ]) 


  let createRealMain (f:file) = 
    let ftype = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f realMainFnName ftype in

    let new_period = 
      Svm.foldShared 
	(fun v sv r -> 
	  let lhs = Svm.getSharedLval sv 0 in 
	  let rhs = Svm.getSharedLval sv (pred !rounds) in 
	  (Set (lhs, Lval rhs, rekloc)) :: r
	) 
	[] 
    in
    
    let initSharedFn = 
      var (findOrCreateFunc Args.file initSharedFnName ftype) in
    let userInitFn = 
      var (findOrCreateFunc Args.file userInitialFnName ftype) in
      
    let prologue = 
      [
	Call (None, Lval initSharedFn, [], rekloc);
	Call (None, Lval userInitFn, [], rekloc)
      ]
    in
    let inst = ref prologue in

    let __mainfn = var (findOrCreateFunc Args.file mainFnName ftype) in

    for i = 1 to !hyperPeriods do
      inst := !inst @ 
	[ Set (var hyperPeriodVar, integer (pred i), rekloc); 
	  Call (None, Lval  __mainfn, [], rekloc) ] ;
      if (i != !hyperPeriods) then inst := !inst @ new_period
    done ;
    fd.sbody <- mkBlock [mkStmt (Instr !inst)]
      
      
  (*********************************************************************)
  (*** registering global variables **)
  (*********************************************************************)      
  (** register all globals created so far with the file *)
  let registerGlobals unit : unit = 
    (* registers prototypes and definitions of reader/writer
       functions *)

    let createRW map : unit = 
    
      (* register prototypes near variables they mask. This ensures
	 that all appropriate types have already been declared *)
      let fn res g = 
	let out = g :: res in 
	match g with 
	  | GVar (v, _, _) when VM.mem v map -> 
	    let fd = VM.find v map in
	    GVarDecl(fd.svar,rekloc) ::  out
	  | _ -> out
      in
      Args.file.globals <- List.fold_left fn [] (List.rev Args.file.globals);
      
      VM.iter 
	(fun v fd -> Args.file.globals <- 
	  Args.file.globals @ [GFun(fd, rekloc)]) map ;
      
    in

    
    
    (* register readers *)
    createRW !rfnMap ;
    (* register writers *)
    createRW !wfnMap ;

    Svm.registerGlobals Args.file; 
    Sched.registerGlobals Args.file;
    Asrt.registerGlobals Args.file
    
      
  let inline unit : unit =
    let inlineWhat (v:varinfo) : fundec option =
      findFundec Args.file v.vname
    in	
    iterGlobals Args.file 
      (fun g -> 
	match g with 
	  | GFun (fd, _) when fd.svar.vname = mainFnName -> 
	    let anyInlining = ref true in 
	    while !anyInlining do
	      anyInlining := false;
	      Inliner.doFunction fd inlineWhat anyInlining
	    done
	  | _ -> ()
      )

  (** map from hidden shared to what they are hiding *)
  let hiddenShared : varinfo VM.t ref = ref VM.empty 

  (** Rename shared variables to ensure that they cannot be used directly*)
  let hideShared unit : unit = 
    Args.file.globals <- 
      List.map 
      (fun g ->
	  match g with 
	      GVar (v, init, loc) when Svm.mem v -> 
		let u = copyVarinfo v ("__startrek_hidden_" ^ v.vname) in
		  hiddenShared := VM.add u v !hiddenShared; 
		  GVar (u, init, loc)
	    | _ -> g)
      Args.file.globals


    
  let createInitShared unit : unit = 
    let initType = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec Args.file initSharedFnName initType in
    fd.svar.vinline <- true;

    let body = VM.fold 
      (fun k v res -> 
	 let sv = Svm.getSharedVar v in 
	 let shared = Svm.getSharedLval sv 0  in
	   Set (shared, Lval (var k), rekloc) :: res
      )
      !hiddenShared 
      []
    in
      fd.sbody.bstmts <- [ mkStmt (Instr body) ]
	

  (** remove all globals that we don't like *)
  let cleanupGlobals unit : unit = 
    Args.file.globals <- 
      List.filter 
      (fun g -> 
	match g with 
	    (* remove all prototypes for functions used to mark shared
	       reads and writes*)
	    GVarDecl (v, _) -> 
	      (not (Util.hasPrefix sharedReadPrefix v.vname)) && 
		(not (Util.hasPrefix sharedWritePrefix v.vname))
	  | _ -> true
      ) 
      Args.file.globals


    
  (*********************************************************************)
  (*** Entry point **)
  (*********************************************************************)      

  (** run everything *)
  let run unit : unit = 
    (* create global variables *)
    createGlobalVars Args.file;
    if (!dopo) then createRWFlags ();
    (* schedule jobs *)
    createScheduleJobs ();
    (* create init function *)
    createInit Args.file;
    (* create context switch function *)
    createCS Args.file;
    (* create the tasks *)
    createTasks Args.file;
    (* create check assumptions function *)
    createCheckAssumption Args.file;
    (* create user initial function *)
    createUserInitial Args.file;
    (* create user final function *)
    createUserFinal Args.file;
    (* create check assertionsfunction *)
    createCheckAssertions Args.file; 
    (* create main function *)
    createMain Args.file; 
    createRealMain Args.file;
    cleanupGlobals ();
    (* create globals *)
    registerGlobals ();

    hideShared ();
    (* must hapen after the shared variables are hidden *)
    createInitShared ();

    if (!inlineAll) then inline (); 
end;;

(*********************************************************************)
(*** the Cil feature, where everything begins *)
(*********************************************************************)

(* command line option for first pass *)
let feature : featureDescr = { 
  fd_name = "rekcba";              
  fd_enabled = ref false;
  fd_description = "CBA for RTES";
  fd_extraopt = [];
  fd_doit = (
    function (f: file) -> 
      print_string ("Analyzing " ^ f.fileName ^ "\n");

      (* compute task specific info *)
      initTaskList f ;
      
      (* #rounds == total # of jobs *)
      rounds := List.fold_left (fun res t -> res + t.tjobs) 0 !tasks ;


      (* package arguments into a module *)
      let module A = struct 
	let file = f 
	let taskVar = makeGlobalVar "__startrek_task" ucharType
	let jobVar = makeGlobalVar "__startrek_job" ucharType
	let jobEndVar = makeGlobalVar "__startrek_job_end" ucharType
	let errorRoundVar = makeGlobalVar "__startrek_error_round" ucharType
	let roundVar = makeGlobalVar "__startrek_round" ucharType
	let lockVar = makeGlobalVar "__startrek_lock" boolType
	let isfirstcsVar = makeGlobalVar "__startrek_is_first_cs" boolType
	let hyperPeriodVar = makeGlobalVar "__startrek_hyper_period" ucharType
      end in

	(* register global variables created so far with the file *)
	f.globals <- 
	  (GVar (A.taskVar, {init = None}, rekloc))::
	  (GVar (A.jobVar, {init = None}, rekloc))::
	  (GVar (A.jobEndVar, {init = None}, rekloc))::
	  (GVar (A.errorRoundVar, {init = None}, rekloc))::
	  (GVar (A.roundVar, {init = None}, rekloc))::
	  (GVar (A.lockVar, {init = Some (makeZeroInit boolType)}, 
		 rekloc))::
	  (GVar (A.isfirstcsVar, {init = None}, rekloc))::
	  (GVar (A.hyperPeriodVar, {init = None}, rekloc))::
	  f.globals ;
	

      if !noarray then 
	let module M = 
	  RekCba 
	    (A) 
	    (NoArraySharedVarManager (A)) 
	    (NoArraySchedVarManager (A))
	    (ArrayAssertVarManager (A)) in
	  M.run ()
      else
	let module M = 
	  RekCba 
	    (A) 
	    (ArraySharedVarManager (A)) 
	    (ArraySchedVarManager (A))
	    (ArrayAssertVarManager (A)) in
	  M.run ()
  );
  fd_post_check = false
} 
