open Cil
open Rtcbautil
open Rtcbainput
open Rekoptions
open Rekargs
open Reksvm
open Rekhtasks

(** whether to do unbounded sequentialization -- default is false *)
let unboundedSeq = ref false

(*********************************************************************)
(* Signatures *)
(*********************************************************************)

module type ASSERT_VAR_MANAGER = sig
  val getAssertLval : int -> int -> lval
  val mkAssertWriteStmt : int -> task -> exp -> stmt 
  val newAssertFun : file -> varinfo
  val registerGlobals : file -> unit
  val map : (Cil.lval -> 'a) -> 'a list
  val map_task : (Cil.lval -> 'a) -> int -> 'a list
end

module type SCHED_VAR_MANAGER = sig
  type schedvars
  val jobStartLval :  int -> lval
  val jobEndLval :  int -> lval
  val jobMaxLval :  int -> lval
  val jobMinLval :  int -> lval
  val registerGlobals : file -> unit

end

(*********************************************************************)
(* Modules *)
(*********************************************************************)

(*********************************************************************)
(* Managing assert variables  *)
(*********************************************************************)
module AssertVarManager (Args : ARGS) =
struct
  let taskVar = Args.taskVar
  let roundVar = Args.roundVar

  (* assert variables. the list is indexed by assert ids; list
     entries are indexed by task id *)
  let assertVars : varinfo array list ref = ref []

  let assertVarName (aid:int) (tid:int) : string = 
    "__startrek_Assert_t" ^ (string_of_int tid) ^ "_i" ^ (string_of_int aid)

  let getAssertVar (aid:int) (tid:int) = (List.nth !assertVars aid).(tid) 
  let getAssertLval (aid : int) (tid : int) : lval = var (getAssertVar aid tid)
      
  let mkAssertWriteStmt (aid:int) (t:task) (arg:exp) : stmt = 
    Formatcil.cStmt "a = %e:arg;"
      (fun n t -> assert false)
      builtinLoc
      [
	("a", Fv (getAssertVar aid t.tid)); 
	("arg", Fe arg);
      ]
		   
   
  (** Creates an assert variable. Creates an assert function. Returns
      varinfo of the created function*)
  let newAssertFun file : varinfo = 

    let createAssertFunc (id:int) : varinfo =

      let fd = findOrCreateFundec file
	("__startrek_assert_i" ^ (string_of_int id)) assertType in

      let switchStmt = 
	mkSwitchStmt (Lval (var taskVar)) 0 (List.length !tasks) 
	  (fun i -> [mkAssertWriteStmt id (taskByTid i) zero]) in

      let b = Formatcil.cStmts 
	"if (arg) return ;  %s:switch"
	(fun n t -> assert false)
	builtinLoc
	[
	  ("arg", Fv (List.hd fd.sformals));
	  ("switch", Fs switchStmt);
	]
      in	
      fd.svar.vinline <- true;
      fd.svar.vstorage <- Static;
      fd.sbody <- mkBlock b;
      (* XXX A hack to get a prototype *)
      Args.file.globals <- GVarDecl (fd.svar,builtinLoc) :: Args.file.globals;
      fd.svar
    in

    let id = List.length !assertVars in

    (* create assert variables for task t *) 
    let entry =
      List.map 
	(fun t -> 
	  makeGlobalVar (assertVarName id t.tid) boolType
	)
	!tasks 
    in
    assertVars := !assertVars @ [Array.of_list entry] ; 
    createAssertFunc id
      
  let registerGlobals file : unit = 
    List.iter 
      (fun entry -> 
	Array.iteri 
	  (fun (tid:int) (av : varinfo) -> 
	    file.globals <- 
	      GVar (av, {init = Some (SingleInit one)}, builtinLoc) :: 
	      file.globals
	  )
	  entry
      ) 
      !assertVars
  

  let map  (fn : lval -> 'a) : 'a list =      

    (* fold over all tasks *)
    let rec fold_entry b i res =
      if (i == Array.length b) then res 
      else fold_entry b (i + 1) ( (fn (var b.(i))) :: res)
    in
    
      (* fold over all assert variable entries *)
      List.fold_left (fun r e -> fold_entry e 0 r) [] !assertVars

  let map_task (fn : lval -> 'a) (tid: int) : 'a list = 
    List.map (fun e -> fn (var e.(tid))) !assertVars
end

(*********************************************************************)
(* Managing job scheduling variables  *)
(*********************************************************************)
module ArraySchedVarManager (Args : ARGS) =
struct
  (* sv - start, ev - end, maxv - max, minv - min *) 
  type schedvars = { sv: varinfo; ev: varinfo; maxv: varinfo; minv: varinfo }

  let jvtype = Args.roundVar.vtype

  let jobVarName (vname: string)  : string = 
    "__startrek_" ^ vname  

  let jobStartVarName = jobVarName "start" 
  let jobEndVarName = jobVarName "end" 
  let jobMaxVarName  = jobVarName "max" 
  let jobMinVarName = jobVarName "min" 

    
  let schedVars = 
    let atype = Formatcil.cType "%t:T[%e:j]"
      [
	("T", Ft jvtype);
	("j", Fe (integer (numOfJobs ()) ));
      ] in
    {
      sv =  makeGlobalVar jobStartVarName atype;
      ev = makeGlobalVar jobEndVarName atype;
      maxv = makeGlobalVar jobMaxVarName atype;
      minv = makeGlobalVar jobMinVarName atype;
    }


  let jobStartLval (job:int) : lval = 
    Formatcil.cLval "sv[%e:job]" 
      [("sv", Fv schedVars.sv);
       ("job", Fe (integer job))
      ]

  let jobEndLval (job:int) : lval = 
    Formatcil.cLval "ev[%e:job]" 
      [("ev", Fv schedVars.ev);
       ("job", Fe (integer job))
      ]

  let jobMaxLval (job:int) : lval = 
    Formatcil.cLval "maxv[%e:job]" 
      [("maxv", Fv schedVars.maxv);
       ("job", Fe (integer job))
      ]


  let jobMinLval (job:int) : lval = 
    Formatcil.cLval "minv[%e:job]" 
      [("minv", Fv schedVars.minv);
       ("job", Fe (integer job))
      ]
    

  let registerGlobals f : unit = 
    f.globals <- 
      GVar (schedVars.sv, {init = None}, builtinLoc) ::
      GVar (schedVars.ev, {init = None}, builtinLoc) ::
      GVar (schedVars.maxv, {init = None}, builtinLoc) ::
      GVar (schedVars.minv, {init = None}, builtinLoc) ::
      f.globals
end


(** Time Bounded Analysis for Harmonic Tasks *)
module RekH
  (Args : ARGS) 
  (Svm :  SHARED_VAR_MANAGER) 
  (Sched : SCHED_VAR_MANAGER)
  (Asrt : ASSERT_VAR_MANAGER) 
  
  = 
struct


  (** map from varinfo to corresponding read function *)
  let rfnMap : fundec VM.t ref = ref VM.empty

  (** map from varinfo to corresponding write function *)
  let wfnMap : fundec VM.t ref = ref VM.empty


  let getReadFn v = VM.find v !rfnMap 
  let setReadFn v fd : unit =  rfnMap := VM.add v fd !rfnMap

  let setWriteFn v fd = wfnMap := VM.add v fd !wfnMap
  let getWriteFn v = VM.find v !wfnMap
    

  (*********************************************************************)
  (* module-global constants *)
  (*********************************************************************)

  (*********************************************************************)
  (* Arguments *)
  (*********************************************************************)
  let rounds = Args.rounds
  let taskVar = Args.taskVar
  let jobVar = Args.jobVar
  let jobEndVar = Args.jobEndVar
  let roundVar = Args.roundVar
  let lockVar = Args.lockVar
  let hyperPeriodVar = Args.hyperPeriodVar


  (*********************************************************************)
  (* Common functions and types  *)
  (*********************************************************************)

  let csType = TFun (boolType,Some [],false,[])
  let assumeFn = findOrCreateFunc Args.file "__CPROVER_assume" assumeType
  let assertFn = findOrCreateFunc Args.file "assert" assertType 

  (*********************************************************************)
  (* Names  *)
  (*********************************************************************)
  let initSharedFnName = "__startrek_init_shared"
  let initGlobalsFnName = "__startrek_init_globals"
  let scheduleJobsFnName = "__startrek_schedule_jobs"
  let userFinalFnName = "__startrek_user_final"
  let userInitialFnName = "__startrek_user_init"
  let hyperFnName = "__startrek_hyperperiod"
  let mainFnName = "main"

  let sharedReadPrefix = "__startrek_shared_read_"
  let sharedWritePrefix = "__startrek_shared_write_"


(** returns a statement that assumes x < y ===> y > z **)
  let notBetweenStmt (x:lval) (y:lval) (z:lval) : stmt =
    (* x < y ==> y > z *)
    Formatcil.cStmt 
      "if (%l:x < %l:y)  assume (%l:y > %l:z); "
      (fun n t -> assert false)
      builtinLoc
      [
	("x", Fl x);
	("y", Fl y);
	("z", Fl z);
	("assume", Fv assumeFn);
      ]

      
  (** returns a statement that assumes x < y ===> y > z **)
      
  let notBetweenStmt2 (x:lval) (y:lval) (z:lval) : stmt =
    (* x < y ==> y > z *)
    (* __CPROVER_assume (x >= y || (z < y)); *)

    let xGey = BinOp (Ge, Lval x, Lval y, boolType) and
	zLty = BinOp (Lt, Lval z, Lval y, boolType) in
    let disj = BinOp (LOr, xGey, zLty, boolType) in 
    mkStmtOneInstr (Call (None, Lval (var assumeFn), [disj], builtinLoc))
      
    
      

  (*********************************************************************)
  (* Read/Write flags for Partial-Order Reduction  *)
  (*********************************************************************)

  (** map from shared variables to read flag variables *)
  let readFlags : varinfo VM.t ref = ref VM.empty

  (** map from shared variables to write flag variables *)
  let writeFlags : varinfo VM.t ref = ref VM.empty


  (** create read/write flags for every shared variable *)
  let createRWFlags unit : unit = 
    
    let readFlagName v = "__startrek_rflag_" ^ v.vname in
    let writeFlagName v = "__startrek_wflag_" ^ v.vname in

    (* creaete flag variables for shared variables *)
    let flags = List.fold_left 
      (fun res g -> 
	match g with 
	  | GVar (v, _, _) when Svm.mem v -> 
	    (* found a shared variable *)
	    let readf = makeGlobalVar (readFlagName v) boolType and
		writef = makeGlobalVar (writeFlagName v) boolType in
	    readFlags := VM.add v readf !readFlags;
	    Svm.makeSharedVar readf;

	    writeFlags := VM.add v writef !writeFlags;
	    Svm.makeSharedVar writef; 

	    GVar (readf, {init = None}, builtinLoc) ::
	      GVar (writef, {init = None}, builtinLoc) :: res
	  | _ -> res
      )
      []
      Args.file.globals 
    in
    
    (* add flag variables to globals *)
    Args.file.globals <- flags @ Args.file.globals


  let globalTypes = ref []
  let globalVars = ref []

  let pushGlobal g : unit = 
    pushGlobal g ~types:globalTypes ~variables:globalVars

  let pushGlobals gl = List.iter pushGlobal gl

  (*********************************************************************)
  (* Non-deterministic values functions  *)
  (*********************************************************************)

  let ndbFn = findOrCreateFunc Args.file "nondet_bool" ndbType

  (** Returns a varinfo of a function that returns a non-deterministic value 
      of a given type *)
  let nondetFn (typ:typ) : varinfo =
    let tyname = 
      match typ with
	  TInt (IChar, _) -> "char"
	| TInt (ISChar, _) -> "schar"
	| TInt (IUChar, _) -> "uchar"
	| TInt (IBool, _) -> "_Bool"
	| TInt (IInt, _) -> "int"
	| TInt (IUInt, _) -> "uint"
	| TInt (IShort, _) -> "short"
	| TInt (IUShort, _) -> "ushort"
	| TInt (ILong, _) -> "long"
	| TInt (IULong, _) -> "ulong"
	| TInt (ILongLong, _) -> "long_long"
	| TInt (IULongLong, _) -> "ulong_long"
	| TFloat (FFloat, _) -> "float"
	| TFloat (FDouble, _) -> "double"
	| TFloat (FLongDouble, _) -> "long_double"
	| TNamed (tyinfo, _) -> tyinfo.tname
	| _ -> invalid_arg "Unsupported type"
    in
    findOrCreateFunc 
      Args.file 
      ("nondet_" ^ tyname) 
      (TFun (typ, Some [], false, []))





  (*********************************************************************)
  (* Context Switch Functions  *)
  (*********************************************************************)
      
  let csFnName (t:task) = "__startrek_cs_t" ^ (string_of_int t.tid)

  let getCsFn (t:task) : varinfo = 
    findOrCreateFunc Args.file (csFnName t) csType

  (*********************************************************************)
  (*** creating  context switch functions **)
  (*********************************************************************)
  let createCS (f:file) =


    (*create CS for a task *)
    let createCSTask (t:task) : unit =
      let fd = findOrCreateFundec f (csFnName t) csType in
	
      (* the body of context switch *)
      let body = 
	if isTopTask t then 
	  [mkStmt (Return (Some zero, builtinLoc))]
	else
	  begin
	    (* switch over all jobs of higher-priority tasks. Note
	       that t is not the highest priority, so it has children
	       in the job tree *)
	    let fstJob = firstJobOfTask t in
	    let swBody = BT.fold_left_generation 
	      (fun b tree node -> 
		let cbody = BT.fold_dfs_siblings 
		  (fun r tree node ->
		    let rnd = var roundVar in
		    let id = BT.id tree node in
		    let slv = Sched.jobStartLval id in
		    let elv = Sched.jobEndLval id in
		    (notBetweenStmt slv rnd elv) :: r
		  ) 
		  [] tree (BT.first_child tree node) 
		in
		let hd = List.hd cbody in
		hd.labels <- [Case (integer (BT.id tree node), rekloc)];
		cbody @ ((mkStmt (Break builtinLoc)) :: b)
	      )
	      [] !jobTree fstJob 
	    in
	    let stmts = if !nopriorcs then [] else
		[mkStmt (Switch (Lval (var jobVar), 
				 mkBlock swBody, swBody, rekloc))] 
	    in
	    Formatcil.cStmts 
              "
                if (lock) return 0;
                %t:Bool c = ndb (); if (c) return 0;
                
                %t:rty o = round; round = nd();
                assume (round > o); assume (round <= ej); 

                if (round != ej) { %S:w } return 1;
              "
	      (fun n t -> makeTempVar fd ~name:n t)
	      rekloc
	      [
		("rty", Ft roundVar.vtype);
		("round", Fv roundVar);
		("nd", Fv (nondetFn roundVar.vtype));
		("Bool", Ft boolType);
		("ndb", Fv ndbFn);
		("assume", Fv assumeFn);
		("lock", Fv lockVar);
		("ej", Fv jobEndVar);
		("w", FS stmts);
	      ]
	  end
      in
      fd.sbody.bstmts <- body @ fd.sbody.bstmts ; 
      fd.svar.vinline <- true;
      fd.svar.vstorage <- Static;
      Args.file.globals <- 
	GVarDecl (fd.svar, builtinLoc) :: Args.file.globals 
    in
    (* create context-switch for each task *)
    List.iter createCSTask !tasks


  (*********************************************************************)
  (*** shared variable read/write functions **)
  (*********************************************************************)

  let getSharedReadFn (v:varinfo) = 

    let readFnName (v:varinfo) = 
      "__startrek_read_" ^ v.vname in
    
    let makeSwitchRead readFn = 
      let cvar = makeTempVar readFn ~name:"c" boolType in
      let oldRound = makeTempVar readFn ~name:"or" roundVar.vtype in
      (* list of statements *)
      let slist = 
	
	let fn (t:task) (res:stmt list) : stmt list = 
	  let poblock = 
	    if !dopo then
	      begin
		let wf = VM.find v !writeFlags and
		    rf = VM.find v !readFlags in
		let svwf = Svm.getSharedVar wf and
		    svrf = Svm.getSharedVar rf in

		let wfVar = makeTempVar readFn ~name:"wf" boolType in
		
		let orndE = (Lval (var oldRound)) and 
		    rndE = (Lval (var roundVar)) in 
		
		Formatcil.cStmts
		  "if (c) { %s:owf %s:wfread assume (wf); } %s:rf"
		  (fun n t -> assert false)
		  builtinLoc
		  [
		    ("c", Fv cvar);
		    ("owf", Fs (Svm.mkSharedWriteStmt svwf orndE zero));
		    ("wfread", Fs (Svm.mkSharedReadStmt (var wfVar) svwf rndE));
		    ("wf", Fv wfVar);
		    ("rf", Fs (Svm.mkSharedWriteStmt svrf rndE one));
		    ("assume", Fv assumeFn)
		  ]
	      end
	    else [] 
	  in
	  let caseBody = 
	    mkStmt 
	      (Block 
		 (mkBlock
		    (Formatcil.cStmts 
		       "or = r; c = cs(); { %S:po }"
		       (fun n t -> assert false)
		       builtinLoc
		       [
			 ("cs", Fv (getCsFn t));
			 ("or", Fv oldRound);
			 ("r", Fv roundVar);
			 ("c", Fv cvar);
			 ("po", FS poblock)
		       ]
		    ))) 
	  in
	  
	  (* add a case label *)
	  caseBody.labels <- 
	    (Case (integer t.tid, builtinLoc)) :: caseBody.labels ;
	  caseBody :: mkStmt (Break builtinLoc) :: res
	    
	in
	List.fold_right fn !tasks [] 
      in
      mkStmt (Switch (Lval (var taskVar),mkBlock slist, slist, builtinLoc))
    in
    try (getReadFn v).svar with 
      (* no read function found, create one *)
      | Not_found -> 
	begin
	  let readFnType = TFun (v.vtype, Some [], false, []) in
	  let readFn = emptyFunction (readFnName v) in 
	  
	  setFunctionTypeMakeFormals readFn readFnType ; 
	  readFn.svar.vinline <- true ;
	  readFn.svar.vstorage <- Static ;
	  
	  let rvar = makeTempVar readFn ~name:"r" v.vtype in
	  let sv = Svm.getSharedVar v in
	  let rstmt = 
	    Svm.mkSharedReadStmt (var rvar) sv (Lval (var roundVar)) in
	  let retStmt = mkStmt (Return (Some (Lval (var rvar)), builtinLoc)) in
	  readFn.sbody <- mkBlock ([makeSwitchRead readFn ; rstmt ; retStmt]);
	  setReadFn v readFn ; 
	  readFn.svar
	end
	  
  let getSharedWriteFn (v:varinfo)  = 

    let writeFnName (v:varinfo) = 
      "__startrek_write_" ^ v.vname in

    let makeSwitchWrite writeFn = 
      
      let cvar = makeTempVar writeFn ~name:"c" boolType in
      let oldRound = makeTempVar writeFn ~name:"or" roundVar.vtype in

      let slist =
	let fn (t:task) (l:stmt list) : stmt list = 

	  let poblock = 
	    if !dopo then 
	      let wf = VM.find v !writeFlags and
		  rf = VM.find v !readFlags in
	      let svwf = Svm.getSharedVar wf  and
		  svrf = Svm.getSharedVar rf  in
	      
	      let wfVar = makeTempVar writeFn ~name:"wf" boolType and
		  rfVar = makeTempVar writeFn ~name:"rf" boolType in
	      
	      let orndE = (Lval (var oldRound)) and 
		  rndE = (Lval (var roundVar)) in 

	      Formatcil.cStmts
		"if (c)  
                    { %s:orf %s:owf %s:readrf %s:readwf assume (wf | rf); } 
                  %s:writewf" 
		(fun n t -> assert false)
		builtinLoc
		[
		  ("c", Fv cvar);
		  ("orf", Fs (Svm.mkSharedWriteStmt svrf orndE zero));
		  ("owf", Fs (Svm.mkSharedWriteStmt svwf orndE zero));
		  ("readrf", Fs (Svm.mkSharedReadStmt (var rfVar) svrf rndE));
		  ("readwf", Fs (Svm.mkSharedReadStmt (var wfVar) svwf rndE));
		  ("wf", Fv wfVar);
		  ("rf", Fv rfVar);
		  ("writewf", Fs (Svm.mkSharedWriteStmt svwf rndE one));
		  ("assume", Fv assumeFn);
		]	      
	    else
	      []
	  in
	  let caseBody = 
	    mkStmt 
	      (Block (
		mkBlock 
		  (Formatcil.cStmts
                     "or = r; c = cs(); { %S:po } "
		     (fun n t -> makeTempVar writeFn ~name:n t)
		     builtinLoc
		     [
		       ("or", Fv oldRound);
		       ("r", Fv roundVar);
		       ("c", Fv cvar);
		       ("cs", Fv (getCsFn t));
		       ("po", FS poblock)
		     ])))
	  in
	  caseBody.labels <- (Case (integer t.tid, builtinLoc)) :: 
	    caseBody.labels;
	  caseBody :: mkStmt (Break builtinLoc) :: l 
	in
	List.fold_right fn !tasks []
      in
      mkStmt (Switch (Lval (var taskVar),mkBlock slist, slist, builtinLoc)) 
    in

    try (getWriteFn v).svar with
      (* no write function found, create one *)
      | Not_found -> 
	begin
	  let writeFnType = 
	    TFun (voidType,Some [("arg",v.vtype,[])],false,[]) in
	  let writeFn = emptyFunction (writeFnName v) in 
	  
	  setFunctionTypeMakeFormals writeFn writeFnType ; 
	  writeFn.svar.vinline <- true ;
	  writeFn.svar.vstorage <- Static ;


	  let arg = Lval (var (List.hd writeFn.sformals)) in
	  let sv = Svm.getSharedVar v in
	  let wstmt = Svm.mkSharedWriteStmt sv (Lval (var roundVar)) arg in
	  writeFn.sbody <- mkBlock ([makeSwitchWrite writeFn; wstmt]) ;
	  
	  setWriteFn v writeFn ; 
	  writeFn.svar
	end

  (*********************************************************************)
  (* Visitors  *)
  (*********************************************************************)
    
  (** Finds and registers shared variables *)
  class sharedVariableFinder = object (self)
    inherit nopCilVisitor as super
      
    method vinst (ins : instr) = begin
      match ins with
          (* read a global variable -- must be of the form x = R(&v) *)
          Call(Some(x),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when 
	      Util.hasPrefix sharedReadPrefix fn && 
		(not (Svm.mem v)) -> 
		  Svm.makeSharedVar v ; SkipChildren

		    
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when 
	      Util.hasPrefix sharedWritePrefix fn && 
		(not (Svm.mem v)) -> 
	  Svm.makeSharedVar v ; SkipChildren
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren

    end
  end
	  
  (** rewrites accesses to shared variables and lock functions *)
  class globalReplacer = 
  object(self) 
    inherit nopCilVisitor as super
      
    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
	  (* read a global variable -- must be of the form x = R(&v) *)
	  Call(Some(Var x, NoOffset),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when Util.hasPrefix sharedReadPrefix fn -> 
	      x.vtype <- v.vtype;
	      ChangeTo 
		[Call (Some(Var x, NoOffset), 
		       Lval (var (getSharedReadFn v)), [], builtinLoc)]
		
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when Util.hasPrefix sharedWritePrefix fn ->
	  ChangeTo 
	    ([Call (None, Lval (var (getSharedWriteFn v)), [e], builtinLoc)])

	| Call (None, Lval (Var {vname = fn}, NoOffset), [], loc) 
	    when fn = cpuLockFnName ->
	  ChangeTo ([Set (var lockVar, one, builtinLoc)])
	| Call (None, Lval (Var {vname = fn}, NoOffset), [], loc) 
	    when fn = cpuUnlockFnName ->
	  ChangeTo ([Set (var lockVar, zero, builtinLoc)])
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end



  (**  pulls out assertions from task bodies *)
  class assertReplacer = object(self) inherit nopCilVisitor as super
            
    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
          (* read a global variable -- must be of the form x = R(&v) *)
          Call(None,
	       Lval(Var {vname = fn}, NoOffset), [e], loc)  
	    when fn = "assert" -> 
	      ChangeTo 
		[Call (None, 
		       Lval (var (Asrt.newAssertFun Args.file)), 
		       [e], 
		       builtinLoc)]
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end



  (** replaces accesses to shared variables in user initializer *)
  class initialReplacer = 
  object(self) 
    inherit nopCilVisitor as super
      
    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
	  (* read a global variable -- must be of the form x = R(&v) *)
	  Call(Some(Var x, NoOffset),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when Util.hasPrefix sharedReadPrefix fn -> 
	      let sv = Svm.getSharedVar v in
	      x.vtype <- v.vtype;
	      ChangeTo [Set (var x, 
			     Lval (Svm.getSharedLval sv 0), builtinLoc)]

		
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when Util.hasPrefix sharedWritePrefix fn ->
	  let sv = Svm.getSharedVar v in
	  ChangeTo [Set (Svm.getSharedLval sv  0, e, builtinLoc)]
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end

  (** replaces accesses to shared variables in user finalizer *)
  class finalReplacer = 
  object(self) 
    inherit nopCilVisitor as super
      
    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
          (* read a global variable -- must be of the form x = R(&v) *)
          Call(Some(x),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when Util.hasPrefix sharedReadPrefix fn -> 
	      let sv = Svm.getSharedVar v in
	      ChangeTo 
		[Set (x, Lval (Svm.getSharedLval sv (pred rounds)), 
			       builtinLoc)]
		
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when Util.hasPrefix sharedWritePrefix fn ->
	  let sv = Svm.getSharedVar v in
	  ChangeTo 
	    [Set (Svm.getSharedLval sv (pred rounds), e, builtinLoc)]
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end


  (*********************************************************************)
  (* Creation of shared variables  *)
  (*********************************************************************)
 
  (** create variables for initial values *)
  let createInitVars (f:file) = 
    let svf = new sharedVariableFinder in
    iterGlobals 
      f 
      (fun g -> 
	match g with
	  | GFun(fd, _) -> fd.sbody <- visitCilBlock svf fd.sbody
	  | _ -> ()
      )
      
  (** create all global variables *)
  let createGlobalVars = createInitVars


    (** generate reset assertions block for a task *)
  let mkResetAssertBlock = Asrt.map_task (fun lv -> Set (lv, one, rekloc)) 

  (** check-assertions block *)
  let mkCheckAssertBlock = 
    Asrt.map_task 
      (fun lv -> Call (None, Lval (var assertFn), [Lval lv], rekloc))
    
  (** Deprecated check-assumptions block *)
  let mkCheckAssumeBlock () =  
    let succRound = 
      BinOp (PlusA, (Lval (var roundVar)), one, roundVar.vtype) in
    let assumeBlock = 
      Svm.foldShared (fun v sv res -> 
	(Formatcil.cInstr
	   "assume (%l:g == %l:v);"
	   rekloc
	   [
	     ("assume", Fv assumeFn);
	     ("g", Fl (Svm.getSharedExpLval sv (Lval (var roundVar))));
	     ("v", Fl (Svm.getSharedInitExpLval sv succRound));
	   ]) :: res) [] in
    
    let thenBlock = mkBlock [mkStmt (Instr assumeBlock)] in
    let x = Lval (var roundVar) in
    let y = (integer (pred rounds)) in
    mkStmt (If (BinOp (Lt, x, y, boolType), thenBlock, mkBlock [], rekloc))


  (** check-assumptions block specific to a given job *)
  let mkJobCheckAssumeBlock (job:jobt) : Cil.stmt =  
    let low = jobLowRound job in 
    let high = let x = jobHighRound job in 
	       if x = pred rounds then pred x else x in

    let mkAssumeBlock (n:int) = 
      Svm.foldShared (fun v sv res -> 
	(Formatcil.cInstr
	   "assume (%l:g == %l:v);"
	   rekloc
	   [
	     ("assume", Fv assumeFn);
	     ("g", Fl (Svm.getSharedLval sv n));
	     ("v", Fl (Svm.getSharedInitLval sv (succ n)));
	   ]) :: res) [] 
    in

    (* high < low implies that low is in the last round *)
    if high < low then mkEmptyStmt () else
      mkSwitchStmt (Lval (var jobEndVar)) low (high - low + 1) 
	(fun n -> [mkStmt (Instr (mkAssumeBlock n))])

  (*********************************************************************)
  (*** creating job scheduling function **)
  (*********************************************************************)



  (** create schedule_jobs  function **)
  let createScheduleJobs unit : unit =
    let ftype =  TFun (voidType,Some [],false,[]) in
    let fd = findOrCreateFundec Args.file scheduleJobsFnName ftype in
    fd.svar.vinline <- true;

    (** Returns nesting constraints for job j2 of a low priority task
	t2 and job j1 of a higher priority task t1 **)
    let nestingConstraints j1 j2 =
      let 
	  jid1 = BT.id !jobTree j1 and
	  jid2 = BT.id !jobTree j2 and
	  tid1 = BT.level_of j1 and
	  tid2 = BT.level_of j2 
      in
      let
    	  s1 = Sched.jobStartLval jid1 and
    	  s2 = Sched.jobStartLval jid2 and
    	  e1 = Sched.jobEndLval jid1 and
    	  e2 = Sched.jobEndLval jid2 in
      
      let pb = if !nopbounds then 0 else
      	  !taskArray.(tid2).pbounds.(tid1 - tid2 - 1) in
      let pbound = 
	if !nopbounds then []
	else
	  begin
	    try
	      let priorJob = BT.stepInGeneration !jobTree j2 j1 (-pb) in
	      Formatcil.cStmts
		"assume (%l:je < %l:s2);"
		(fun n t -> assert false)
		builtinLoc
		[
		  ("assume", Fv assumeFn);
		  ("s2", Fl s2);
		  ("je", Fl (Sched.jobEndLval (BT.id !jobTree priorJob) ));
		]
	    with Invalid_argument _ -> []
	  end
      in 

      if jobPreceeds j1 j2 then
	Formatcil.cStmts
	  "assume(%l:e1 < %l:s2);"
	  (fun n t -> assert false)
	  rekloc
	  [ ("e1", Fl e1); ("s2", Fl s2); ("assume", Fv assumeFn);]
      else if jobFollows j1 j2 then
	Formatcil.cStmts
	  "assume(%l:e2 < %l:s1);"
	  (fun n t -> assert false)
	  rekloc
	  [ ("e2", Fl e2); ("s1", Fl s1); ("assume", Fv assumeFn);]
      else
	(* let p1 = if jobMayNotPreceed j1 j2 then
	    Formatcil.cStmts
	      "assume(%l:s2 <= %l:s1);"
	      (fun n t -> assert false)
	      rekloc
	      [("s1", Fl s1); ("s2", Fl s2); ("assume", Fv assumeFn);]
	  else [] in *)
	let p2 = if jobMayNotFollow j1 j2 then
	    Formatcil.cStmts
	      "assume(%l:e1 <= %l:e2);"
	      (fun n t -> assert false)
	      rekloc
	      [("e1", Fl e1); ("e2", Fl e2); ("assume", Fv assumeFn);]
	  else [] in
	p2 @
	  Formatcil.cStmts
      	    "if (%l:s2 <= %l:e1)
               if (%l:s1 <= %l:e2)
                 {
                   assume (%l:s2 <= %l:s1);
                   assume (%l:e1 < %l:e2);
                   %S:pbound
                 }
            "
    	  (fun n t -> assert false)
    	  builtinLoc
    	  [
    	    ("s1", Fl s1);
    	    ("s2", Fl s2);
    	    ("e1", Fl e1);
    	    ("e2", Fl e2);
	    ("pbound", FS pbound);
    	    ("assume", Fv assumeFn);
    	  ]	  
    in
    
    let minStmt x y z =
      Formatcil.cStmt
	"if (%l:y < %l:z)  assume(%l:x == %l:y); else assume(%l:x == %l:z);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("z", Fl z);
	  ("assume", Fv assumeFn)
	] in

    let maxStmt x y z =
      Formatcil.cStmt
	"if (%l:y > %l:z)  assume(%l:x == %l:y); else assume(%l:x == %l:z);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("z", Fl z);
	  ("assume", Fv assumeFn)
	] in

    let leqStmt x y = 
      Formatcil.cStmt
	"assume (%l:x <= %l:y);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("assume", Fv assumeFn)
	] in

    let ltStmt x y = 
      Formatcil.cStmt
	"assume (%l:x < %l:y);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("assume", Fv assumeFn)
	] in

    let eqStmt x y = 
      Formatcil.cStmt
	"assume (%l:x == %l:y);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("assume", Fv assumeFn)
	] in
	
    let perJob tree job : bool =
      let jid = BT.id tree job in
      
      let minV = Sched.jobMinLval jid in
      let maxV = Sched.jobMaxLval jid in
      let startV = Sched.jobStartLval jid in
      let endV = Sched.jobEndLval jid  in
      
      if BT.isLeaf tree job then 
	fd.sbody.bstmts <- 
	  (eqStmt startV endV) ::
	  (eqStmt minV startV) :: 
	  (eqStmt maxV  endV) :: 
	  fd.sbody.bstmts
      else
	begin
	  let cminV = Sched.jobMinLval (BT.id tree (BT.first_child tree job)) in
	  let cmaxV = Sched.jobMaxLval (BT.id tree (BT.last_child tree job)) in
	(* add min/max statements for the current job *)
	  fd.sbody.bstmts <- 
	    (leqStmt startV endV) ::
	    (minStmt minV startV cminV) ::
	    (maxStmt maxV endV cmaxV) :: fd.sbody.bstmts 
	end
      ;
	(* ensure that siblings start/end times  are ordered *)
      if BT.hasNext tree job then 
	begin
	  let minNextV = Sched.jobMinLval (BT.id tree (BT.next tree job)) in
	  fd.sbody.bstmts <-
	    (ltStmt maxV minNextV) :: fd.sbody.bstmts 
	end
      ;

      fd.sbody.bstmts <- 
	(Formatcil.cStmts
	   "assume (%e:job_low <= %l:start);
            assume (%l:end <= %e:job_high);"
	   (fun n t -> assert false)
	   rekloc
	   [
	     ("assume", Fv assumeFn);
	     ("end", Fl endV);
	     ("start", Fl startV);
	     ("job_low", Fe (integer (jobLowRound job)));
	     ("job_high", Fe (integer (jobHighRound job)));
	   ]) @ fd.sbody.bstmts ;


      if !nesting && not (BT.isRoot job) then 
	begin
	  BT.iter_parents  
	    (fun t j2 -> 
	      fd.sbody.bstmts <- 
		fd.sbody.bstmts @ nestingConstraints job j2 
	    ) 
	    tree (BT.parent_of job)
	end
      ;
      true
    in
    BT.iter_dfs perJob !jobTree (BT.root !jobTree)  ; 
    (* initialize all variables non-deterministically *)
    BT.iter_dfs 
      (fun tree job -> 
	let jid = BT.id tree job in 
	fd.sbody.bstmts <- 
	  (Formatcil.cStmts 
	  "%l:start=ndc(); %l:end=ndc (); %l:min=ndc(); %l:max=ndc ();"
	  (fun n t -> assert false)
	  rekloc
	  [
	    ("start", Fl (Sched.jobStartLval jid));
	    ("end", Fl (Sched.jobEndLval jid));
	    ("min", Fl (Sched.jobMinLval jid));
	    ("max", Fl (Sched.jobMaxLval jid));
	    ("ndc", Fv (nondetFn ucharType));
	  ]) @ fd.sbody.bstmts 
	;
	true

      ) 
      !jobTree (BT.root !jobTree)

  (*********************************************************************)
  (*** creating the init function **)
  (*********************************************************************)
    
  (*
    vi3[j][s] = ndVi (); vi2[j][s] = vi3[j][s]
  *)

  (** create a init function *)
  let createInit (f:file) =
    let initType = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f initGlobalsFnName initType in
    fd.svar.vinline <- true;

    (* for every shared variable *)
    Svm.iterShared
      (fun v sv ->
	(* for every round, except the first one *)
	for r = pred rounds downto 1 do
	  fd.sbody.bstmts <-
	    Formatcil.cStmts
	    "%l:sc = nd () ; %l:xs = %l:sc;"
	    (fun n t -> makeTempVar fd ~name:n t)
	    builtinLoc
	    [("sc", Fl (Svm.getSharedInitLval sv r)); 
	     ("xs", Fl (Svm.getSharedLval sv r));
	     ("nd", Fv (nondetFn v.vtype))
	    ] @
	    fd.sbody.bstmts
	done)

	
	
  (*********************************************************************)
  (*** creating the tasks **)
  (*********************************************************************)

  (** rewrites all accesses to shared variables, lock functions, and
      asserts *)
  let createTasks (f:file) = 
    let glblr = new globalReplacer and
	asrtr = new assertReplacer in
      iterGlobals 
	f 
	(fun g -> 
	  match g with 
	    | GFun(fd, _) -> 
	      if (not (fd.svar.vname = userFinalFnName) &&
		    not (fd.svar.vname = userInitialFnName)) then
		fd.sbody <- visitCilBlock asrtr (visitCilBlock glblr fd.sbody)
	    | _ -> ()
	)
	

	
  (*********************************************************************)
  (*** rewritting user-supplied init and final functions **)
  (*********************************************************************)

  let createUserInitial (f:file) =
    let finalType = TFun (voidType, Some [], false,[]) in
    let fd = findOrCreateFundec f userInitialFnName finalType in
    fd.svar.vinline <- true;

    
    fd.sbody <-  visitCilBlock (new initialReplacer) fd.sbody	

  (* create a final function *)
  let createUserFinal (f:file) =
    let finalType = TFun (voidType, Some [], false,[]) in
    let fd = findOrCreateFundec f userFinalFnName finalType in
    fd.svar.vinline <- true;
    fd.sbody <-  visitCilBlock (new finalReplacer) fd.sbody



      
  (*********************************************************************)
  (*** creating the main function **)
  (*********************************************************************)
  (** Encoding of a hyper-period when arrivals are fixed *)
  let createHyperPeriod_NoRange (f: file) = 
    let ftype = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f hyperFnName ftype in

    let jobOrder = (fun a b -> 
      if a = b then 0 
      else if jobPriority a >= jobPriority b then
	if jobPreceeds a b then -1 else 1
      else (if jobPreceeds b a then 1 else -1)
    ) in
    let orderedJobs = List.sort jobOrder 
      (BT.fold_dfs (fun res t n -> n ::res) [] !jobTree (BT.root !jobTree))
    in

    List.iter 
      (fun j -> 
	print_string ((string_of_int (BT.id !jobTree j)) ^ " ")) orderedJobs ;
    print_string "\n";
    
    let encode (job:jobt) : stmt list = 
      let jid = BT.id !jobTree job in
      let tid = BT.level_of job in

	
      let (assertedJobs:jobt list) =
	BT.fold_parents 
	  (fun res t low -> 
	    if (job = low || jobMayPreempt job low) &&
	      BT.forall 
	      (fun t med -> 
		(not (jobMayPreempt med low)) || (jobOrder med job) <= 0)
	      t low 
	    then 
	      low :: res
	    else res
	  )
	  [] !jobTree job
      in
      print_string ("assrt after " ^ (string_of_int jid) ^ ": ");
      List.iter (fun j -> 
	print_string ((string_of_int (BT.id !jobTree j)) ^ " ")) assertedJobs;
      print_string "\n";
      let asserts = 
	List.fold_left 
	  (fun res job -> (mkCheckAssertBlock (BT.level_of job)) @ res)
	  []
	  assertedJobs 
      in
      Formatcil.cStmts
	"
           round = %l:mystart;
           task = %e:mytask;
           end = %l:myend;
           job = %e:jid;

           %I:resetAssert
           
           task_entry_pt ();
           assume (round == end);
           
           %s:checkAssume

           %I:checkAssert
        "
	(fun n t -> assert false)
	rekloc
	[
	  ("assume", Fv assumeFn);
	  ("task", Fv taskVar);
	  ("mytask", Fe (integer tid));
	  ("round", Fv roundVar);
	  ("end", Fv jobEndVar);
	  ("mystart", Fl (Sched.jobStartLval jid));
	  ("myend", Fl (Sched.jobEndLval jid));
	  ("job", Fv jobVar);
	  ("jid", Fe (integer jid));
	  ("resetAssert", FI (mkResetAssertBlock tid));
	  ("task_entry_pt", Fv (oget (!taskArray.(tid).entryPt)).svar);
	  ("checkAssume", Fs (mkJobCheckAssumeBlock job));  
	  (* ("checkAssume", Fs (mkCheckAssumeBlock ())); *)
	  (* ("checkAssert", FI (mkCheckAssertBlock tid)); *)
	  ("checkAssert", FI asserts);
	]

    in

    let jobs = List.fold_left (fun r j -> r @ (encode j) ) []  orderedJobs
    in

    let sjobs = 
      mkStmtOneInstr 
	(Call (None, 
	       Lval (var (findOrCreateFunc f scheduleJobsFnName ftype)), 
	       [], rekloc)) in
    let initfn = 
      mkStmtOneInstr
	(Call (None, 
	       Lval (var (findOrCreateFunc f initGlobalsFnName ftype)), 
	       [], rekloc)) in
    fd.sbody <- mkBlock (sjobs :: initfn :: jobs)



       
  (** Encoding of a hyper-period with arival times range *)
  let createHyperPeriod_WithRange (f: file) = 
    let ftype = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f hyperFnName ftype in
    
    let rec encode_siblings job res = 
      let x = res @ (encode job) in
      if BT.hasNext !jobTree job then  
	encode_siblings (BT.next !jobTree job) x
      else x

    and encode job = 
      let kids = if BT.isLeaf !jobTree job then [] else
	  encode_siblings (BT.first_child !jobTree job) [] in
      
      let jid = BT.id !jobTree job in
      let tid = BT.level_of job in


      Formatcil.cStmts
	"
           round = %l:mystart;
           task = %e:mytask;
           end = %l:myend;
           job = %e:jid;

           %I:resetAssert
           
           task_entry_pt ();
           assume (round == end);
           
           %s:checkAssume

           { /* KIDS */ %S:kids  }
 
           %I:checkAssert
        "
	(fun n t -> assert false)
	rekloc
	[
	  ("assume", Fv assumeFn);
	  ("task", Fv taskVar);
	  ("mytask", Fe (integer tid));
	  ("round", Fv roundVar);
	  ("end", Fv jobEndVar);
	  ("mystart", Fl (Sched.jobStartLval jid));
	  ("myend", Fl (Sched.jobEndLval jid));
	  ("job", Fv jobVar);
	  ("jid", Fe (integer jid));
	  ("resetAssert", FI (mkResetAssertBlock tid));
	  ("task_entry_pt", Fv (oget (!taskArray.(tid).entryPt)).svar);
	  ("checkAssume", Fs (mkJobCheckAssumeBlock job));  
	  (* ("checkAssume", Fs (mkCheckAssumeBlock ())); *)
	  ("kids", FS kids);
	  ("checkAssert", FI (mkCheckAssertBlock tid));
	]
    in
    let sjobs = 
      mkStmtOneInstr 
	(Call (None, 
	       Lval (var (findOrCreateFunc f scheduleJobsFnName ftype)), 
	       [], rekloc)) in
    let initfn = 
      mkStmtOneInstr
	(Call (None, 
	       Lval (var (findOrCreateFunc f initGlobalsFnName ftype)), 
	       [], rekloc)) in
    fd.sbody <- mkBlock (sjobs :: initfn :: encode (BT.root !jobTree))

  let createHyperPeriod =  
    if !rangesInArrival then createHyperPeriod_WithRange 
    else createHyperPeriod_NoRange
	
  let createMain (f:file) = 
    let ftype = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f mainFnName ftype in

    let new_period = 
      Svm.foldShared 
	(fun v sv r -> 
	  let lhs = Svm.getSharedLval sv 0 in 
	  let rhs = Svm.getSharedLval sv (pred rounds) in 
	  mkStmtOneInstr (Set (lhs, Lval rhs, rekloc)) :: r
	) 
	[] 
    in
    
    let initSharedFn = findOrCreateFunc Args.file initSharedFnName ftype in
    let userInitFn = findOrCreateFunc Args.file userInitialFnName ftype in

    let stmts = ref (Formatcil.cStmts "initSharedFn(); userInitFn();"
           (fun n t -> makeTempVar fd ~name:n t) rekloc
           [ ("initSharedFn", Fv initSharedFn); ("userInitFn", Fv userInitFn); ])
    in
      
    let hyperFn = findOrCreateFunc Args.file hyperFnName ftype in
      if (!unboundedSeq) then (
        let loopBody = (Formatcil.cStmts "hyperFn();"
          (fun n t -> makeTempVar fd ~name:n t) rekloc
          [ ("hyperFn", Fv hyperFn); ]) @ new_period 
        in
        stmts := !stmts @ (mkWhile (integer 1) loopBody)
      ) else (
        for i = 1 to !hyperPeriods do
          stmts := !stmts @ Formatcil.cStmts "%v:hv = %d:i; hyperFn();"
            (fun n t -> makeTempVar fd ~name:n t) rekloc
            [ ("hv", Fv hyperPeriodVar); ("i", Fd (i - 1)); ("hyperFn", Fv hyperFn); ];
          if (i != !hyperPeriods) then stmts := !stmts @ new_period;
        done ;
      );
      let finalFn = findOrCreateFunc Args.file userFinalFnName ftype in
        stmts := !stmts @ Formatcil.cStmts "finalFn();"
           (fun n t -> makeTempVar fd ~name:n t) rekloc
           [ ("finalFn", Fv finalFn); ];
        fd.sbody <- mkBlock !stmts      
      
  (*********************************************************************)
  (*** registering global variables **)
  (*********************************************************************)      
  (** register all globals created so far with the file *)
  let registerGlobals unit : unit = 
    (* registers prototypes and definitions of reader/writer
       functions *)

    let createRW map : unit = 
    
      (* register prototypes near variables they mask. This ensures
	 that all appropriate types have already been declared *)
      let fn res g = 
	let out = g :: res in 
	match g with 
	  | GVar (v, _, _) when VM.mem v map -> 
	    let fd = VM.find v map in
	    GVarDecl(fd.svar,builtinLoc) ::  out
	  | _ -> out
      in
      Args.file.globals <- List.fold_left fn [] (List.rev Args.file.globals);
      
      VM.iter 
	(fun v fd -> Args.file.globals <- 
	  Args.file.globals @ [GFun(fd, builtinLoc)]) map ;
      
    in

    
    
    (* register readers *)
    createRW !rfnMap ;
    (* register writers *)
    createRW !wfnMap ;

    Svm.registerGlobals Args.file; 
    Sched.registerGlobals Args.file;
    Asrt.registerGlobals Args.file
    
      
  let inline unit : unit =
    let inlineWhat (v:varinfo) : fundec option =
      findFundec Args.file v.vname
    in	
    iterGlobals Args.file 
      (fun g -> 
	match g with 
	  | GFun (fd, _) when fd.svar.vname = mainFnName -> 
	    let anyInlining = ref true in 
	    while !anyInlining do
	      anyInlining := false;
	      Inliner.doFunction fd inlineWhat anyInlining
	    done
	  | _ -> ()
      )

  (** map from hidden shared to what they are hiding *)
  let hiddenShared : varinfo VM.t ref = ref VM.empty 

  (** Rename shared variables to ensure that they cannot be used directly*)
  let hideShared unit : unit = 
    Args.file.globals <- 
      List.map 
      (fun g ->
	  match g with 
	      GVar (v, init, loc) when Svm.mem v -> 
		let u = copyVarinfo v ("__startrek_hidden_" ^ v.vname) in
		  hiddenShared := VM.add u v !hiddenShared; 
		  GVar (u, init, loc)
	    | _ -> g)
      Args.file.globals


    
  let createInitShared unit : unit = 
    let initType = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec Args.file initSharedFnName initType in
    fd.svar.vinline <- true;

    let body = VM.fold 
      (fun k v res -> 
	 let sv = Svm.getSharedVar v in 
	 let shared = Svm.getSharedLval sv 0  in
	   Set (shared, Lval (var k), builtinLoc) :: res
      )
      !hiddenShared 
      []
    in
      fd.sbody.bstmts <- [ mkStmt (Instr body) ]
	

  (** remove all globals that we don't like *)
  let cleanupGlobals unit : unit = 
    Args.file.globals <- 
      List.filter 
      (fun g -> 
	match g with 
	    (* remove all prototypes for functions used to mark shared
	       reads and writes*)
	    GVarDecl (v, _) -> 
	      (not (Util.hasPrefix sharedReadPrefix v.vname)) && 
		(not (Util.hasPrefix sharedWritePrefix v.vname))
	  | _ -> true
      ) 
      Args.file.globals


    
  (*********************************************************************)
  (*** Entry point **)
  (*********************************************************************)      

  (** run everything *)
  let run unit : unit = 
    (* create global variables *)
    createGlobalVars Args.file;
    if (!dopo) then createRWFlags ();
    (* schedule jobs *)
    createScheduleJobs ();
    (* create init function *)
    createInit Args.file;
    (* create context switch function *)
    createCS Args.file;
    (* create the tasks *)
    createTasks Args.file;
    (* create user initial function *)
    createUserInitial Args.file;
    (* create user final function *)
    createUserFinal Args.file;
    (* create hyper-period function *)
    createHyperPeriod Args.file; 
    createMain Args.file;
    cleanupGlobals ();
    (* create globals *)
    registerGlobals ();

    hideShared ();
    (* must hapen after the shared variables are hidden *)
    createInitShared ();

    if (!inlineAll) then inline (); 

    
end;;
(*********************************************************************)
(*** the Cil feature, where everything begins *)
(*********************************************************************)


(* command line option for first pass *)
let feature : featureDescr = { 
  fd_name = "rekh";              
  fd_enabled = ref false;
  fd_description = "Harmonic REK for RTES";
  fd_extraopt = 
    [
      "--rekhunbounded", Arg.Set unboundedSeq, " do unbounded sequentialization";
    ];
  fd_doit = (
    function (f: file) -> 
      print_string ("Analyzing " ^ f.fileName ^ "\n");

      (* compute task specific info *)
      initTaskList f ;
      
      (* package arguments into a module *)
      let module A = struct 
	let file = f 
        let rounds = List.fold_left (fun res t -> res + t.tjobs) 0 !tasks
	let hyperPeriodVar = findOrCreateGlobalVar f "__startrek_hyper_period" ucharType {init = None} builtinLoc
	let lockVar = findOrCreateGlobalVar f "__startrek_lock" boolType {init = Some (makeZeroInit boolType)} builtinLoc
	let jobEndVar = findOrCreateGlobalVar f "__startrek_job_end" ucharType {init = None} builtinLoc
	let jobVar = findOrCreateGlobalVar f "__startrek_job" ucharType {init = None} builtinLoc
	let taskVar = findOrCreateGlobalVar f "__startrek_task" ucharType {init = None} builtinLoc
	let roundVar = findOrCreateGlobalVar f "__startrek_round" ucharType {init = None} builtinLoc
      end in
      let module M = 
	    RekH
	      (A) 
	      (ArraySharedVarManager (A)) 
	      (ArraySchedVarManager (A))
	      (AssertVarManager (A)) in
      M.run ()
  );
  fd_post_check = false
} 
