open Cil
open Rtcbautil
open Rtcbainput
open Rekoptions
open Rekargs
open Rekhtasks
open Reksvm

(** number of rounds to sequentialize over *)
let numRounds = ref (-1)

(** whether to print number of jobs and exit *)
let printNumJobs = ref false

(** whether to do unbounded sequentialization -- default is false *)
let unboundedSeq = ref false

(** whether to check if there are enough rounds **)
let checkRounds = ref false

(** whether to perform a sanity check *)
let checkSanity = ref false

(*********************************************************************)
(* globals *)
(*********************************************************************)
(* number of PIP locks *)
let numPIPLocks = 2

(* name, type and initializer of E array *)
let eVarName = "E"
let eVarType = boolType
let eVarInit = {init = Some(makeZeroInit eVarType)}

(* name, type and initializer of S array *)
let sVarName = "S"
let sVarType = Formatcil.cType "char [%e:x]" [("x", Fe(integer numPIPLocks))]
let sVarInit =
  let x = ref [] in
  for i = 0 to (numPIPLocks - 1) do
    x := !x @ [(Index(integer i,NoOffset),(SingleInit mone))]
  done;
  {init = Some(CompoundInit(sVarType,!x))}

(* name, type and initializer of RT array *)
let rtVarName = "RT"
let rtVarType unit : typ = Formatcil.cType "unsigned char [%e:x]" [("x", Fe(integer (numOfTasks())))]
let rtVarInit unit = {init = Some(makeZeroInit (rtVarType()))}

(* name, type and initializer of RL array *)
let rlVarName = "RL"
let rlVarType = Formatcil.cType "unsigned char [%e:x]" [("x", Fe(integer numPIPLocks))]
let rlVarInit = {init = Some(makeZeroInit rlVarType)}

(* name and type of the current priority variable *)
let cpVarName = "__startrek_current_priority"
let cpVarType = charType

(*********************************************************************)
(* Signatures *)
(*********************************************************************)
module type SCHED_VAR_MANAGER = sig
  type schedvars
  val jobStartLval :  int -> lval
  val jobEndLval :  int -> lval
  val jobMaxLval :  int -> lval
  val jobMinLval :  int -> lval
  val roundBlockLval : exp -> lval
  val registerGlobals : file -> unit
end

(* extend ARGS To PIP args *)
module type PIPARGS = sig
  include ARGS
  val eVar : varinfo
  val sVar : varinfo
  val rtVar : varinfo
  val rlVar : varinfo
  val cpVar : varinfo
end

(*********************************************************************)
(* Modules *)
(*********************************************************************)

module ArraySchedVarManager (Args : ARGS) =
struct
  (*********************************************************************)
  (* Managing job scheduling variables  *)
  (*********************************************************************)
  (* sv - start, ev - end, maxv - max, minv - min *) 
  type schedvars = { sv: varinfo; ev: varinfo; maxv: varinfo; minv: varinfo; bv: varinfo }

  let jvtype = Args.roundVar.vtype
  let bvtype = boolType

  let jobVarName (vname: string)  : string = 
    "__startrek_" ^ vname  

  let jobStartVarName = jobVarName "start" 
  let jobEndVarName = jobVarName "end" 
  let jobMaxVarName  = jobVarName "max" 
  let jobMinVarName = jobVarName "min" 
  let roundBlockVarName = jobVarName "block" 
    
  let schedVars = 
    let atype = Formatcil.cType "%t:T[%e:j]"
      [
	("T", Ft jvtype);
	("j", Fe (integer (numOfJobs ()) ));
      ] in
    let btype = Formatcil.cType "%t:T[%e:j]"
      [
	("T", Ft bvtype);
	("j", Fe (integer (Args.rounds) ));
      ] in
    {
      sv =  makeGlobalVar jobStartVarName atype;
      ev = makeGlobalVar jobEndVarName atype;
      maxv = makeGlobalVar jobMaxVarName atype;
      minv = makeGlobalVar jobMinVarName atype;
      bv = makeGlobalVar roundBlockVarName btype;
    }

  let jobStartLval (job:int) : lval = 
    Formatcil.cLval "sv[%e:job]" 
      [("sv", Fv schedVars.sv);
       ("job", Fe (integer job))
      ]

  let jobEndLval (job:int) : lval = 
    Formatcil.cLval "ev[%e:job]" 
      [("ev", Fv schedVars.ev);
       ("job", Fe (integer job))
      ]

  let jobMaxLval (job:int) : lval = 
    Formatcil.cLval "maxv[%e:job]" 
      [("maxv", Fv schedVars.maxv);
       ("job", Fe (integer job))
      ]

  let jobMinLval (job:int) : lval = 
    Formatcil.cLval "minv[%e:job]" 
      [("minv", Fv schedVars.minv);
       ("job", Fe (integer job))
      ]

  let roundBlockLval (rnd:exp) : lval = 
    Formatcil.cLval "bv[%e:rnd]" 
      [("bv", Fv schedVars.bv);
       ("rnd", Fe (rnd))
      ]

  let registerGlobals f : unit = 
    let bvinit =
      let x = ref [] in
      for i = 0 to (Args.rounds - 1) do
        x := !x @ [(Index(integer i,NoOffset),(SingleInit one))]
      done;
      {init = Some(CompoundInit(bvtype,!x))}
    in
    f.globals <- 
      GVar (schedVars.sv, {init = None}, builtinLoc) ::
      GVar (schedVars.ev, {init = None}, builtinLoc) ::
      GVar (schedVars.maxv, {init = None}, builtinLoc) ::
      GVar (schedVars.minv, {init = None}, builtinLoc) ::
      GVar (schedVars.bv, bvinit, builtinLoc) ::
      f.globals
end

(** Time Bounded Analysis for Harmonic Tasks with PIP locks *)
module RekHPIP
  (Args : PIPARGS) 
  (Svm :  SHARED_VAR_MANAGER) 
  (Sched : SCHED_VAR_MANAGER)

  = 

struct
  (** map from varinfo to corresponding read function *)
  let rfnMap : fundec VM.t ref = ref VM.empty

  (** map from varinfo to corresponding write function *)
  let wfnMap : fundec VM.t ref = ref VM.empty

  let getReadFn v = VM.find v !rfnMap 
  let setReadFn v fd : unit =  rfnMap := VM.add v fd !rfnMap

  let setWriteFn v fd = wfnMap := VM.add v fd !wfnMap
  let getWriteFn v = VM.find v !wfnMap

  (*********************************************************************)
  (* module-global constants *)
  (*********************************************************************)

  (*********************************************************************)
  (* Arguments *)
  (*********************************************************************)
  let rounds = Args.rounds
  let taskVar = Args.taskVar
  let jobVar = Args.jobVar
  let jobEndVar = Args.jobEndVar
  let roundVar = Args.roundVar
  let lockVar = Args.lockVar
  let hyperPeriodVar = Args.hyperPeriodVar

  (*********************************************************************)
  (* Names  *)
  (*********************************************************************)
  let assumeFnName = "__CPROVER_assume"
  let doCSFnName = "__startrek_do_cs"
  let ipFnName = "__startrek_inherited_priority"
  let abortFnName = "__startrek_abort"
  let unblockFnName = "__startrek_unblock"
  let pipLockFnName = "__startrek_pip_lock"
  let pipUnlockFnName = "__startrek_pip_unlock"
  let initZeroFnName = "__startrek_init_zero"
  let initNonZeroFnName = "__startrek_init_non_zero"
  let scheduleJobsFnName = "__startrek_schedule_jobs"
  let userFinalFnName = "__startrek_user_final"
  let userInitialFnName = "__startrek_user_init"
  let checkAssumpFnName = "__startrek_check_assumptions"
  let checkAssertFnName = "__startrek_check_assertions"
  let hyperFnName = "__startrek_hyperperiod"
  let mainFnName = "main"

  (*********************************************************************)
  (* Common functions and types  *)
  (*********************************************************************)
  let csType = TFun (boolType,Some [],false,[])
  let doCSFnType = csType
  let abortFnType = TFun(voidType,Some [], false, [])
  let unblockFnType = TFun (voidType,Some [("lock",ucharType,[])],false,[])
  let pipLockFnType = TFun (voidType,Some [("lock",ucharType,[])],false,[])
  let pipUnlockFnType = TFun (voidType,Some [("lock",ucharType,[])],false,[])
  let ipFnType = TFun (charType,Some [("base_prio",charType,[])],false,[])
  let checkAssumpFnType = TFun(voidType,Some [], false, [])
  let checkAssertFnType = TFun(voidType,Some [], false, [])
  let assumeFn = findOrCreateFunc Args.file assumeFnName assumeType
  let assertFn = findOrCreateFunc Args.file "assert" assertType 

  let sharedReadPrefix = "__startrek_shared_read_"
  let sharedWritePrefix = "__startrek_shared_write_"
  let sharedVarPrefix = "__startrek_shared_var_"

  (** returns a statement that assumes x < y ===> y > z **)
  let notBetweenStmt (x:lval) (y:lval) (z:lval) : stmt =
    (* x < y ==> y > z *)
    Formatcil.cStmt 
      "if (%l:x < %l:y)  assume (%l:y > %l:z); "
      (fun n t -> assert false)
      builtinLoc
      [
	("x", Fl x);
	("y", Fl y);
	("z", Fl z);
	("assume", Fv assumeFn);
      ]
      
  (** returns a statement that assumes x < y ===> y > z **)
      
  let notBetweenStmt2 (x:lval) (y:lval) (z:lval) : stmt =
    (* x < y ==> y > z *)
    (* __CPROVER_assume (x >= y || (z < y)); *)

    let xGey = BinOp (Ge, Lval x, Lval y, boolType) and
	zLty = BinOp (Lt, Lval z, Lval y, boolType) in
    let disj = BinOp (LOr, xGey, zLty, boolType) in 
    mkStmtOneInstr (Call (None, Lval (var assumeFn), [disj], builtinLoc))

  (*********************************************************************)
  (* Read/Write flags for Partial-Order Reduction  *)
  (*********************************************************************)

  (** map from shared variables to read flag variables *)
  let readFlags : varinfo VM.t ref = ref VM.empty

  (** map from shared variables to write flag variables *)
  let writeFlags : varinfo VM.t ref = ref VM.empty


  (** create read/write flags for every shared variable *)
  let createRWFlags unit : unit = 
    
    let readFlagName v = "__startrek_rflag_" ^ v.vname in
    let writeFlagName v = "__startrek_wflag_" ^ v.vname in

    (* create flag variables for shared variables *)
    let flags = List.fold_left 
      (fun res g -> 
	match g with 
	  | GVar (v, _, _) when Svm.mem v -> 
	    (* found a shared variable *)
	    let readf = makeGlobalVar (readFlagName v) boolType and
		writef = makeGlobalVar (writeFlagName v) boolType in
	    readFlags := VM.add v readf !readFlags;
	    Svm.makeSharedVar readf;

	    writeFlags := VM.add v writef !writeFlags;
	    Svm.makeSharedVar writef; 

	    GVar (readf, {init = None}, builtinLoc) ::
	      GVar (writef, {init = None}, builtinLoc) :: res
	  | _ -> res
      )
      []
      Args.file.globals 
    in
    
    (* add flag variables to globals *)
    Args.file.globals <- flags @ Args.file.globals

  let globalTypes = ref []
  let globalVars = ref []

  let pushGlobal g : unit = 
    pushGlobal g ~types:globalTypes ~variables:globalVars

  let pushGlobals gl = List.iter pushGlobal gl

  (*********************************************************************)
  (* Non-deterministic values functions  *)
  (*********************************************************************)

  let ndbFn = findOrCreateFunc Args.file "nondet_bool" ndbType

  (** Returns a varinfo of a function that returns a non-deterministic value 
      of a given type *)
  let nondetFn (typ:typ) : varinfo =
    let tyname = 
      match typ with
	  TInt (IChar, _) -> "char"
	| TInt (ISChar, _) -> "schar"
	| TInt (IUChar, _) -> "uchar"
	| TInt (IBool, _) -> "_Bool"
	| TInt (IInt, _) -> "int"
	| TInt (IUInt, _) -> "uint"
	| TInt (IShort, _) -> "short"
	| TInt (IUShort, _) -> "ushort"
	| TInt (ILong, _) -> "long"
	| TInt (IULong, _) -> "ulong"
	| TInt (ILongLong, _) -> "long_long"
	| TInt (IULongLong, _) -> "ulong_long"
	| TFloat (FFloat, _) -> "float"
	| TFloat (FDouble, _) -> "double"
	| TFloat (FLongDouble, _) -> "long_double"
	| TNamed (tyinfo, _) -> tyinfo.tname
	| _ -> invalid_arg "Unsupported type"
    in
    findOrCreateFunc 
      Args.file 
      ("nondet_" ^ tyname) 
      (TFun (typ, Some [], false, []))

  (** condition to check for exception *)
  let abortCond unit : exp =
    let sv = Svm.getSharedVar Args.eVar in
    Lval (Svm.getSharedExpLval sv (Lval (var roundVar)))
      
  (** a handy function to create statements -- has a set of things
      hardwired *)
  let createStmtsLoc (s:string) (loc:location) fmt : stmt list = 
    Formatcil.cStmts s (fun n t -> assert false) loc
      ([
        ("do_cs", Fv(findOrCreateFunc Args.file doCSFnName doCSFnType));
        ("aborted", Fe(abortCond ()));
	("assume", Fv assumeFn);
	("assert", Fv assertFn);
        ("pipLock", Fv(findOrCreateFunc Args.file pipLockFnName pipLockFnType));
        ("pipUnlock", Fv(findOrCreateFunc Args.file pipUnlockFnName pipUnlockFnType));
      ] @ fmt)

  let createStmts (s:string) fmt : stmt list = 
    createStmtsLoc s builtinLoc fmt
    
  (*********************************************************************)
  (* Context Switch Functions  *)
  (*********************************************************************)
  let csFnName (t:task) = "__startrek_cs_t" ^ (string_of_int t.tid)
    
  let getCsFn (t:task) : varinfo = 
    findOrCreateFunc Args.file (csFnName t) csType

  (*********************************************************************)
  (*** creating  context switch functions **)
  (*********************************************************************)
  let createCS (f:file) =

    (*create CS for a task *)
    let createCSTask (t:task) : unit =
      let fd = findOrCreateFundec f (csFnName t) csType in
      let pvar = makeTempVar fd ~name:"p" charType in
	
      (* the body of context switch *)
      let body = 
	if isTopTask t then 
	  [mkStmt (Return (Some zero, builtinLoc))]
	else
          let sv = Svm.getSharedVar Args.cpVar in 
          let cp = Lval (Svm.getSharedExpLval sv (Lval (var roundVar))) in
	  Formatcil.cStmts 
            "
              if (lock) return 0;
              %t:Bool c = ndb (); if (c) return 0;
                
              %t:rty o = round; round = nd();
              assume (round > o); assume (round <= ej); 
             
              %v:pvar = ip (%e:tid);
              assume (%e:cp == %v:pvar);

              if (%e:aborted) abort();

              return 1;
            "
	    (fun n t -> makeTempVar fd ~name:n t)
	    rekloc
	    [
	      ("rty", Ft roundVar.vtype);
	      ("round", Fv roundVar);
	      ("nd", Fv (nondetFn roundVar.vtype));
	      ("Bool", Ft boolType);
	      ("ndb", Fv ndbFn);
	      ("assume", Fv assumeFn);
	      ("lock", Fv lockVar);
	      ("ej", Fv jobEndVar);
              ("aborted", Fe(abortCond ()));
              ("abort", Fv(findOrCreateFunc Args.file abortFnName abortFnType));
              ("cp", Fe(cp));
              ("tid", Fe(integer t.tid));
              ("pvar", Fv(pvar));
              ("ip", Fv(findOrCreateFunc Args.file ipFnName ipFnType));
	    ]
      in
      fd.sbody.bstmts <- body @ fd.sbody.bstmts ; 
      fd.svar.vinline <- true;
      fd.svar.vstorage <- Static;
      Args.file.globals <- 
	GVarDecl (fd.svar, builtinLoc) :: Args.file.globals 
    in
    (* create context-switch for each task *)
    List.iter createCSTask !tasks

  (*********************************************************************)
  (*** shared variable read/write functions **)
  (*********************************************************************)

  (** returns true if there should be no context switch before access
      to a shared variable -- currently, this is only true for
      __startrek_current_priority *)
  let noCSAccess (v:varinfo) =
    (v.vname = "__startrek_current_priority")

  let getSharedReadFn (v:varinfo) = 

    let readFnName (v:varinfo) = "__startrek_read_" ^ v.vname in

    try (getReadFn v).svar with 
      (* no read function found, create one *)
      | Not_found -> 
	begin
	  let readFnType = TFun (v.vtype, Some [], false, []) in
	  let readFn = emptyFunction (readFnName v) in 
	  
	  setFunctionTypeMakeFormals readFn readFnType ; 
	  readFn.svar.vinline <- true ;
	  readFn.svar.vstorage <- Static ;
	  
	  let rvar = makeTempVar readFn ~name:"r" v.vtype in
          let csStmt = createStmts 
	    (if (noCSAccess v) then "" else "do_cs(); if (%e:aborted) return %e:rv;")
            [("rv", Fe(Lval(var rvar)))]
	  in
	  let sv = Svm.getSharedVar v in
	  let rstmt = Svm.mkSharedReadStmt (var rvar) sv (Lval (var roundVar)) in
	  let retStmt = mkStmt (Return (Some (Lval (var rvar)), builtinLoc)) in
	  readFn.sbody <- mkBlock (csStmt @ [rstmt ; retStmt]);
	  setReadFn v readFn ; 
	  readFn.svar
	end
	  
  let getSharedWriteFn (v:varinfo)  = 

    let writeFnName (v:varinfo) = "__startrek_write_" ^ v.vname in

    try (getWriteFn v).svar with
      (* no write function found, create one *)
      | Not_found -> 
	begin
	  let writeFnType = 
	    TFun (voidType,Some [("arg",v.vtype,[])],false,[]) in
	  let writeFn = emptyFunction (writeFnName v) in 
	  
	  setFunctionTypeMakeFormals writeFn writeFnType ; 
	  writeFn.svar.vinline <- true ;
	  writeFn.svar.vstorage <- Static ;
	  
          let csStmt = createStmts (if (noCSAccess v) then "" else "do_cs(); if (%e:aborted) return;") [] in
	  let arg = Lval (var (List.hd writeFn.sformals)) in
	  let sv = Svm.getSharedVar v in
	  let wstmt = Svm.mkSharedWriteStmt sv (Lval (var roundVar)) arg in
	  writeFn.sbody <- mkBlock (csStmt @ [wstmt]) ;	  
	  setWriteFn v writeFn ; 
	  writeFn.svar
	end

  (*********************************************************************)
  (* Visitors  *)
  (*********************************************************************)
    
  (** Finds and registers shared variables *)
  class sharedVariableFinder = object (self)
    inherit nopCilVisitor as super
    (* all variables that have been accessed using specially named
       function are shared *)
    method vinst (ins : instr) = begin
      match ins with
          (* read a global variable -- must be of the form x = R(&v) *)
          Call(Some(x),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when 
	      Util.hasPrefix sharedReadPrefix fn && 
		(not (Svm.mem v)) -> 
		  Svm.makeSharedVar v ; SkipChildren
		    
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when 
	      Util.hasPrefix sharedWritePrefix fn && 
		(not (Svm.mem v)) -> 
	  Svm.makeSharedVar v ; SkipChildren
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren

    end
  end
      
  (** rewrites accesses to shared variables and lock functions *)
  class globalReplacer (_checkAbort:bool) = 
  object(self) 
    inherit nopCilVisitor as super
      
    val checkAbort = _checkAbort

    (* visit a statement -- we will go through the instructions, find
       function calls, and change them appropriately *)
    method vstmt (st:stmt) = begin
      let processInstr (ins:instr) (sl:stmt list) : stmt list =
        let newStmts = 
          match ins with
	      (* read a global variable -- must be of the form x =
	         R(&v) *)
	      Call(Some(Var x, NoOffset),
	           Lval(Var {vname = fn}, NoOffset),
	           [AddrOf(Var v, NoOffset)], loc) 
	        when Util.hasPrefix sharedReadPrefix fn -> 
	          x.vtype <- v.vtype;
	          createStmts 
                    (
                      if (checkAbort) then
                        "%l:x = read (); if (%e:aborted) return;" 
                      else
                        "%l:x = read ();" 
                    )
                    [("x", Fl(Var x,NoOffset));
                     ("read",Fv(getSharedReadFn v))]
		    
	    (* write a global variable -- must be of the form W(&v,e) *)
	    | Call(None,Lval(Var {vname = fn}, NoOffset),
	           [AddrOf(Var v, NoOffset); e],loc) 
	        when Util.hasPrefix sharedWritePrefix fn ->
	      createStmts 
                (
                  if (checkAbort) then
                    "write (%e:e); if (%e:aborted) return;" 
                  else
                    "write (%e:e);" 
                )
                [("e", Fe(e));
                 ("write",Fv(getSharedWriteFn v))]

            (* cpu lock *)
	    | Call (None, Lval (Var {vname = fn}, NoOffset), [], loc) 
	        when fn = cpuLockFnName ->
	      [mkStmtOneInstr(Set (var lockVar, one, builtinLoc))]
            (* cpu unlock *)
	    | Call (None, Lval (Var {vname = fn}, NoOffset), [], loc) 
	        when fn = cpuUnlockFnName ->
	      [mkStmtOneInstr(Set (var lockVar, zero, builtinLoc))]

            (* pip lock *)
	    | Call (None, Lval (Var {vname = fn}, NoOffset), [l], loc) 
                when (fn = pipLockFnName) ->
              createStmts 
                (
                  if (checkAbort) then
                    "do_cs(); if (%e:aborted) return; pipLock (%e:l); if (%e:aborted) return;" 
                  else
                    "do_cs(); pipLock (%e:l);" 
                )
                [("l", Fe(l))]
            (* pip unlock *)
	    | Call (None, Lval (Var {vname = fn}, NoOffset), [l], loc) 
                when (fn = pipUnlockFnName) ->
	      createStmts 
                (
                  if (checkAbort) then
                    "do_cs(); if (%e:aborted) return; pipUnlock (%e:l);" 
                  else
                    "do_cs(); pipUnlock (%e:l);" 
                )
                [("l", Fe(l))]

            (* assert *)
	    | Call (None, Lval (Var {vname = fn}, NoOffset), [e], loc) 
                when (fn = "assert") -> createStmts 
              (
                if (!checkRounds) then
                  "assume (%e:e);"
                else
                  "if (! (%e:e)) { %l:evar = 1; abort(); return; }" 
              )
              [
                ("e", Fe(e));
                ("evar", Fl(Svm.getSharedExpLval (Svm.getSharedVar Args.eVar) (Lval (var roundVar))));
                ("abort", Fv(findOrCreateFunc Args.file abortFnName abortFnType));
	        ("assume", Fv assumeFn);
              ]

            (** ignore special calls *)
            | Call(_,Lval (Var {vname = fn}, NoOffset),_,_) 
                when (Util.hasPrefix "nondet_" fn || fn = assumeFnName ||
                    fn = abortFnName || fn = ipFnName)
                  -> [mkStmtOneInstr ins]

            (** other calls -- check for exceptions afterwards *)
            | Call(_,_,_,_) -> [mkStmtOneInstr ins] @
              (
                if (checkAbort) then
	          (createStmts "if (%e:aborted) return;" [])
                else
                  []
              )

	    (* otherwise, do nothing *)
	    | _ -> [mkStmtOneInstr ins]
        in 
        newStmts @ sl
      in
      match st.skind with
          Instr(il) ->
            st.skind <- Block(mkBlock (compactStmts (List.fold_right processInstr il [])));
            SkipChildren
        | _ -> DoChildren
    end
  end

  (** replaces accesses to shared variables in user initializer *)
  class initialReplacer = 
  object(self) 
    inherit nopCilVisitor as super
      
    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
	  (* read a global variable -- must be of the form x = R(&v) *)
	  Call(Some(Var x, NoOffset),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when Util.hasPrefix sharedReadPrefix fn -> 
	      let sv = Svm.getSharedVar v in
	      x.vtype <- v.vtype;
	      ChangeTo [Set (var x, 
			     Lval (Svm.getSharedLval sv 0), builtinLoc)]

		
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when Util.hasPrefix sharedWritePrefix fn ->
	  let sv = Svm.getSharedVar v in
	  ChangeTo [Set (Svm.getSharedLval sv  0, e, builtinLoc)]
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end

  (** replaces accesses to shared variables in user finalizer *)
  class finalReplacer = 
  object(self) 
    inherit nopCilVisitor as super
      
    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
          (* read a global variable -- must be of the form x = R(&v) *)
          Call(Some(x),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when Util.hasPrefix sharedReadPrefix fn -> 
	      let sv = Svm.getSharedVar v in
	      ChangeTo 
		[Set (x, Lval (Svm.getSharedLval sv (pred rounds)), 
			       builtinLoc)]
		
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when Util.hasPrefix sharedWritePrefix fn ->
	  let sv = Svm.getSharedVar v in
	  ChangeTo 
	    [Set (Svm.getSharedLval sv (pred rounds), e, builtinLoc)]
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end

  (*********************************************************************)
  (* Creation of shared variables  *)
  (*********************************************************************)
 
  (** create variables for initial values *)
  let createInitVars (f:file) = 
    let svf = new sharedVariableFinder in
    iterGlobals 
      f 
      (fun g -> 
	match g with
            GVar(vi,_,_)
              when (vi.vname = eVarName || vi.vname = sVarName ||
                  vi.vname = rtVarName || vi.vname = rlVarName) -> 
                Svm.makeSharedVar vi
	  | GFun(fd, _) -> fd.sbody <- visitCilBlock svf fd.sbody
	  | _ -> ()
      )

  (** create all global variables *)
  let createGlobalVars = createInitVars

  (** check-assumptions block *)
  let mkCheckAssumeBlock (r:int) : stmt =  
    (* recursive function to check assumption for arrays *)
    let rec assumeRec (l1:lval) (l2:lval) (il:instr list) : instr list =
      match (typeOfLval l1) with
          TArray(_,eo,_) ->
            let ilref = ref il in
            for i = 0 to ((lenOfArray eo) - 1) do
              let os = Index(integer i,NoOffset) in
              ilref := assumeRec (addOffsetLval os l1) (addOffsetLval os l2) !ilref
            done;
            !ilref
        | _ -> il @
          [(Formatcil.cInstr
	      "assume (%l:l1 == %l:l2);"
	      rekloc
	      [
	        ("assume", Fv assumeFn);
	        ("l1", Fl (l1));
	        ("l2", Fl (l2));
	      ]
          )]
    in
    let assumeBlock = 
      Svm.foldShared 
        (fun v sv res -> 
          let l1 = Svm.getSharedLval sv r in
          let l2 = Svm.getSharedInitLval sv (r+1) in
          assumeRec l1 l2 res
        ) []
    in    
    mkStmt (Instr assumeBlock)

  (** Deprecated check-assumptions block *)
  let mkCheckAssumeBlock_old () =  
    let succRound = 
      BinOp (PlusA, (Lval (var roundVar)), one, roundVar.vtype) in
    let assumeBlock = 
      Svm.foldShared (fun v sv res -> 
	(Formatcil.cInstr
	   "assume (%l:g == %l:v);"
	   rekloc
	   [
	     ("assume", Fv assumeFn);
	     ("g", Fl (Svm.getSharedExpLval sv (Lval (var roundVar))));
	     ("v", Fl (Svm.getSharedInitExpLval sv succRound));
	   ]) :: res) [] in
    
    let thenBlock = mkBlock [mkStmt (Instr assumeBlock)] in
    let x = Lval (var roundVar) in
    let y = (integer (pred rounds)) in
    mkStmt (If (BinOp (Lt, x, y, boolType), thenBlock, mkBlock [], rekloc))


  (** check-assumptions block specific to a given job *)
  let mkJobCheckAssumeBlock (job:jobt) : Cil.stmt =  
    let low = jobLowRound job in 
    let high = let x = jobHighRound job in 
	       if x = pred rounds then pred x else x in

    (* recursive function to check assumption for arrays *)
    let rec assumeRec (l1:lval) (l2:lval) (il:instr list) : instr list =
      match (typeOfLval l1) with
          TArray(_,eo,_) ->
            let ilref = ref il in
            for i = 0 to ((lenOfArray eo) - 1) do
              let os = Index(integer i,NoOffset) in
              ilref := assumeRec (addOffsetLval os l1) (addOffsetLval os l2) !ilref
            done;
            !ilref
        | _ -> il @
          [(Formatcil.cInstr
	      "assume (%l:l1 == %l:l2);"
	      rekloc
	      [
	        ("assume", Fv assumeFn);
	        ("l1", Fl (l1));
	        ("l2", Fl (l2));
	      ]
          )]
    in

    let mkAssumeBlock (n:int) = 
      Svm.foldShared 
        (fun v sv res -> 
          let l1 = Svm.getSharedLval sv n in
          let l2 = Svm.getSharedInitLval sv (n+1) in
          assumeRec l1 l2 res
        ) []
    in

    (* high < low implies that low is in the last round *)
    if high < low then mkEmptyStmt () else
      mkSwitchStmt (Lval (var jobEndVar)) low (high - low + 1) 
	(fun n -> [mkStmt (Instr (mkAssumeBlock n))])

  (*********************************************************************)
  (** create a bunch of utility functions *)
  (*********************************************************************)

  let createAbort unit : unit =
    let fd = findOrCreateFundec Args.file abortFnName abortFnType in
    fd.svar.vinline <- true;
    fd.svar.vstorage <- Static; 
    let fmt = 
      [
        ("taskvar", Fl(var taskVar));
	("lock", Fv lockVar);
        ("pipunlock", Fv(findOrCreateFunc Args.file pipUnlockFnName pipUnlockFnType));
	]
      in
    (* the body of abort *)
    let body = ref(Formatcil.cStmts "%l:lock = 0;"
	             (fun n t -> makeTempVar fd ~name:n t)
	             rekloc fmt)
    in
    for i = 0 to (numPIPLocks - 1) do
      let svar = addOffsetLval (Index(integer i,NoOffset))
        (Svm.getSharedExpLval (Svm.getSharedVar Args.sVar) (Lval (var roundVar))) 
      in
      body := !body @ Formatcil.cStmts
        "if (%l:svar == %l:taskvar) pipunlock (%e:l);"
	(fun n t -> makeTempVar fd ~name:n t)
	rekloc (fmt @ [("svar", Fl(svar)); ("l", Fe(integer i))])
    done;
    fd.sbody.bstmts <- !body @ fd.sbody.bstmts

  let createUnblock unit : unit =
    let fd = findOrCreateFundec Args.file unblockFnName unblockFnType in
    fd.svar.vinline <- true;
    fd.svar.vstorage <- Static;
    let par = List.hd (fd.sformals) in
    let ovar = makeTempVar fd ~name:"o" roundVar.vtype in
    
    (* the body of unblock *)
    let body = 
      let sv = Svm.getSharedVar Args.cpVar in 
      let cpold = Lval (Svm.getSharedExpLval sv (Lval (var ovar))) in
      let cpnew = Lval (Svm.getSharedExpLval sv (Lval (var roundVar))) in
      let svar = addOffsetLval (Index(Lval(var par),NoOffset))
        (Svm.getSharedExpLval (Svm.getSharedVar Args.sVar) (Lval (var roundVar))) 
      in
      Formatcil.cStmts 
        "
          %v:ovar = round; round = nd();
          assume (round > ovar); assume (round <= ej);              
          assume (%e:cpnew == %e:cpold);
          assume (%e:svar == -1);
          if (%e:aborted) abort();
          return 1;
        "
	(fun n t -> makeTempVar fd ~name:n t)
	rekloc
	[
	  ("rty", Ft roundVar.vtype);
	  ("round", Fv roundVar);
	  ("ovar", Fv ovar);
	  ("nd", Fv (nondetFn roundVar.vtype));
	  ("assume", Fv assumeFn);
	  ("ej", Fv jobEndVar);
          ("cpold", Fe(cpold));
          ("cpnew", Fe(cpnew));
          ("svar", Fe(Lval(svar)));
          ("aborted", Fe(abortCond ()));
          ("abort", Fv(findOrCreateFunc Args.file abortFnName abortFnType));
	]
    in
    fd.sbody.bstmts <- body @ fd.sbody.bstmts

  let createPipLock unit : unit =
    let fd = findOrCreateFundec Args.file pipLockFnName pipLockFnType in
    fd.svar.vinline <- true;
    fd.svar.vstorage <- Static;
    let par = List.hd (fd.sformals) in

    (* the body of context switch *)
    let body = 
      let rtvar = addOffsetLval (Index(Lval(var taskVar),NoOffset))
        (Svm.getSharedExpLval (Svm.getSharedVar Args.rtVar) (Lval (var roundVar)))
      in
      let rlvar = addOffsetLval (Index(Lval(var par),NoOffset))
        (Svm.getSharedExpLval (Svm.getSharedVar Args.rlVar) (Lval (var roundVar)))
      in 
      let svar = addOffsetLval (Index(Lval(var par),NoOffset))
        (Svm.getSharedExpLval (Svm.getSharedVar Args.sVar) (Lval (var roundVar))) 
      in
      let rtsvar = addOffsetLval (Index(Lval svar,NoOffset))
        (Svm.getSharedExpLval (Svm.getSharedVar Args.rtVar) (Lval (var roundVar)))
      in
      let evar = Svm.getSharedExpLval (Svm.getSharedVar Args.eVar) (Lval (var roundVar)) in
      let bvar = Sched.roundBlockLval (Lval (var roundVar)) in
      let fmt = 
	[
          ("par", Fv(par));
          ("rtvar", Fl(rtvar));
          ("rlvar", Fl(rlvar));
          ("svar", Fl(svar));
          ("rtsvar", Fl(rtsvar));
          ("evar", Fl(evar));
          ("bvar", Fl(bvar));
          ("taskvar", Fl(var taskVar));
	  ("assume", Fv assumeFn);
	  ("lock", Fv lockVar);
	  ("abort", Fv (findOrCreateFunc Args.file abortFnName abortFnType));
	  ("unblock", Fv (findOrCreateFunc Args.file unblockFnName unblockFnType));
          ("aborted", Fe(abortCond ()));
	]
      in
      let st1 = Formatcil.cStmts 
        "
          if (%l:svar == -1) {
            %l:svar = %l:taskvar;
            %l:rtvar = %l:rtvar | %l:rlvar;
            return;
          }
        "
	(fun n t -> makeTempVar fd ~name:n t)
	rekloc fmt
      in      
      (** check if we need more rounds *)
      let roundCond = ref (Lval(Sched.roundBlockLval zero)) in
      for i = 1 to (rounds - 1) do
        roundCond := BinOp(LOr, !roundCond, Lval(Sched.roundBlockLval (integer i)), boolType)
      done;
      let roundsCheck = Formatcil.cStmts 
        (
          if (!checkRounds) then
            "
              if (! (%e:rc)) {
                %l:evar = 1;
                abort();
                return;
              }      
            "
          else
            "
            "
        )
	(fun n t -> makeTempVar fd ~name:n t)
	rekloc (fmt @ [("rc", Fe(!roundCond))])
      in
      (** deadlock check 1 *) 
      let dead1 = Formatcil.cStmts 
        (
          if (!checkRounds) then
            "
              assume(! %v:lock);
            "
          else
            "
              if (%v:lock) {
                %l:evar = 1;
                abort();
                return;
              }      
            "
        )
	(fun n t -> makeTempVar fd ~name:n t)
	rekloc fmt
      in
      (** deadlock check 2 *) 
      let dead2 = Formatcil.cStmts 
        (
          if (!checkRounds) then
            "
              assume(! (%l:rtvar & (0x1 << %l:svar)));
            "
          else
            "
              if (%l:rtvar & (0x1 << %l:svar)) {
                %l:evar = 1;
                abort();
                return;
              }      
            "
        )
	(fun n t -> makeTempVar fd ~name:n t)
	rekloc fmt
      in
      let st2 = Formatcil.cStmts 
        "
          assume (%l:bvar);
          %l:bvar = 0;

          %l:rlvar = %l:rlvar | %l:rtvar | (0x1 << %l:taskvar);
          %l:rtsvar = %l:rtsvar | %l:rtvar | (0x1 << %l:taskvar);

          unblock (%v:par);

          %l:rlvar = %l:rlvar & (~ %l:rtvar) & (~(0x1 << %l:taskvar));

          if (%e:aborted) return;

          %l:svar = %l:taskvar;
          %l:rtvar = %l:rtvar | %l:rlvar;
        "
	(fun n t -> makeTempVar fd ~name:n t)
	rekloc fmt
      in
      st1 @ roundsCheck @ dead1 @ dead2 @ st2
    in
    fd.sbody.bstmts <- body @ fd.sbody.bstmts

  let createPipUnlock unit : unit =
    let fd = findOrCreateFundec Args.file pipUnlockFnName pipUnlockFnType in
    fd.svar.vinline <- true;
    fd.svar.vstorage <- Static;
    let par = List.hd (fd.sformals) in

    (* the body of context switch *)
    let body = 
      let rtvar = addOffsetLval (Index(Lval(var taskVar),NoOffset))
        (Svm.getSharedExpLval (Svm.getSharedVar Args.rtVar) (Lval (var roundVar)))
      in
      let rlvar = addOffsetLval (Index(Lval(var par),NoOffset))
        (Svm.getSharedExpLval (Svm.getSharedVar Args.rlVar) (Lval (var roundVar)))
      in 
      let svar = addOffsetLval (Index(Lval(var par),NoOffset))
        (Svm.getSharedExpLval (Svm.getSharedVar Args.sVar) (Lval (var roundVar))) 
      in
      Formatcil.cStmts 
        "
          %l:svar = -1;
          %l:rtvar = %l:rtvar & ~ %l:rlvar;
        "
	(fun n t -> makeTempVar fd ~name:n t)
	rekloc
	[
          ("rtvar", Fl(rtvar));
          ("rlvar", Fl(rlvar));
          ("svar", Fl(svar));
	]
    in
    fd.sbody.bstmts <- body @ fd.sbody.bstmts

  let createDoCS unit : unit =
    let fd = findOrCreateFundec Args.file doCSFnName doCSFnType in
    fd.svar.vinline <- true;
    fd.svar.vstorage <- Static;
    let cvar = makeTempVar fd ~name:"c" boolType in
    let retStmt = mkStmt (Return (Some (Lval (var cvar)), builtinLoc)) in
    let makeSwitchRead = 
      (* list of statements *)
      let slist =       
        let fn (t:task) (res:stmt list) : stmt list = 
	  let caseBody = 
	    mkStmt 
	      (Block 
	         (mkBlock
		    (createStmts
		       "c = task_cs(); break;"
		       [("task_cs", Fv (getCsFn t)); ("c", Fv cvar)]
		    ))) 
	  in	
	(* add a case label *)
	  caseBody.labels <- 
	    (Case (integer t.tid, builtinLoc)) :: caseBody.labels ;
	  caseBody :: res	    
        in
        List.fold_right fn !tasks [] 
      in
      mkStmt (Switch (Lval (var taskVar),mkBlock slist, slist, builtinLoc))
    in
    fd.sbody <- mkBlock ([makeSwitchRead; retStmt])

  let createIP unit : unit =
    let fd = findOrCreateFundec Args.file ipFnName ipFnType in
    fd.svar.vinline <- true;
    fd.svar.vstorage <- Static;
    let par = List.hd (fd.sformals) in
    let sv = Svm.getSharedVar Args.rtVar in 
    let rt = Svm.getSharedExpLval sv (Lval (var roundVar)) in
    let rtt = addOffsetLval (Index(Lval(var taskVar),NoOffset)) rt in
    let st = ref [] in
    for i = (numOfTasks() - 1) downto 1 do 
      st := !st @
        createStmts
          "if (%e:bp == %e:i %b:or %e:rtt & (0x1 << %e:i)) return %e:i;"
          [("bp", Fe (Lval(var par)));
	   ("i", Fe (integer i));
           ("rtt",Fe(Lval(rtt)));
           ("or",Fb(LOr))]
    done;
    st := !st @ [mkStmt (Return (Some(zero), builtinLoc))];
    fd.sbody <- mkBlock (!st)
    
  let createCheckAssump unit : unit =
    let fd = findOrCreateFundec Args.file checkAssumpFnName checkAssumpFnType in
    fd.svar.vinline <- true;
    fd.svar.vstorage <- Static;
    let st = ref [] in
    for i = 0 to (rounds - 2) do
      st := !st @ [mkCheckAssumeBlock i]
    done;
    for i = 0 to (rounds - 1) do
      let bvar = Sched.roundBlockLval (integer i) in
      st := !st @ createStmts "assume (%l:bvar == 0);" [("bvar", Fl(bvar))]
    done;
    fd.sbody <- mkBlock (!st)

  let createCheckAssert unit : unit =
    let fd = findOrCreateFundec Args.file checkAssertFnName checkAssertFnType in
    fd.svar.vinline <- true;
    fd.svar.vstorage <- Static;
    let st = createStmts
      ( (if (!checkSanity) then "assert (0);" else "") ^ "assume (%l:evar); assert (0);" )
      [("evar", Fl(Svm.getSharedLval (Svm.getSharedVar Args.eVar) (rounds - 1)));]
    in
    fd.sbody <- mkBlock (st)

  let createUtilFuncs unit : unit =
    createAbort();
    createUnblock();
    createPipLock();
    createPipUnlock();
    createDoCS ();
    createIP ();
    createCheckAssump ();
    createCheckAssert ()

  (*********************************************************************)
  (*** creating job scheduling function **)
  (*********************************************************************)

  (** create schedule_jobs  function **)
  let createScheduleJobs unit : unit =
    let ftype =  TFun (voidType,Some [],false,[]) in
    let fd = findOrCreateFundec Args.file scheduleJobsFnName ftype in
    fd.svar.vinline <- true;

    (** Returns nesting constraints for job j2 of a low priority task
	t2 and job j1 of a higher priority task t1 **)
    let nestingConstraints j1 j2 =
      let 
	  jid1 = BT.id !jobTree j1 and
	  jid2 = BT.id !jobTree j2 and
	  tid1 = BT.level_of j1 and
	  tid2 = BT.level_of j2 
      in
      let
    	  s1 = Sched.jobStartLval jid1 and
    	  s2 = Sched.jobStartLval jid2 and
    	  e1 = Sched.jobEndLval jid1 and
    	  e2 = Sched.jobEndLval jid2 in
      
      let pb = if !nopbounds then 0 else
      	  !taskArray.(tid2).pbounds.(tid1 - tid2 - 1) in
      let pbound = 
	if !nopbounds then []
	else
	  begin
	    try
	      let priorJob = BT.stepInGeneration !jobTree j2 j1 (-pb) in
	      createStmts
		"assume (%l:je < %l:s2);"
		[
		  ("s2", Fl s2);
		  ("je", Fl (Sched.jobEndLval (BT.id !jobTree priorJob) ));
		]
	    with Invalid_argument _ -> []
	  end
      in 

      if jobPreceeds j1 j2 then
	createStmtsLoc "assume(%l:e1 < %l:s2);" rekloc
	  [ ("e1", Fl e1); ("s2", Fl s2) ]
      else if jobFollows j1 j2 then
	createStmtsLoc "assume(%l:e2 < %l:s1);" rekloc
	  [ ("e2", Fl e2); ("s1", Fl s1); ]
      else
	(* 
           let p1 = if jobMayNotPreceed j1 j2 then
	    Formatcil.cStmts
	      "assume(%l:s2 <= %l:s1);"
	      (fun n t -> assert false)
	      rekloc
	      [("s1", Fl s1); ("s2", Fl s2); ("assume", Fv assumeFn);]
	  else [] in 
        *)
	let p2 = if jobMayNotFollow j1 j2 then
            createStmtsLoc "assume(%l:e1 <= %l:e2);" rekloc
	      [("e1", Fl e1); ("e2", Fl e2)]
	  else [] in
	p2 @
	  createStmts
      	    "if (%l:s2 <= %l:e1)
               if (%l:s1 <= %l:e2)
                 {
                   assume (%l:s2 <= %l:s1);
                   assume (%l:e1 < %l:e2);
                   %S:pbound
                 }
            "
    	  [
    	    ("s1", Fl s1);
    	    ("s2", Fl s2);
    	    ("e1", Fl e1);
    	    ("e2", Fl e2);
	    ("pbound", FS pbound);
    	  ]	  
    in
    
    let minStmt x y z =
      Formatcil.cStmt
	"if (%l:y < %l:z)  assume(%l:x == %l:y); else assume(%l:x == %l:z);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("z", Fl z);
	  ("assume", Fv assumeFn)
	] in

    let maxStmt x y z =
      Formatcil.cStmt
	"if (%l:y > %l:z)  assume(%l:x == %l:y); else assume(%l:x == %l:z);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("z", Fl z);
	  ("assume", Fv assumeFn)
	] in

    let leqStmt x y = 
      Formatcil.cStmt
	"assume (%l:x <= %l:y);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("assume", Fv assumeFn)
	] in

    let ltStmt x y = 
      Formatcil.cStmt
	"assume (%l:x < %l:y);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("assume", Fv assumeFn)
	] in

    let eqStmt x y = 
      Formatcil.cStmt
	"assume (%l:x == %l:y);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("assume", Fv assumeFn)
	] in
	
    let perJob tree job : bool =
      let jid = BT.id tree job in
      
      let minV = Sched.jobMinLval jid in
      let maxV = Sched.jobMaxLval jid in
      let startV = Sched.jobStartLval jid in
      let endV = Sched.jobEndLval jid  in
      
      if BT.isLeaf tree job then 
	fd.sbody.bstmts <- 
	  (leqStmt startV endV) ::
	  (eqStmt minV startV) :: 
	  (eqStmt maxV  endV) :: 
	  fd.sbody.bstmts
      else
	begin
	  let cminV = Sched.jobMinLval (BT.id tree (BT.first_child tree job)) in
	  let cmaxV = Sched.jobMaxLval (BT.id tree (BT.last_child tree job)) in
	(* add min/max statements for the current job *)
	  fd.sbody.bstmts <- 
	    (leqStmt startV endV) ::
	    (minStmt minV startV cminV) ::
	    (maxStmt maxV endV cmaxV) :: fd.sbody.bstmts 
	end
      ;
	(* ensure that siblings start/end times  are ordered *)
      if BT.hasNext tree job then 
	begin
	  let minNextV = Sched.jobMinLval (BT.id tree (BT.next tree job)) in
	  fd.sbody.bstmts <-
	    (ltStmt maxV minNextV) :: fd.sbody.bstmts 
	end
      ;

      fd.sbody.bstmts <- 
	(Formatcil.cStmts
	   "assume (%e:job_low <= %l:start);
            assume (%l:end <= %e:job_high);"
	   (fun n t -> assert false)
	   rekloc
	   [
	     ("assume", Fv assumeFn);
	     ("end", Fl endV);
	     ("start", Fl startV);
	     ("job_low", Fe (integer (jobLowRound job)));
	     (* ("job_high", Fe (integer (jobHighRound job))); *)
	     ("job_high", Fe (integer (rounds - 1)));
	   ]) @ fd.sbody.bstmts ;


      if !nesting && not (BT.isRoot job) then 
	begin
	  BT.iter_parents  
	    (fun t j2 -> 
	      fd.sbody.bstmts <- 
		fd.sbody.bstmts @ nestingConstraints job j2 
	    ) 
	    tree (BT.parent_of job)
	end
      ;
      true
    in
    BT.iter_dfs perJob !jobTree (BT.root !jobTree)  ; 
    (* initialize all variables non-deterministically *)
    BT.iter_dfs 
      (fun tree job -> 
	let jid = BT.id tree job in 
        let bvar = Sched.roundBlockLval (Lval (Sched.jobEndLval jid)) in
	fd.sbody.bstmts <- 
	  (Formatcil.cStmts 
	  "%l:start=ndc(); %l:end=ndc (); %l:min=ndc(); %l:max=ndc (); %l:bvar = 0;"
	  (fun n t -> assert false)
	  rekloc
	  [
	    ("start", Fl (Sched.jobStartLval jid));
	    ("end", Fl (Sched.jobEndLval jid));
	    ("min", Fl (Sched.jobMinLval jid));
	    ("max", Fl (Sched.jobMaxLval jid));
	    ("ndc", Fv (nondetFn ucharType));
            ("bvar", Fl(bvar));
	  ]) @ fd.sbody.bstmts 
	;
	true

      ) 
      !jobTree (BT.root !jobTree)

  (*********************************************************************)
  (*** creating the init function **)
  (*********************************************************************)
    
  (*
    vi3[j][s] = ndVi (); vi2[j][s] = vi3[j][s]
  *)

  (** create a function that initializes rounds 1 to R-1 *)
  let createInitNonZero (f:file) =
    let initType = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f initNonZeroFnName initType in
    fd.svar.vinline <- true;

    (* recursive function to initialize arrays *)
    let rec initRec (l1:lval) (l2:lval) : unit =
      let ltype = typeOfLval l1 in
      match ltype with
          TArray(_,eo,_) ->
            for i = 0 to ((lenOfArray eo) - 1) do
              let os = Index(integer i,NoOffset) in
              initRec (addOffsetLval os l1) (addOffsetLval os l2)
            done
        | _ -> 
	  fd.sbody.bstmts <-
	    fd.sbody.bstmts @
	    Formatcil.cStmts
	    "%l:sc = nd () ; %l:xs = %l:sc;"
	    (fun n t -> makeTempVar fd ~name:n t)
	    builtinLoc
	    [("sc", Fl (l1)); 
	     ("xs", Fl (l2));
	     ("nd", Fv (nondetFn ltype))
	    ]
    in

    (* for every shared variable *)
    Svm.iterShared
      (
        fun v sv ->
	(* for every round, except the first one *)
	  for r = 1 to (rounds-1) do
            let initVal = Svm.getSharedInitLval sv r in
            let sharedVal = Svm.getSharedLval sv r in
            initRec initVal sharedVal
	  done
      )
	
  (*********************************************************************)
  (*** creating the tasks **)
  (*********************************************************************)

  (** rewrites all accesses to shared variables, lock functions, and
      asserts *)
  let createTasks (f:file) = 
    iterGlobals 
      f 
      (fun g -> 
	match g with 
	  | GFun(fd, _) -> 
            let fn = fd.svar.vname in
            let checkAbort = 
              (not (Util.hasPrefix "__startrek_entry_pt" fn)) &&
              (not (Util.hasPrefix "cil_keeper" fn))
            in
	    if (not (fn = userFinalFnName) && not (fn = userInitialFnName)) then
              let glblr = new globalReplacer checkAbort in              
	      fd.sbody <- visitCilBlock glblr fd.sbody
	  | _ -> ()
      )
	
  (*********************************************************************)
  (*** rewritting user-supplied init and final functions **)
  (*********************************************************************)

  let createUserInitial (f:file) =
    let finalType = TFun (voidType, Some [], false,[]) in
    let fd = findOrCreateFundec f userInitialFnName finalType in
    fd.svar.vinline <- true;

    
    fd.sbody <-  visitCilBlock (new initialReplacer) fd.sbody	

  (* create a final function *)
  let createUserFinal (f:file) =
    let finalType = TFun (voidType, Some [], false,[]) in
    let fd = findOrCreateFundec f userFinalFnName finalType in
    fd.svar.vinline <- true;
    fd.sbody <-  visitCilBlock (new finalReplacer) fd.sbody
      
  (*********************************************************************)
  (*** creating the main function **)
  (*********************************************************************)
  (** Encoding of a hyper-period when arrivals are fixed *)
  let createHyperPeriod_NoRange (f: file) = 
    let ftype = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f hyperFnName ftype in

    let jobOrder = (fun a b -> 
      if a = b then 0 
      else if jobPriority a >= jobPriority b then
	if jobPreceeds a b then -1 else 1
      else (if jobPreceeds b a then 1 else -1)
    ) in
    let orderedJobs = List.sort jobOrder 
      (BT.fold_dfs (fun res t n -> n ::res) [] !jobTree (BT.root !jobTree))
    in

    List.iter 
      (fun j -> 
	print_string ((string_of_int (BT.id !jobTree j)) ^ " ")) orderedJobs ;
    print_string "\n";
    
    let encode (job:jobt) : stmt list = 
      let jid = BT.id !jobTree job in
      let tid = BT.level_of job in
      Formatcil.cStmts
	"
           task = %e:mytask;
           job = %e:jid;
           round = %l:mystart;

           if (! %e:aborted) {
             end = %l:myend;
             task_entry_pt ();
             %v:lock = 0;
             assume (round == end);
          }           
        "
	(fun n t -> assert false)
	rekloc
	[
	  ("assume", Fv assumeFn);
	  ("task", Fv taskVar);
	  ("mytask", Fe (integer tid));
	  ("round", Fv roundVar);
	  ("end", Fv jobEndVar);
	  ("mystart", Fl (Sched.jobStartLval jid));
	  ("myend", Fl (Sched.jobEndLval jid));
	  ("job", Fv jobVar);
	  ("jid", Fe (integer jid));
          ("aborted", Fe(abortCond ()));
	  ("task_entry_pt", Fv (oget (!taskArray.(tid).entryPt)).svar);
          ("lock", Fv lockVar);
	]

    in

    let jobs = List.fold_left (fun r j -> r @ (encode j) ) []  orderedJobs
    in

    let sjobs = mkStmtOneInstr (Call (None, Lval (var (findOrCreateFunc f scheduleJobsFnName ftype)), [], rekloc)) in
    let initfn = mkStmtOneInstr (Call (None, Lval (var (findOrCreateFunc f initNonZeroFnName ftype)), [], rekloc)) in
    let checkAssump = mkStmtOneInstr (Call (None, Lval (var (findOrCreateFunc f checkAssumpFnName checkAssumpFnType)), [], rekloc)) in
    let checkAssert = mkStmtOneInstr (Call (None, Lval (var (findOrCreateFunc f checkAssertFnName checkAssertFnType)), [], rekloc)) in

    fd.sbody <- mkBlock ((sjobs :: initfn :: jobs) @ [checkAssump; checkAssert])

  (** Encoding of a hyper-period when arrivals are fixed *)
  let createHyperPeriod_NoRange_old (f: file) = 
    let ftype = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f hyperFnName ftype in

    let jobOrder = (fun a b -> 
      if a = b then 0 
      else if jobPriority a >= jobPriority b then
	if jobPreceeds a b then -1 else 1
      else (if jobPreceeds b a then 1 else -1)
    ) in
    let orderedJobs = List.sort jobOrder 
      (BT.fold_dfs (fun res t n -> n ::res) [] !jobTree (BT.root !jobTree))
    in

    List.iter 
      (fun j -> 
	print_string ((string_of_int (BT.id !jobTree j)) ^ " ")) orderedJobs ;
    print_string "\n";
    
    let encode (job:jobt) : stmt list = 
      let jid = BT.id !jobTree job in
      let tid = BT.level_of job in

	
      let (assertedJobs:jobt list) =
	BT.fold_parents 
	  (fun res t low -> 
	    if (job = low || jobMayPreempt job low) &&
	      BT.forall 
	      (fun t med -> 
		(not (jobMayPreempt med low)) || (jobOrder med job) <= 0)
	      t low 
	    then 
	      low :: res
	    else res
	  )
	  [] !jobTree job
      in
      print_string ("assrt after " ^ (string_of_int jid) ^ ": ");
      List.iter (fun j -> 
	print_string ((string_of_int (BT.id !jobTree j)) ^ " ")) assertedJobs;
      print_string "\n";
      let asserts = 
        let acond = Svm.getSharedExpLval (Svm.getSharedVar Args.eVar) (Lval (var roundVar)) in
        [Call (None, Lval (var assertFn), [Lval acond], rekloc)]
      in
      Formatcil.cStmts
	"
           round = %l:mystart;
           task = %e:mytask;
           end = %l:myend;
           job = %e:jid;

           task_entry_pt ();
           assume (round == end);
           
           %s:checkAssume

           %I:checkAssert
        "
	(fun n t -> assert false)
	rekloc
	[
	  ("assume", Fv assumeFn);
	  ("task", Fv taskVar);
	  ("mytask", Fe (integer tid));
	  ("round", Fv roundVar);
	  ("end", Fv jobEndVar);
	  ("mystart", Fl (Sched.jobStartLval jid));
	  ("myend", Fl (Sched.jobEndLval jid));
	  ("job", Fv jobVar);
	  ("jid", Fe (integer jid));
	  ("task_entry_pt", Fv (oget (!taskArray.(tid).entryPt)).svar);
	  ("checkAssume", Fs (mkJobCheckAssumeBlock job));  
	  (* ("checkAssume", Fs (mkCheckAssumeBlock ())); *)
	  (* ("checkAssert", FI (mkCheckAssertBlock tid)); *)
	  ("checkAssert", FI asserts);
	]

    in

    let jobs = List.fold_left (fun r j -> r @ (encode j) ) []  orderedJobs
    in

    let sjobs = mkStmtOneInstr (Call (None, Lval (var (findOrCreateFunc f scheduleJobsFnName ftype)), [], rekloc)) in
    let initfn = mkStmtOneInstr (Call (None, Lval (var (findOrCreateFunc f initNonZeroFnName ftype)), [], rekloc)) in
    let checkAssump = mkStmtOneInstr (Call (None, Lval (var (findOrCreateFunc f checkAssumpFnName checkAssumpFnType)), [], rekloc)) in
    let checkAssert = mkStmtOneInstr (Call (None, Lval (var (findOrCreateFunc f checkAssertFnName checkAssertFnType)), [], rekloc)) in

    fd.sbody <- mkBlock ((sjobs :: initfn :: jobs) @ [checkAssump; checkAssert])

  (** Encoding of a hyper-period when arrivals are fixed *)
  let createHyperPeriod_NoRange_old (f: file) = 
    let ftype = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f hyperFnName ftype in

    let jobOrder = (fun a b -> 
      if a = b then 0 
      else if jobPriority a >= jobPriority b then
	if jobPreceeds a b then -1 else 1
      else (if jobPreceeds b a then 1 else -1)
    ) in
    let orderedJobs = List.sort jobOrder 
      (BT.fold_dfs (fun res t n -> n ::res) [] !jobTree (BT.root !jobTree))
    in

    List.iter 
      (fun j -> 
	print_string ((string_of_int (BT.id !jobTree j)) ^ " ")) orderedJobs ;
    print_string "\n";
    
    let encode (job:jobt) : stmt list = 
      let jid = BT.id !jobTree job in
      let tid = BT.level_of job in

	
      let (assertedJobs:jobt list) =
	BT.fold_parents 
	  (fun res t low -> 
	    if (job = low || jobMayPreempt job low) &&
	      BT.forall 
	      (fun t med -> 
		(not (jobMayPreempt med low)) || (jobOrder med job) <= 0)
	      t low 
	    then 
	      low :: res
	    else res
	  )
	  [] !jobTree job
      in
      print_string ("assrt after " ^ (string_of_int jid) ^ ": ");
      List.iter (fun j -> 
	print_string ((string_of_int (BT.id !jobTree j)) ^ " ")) assertedJobs;
      print_string "\n";
      let asserts = 
        let acond = Svm.getSharedExpLval (Svm.getSharedVar Args.eVar) (Lval (var roundVar)) in
        [Call (None, Lval (var assertFn), [Lval acond], rekloc)]
      in
      Formatcil.cStmts
	"
           round = %l:mystart;
           task = %e:mytask;
           end = %l:myend;
           job = %e:jid;

           task_entry_pt ();
           assume (round == end);
           
           %s:checkAssume

           %I:checkAssert
        "
	(fun n t -> assert false)
	rekloc
	[
	  ("assume", Fv assumeFn);
	  ("task", Fv taskVar);
	  ("mytask", Fe (integer tid));
	  ("round", Fv roundVar);
	  ("end", Fv jobEndVar);
	  ("mystart", Fl (Sched.jobStartLval jid));
	  ("myend", Fl (Sched.jobEndLval jid));
	  ("job", Fv jobVar);
	  ("jid", Fe (integer jid));
	  ("task_entry_pt", Fv (oget (!taskArray.(tid).entryPt)).svar);
	  ("checkAssume", Fs (mkJobCheckAssumeBlock job));  
	  (* ("checkAssume", Fs (mkCheckAssumeBlock ())); *)
	  (* ("checkAssert", FI (mkCheckAssertBlock tid)); *)
	  ("checkAssert", FI asserts);
	]

    in

    let jobs = List.fold_left (fun r j -> r @ (encode j) ) []  orderedJobs
    in

    let sjobs = 
      mkStmtOneInstr 
	(Call (None, 
	       Lval (var (findOrCreateFunc f scheduleJobsFnName ftype)), 
	       [], rekloc)) in
    let initfn = 
      mkStmtOneInstr
	(Call (None, 
	       Lval (var (findOrCreateFunc f initNonZeroFnName ftype)), 
	       [], rekloc)) in
    fd.sbody <- mkBlock (sjobs :: initfn :: jobs)
       
  (** Encoding of a hyper-period with arival times range *)
  let createHyperPeriod_WithRange (f: file) = 
    let ftype = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f hyperFnName ftype in
    
    let rec encode_siblings job res = 
      let x = res @ (encode job) in
      if BT.hasNext !jobTree job then  
	encode_siblings (BT.next !jobTree job) x
      else x

    and encode job = 
      let kids = if BT.isLeaf !jobTree job then [] else
	  encode_siblings (BT.first_child !jobTree job) [] in
      
      let jid = BT.id !jobTree job in
      let tid = BT.level_of job in
      let asserts = [Call (None, Lval (var assertFn), [zero], rekloc)] in

      Formatcil.cStmts
	"
           round = %l:mystart;
           task = %e:mytask;
           end = %l:myend;
           job = %e:jid;

           task_entry_pt ();
           assume (round == end);
           
           %s:checkAssume

           { /* KIDS */ %S:kids  }
 
           %I:checkAssert
        "
	(fun n t -> assert false)
	rekloc
	[
	  ("assume", Fv assumeFn);
	  ("task", Fv taskVar);
	  ("mytask", Fe (integer tid));
	  ("round", Fv roundVar);
	  ("end", Fv jobEndVar);
	  ("mystart", Fl (Sched.jobStartLval jid));
	  ("myend", Fl (Sched.jobEndLval jid));
	  ("job", Fv jobVar);
	  ("jid", Fe (integer jid));
	  ("task_entry_pt", Fv (oget (!taskArray.(tid).entryPt)).svar);
	  ("checkAssume", Fs (mkJobCheckAssumeBlock job));  
	  (* ("checkAssume", Fs (mkCheckAssumeBlock ())); *)
	  ("kids", FS kids);
	  ("checkAssert", FI asserts);
	]
    in
    let sjobs = 
      mkStmtOneInstr 
	(Call (None, 
	       Lval (var (findOrCreateFunc f scheduleJobsFnName ftype)), 
	       [], rekloc)) in
    let initfn = 
      mkStmtOneInstr
	(Call (None, 
	       Lval (var (findOrCreateFunc f initNonZeroFnName ftype)), 
	       [], rekloc)) in
    fd.sbody <- mkBlock (sjobs :: initfn :: encode (BT.root !jobTree))

  let createHyperPeriod =  
    if !rangesInArrival then createHyperPeriod_WithRange 
    else createHyperPeriod_NoRange
	
  let createMain (f:file) = 
    let ftype = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec f mainFnName ftype in

    let new_period = 
      Svm.foldShared 
	(fun v sv r -> 
	  let lhs = Svm.getSharedLval sv 0 in 
	  let rhs = Svm.getSharedLval sv (pred rounds) in 
	  mkStmtOneInstr (Set (lhs, Lval rhs, rekloc)) :: r
	) 
	[] 
    in
    
    let initSharedFn = findOrCreateFunc Args.file initZeroFnName ftype in
    let userInitFn = findOrCreateFunc Args.file userInitialFnName ftype in

    let stmts = ref (Formatcil.cStmts "initSharedFn(); userInitFn();"
           (fun n t -> makeTempVar fd ~name:n t) rekloc
           [ ("initSharedFn", Fv initSharedFn); ("userInitFn", Fv userInitFn); ])
    in
      
    let hyperFn = findOrCreateFunc Args.file hyperFnName ftype in
      if (!unboundedSeq) then (
        let loopBody = (Formatcil.cStmts "hyperFn();"
          (fun n t -> makeTempVar fd ~name:n t) rekloc
          [ ("hyperFn", Fv hyperFn); ]) @ new_period 
        in
        stmts := !stmts @ (mkWhile (integer 1) loopBody)
      ) else (
        for i = 1 to !hyperPeriods do
          stmts := !stmts @ Formatcil.cStmts "%v:hv = %d:i; hyperFn();"
            (fun n t -> makeTempVar fd ~name:n t) rekloc
            [ ("hv", Fv hyperPeriodVar); ("i", Fd (i - 1)); ("hyperFn", Fv hyperFn); ];
          if (i != !hyperPeriods) then stmts := !stmts @ new_period;
        done ;
      );
      let finalFn = findOrCreateFunc Args.file userFinalFnName ftype in
        stmts := !stmts @ Formatcil.cStmts "finalFn();"
           (fun n t -> makeTempVar fd ~name:n t) rekloc
           [ ("finalFn", Fv finalFn); ];
        fd.sbody <- mkBlock !stmts      
      
  (*********************************************************************)
  (*** registering global variables **)
  (*********************************************************************)      
  (** register all globals created so far with the file *)
  let registerGlobals unit : unit = 
    (* registers prototypes and definitions of reader/writer
       functions *)

    let createRW map : unit = 
    
      (* register prototypes near variables they mask. This ensures
	 that all appropriate types have already been declared *)
      let fn res g = 
	let out = g :: res in 
	match g with 
	  | GVar (v, _, _) when VM.mem v map -> 
	    let fd = VM.find v map in
	    GVarDecl(fd.svar,builtinLoc) ::  out
	  | _ -> out
      in
      Args.file.globals <- List.fold_left fn [] (List.rev Args.file.globals);
      
      VM.iter 
	(fun v fd -> Args.file.globals <- 
	  Args.file.globals @ [GFun(fd, builtinLoc)]) map ;
      
    in
    
    (* register readers *)
    createRW !rfnMap ;
    (* register writers *)
    createRW !wfnMap ;

    Svm.registerGlobals Args.file; 
    Sched.registerGlobals Args.file
      
  let inline unit : unit =
    let inlineWhat (v:varinfo) : fundec option =
      findFundec Args.file v.vname
    in	
    iterGlobals Args.file 
      (fun g -> 
	match g with 
	  | GFun (fd, _) when fd.svar.vname = mainFnName -> 
	    let anyInlining = ref true in 
	    while !anyInlining do
	      anyInlining := false;
	      Inliner.doFunction fd inlineWhat anyInlining
	    done
	  | _ -> ()
      )

  (** map from hidden shared to what they are hiding *)
  let hiddenShared : varinfo VM.t ref = ref VM.empty 

  (** Rename shared variables to ensure that they cannot be used directly*)
  let hideShared unit : unit = 
    Args.file.globals <- 
      List.map 
      (fun g ->
	  match g with 
	      GVar (v, init, loc) when Svm.mem v -> 
		let u = copyVarinfo v ("__startrek_hidden_" ^ v.vname) in
		  hiddenShared := VM.add u v !hiddenShared; 
		  GVar (u, init, loc)
	    | _ -> g)
      Args.file.globals
    
  (** create function that initializes round zero *)
  let createInitZero unit : unit = 
    let initType = TFun (voidType, Some [], false, []) in
    let fd = findOrCreateFundec Args.file initZeroFnName initType in
    fd.svar.vinline <- true;

    (* recursive function to initialize arrays *)
    let rec initRec (l1:lval) (l2:lval) (il:instr list) : instr list =
      match (typeOfLval l1) with
          TArray(_,eo,_) ->
            let ilref = ref il in
            for i = 0 to ((lenOfArray eo) - 1) do
              let os = Index(integer i,NoOffset) in
              ilref := initRec (addOffsetLval os l1) (addOffsetLval os l2) !ilref
            done;
            !ilref
        | _ -> 
          il @ [Set(l1,Lval(l2),builtinLoc)]
    in

    let body = VM.fold 
      (fun k v res -> 
	let sv = Svm.getSharedVar v in 
	let shared = Svm.getSharedLval sv 0  in
        initRec shared (var k) res
      )
      !hiddenShared 
      []
    in
    fd.sbody.bstmts <- [ mkStmt (Instr body) ]

  (** remove all globals that we don't like *)
  let cleanupGlobals unit : unit = 
    Args.file.globals <- 
      List.filter 
      (fun g -> 
	match g with 
	    (* remove all prototypes for functions used to mark shared
	       reads and writes*)
	    GVarDecl (v, _) -> 
	      (not (Util.hasPrefix sharedReadPrefix v.vname)) && 
		(not (Util.hasPrefix sharedWritePrefix v.vname))
	  | _ -> true
      ) 
      Args.file.globals
    
  (*********************************************************************)
  (*** Entry point **)
  (*********************************************************************)      

  (** run everything *)
  let run unit : unit = 
    (* create global variables *)
    createGlobalVars Args.file;
    if (!dopo) then createRWFlags ();
    (* schedule jobs *)
    createScheduleJobs ();
    (* create function that initializes rounds 1 to R-1 *)
    createInitNonZero Args.file;
    (* create context switch function *)
    createCS Args.file;
    (* create the tasks *)
    createTasks Args.file;
    (* create user initial function *)
    createUserInitial Args.file;
    (* create user final function *)
    createUserFinal Args.file;
    (* create hyper-period function *)
    createHyperPeriod Args.file; 
    createMain Args.file;
    cleanupGlobals ();
    (* create globals *)
    registerGlobals ();
    (* hide shared variables by renaming them *)
    hideShared ();
    (* create function that initializes round zero. must hapen after
       the shared variables are hidden *)
    createInitZero ();
    (* create a bunch of utility functions *)
    createUtilFuncs ();

    if (!inlineAll) then inline ();     
end;;

(*********************************************************************)
(*** the Cil feature, where everything begins *)
(*********************************************************************)

(* command line option for first pass *)
let feature : featureDescr = { 
  fd_name = "rekhpip";              
  fd_enabled = ref false;
  fd_description = "Harmonic REK for RTES with PIP locks";
  fd_extraopt = 
    [
      "--rekhPipRounds", Arg.Set_int numRounds, " number of sequentialization rounds";
      "--rekhPipPrintNumJobs", Arg.Set printNumJobs, " print number of jobs and exit";
      "--rekhPipCheckRounds", Arg.Set checkRounds, " check if there are enough rounds";
      "--rekhPipCheckSanity", Arg.Set checkSanity, " check sanity, i.e., if a path exists";
      "--rekhPipUnbounded", Arg.Set unboundedSeq, " do unbounded sequentialization";
    ];
  fd_doit = (
    function (f: file) -> 
      print_string ("Analyzing " ^ f.fileName ^ "\n");

      (* compute task specific info *)
      initTaskList f ;
      
      (** print number of jobs and exit if necessary *)
      let jobNum = List.fold_left (fun res t -> res + t.tjobs) 0 !tasks in
      print_string ("number of jobs = " ^ (string_of_int jobNum) ^ "\n");
      if (!printNumJobs) then exit 0;

      (** print number of rounds and check if they are enough *)
      print_string ("number of rounds = " ^ (string_of_int !numRounds) ^ "\n");
      if (!numRounds < jobNum) then (
        print_string ("ERROR: must have at least " ^ (string_of_int jobNum) ^ " rounds ...\n");
        exit 1
      );

      (* package arguments into a module *)
      let module A = struct 
	let file = f 
        let rounds = !numRounds (* List.fold_left (fun res t -> res + t.tjobs) 0 !tasks *)
	let hyperPeriodVar = findOrCreateGlobalVar f "__startrek_hyper_period" ucharType {init = None} builtinLoc
	let lockVar = findOrCreateGlobalVar f "__startrek_lock" boolType {init = Some (makeZeroInit boolType)} builtinLoc
	let jobEndVar = findOrCreateGlobalVar f "__startrek_job_end" ucharType {init = None} builtinLoc
	let jobVar = findOrCreateGlobalVar f "__startrek_job" ucharType {init = None} builtinLoc
	let taskVar = findOrCreateGlobalVar f "__startrek_task" ucharType {init = None} builtinLoc
	let roundVar = findOrCreateGlobalVar f "__startrek_round" ucharType {init = None} builtinLoc
	let eVar = findOrCreateGlobalVar f eVarName eVarType eVarInit builtinLoc
	let sVar = findOrCreateGlobalVar f sVarName sVarType sVarInit builtinLoc
	let rtVar = findOrCreateGlobalVar f rtVarName (rtVarType()) (rtVarInit()) builtinLoc
	let rlVar = findOrCreateGlobalVar f rlVarName rlVarType rlVarInit builtinLoc
	let cpVar = findOrCreateGlobalVar f cpVarName cpVarType {init=None} builtinLoc
      end in
      let module M = 
	    RekHPIP
	      (A) 
	      (ArraySharedVarManager (A)) 
	      (ArraySchedVarManager (A))
      in
      M.run ()
  );
  fd_post_check = false
} 
