open Cil
open Rtcbautil
open Rtcbainput
open Rekoptions

(* open Graph.Pack *)

(* string map *)
module SM = Map.Make(struct type t = string let compare = compare end)

(** Balanced Tree to keep track of jobs *)
module BT = BalancedTree

(** type for the tree of jobs *)
type jobtree  = BT.t
(** type of a job *)
type jobt = BT.n

(*********************************************************************)
(* Task Guessing code *)
(*********************************************************************)

(** number of hyper-periods to analyze *)
let hyperPeriods : int ref = ref 1

(** the tree of jobs *)
let jobTree : jobtree ref = ref (BT.make [])

let numOfJobs (unit) : int = BT.size !jobTree (BT.root !jobTree)
let jobLowRound (job:jobt) = (BT.id !jobTree job) - (BT.level_of job)
let jobHighRound (job:jobt) = (BT.max_id !jobTree job)

(** first job of a task *)
let firstJobOfTask (t:task) : jobt = 
  BT.firstInGeneration !jobTree t.tid
    
type taskInfo =
    {
      taskName : string;
      
      (** priority of the task. Priorities are unique. Priority of a
	  task is also its id *)
      mutable priority : int option;
      
      (** entry point *)
      mutable entryPt : fundec option;
      
      (** worst case execution time *)
      mutable wcet : int option;

      (** arival time *)
      mutable arrival : range;
      
      (** period *)
      mutable period : int option;
      
      mutable latency : int;
      
      mutable pbounds : int array
    }
      
(** array of taskInfo sorted by priorities *)
let taskArray : taskInfo array ref = ref [||]

let arr_leq x y =
  if x.rmax = -1 || y.rmin = -1 then false
  else x.rmax <= y.rmin

let arr_lt x y = 
  if x.rmax = -1 || y.rmin = -1 then false
  else x.rmax < y.rmin

let arr_add x y = 
  if x.rmax = -1 || x.rmin = -1 then {rmin = -1; rmax = -1;}
  else {rmin = x.rmin + y; rmax = x.rmax + y}

let jobPriority = BT.level_of 
let jobTask = BT.level_of

let jobArrival (job:jobt) =
  let tid = jobTask job in
  let arrival = !taskArray.(tid).arrival in
  let period = oget (!taskArray.(tid).period) in
  arr_add arrival ((BT.generationId !jobTree job) * period)

let jobRT (job:jobt) = 
  let tid = BT.level_of job in  !taskArray.(tid).latency

let jobPreceeds high low = 
  let jh = (jobArrival high) in
  let jl = (jobArrival low) in
  (*int_string ("ja_high " ^ (string_of_int jh.rmax) ^ "\n");
  print_string ("ja_low " ^ (string_of_int jl.rmin) ^ "\n"); *)
  let r = arr_leq  jh jl
  in 
  (*print_string 
    ((string_of_int (BT.id !jobTree high)) ^ 
	"<=" ^ 
	(string_of_int (BT.id !jobTree low)) ^ 
	" is " ^ 
	(string_of_bool r) ^ 
	"\n")
  ;*)
  r

let jobFollows high low = 
  let low_completion = arr_add (jobArrival low) (jobRT low) in
  arr_leq low_completion (jobArrival high)
    
let jobMayPreempt high low = 
  let x = not ( (jobPreceeds high low) || jobFollows high low)   
  in 
  (*let jid1 = BT.id !jobTree high in
  let jid2 = BT.id !jobTree low in
  print_string 
    ((string_of_int jid1) ^
	" may preempt " ^
	(string_of_int jid2) ^
	": " ^
	(string_of_bool x) ^ "\n");*)
  x

let jobMayNotPreceed high low = 
  arr_lt (jobArrival low) (jobArrival high)

let jobMayNotFollow high low =
  let low_completion = arr_add (jobArrival low) (jobRT low) in  
  arr_lt (jobArrival high) low_completion

(** compute a hyper-period for a task array*)
let hyperPeriod ta = 
  Array.fold_left (fun res ti -> lcm res (oget ti.period) ) 1 ta

(** convert task array to a list of tasks for rtcbautil *)
let taskArrayToList ta : task list = 
  (* number of jobs in previous iteration *)
  let res =  Array.mapi 
    (fun i a -> 
      { 
	name = a.taskName; 
	tid = i;
        tperiod = oget a.period;
	tslices = 1;
	tjobs = intceil !timeBound (oget a.period)
      }
    ) 
    ta 
  in Array.to_list res
	   
let guessTasks file : taskInfo array  = 
  print_string "Guessing tasks\n";

  (** map from task names to task information *)
  let taskMap : taskInfo SM.t ref = ref SM.empty in


  let taskName vn p =
    let vnLen = String.length vn and
	pLen = String.length p in
    String.sub vn pLen (vnLen - pLen) 
  in
  
  let findOrCreateTask n = 
    try 
      SM.find n !taskMap
    with 
      | Not_found ->
	let ti = {taskName = n; 
		  priority = None; 
		  entryPt = None;
		  wcet = None;
		  arrival = { rmin = -1; rmax = -1 };
		  period = None;
		  latency = 0;
		  pbounds = [||]
		 } in
	taskMap := SM.add n ti !taskMap; 
	ti
  in
  
  let intFromInit init = 
    match init.init with
	Some (SingleInit e) when isConstant e ->
	  (match isInteger e with
	      Some int64 -> i64_to_int int64
	    | None -> invalid_arg "expecting an int")
      | _ -> invalid_arg "missing init"
  in
  let 
      taskEntryPt = "__startrek_entry_pt_" and
      basePriority = "__startrek_base_priority_" and
      taskWcet = "__startrek_wcet_" and
      taskPeriod = "__startrek_period_" and
      sTimeBound = "__startrek_time_bound" and
      taskLatency = "__startrek_latency_" and
      taskArrivalMin = "__startrek_arrival_min_" and
      taskArrivalMax = "__startrek_arrival_max_"
  in
  iterGlobals 
    file
    (fun g -> 
      match g with
	| GFun (fd, loc) when 
	    Util.hasPrefix taskEntryPt fd.svar.vname  -> 
	  let name = taskName fd.svar.vname taskEntryPt in
	  (findOrCreateTask name).entryPt <- Some fd
	| GVar (v, init, loc) when 
	    Util.hasPrefix basePriority v.vname  -> 
	  let name = taskName v.vname basePriority and
	      priority = intFromInit init in
	  (findOrCreateTask name).priority <- Some (priority) 
	| GVar (v, init, _) when
	    Util.hasPrefix taskWcet v.vname ->
	  let name = taskName v.vname taskWcet and
	      wcet = intFromInit init in
	  (findOrCreateTask name).wcet <- Some (wcet)
	| GVar (v, init, _) when
	    Util.hasPrefix taskPeriod v.vname ->
	  let name = taskName v.vname taskPeriod and
	      period = intFromInit init in
	  (findOrCreateTask name).period <- Some (period)
	| GVar (v, init, _) when
	    Util.hasPrefix taskLatency v.vname ->
	  let name = taskName v.vname taskLatency and
	      latency = intFromInit init in
	  (findOrCreateTask name).latency <- latency
	| GVar (v, init, _) when v.vname = sTimeBound ->
	  setTimeBound (intFromInit init)
	| GVar (v, init, _) when
	    Util.hasPrefix taskArrivalMin v.vname ->
	  let name = taskName v.vname taskArrivalMin and
	      minArrival = intFromInit init in
	  let task = findOrCreateTask name in
	  task.arrival <- { rmin = minArrival; rmax = task.arrival.rmax}
	| GVar (v, init, _) when
	    Util.hasPrefix taskArrivalMax v.vname ->
	  let name = taskName v.vname taskArrivalMax and
	      maxArrival = intFromInit init in
	  let task = findOrCreateTask name in
	  task.arrival <- { rmin = task.arrival.rmin; rmax = maxArrival}
	    
	| _ -> () 
    ) ;
  let taskArray = 
    Array.of_list (SM.fold (fun k d r -> d :: r ) !taskMap [])  in
  Array.sort (fun t1 t2 -> compare t1.priority t2.priority) taskArray; 
  taskArray
	  
let sanityCheck ta : unit = 
  for i = 0 to pred (Array.length ta) do
    begin
      (* All priorities are specified *)
      (match ta.(i).priority with 
	  None -> invalid_arg "No priority!"
	| _ -> ()
      );
      
      (* Rate Monotonic Priority Assignment *)
      if i < pred (Array.length ta) then
	begin
	  let x = oget ta.(i).period and
	      y = oget ta.(succ i).period in
	  if x < y then invalid_arg "Not RMS!" ;
	  if x mod y <> 0 then invalid_arg "Not harmonic"
	end
    end
  done


(*********************************************************************)
(* Globals *)
(*********************************************************************)
let rangesInArrival : bool ref = ref false

(*********************************************************************)
(*** Guessing task arguments **)
(*********************************************************************)      
let initTaskList file : unit =

  let checkArrivals ta =
    let x = Array.fold_left  
      (fun r t -> 
	(* compute arrival range if not set *)
	if (t.arrival.rmax == -1 || t.arrival.rmin = -1) then
	  begin
	    let m = (oget t.period) - t.latency in
	    let rng = { rmin = max 0 t.arrival.rmin; 
			rmax = if t.arrival.rmax == -1 then m 
			  else t.arrival.rmax; }
	    in t.arrival <- rng
	  end;
	if (t.arrival.rmax != t.arrival.rmin) then true
	else r
      )
      false
      ta
    in
    if x then
      begin
	print_string "RANGES found\n";
	rangesInArrival := true
      end
    else
      rangesInArrival := false
  in
  let computeJobTree ta = 
    let r = ref [] in
    print_string "JOB_TREE: start\n" ;
    for i = (Array.length ta) - 2 downto 0 do
      let d = (oget ta.(i).period) / (oget ta.(succ i).period) in
      r :=  d :: !r  ;
      print_string ("\tLevel " ^ (string_of_int i) ^ " has degree " ^ 
		       (string_of_int d) ^ "\n")
    done ; 
    print_string "JOB_TREE: done\n";
    BT.make !r 
  in

  taskArray := guessTasks file;
  sanityCheck !taskArray;
  
  (* normalize timebound to the closest hyper period *)
  let hp = hyperPeriod !taskArray in
  hyperPeriods := intceil !timeBound hp;
  setTimeBound (hp);
  print_string ("Using time bound: " ^ (string_of_int !timeBound) ^ "\n");
  print_string ("      with hyper-period: " ^ (string_of_int hp) ^ "\n");  

  begin 
    try 
      (* compute PB *)
      
      (* compute the rtInput from the arguments *)
      let rti = new rtInput in
      let ti2rt (ti:taskInfo) : unit = 
	let rt : rt_task = {
	rtname = "";
	rtwcet = oget ti.wcet;
	rtperiod = oget ti.period;
      } in rti#addTask (rt)
      in
      Array.iter ti2rt !taskArray;

      (* compute pbInput from rtInput *)
      let pbi = new pbInput in
      pbi#populate(rti);
      (* set latency and pbounds *)
      for i = 0 to pred (Array.length !taskArray) do
	let pbt = List.nth (pbi#getTasks) i in
	!taskArray.(i).pbounds <- Array.of_list pbt.pbpbounds;
	!taskArray.(i).latency <- pbt.pblatency;
      done;
    with invalid_arg ->nopbounds := true
  end
  ;
  tasks := taskArrayToList !taskArray ; 
  jobTree := computeJobTree !taskArray;
  checkArrivals !taskArray

(*********************************************************************)
(** Dummy Feature description for CIL compilation *)
(*********************************************************************)
let feature : featureDescr = 
{
  fd_name = "rekhtasks";
  fd_enabled = ref false;
  fd_description = "rekhtasks";
  fd_extraopt = [ ];
  fd_doit = (fun f -> ());
  fd_post_check = false;
}
