open Cil
open Rtcbautil
open Rtcbainput
open Rekoptions
open Rekargs
open Rekhtasks

(** an exception *)
exception RekSCError of string

(** number of rounds to sequentialize over *)
let numRounds = ref (-1)

(** whether to print number of jobs and exit *)
let printNumJobs = ref false

(** whether to do unbounded sequentialization -- default is false *)
let unboundedSeq = ref false

(** shorthand name *)
let soi = string_of_int

(*********************************************************************)
(** ordering between strings *)
(*********************************************************************)
module StrOrder = struct
  type t = string
  let compare = Pervasives.compare
end

module StrSet = Set.Make(StrOrder)

(* set with varinfo key *)
module ViSet = Set.Make(ViOrder)

(*********************************************************************)
(* Signatures *)
(*********************************************************************)

module type SCHED_VAR_MANAGER = sig
  type schedvars
  val jobStartLval :  int -> lval
  val jobEndLval :  int -> lval
  val jobMaxLval :  int -> lval
  val jobMinLval :  int -> lval
  val registerGlobals : file -> unit
end

(* extend ARGS To SC args *)
module type SCARGS = sig
  include ARGS
  val jobNum : int
  val clockVar : varinfo
end

(*********************************************************************)
(* Managing job scheduling variables  *)
(*********************************************************************)
module ArraySchedVarManager (Args : ARGS) =
struct
  (* sv - start, ev - end, maxv - max, minv - min *) 
  type schedvars = { sv: varinfo; ev: varinfo; maxv: varinfo; minv: varinfo }

  let jvtype = Args.roundVar.vtype

  let jobVarName (vname: string)  : string = 
    "__startrek_" ^ vname  

  let jobStartVarName = jobVarName "start" 
  let jobEndVarName = jobVarName "end" 
  let jobMaxVarName  = jobVarName "max" 
  let jobMinVarName = jobVarName "min" 
    
  let schedVars = 
    let atype = Formatcil.cType "%t:T[%e:j]"
      [
	("T", Ft jvtype);
	("j", Fe (integer (numOfJobs ()) ));
      ] in
    {
      sv =  makeGlobalVar jobStartVarName atype;
      ev = makeGlobalVar jobEndVarName atype;
      maxv = makeGlobalVar jobMaxVarName atype;
      minv = makeGlobalVar jobMinVarName atype;
    }

  let jobStartLval (job:int) : lval = 
    Formatcil.cLval "sv[%e:job]" 
      [("sv", Fv schedVars.sv);
       ("job", Fe (integer job))
      ]

  let jobEndLval (job:int) : lval = 
    Formatcil.cLval "ev[%e:job]" 
      [("ev", Fv schedVars.ev);
       ("job", Fe (integer job))
      ]

  let jobMaxLval (job:int) : lval = 
    Formatcil.cLval "maxv[%e:job]" 
      [("maxv", Fv schedVars.maxv);
       ("job", Fe (integer job))
      ]

  let jobMinLval (job:int) : lval = 
    Formatcil.cLval "minv[%e:job]" 
      [("minv", Fv schedVars.minv);
       ("job", Fe (integer job))
      ]
    
  let registerGlobals f : unit = 
    f.globals <- 
      GVar (schedVars.sv, {init = None}, builtinLoc) ::
      GVar (schedVars.ev, {init = None}, builtinLoc) ::
      GVar (schedVars.maxv, {init = None}, builtinLoc) ::
      GVar (schedVars.minv, {init = None}, builtinLoc) ::
      f.globals
end

(** Time Bounded Analysis for Harmonic Tasks with PIP locks *)
module RekSC
  (Args : SCARGS) 
  (Sched : SCHED_VAR_MANAGER)

  = 

struct
  (** the total number of jobs *)
  let jobNum = Args.jobNum

  (** set of names of useful functions *)
  let usefulFuncs = ref StrSet.empty

  (** add a useful function *)
  let addUsefulFunc (s:string) : unit =
    usefulFuncs := StrSet.add s !usefulFuncs

  (** return true if the argument is a useful function *)
  let isUsefulFunc (s:string) : bool =
    StrSet.mem s !usefulFuncs

  (** the list of global variables created during sequentialization *)
  let createdVars : varinfo list ref = ref []

  (** the array of job functions *)
  let jobFuncs : fundec array = Array.make Args.jobNum dummyFunDec

  (** the array of job assume functions *)
  let jobAssumeFuncs : fundec array = Array.make Args.jobNum dummyFunDec

  (** the next symbolic constant id *)
  let scid = ref 0

  (** return the next symbolic constant, and add it to the file as
      a global variable *)
  let nextSC (t:typ) : varinfo =
    let e = !scid in
    scid := !scid + 1;
    let r = makeGlobalVar ("_sc_" ^ (soi e)) t in
    createdVars := !createdVars@[r]; r

  (** array such that i-th element maps shared variable x to symbolic
      constants that correspond to reading x in job i *)
  let readSC : varinfo list ViMap.t array = Array.make Args.jobNum ViMap.empty

  (** array such that i-th element maps shared variable x to symbolic
      constants that correspond to writing x in job i *)
  let writeSC : varinfo list ViMap.t array = Array.make Args.jobNum ViMap.empty

  (** the set of all shared variables *)
  let sharedVars  = ref ViSet.empty

  (** add a read symbolic constant *)
  let addReadSC (jid:int) (sv:varinfo) (sc:varinfo) : unit =
    sharedVars := ViSet.add sv !sharedVars;
    let m = readSC.(jid) in
    try 
      readSC.(jid) <- ViMap.add sv (sc::(ViMap.find sv m)) m
    with Not_found ->
      readSC.(jid) <- ViMap.add sv [sc] m

  (** add a write symbolic constant *)
  let addWriteSC (jid:int) (sv:varinfo) (sc:varinfo) : unit =
    sharedVars := ViSet.add sv !sharedVars;
    let m = writeSC.(jid) in
    try 
      writeSC.(jid) <- ViMap.add sv (sc::(ViMap.find sv m)) m
    with Not_found ->
      writeSC.(jid) <- ViMap.add sv [sc] m

  (** map from symbolic constants to lamport clock variables *)
  let scLCMap : varinfo ViMap.t ref = ref ViMap.empty

  (** return the lamport clock variable for a symbolic constant --
      create one if necessary *)
  let scLCVar (vi:varinfo) : varinfo =
    try ViMap.find vi !scLCMap
    with Not_found ->
      let x = makeGlobalVar ("_lc" ^ vi.vname) uintType in
      scLCMap := ViMap.add vi x !scLCMap;
      createdVars := !createdVars@[x];
      x

  (** the array of assert variables *)
  let jobAsserts = Array.make Args.jobNum dummyFunDec.svar

  (** return the assert variable for job jid -- create one if necessary *)
  let jobAssert (jid:int) : varinfo =
    if (jobAsserts.(jid).vid = dummyFunDec.svar.vid) then (
      jobAsserts.(jid) <- makeGlobalVar ("_job_asrt_" ^ (soi jid)) boolType;
      createdVars := !createdVars@[jobAsserts.(jid)]
    );
    jobAsserts.(jid)

  (*********************************************************************)
  (* module-global constants *)
  (*********************************************************************)

  (*********************************************************************)
  (* Arguments *)
  (*********************************************************************)
  let rounds = Args.rounds
  let taskVar = Args.taskVar
  let jobVar = Args.jobVar
  let jobEndVar = Args.jobEndVar
  let roundVar = Args.roundVar
  let clockVar = Args.clockVar
  let lockVar = Args.lockVar
  let hyperPeriodVar = Args.hyperPeriodVar

  (*********************************************************************)
  (* Names  *)
  (*********************************************************************)
  let assumeFnName = "__CPROVER_assume"
  let doCSFnName = "__startrek_do_cs"
  let scheduleJobsFnName = "__startrek_schedule_jobs"
  let userFinalFnName = "__startrek_user_final"
  let userInitialFnName = "__startrek_user_init"
  let hyperFnName = "__startrek_hyperperiod"
  let mainFnName = "main"

  (*********************************************************************)
  (* Common functions and types  *)
  (*********************************************************************)
  let csType = TFun (boolType,Some [],false,[])
  let doCSFnType = csType
  let assumeFn = findOrCreateFunc Args.file assumeFnName assumeType
  let assertFn = findOrCreateFunc Args.file "assert" assertType 

  let sharedReadPrefix = "__startrek_shared_read_"
  let sharedWritePrefix = "__startrek_shared_write_"
  let sharedVarPrefix = "__startrek_shared_var_"

  (** returns a statement that assumes x < y ===> y > z **)
  let notBetweenStmt (x:lval) (y:lval) (z:lval) : stmt =
    (* x < y ==> y > z *)
    Formatcil.cStmt 
      "if (%l:x < %l:y)  assume (%l:y > %l:z); "
      (fun n t -> assert false)
      builtinLoc
      [
	("x", Fl x);
	("y", Fl y);
	("z", Fl z);
	("assume", Fv assumeFn);
      ]
      
  (** returns a statement that assumes x < y ===> y > z **)
      
  let notBetweenStmt2 (x:lval) (y:lval) (z:lval) : stmt =
    (* x < y ==> y > z *)
    (* __CPROVER_assume (x >= y || (z < y)); *)

    let xGey = BinOp (Ge, Lval x, Lval y, boolType) and
	zLty = BinOp (Lt, Lval z, Lval y, boolType) in
    let disj = BinOp (LOr, xGey, zLty, boolType) in 
    mkStmtOneInstr (Call (None, Lval (var assumeFn), [disj], builtinLoc))

  (*********************************************************************)
  (* Non-deterministic values functions  *)
  (*********************************************************************)

  let ndbFn = findOrCreateFunc Args.file "nondet_bool" ndbType

  (** Returns a varinfo of a function that returns a non-deterministic value 
      of a given type *)
  let nondetFn (typ:typ) : varinfo =
    let tyname = 
      match typ with
	  TInt (IChar, _) -> "char"
	| TInt (ISChar, _) -> "schar"
	| TInt (IUChar, _) -> "uchar"
	| TInt (IBool, _) -> "_Bool"
	| TInt (IInt, _) -> "int"
	| TInt (IUInt, _) -> "uint"
	| TInt (IShort, _) -> "short"
	| TInt (IUShort, _) -> "ushort"
	| TInt (ILong, _) -> "long"
	| TInt (IULong, _) -> "ulong"
	| TInt (ILongLong, _) -> "long_long"
	| TInt (IULongLong, _) -> "ulong_long"
	| TFloat (FFloat, _) -> "float"
	| TFloat (FDouble, _) -> "double"
	| TFloat (FLongDouble, _) -> "long_double"
	| TNamed (tyinfo, _) -> tyinfo.tname
	| _ -> invalid_arg "Unsupported type"
    in
    findOrCreateFunc 
      Args.file 
      ("nondet_" ^ tyname) 
      (TFun (typ, Some [], false, []))

  (** a handy function to create statements -- has a set of things
      hardwired *)
  let createStmtsLoc (s:string) (loc:location) fmt : stmt list = 
    Formatcil.cStmts s (fun n t -> assert false) loc
      ([
        ("do_cs", Fv(findOrCreateFunc Args.file doCSFnName doCSFnType));
	("assume", Fv assumeFn);
	("assert", Fv assertFn);
      ] @ fmt)

  let createStmts (s:string) fmt : stmt list = 
    createStmtsLoc s builtinLoc fmt

  (*********************************************************************)
  (** create user initial function *)
  (*********************************************************************)

  (** replaces accesses to shared variables with direct
      assignments. this visitor is applied to the body of the user
      initial and final functions where such assignment is legal. *)
  class initialFinalReplacer = 
  object(self) 
    inherit nopCilVisitor as super
      
    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
	  (* read a global variable -- must be of the form x = R(&v) *)
	  Call(Some(Var x, NoOffset),
	       Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset)], loc) 
	    when Util.hasPrefix sharedReadPrefix fn -> 
	      ChangeTo [Set (var x, Lval(var v), loc) ]
		
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var {vname = fn}, NoOffset),
	       [AddrOf(Var v, NoOffset); e],loc) 
	    when Util.hasPrefix sharedWritePrefix fn ->
	  ChangeTo [Set (var v, e, loc)]
	    
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end

  let createUserInitialOrFinal (fname:string) (f:file) =
    let ftype = TFun (voidType, Some [], false,[]) in
    addUsefulFunc fname;
    let fd = findOrCreateFundec f fname ftype in
    (** helper function needed for inlining *)
    let inlineWhat (v:varinfo) : fundec option =
      findFundec f v.vname
    in
    (** inline calls made by user init copy *)
    let anyInlining = ref true in 
    while !anyInlining do
      anyInlining := false;
      Inliner.doFunction fd inlineWhat anyInlining
    done;
    fd.svar.vinline <- true;   
    fd.sbody <-  visitCilBlock (new initialFinalReplacer) fd.sbody

  let createUserInitial (f:file) = createUserInitialOrFinal userInitialFnName f
  let createUserFinal (f:file) = createUserInitialOrFinal userFinalFnName f    

  (*********************************************************************)
  (* Context Switch Functions  *)
  (*********************************************************************)
      
  let csFnName (t:task) = "__startrek_cs_t" ^ (soi t.tid)

  let getCsFn (t:task) : varinfo = 
    findOrCreateFunc Args.file (csFnName t) csType

  (*********************************************************************)
  (*** creating  context switch functions -- one per task **)
  (*********************************************************************)
  let createCS (f:file) =

    (*create CS for a task *)
    let createCSTask (t:task) : unit =
      addUsefulFunc (csFnName t);
      let fd = findOrCreateFundec f (csFnName t) csType in
	
      (* the body of context switch *)
      let body = 
	if isTopTask t then 
	  [mkStmt (Return (Some zero, builtinLoc))]
	else
	  begin
	    (* switch over all jobs of higher-priority tasks. Note
	       that t is not the highest priority, so it has children
	       in the job tree *)
	    let fstJob = firstJobOfTask t in
	    let swBody = BT.fold_left_generation 
	      (fun b tree node -> 
		let cbody = BT.fold_dfs_siblings 
		  (fun r tree node ->
		    let rnd = var roundVar in
		    let id = BT.id tree node in
		    let slv = Sched.jobStartLval id in
		    let elv = Sched.jobEndLval id in
		    (notBetweenStmt slv rnd elv) :: r
		  ) 
		  [] tree (BT.first_child tree node) 
		in
		let hd = List.hd cbody in
		hd.labels <- [Case (integer (BT.id tree node), rekloc)];
		cbody @ ((mkStmt (Break builtinLoc)) :: b)
	      )
	      [] !jobTree fstJob 
	    in
	    let stmts = if !nopriorcs then [] else
		[mkStmt (Switch (Lval (var jobVar), 
				 mkBlock swBody, swBody, rekloc))] 
	    in
	    Formatcil.cStmts 
              "
                if (lock) return 0;
                %t:Bool c = ndb (); if (c) return 0;
                
                %t:rty o = round; round = nd();
                assume (round > o); assume (round <= ej); 

                if (round != ej) { %S:w } return 1;
              "
	      (fun n t -> makeTempVar fd ~name:n t)
	      rekloc
	      [
		("rty", Ft roundVar.vtype);
		("round", Fv roundVar);
		("nd", Fv (nondetFn roundVar.vtype));
		("Bool", Ft boolType);
		("ndb", Fv ndbFn);
		("assume", Fv assumeFn);
		("lock", Fv lockVar);
		("ej", Fv jobEndVar);
		("w", FS stmts);
	      ]
	  end
      in
      fd.sbody.bstmts <- body @ fd.sbody.bstmts ; 
      fd.svar.vinline <- true;
      fd.svar.vstorage <- Static;
      Args.file.globals <- 
	GVarDecl (fd.svar, builtinLoc) :: Args.file.globals 
    in
    (* create context-switch for each task *)
    List.iter createCSTask !tasks

  (*********************************************************************)
  (* Creation of job ordering and may observe relations *)
  (*********************************************************************)

  (** list of jobs by VMCAI'13 ordering *)
  let orderedJobs : jobt list ref = ref []
      
  (** maps a job j to a list of jobs j' such that j is the maximal job
      that can be observed by j'. this means that after executing j,
      we can check assumptions of j'. *)
  let maxObs : int list array = Array.make Args.jobNum []

  (** maps each job j to job j' such that assumption of j can be
      checked after j' has been executed. this is derived from maxObs,
      and constructed along with it. *)
  let assumeAfter : jobt array = Array.make Args.jobNum (BT.root !jobTree)

  (** maps each job id j to list of job ids j' such that assertion for
      j' can be checked after the assumption for j has been checked *)
  let assertAfter : int list array = Array.make Args.jobNum []

  (** job ordering according to VMCAI'13*)
  let jobOrder = (fun a b -> 
    if a = b then 0 
    else if jobPriority a >= jobPriority b then
      if jobPreceeds a b then -1 else 1
    else (if jobPreceeds b a then 1 else -1)
  ) 

  (** given job j and job list jl, return list of jobs in jl that may
      be observed by j. the returned list is in the same order as in
      jl. *)
  let mayObserve (job:jobt) (jl:jobt list) : jobt list =
    List.fold_left
      (* return true if job can observe job2 *)
      (fun (old:jobt list) (job2:jobt) ->
        if (jobPriority job = jobPriority job2) then (
          if (jobPreceeds job2 job) then old@[job2] else old
        ) else if (jobPriority job < jobPriority job2) then (
          if (jobPreceeds job2 job || jobMayPreempt job2 job) then old@[job2] else old
        ) else (
          if (jobFollows job job2 || jobMayPreempt job job2) then (
            (* print_string ((soi jid2) ^ " ==> " ^ (soi (BT.id !jobTree job)) ^ "\n"); *)
            old@[job2] 
          ) else old
        )
      ) [] jl

  (** create data structures related to job ordering and may
      observe dependencies between jobs *)
  let createJobOrdering (unit) : unit =
    let allJobs = (BT.fold_dfs (fun res t n -> n ::res) [] !jobTree (BT.root !jobTree)) in
    (** create list of jobs ordered by VMCAI'13*)
    orderedJobs := List.sort jobOrder allJobs;
    (** create maximal may observe dependencies *)
    List.iter (
      fun (job:jobt) ->
        let jid = BT.id !jobTree job in
        let moj = mayObserve job !orderedJobs in
        print_string ((soi jid) ^ " may observe : ");
        List.iter (fun j -> print_string ((soi (BT.id !jobTree j)) ^ " ")) moj;
        print_string "\n";
        let maxjob = List.hd (List.rev moj) in
        let maxjid = BT.id !jobTree maxjob in
        maxObs.(maxjid) <- maxObs.(maxjid)@[jid];
        print_string ("MAXOBS " ^ (soi jid) ^ " = " ^ (soi maxjid) ^ "\n");
        assumeAfter.(jid) <- maxjob;
        print_string ("ASSUME after " ^ (soi jid) ^ " ==> " ^ (soi maxjid) ^ "\n")
    ) allJobs;
    (** create the assertAfter array *)
    List.iter (
      fun (job:jobt) ->
        (** find the largest job max such that after max is executed,
            and all assumptions that could be checked after max have been
            checked, then the assertion of job can be checked **)
        let moj = mayObserve job !orderedJobs in
        let assumeAfters = List.fold_left
          (fun jl2 job2 -> jl2@[assumeAfter.(BT.id !jobTree job2)])
          [] moj
        in
        let sortedAA = List.sort jobOrder assumeAfters in
        let maxjob = List.hd (List.rev sortedAA) in
        let maxjid = BT.id !jobTree maxjob in
        assertAfter.(maxjid) <- assertAfter.(maxjid)@[BT.id !jobTree job];
        print_string ("ASSERT after " ^ (soi maxjid) ^ " ==> " ^ (soi (BT.id !jobTree job)) ^ "\n")
    ) allJobs

  (*********************************************************************)
  (* Creation of job assumption functions *)
  (*********************************************************************)

  (** given a symbolic constant, return its Lamport clock *)
  let lamportClock (sc:varinfo) : exp = Lval(var (scLCVar sc))

  (** given a symbolic constant, return the expression which
      represents that it has occurred. essentially this means that its
      lamport clock must not be 0xffffffff. *)
  let hasOcc (sc:varinfo) : exp = 
    Formatcil.cExp "~lc" [ ("lc", Fv (scLCVar sc)) ]

  (* given two lamport clocks c1 and c2, return true if c1 == c2 *)
  let cleq = fun clk1 clk2 ->
    Formatcil.cExp "(%e:clk1 == %e:clk2)"
      [
        ("clk1", Fe clk1);
        ("clk2", Fe clk2);
      ]

  (* given two lamport clocks c1 and c2, return true if c1 < c2 *)
  let cllt = fun clk1 clk2 ->
    Formatcil.cExp "(%e:clk1 < %e:clk2)"
      [
        ("clk1", Fe clk1);
        ("clk2", Fe clk2);
      ]
  
  (* given two lamport clocks c1 and c2, return true if c1 <= c2 *)
  let clle = fun clk1 clk2 ->
    Formatcil.cExp "(%e:clk1 <= %e:clk2)"
      [
        ("clk1", Fe clk1);
        ("clk2", Fe clk2);
      ]

  let createJobAssumps (unit) : unit =
    (** iterate over all jobs -- check assumption functions for
        each *)
    List.iter
      (fun (job:jobt) ->
        let moj = mayObserve job !orderedJobs in
        let jid = BT.id !jobTree job in
        let fname = "_job_assume_" ^ (soi jid) in
        let ftype = TFun (voidType, Some [], false,[]) in
        addUsefulFunc fname;
        let fd = findOrCreateFundec Args.file fname ftype in
        let maxLC = makeLocalVar fd "_max_lc" uintType in
        fd.sbody.bstmts <- [];
        (** for every variable read by jid *)
        let forEachRead (rv:varinfo) (rscs:varinfo list) : unit =
          (** collect the symbolic constants that can be observed *)
          let mosc = ref [] in
          List.iter (
            fun (job2:jobt) ->
              let jid2 = BT.id !jobTree job2 in
              let forEachWrite (wv:varinfo) (wscs:varinfo list) : unit =
                if(rv.vid = wv.vid) then mosc := !mosc@wscs
              in
              ViMap.iter forEachWrite writeSC.(jid2)
          ) moj;
          print_string ("read variable = " ^ rv.vname ^ " #rsc = " ^ (soi (List.length rscs)) ^ 
                           " #wsc = " ^ (soi (List.length !mosc)) ^ "\n");
          (** for each read symbolic constant *)
          let forEachReadSC (rsc:varinfo) : unit =
            (** assign max clock non-deterministically *)
            fd.sbody.bstmts <- fd.sbody.bstmts@
              Formatcil.cStmts "lc = ndui();"
              (fun n t -> assert false) rekloc
              [
                ("lc", Fv maxLC);
                ("ndui", Fv (nondetFn uintType));
              ];
            (** start building the upper and lower bound constraints *)
            let ub = ref (integer 0) in
            let lb = ref (integer 1) in
            let ubneg = ref (Formatcil.cExp "rsc == rv" [("rsc", Fv rsc); ("rv", Fv rv)]) in
            (** generate the lower bound constraints *)
            (** for each symbolic constant that jid can observe *)
            List.iter
              (fun (wsc:varinfo) -> 
                let fmt = 
                  [
                    ("occw", Fe (hasOcc wsc));
                    ("or", Fb LOr);
                    ("and", Fb LAnd);
	            ("assume", Fv assumeFn);
                    ("cleq", Fe (cleq (lamportClock wsc) (Lval (var maxLC))));
                    ("clle", Fe (clle (lamportClock wsc) (Lval (var maxLC))));
                    ("cllt", Fe (cllt (lamportClock wsc) (lamportClock rsc)));
                    ("rsc", Fv rsc);
                    ("wsc", Fv wsc);
                  ]
                in
                (** update lower bound constraint *)
                lb := Formatcil.cExp "%e:lb %b:and (!(%e:occw %b:and %e:cllt) %b:or %e:clle)" (("lb", Fe !lb)::fmt);
                (** update the upper bound constraint *)
                ub := Formatcil.cExp "%e:ub %b:or (%e:occw %b:and %e:cllt %b:and %e:cleq %b:and (rsc == wsc))" (("ub", Fe !ub)::fmt);
                ubneg := Formatcil.cExp "%e:ubneg %b:and (!(%e:occw %b:and %e:cllt))" (("ubneg", Fe !ubneg)::fmt)
              )
              !mosc;
            ub := Formatcil.cExp "%e:ubneg %b:or %e:ub" [("ubneg", Fe !ubneg); ("ub", Fe !ub); ("or", Fb LOr)];
            fd.sbody.bstmts <- fd.sbody.bstmts@
              Formatcil.cStmts "assume ((!(%e:occr)) %b:or (%e:lb %b:and %e:ub));"
              (fun n t -> assert false) rekloc
              [
	        ("assume", Fv assumeFn);
                ("occr", Fe (hasOcc rsc));
                ("lb", Fe !lb);
                ("ub", Fe !ub);
                ("and", Fb LAnd);
                ("or", Fb LOr);
              ];
          in
          List.iter forEachReadSC rscs
        in
        ViMap.iter forEachRead readSC.(jid);
        jobAssumeFuncs.(jid) <- fd
      )
      !orderedJobs

  (*********************************************************************)
  (* Creation of job functions *)
  (*********************************************************************)

  (** a useful function to create an instruction, given a symbolic
      constant *)
  let makeInstr ?fmt:(f=[]) (s:string) (l:location) (sc:varinfo) : instr =
    Formatcil.cInstr s l
      ([
        ("sc", Fv sc);
        ("scLC", Fv (scLCVar sc));
        ("round", Fv roundVar);
        ("clock", Fv clockVar);
      ]@f)
      
  (** a visitor that assigns unique ids to access of shared
      variables. also constructs a map from global variables to
      expressions they are assigned to. also replaces cpu_lock() and
      cpu_unlock() with assignments to the global lock flag. *)
  class sharedAccessNumberer (jobInit:jobt) = 
  object(self) 
    inherit nopCilVisitor as super

    (** the job *)
    val job:jobt = jobInit

    (* the job and task id, and the task *)
    val jid = BT.id !jobTree jobInit
    val task = List.nth !tasks (jobTask jobInit)

    (* visit an instruction -- we are only interested in calls to global
       reads and writes *)
    method vinst (ins:instr) = begin
      match ins with
	  (* read a global variable -- must be of the form x = R(&v) *)
	  Call(Some(Var x, NoOffset),Lval(Var vi, NoOffset),[AddrOf(Var rv, NoOffset)],loc) 
	    when Util.hasPrefix sharedReadPrefix vi.vname -> 
              let nsc = nextSC (rv.vtype) in 
              addReadSC jid rv nsc;
              ChangeTo[(*Call(Some(Var x, NoOffset),Lval(var vi),Lval(var nsc)::[AddrOf(Var rv, NoOffset)], loc);*)
                       makeInstr ~fmt:[("cs", Fv(getCsFn task))] "cs();" loc nsc;
                       makeInstr ~fmt:[("nd", Fv(nondetFn nsc.vtype))] "sc = nd();" loc nsc;
                       makeInstr ~fmt:[("x", Fv x)] "x = sc;" loc nsc;
                       makeInstr ("scLC = ((round & 0x3ff) << 22) | ((" ^ (soi (jobPriority job)) ^ " & 0xf) << 18) | (clock & 0x3ffff);") loc nsc;
                       makeInstr "clock = clock + 1;" loc nsc;]
                
	(* write a global variable -- must be of the form W(&v,e) *)
	| Call(None,Lval(Var vi, NoOffset),[AddrOf(Var wv, NoOffset);e],loc) 
	    when Util.hasPrefix sharedWritePrefix vi.vname ->
          let nsc = nextSC (wv.vtype) in 
          addWriteSC jid wv nsc;
          ChangeTo[(*Call(None,Lval(var vi),[Lval(var nsc);AddrOf(Var wv, NoOffset); e], loc);*)
                   makeInstr ~fmt:[("cs", Fv(getCsFn task))] "cs();" loc nsc;
                   makeInstr ~fmt:[("e", Fe e)] "sc = %e:e;" loc nsc;
                   makeInstr ("scLC = ((round & 0x3ff) << 22) | ((" ^ (soi (jobPriority job)) ^ " & 0xf) << 18) | (clock & 0x3ffff);") loc nsc;
                   makeInstr "clock = clock + 1;" loc nsc;]

        (** call cpu lock function *)
	| Call (None, Lval (Var {vname = fn}, NoOffset), [], loc) 
	    when fn = cpuLockFnName ->
	  ChangeTo ([Set (var lockVar, one, builtinLoc)])
        (** call cpu unlock function *)
	| Call (None, Lval (Var {vname = fn}, NoOffset), [], loc) 
	    when fn = cpuUnlockFnName ->
	  ChangeTo ([Set (var lockVar, zero, builtinLoc)])

          (* call to assert *)
        | Call(None,Lval(Var {vname = fn}, NoOffset), [e], loc)  
	    when fn = "assert" ->
	  ChangeTo ([Formatcil.cInstr "asrt = asrt %b:and %e:e;" loc 
                        [
                          ("asrt", Fv (jobAssert jid));
                          ("e", Fe e);
                          ("and", Fb LAnd);
                        ]])
	(* otherwise, do nothing *)
	| _ -> SkipChildren
    end
  end

  (** create job functions *)
  let createJobFuncs (f:file) : unit =
    (** helper function needed for inlining *)
    let inlineWhat (v:varinfo) : fundec option =
      findFundec f v.vname
    in
    (** iterate over all jobs -- creating top-level job functions for
        each *)
    List.iter
      (fun (job:jobt) ->
        let jid = BT.id !jobTree job in
        (** create the assert variable for the job for later use *)
        ignore(jobAssert jid);
        let tid = jobTask job in
        let t = !taskArray.(tid) in
        let jobFuncName = t.taskName ^ "_job_" ^ (soi jid) in
        match t.entryPt with
            Some(taskfd) -> (
              addUsefulFunc jobFuncName;
              let jobfd = copyFunction taskfd jobFuncName in
              (** inline calls made by job copy *)
	      let anyInlining = ref true in 
	      while !anyInlining do
	        anyInlining := false;
	        Inliner.doFunction jobfd inlineWhat anyInlining
	      done;
              (** number shared accesses, and collect the rhs map for
                  the job and store it in the array. *)
              let san = new sharedAccessNumberer job in
              jobfd.sbody <- visitCilBlock san jobfd.sbody;
              (** add job copy to array of job fundecs *)
              jobFuncs.(jid) <- jobfd
            )
          |_ -> raise (RekSCError ("could not find body of " ^ t.taskName))
      )
      !orderedJobs;
    (** add created variables to file *)
    List.iter
      (fun (v:varinfo) -> f.globals <- f.globals@[GVarDecl(v,builtinLoc)])
      !createdVars;
    (** add job functions to file *)
    Array.iter
      (fun (fd:fundec) -> f.globals <- f.globals@[GFun(fd,builtinLoc)])
      jobFuncs
      
  (*********************************************************************)
  (*** creating job scheduling function **)
  (*********************************************************************)

  (** create schedule_jobs  function **)
  let createScheduleJobs unit : unit =
    let ftype =  TFun (voidType,Some [],false,[]) in
    addUsefulFunc scheduleJobsFnName;
    let fd = findOrCreateFundec Args.file scheduleJobsFnName ftype in
    fd.svar.vinline <- true;

    (** Returns nesting constraints for job j2 of a low priority task
	t2 and job j1 of a higher priority task t1 **)
    let nestingConstraints j1 j2 =
      let 
	  jid1 = BT.id !jobTree j1 and
	  jid2 = BT.id !jobTree j2 and
	  tid1 = BT.level_of j1 and
	  tid2 = BT.level_of j2 
      in
      let
    	  s1 = Sched.jobStartLval jid1 and
    	  s2 = Sched.jobStartLval jid2 and
    	  e1 = Sched.jobEndLval jid1 and
    	  e2 = Sched.jobEndLval jid2 in
      
      let pb = if !nopbounds then 0 else
      	  !taskArray.(tid2).pbounds.(tid1 - tid2 - 1) in
      let pbound = 
	if !nopbounds then []
	else
	  begin
	    try
	      let priorJob = BT.stepInGeneration !jobTree j2 j1 (-pb) in
	      Formatcil.cStmts
		"assume (%l:je < %l:s2);"
		(fun n t -> assert false)
		builtinLoc
		[
		  ("assume", Fv assumeFn);
		  ("s2", Fl s2);
		  ("je", Fl (Sched.jobEndLval (BT.id !jobTree priorJob) ));
		]
	    with Invalid_argument _ -> []
	  end
      in 

      if jobPreceeds j1 j2 then
	Formatcil.cStmts
	  "assume(%l:e1 < %l:s2);"
	  (fun n t -> assert false)
	  rekloc
	  [ ("e1", Fl e1); ("s2", Fl s2); ("assume", Fv assumeFn);]
      else if jobFollows j1 j2 then
	Formatcil.cStmts
	  "assume(%l:e2 < %l:s1);"
	  (fun n t -> assert false)
	  rekloc
	  [ ("e2", Fl e2); ("s1", Fl s1); ("assume", Fv assumeFn);]
      else
	(* let p1 = if jobMayNotPreceed j1 j2 then
	    Formatcil.cStmts
	      "assume(%l:s2 <= %l:s1);"
	      (fun n t -> assert false)
	      rekloc
	      [("s1", Fl s1); ("s2", Fl s2); ("assume", Fv assumeFn);]
	  else [] in *)
	let p2 = if jobMayNotFollow j1 j2 then
	    Formatcil.cStmts
	      "assume(%l:e1 <= %l:e2);"
	      (fun n t -> assert false)
	      rekloc
	      [("e1", Fl e1); ("e2", Fl e2); ("assume", Fv assumeFn);]
	  else [] in
	p2 @
	  Formatcil.cStmts
      	    "if (%l:s2 <= %l:e1)
               if (%l:s1 <= %l:e2)
                 {
                   assume (%l:s2 <= %l:s1);
                   assume (%l:e1 < %l:e2);
                   %S:pbound
                 }
            "
    	  (fun n t -> assert false)
    	  builtinLoc
    	  [
    	    ("s1", Fl s1);
    	    ("s2", Fl s2);
    	    ("e1", Fl e1);
    	    ("e2", Fl e2);
	    ("pbound", FS pbound);
    	    ("assume", Fv assumeFn);
    	  ]	  
    in
    
    let minStmt x y z =
      Formatcil.cStmt
	"if (%l:y < %l:z)  assume(%l:x == %l:y); else assume(%l:x == %l:z);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("z", Fl z);
	  ("assume", Fv assumeFn)
	] in

    let maxStmt x y z =
      Formatcil.cStmt
	"if (%l:y > %l:z)  assume(%l:x == %l:y); else assume(%l:x == %l:z);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("z", Fl z);
	  ("assume", Fv assumeFn)
	] in

    let leqStmt x y = 
      Formatcil.cStmt
	"assume (%l:x <= %l:y);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("assume", Fv assumeFn)
	] in

    let ltStmt x y = 
      Formatcil.cStmt
	"assume (%l:x < %l:y);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("assume", Fv assumeFn)
	] in

    let eqStmt x y = 
      Formatcil.cStmt
	"assume (%l:x == %l:y);"
	(fun n t -> assert false)
	rekloc
	[
	  ("x", Fl x);
	  ("y", Fl y);
	  ("assume", Fv assumeFn)
	] in
	
    let perJob tree job : bool =
      let jid = BT.id tree job in
      
      let minV = Sched.jobMinLval jid in
      let maxV = Sched.jobMaxLval jid in
      let startV = Sched.jobStartLval jid in
      let endV = Sched.jobEndLval jid  in
      
      if BT.isLeaf tree job then 
	fd.sbody.bstmts <- 
	  (eqStmt startV endV) ::
	  (eqStmt minV startV) :: 
	  (eqStmt maxV  endV) :: 
	  fd.sbody.bstmts
      else
	begin
	  let cminV = Sched.jobMinLval (BT.id tree (BT.first_child tree job)) in
	  let cmaxV = Sched.jobMaxLval (BT.id tree (BT.last_child tree job)) in
	(* add min/max statements for the current job *)
	  fd.sbody.bstmts <- 
	    (leqStmt startV endV) ::
	    (minStmt minV startV cminV) ::
	    (maxStmt maxV endV cmaxV) :: fd.sbody.bstmts 
	end
      ;
	(* ensure that siblings start/end times  are ordered *)
      if BT.hasNext tree job then 
	begin
	  let minNextV = Sched.jobMinLval (BT.id tree (BT.next tree job)) in
	  fd.sbody.bstmts <-
	    (ltStmt maxV minNextV) :: fd.sbody.bstmts 
	end
      ;

      fd.sbody.bstmts <- 
	(Formatcil.cStmts
	   "assume (%e:job_low <= %l:start);
            assume (%l:end <= %e:job_high);"
	   (fun n t -> assert false)
	   rekloc
	   [
	     ("assume", Fv assumeFn);
	     ("end", Fl endV);
	     ("start", Fl startV);
	     ("job_low", Fe (integer (jobLowRound job)));
	     ("job_high", Fe (integer (jobHighRound job)));
	   ]) @ fd.sbody.bstmts ;


      if !nesting && not (BT.isRoot job) then 
	begin
	  BT.iter_parents  
	    (fun t j2 -> 
	      fd.sbody.bstmts <- 
		fd.sbody.bstmts @ nestingConstraints job j2 
	    ) 
	    tree (BT.parent_of job)
	end
      ;
      true
    in
    BT.iter_dfs perJob !jobTree (BT.root !jobTree)  ; 
    (* initialize all variables non-deterministically *)
    BT.iter_dfs 
      (fun tree job -> 
	let jid = BT.id tree job in 
	fd.sbody.bstmts <- 
	  (Formatcil.cStmts 
	  "%l:start=ndc(); %l:end=ndc (); %l:min=ndc(); %l:max=ndc ();"
	  (fun n t -> assert false)
	  rekloc
	  [
	    ("start", Fl (Sched.jobStartLval jid));
	    ("end", Fl (Sched.jobEndLval jid));
	    ("min", Fl (Sched.jobMinLval jid));
	    ("max", Fl (Sched.jobMaxLval jid));
	    ("ndc", Fv (nondetFn ucharType));
	  ]) @ fd.sbody.bstmts 
	;
	true

      ) 
      !jobTree (BT.root !jobTree)

  (*********************************************************************)
  (** create the function to initialize lamport clocks *)
  (*********************************************************************)
  let initLCFnName = "_init_lc"
  let initLCType = TFun (voidType, Some [], false, [])

  let createInitLC (unit) : unit =
    addUsefulFunc initLCFnName;
    let fd = findOrCreateFundec Args.file initLCFnName initLCType in
    fd.sbody <- mkBlock (
      ViMap.fold (fun sc lc sl -> 
        sl@[Formatcil.cStmt "lc = 0xffffffff;" (fun n t -> assert false)
          builtinLoc [("lc", Fv lc)]]
      ) !scLCMap []
    )

  (*********************************************************************)
  (*** creating the hyperperiod function when there are no ranges in
       the arrival time **)
  (*********************************************************************)

  (** Encoding of a hyper-period when arrivals are fixed *)
  let createHyperPeriod_NoRange (f: file) = 
    let ftype = TFun (voidType, Some [], false, []) in
    addUsefulFunc hyperFnName;
    let fd = findOrCreateFundec f hyperFnName ftype in

    List.iter 
      (fun j -> 
	print_string ((soi (BT.id !jobTree j)) ^ " ")) !orderedJobs ;
    print_string "\n";
    
    (** encode the execution of a single job *)
    let encode (job:jobt) : stmt list = 
      let jid = BT.id !jobTree job in
      let tid = BT.level_of job in

      (** assume jobs *)
      let assumeJobs = List.fold_left 
        (fun (il:instr list) (jid2:int) ->           
          il@[Formatcil.cInstr "jobAssume ();" rekloc
	       [("jobAssume", Fv jobAssumeFuncs.(jid2).svar);]]
        )
        [] maxObs.(jid)
      in

      (** assert jobs *)
      let assertJobs = List.fold_left 
        (fun (il:instr list) (jid2:int) ->           
          il@[Formatcil.cInstr "assert (%e:cond);" rekloc
	    [("assert", Fv assertFn);
	     ("cond", Fe (Lval (var (jobAssert jid2))));]]
        )
        [] assertAfter.(jid)
      in

      Formatcil.cStmts
	"
           round = %l:mystart;
           task = %e:mytask;
           end = %l:myend;
           job = %e:jid;
           jobAssert = 1;
           clock = 0;
           task_entry_pt ();
           assume (round == end);
           %I:assumeJobs
           %I:assertJobs
        "
	(fun n t -> assert false)
	rekloc
	[
	  ("assume", Fv assumeFn);
	  ("task", Fv taskVar);
	  ("mytask", Fe (integer tid));
	  ("round", Fv roundVar);
	  ("end", Fv jobEndVar);
	  ("mystart", Fl (Sched.jobStartLval jid));
	  ("myend", Fl (Sched.jobEndLval jid));
	  ("job", Fv jobVar);
	  ("jobAssert", Fv (jobAssert jid));
	  ("clock", Fv clockVar);
	  ("jid", Fe (integer jid));
	  ("task_entry_pt", Fv (jobFuncs.(jid).svar));
	  ("assumeJobs", FI assumeJobs);  
	  ("assertJobs", FI assertJobs);  
	]

    in
    (** execute all jobs *)
    let jobs = List.fold_left (fun r j -> r @ (encode j) ) [] !orderedJobs in   

    (** create hyper-period body *)
    fd.sbody <- mkBlock (
      Formatcil.cStmts "initLC(); sjobs(); %S:jobs"
	(fun n t -> assert false)
	rekloc
	[
          ("initLC", Fv (findOrCreateFunc f initLCFnName ftype));
          ("sjobs", Fv (findOrCreateFunc f scheduleJobsFnName ftype));
          ("jobs", FS jobs);
        ]
    )

  (*********************************************************************)
  (** create the hyper-period function. throw exception if there are
      ranges in the arrival time *)
  (*********************************************************************)
  let createHyperPeriod =  
    if !rangesInArrival then
      raise (RekSCError "cannot handle ranges in arrival times!!")
    else 
      createHyperPeriod_NoRange

  (*********************************************************************)
  (** create the shared variable snapshot function *)
  (*********************************************************************)
  let snapshotFnName = "_snapshot"
  let snapshotType = TFun (voidType, Some [], false, [])

  let createSnapshot (unit) : unit =
    addUsefulFunc snapshotFnName;
    let fd = findOrCreateFundec Args.file snapshotFnName snapshotType in
    let mo = !orderedJobs in
    let maxLC = makeLocalVar fd "_max_lc" uintType in
    fd.sbody.bstmts <- [];
    (** for every shared variable *)
    let forEachSharedVar (sv:varinfo) : unit =
      (** collect the symbolic constants that correspond to writes to
          this shared variable *)
      let mosc = ref [] in
      List.iter (
        fun (job2:jobt) ->
          let jid2 = BT.id !jobTree job2 in
          let forEachWrite (wv:varinfo) (wscs:varinfo list) : unit =
            if(sv.vid = wv.vid) then mosc := !mosc@wscs
          in
          ViMap.iter forEachWrite writeSC.(jid2)
      ) mo;
      print_string ("shared variable = " ^ sv.vname ^ " #wsc = " ^ (soi (List.length !mosc)) ^ "\n");
      (** save old value of shared variable, assign shared variable
          and max clock non-deterministically *)
      let oldsv = makeLocalVar fd ("_old_" ^ sv.vname) sv.vtype in
      fd.sbody.bstmts <- fd.sbody.bstmts@
        Formatcil.cStmts "oldsv = sv; sv = nd(); lc = ndui();"
        (fun n t -> assert false) rekloc
        [
          ("oldsv", Fv oldsv);
          ("sv", Fv sv);
          ("nd", Fv (nondetFn sv.vtype));
          ("lc", Fv maxLC);
          ("ndui", Fv (nondetFn uintType));
        ];
      (** start building the upper and lower bound constraints *)
      let ub = ref (integer 0) in
      let lb = ref (integer 1) in
      let ubneg = ref (Formatcil.cExp "sv == oldsv" 
                         [("sv", Fv sv); ("oldsv", Fv oldsv)]) 
      in
      (** generate the lower bound constraints *)
      (** for each symbolic constant that jid can observe *)
      List.iter
        (fun (wsc:varinfo) ->
          let fmt =
            [
              ("occ", Fe (hasOcc wsc));
              ("or", Fb LOr);
              ("and", Fb LAnd);
              ("assume", Fv assumeFn);
              ("cleq", Fe (cleq (lamportClock wsc) (Lval (var maxLC))));
              ("clle", Fe (clle (lamportClock wsc) (Lval (var maxLC))));
              ("sv", Fv sv);
              ("wsc", Fv wsc);
            ]
          in
          (** update lower bound constraint *)
          lb := Formatcil.cExp "%e:lb %b:and (!(%e:occ) %b:or %e:clle)" (("lb", Fe !lb)::fmt);
          (** update the upper bound constraint *)
          ub := Formatcil.cExp "%e:ub %b:or (%e:occ %b:and %e:cleq %b:and (sv == wsc))" (("ub", Fe !ub)::fmt);
          ubneg := Formatcil.cExp "%e:ubneg %b:and (!(%e:occ))" (("ubneg", Fe !ubneg)::fmt)
        )
        !mosc;
      ub := Formatcil.cExp "%e:ubneg %b:or %e:ub" [("ubneg", Fe !ubneg); ("ub", Fe !ub); ("or", Fb LOr)];
      fd.sbody.bstmts <- fd.sbody.bstmts@
        Formatcil.cStmts "assume (%e:lb %b:and %e:ub);"
        (fun n t -> assert false) rekloc
        [
          ("assume", Fv assumeFn);
          ("lb", Fe !lb);
          ("ub", Fe !ub);
          ("and", Fb LAnd);
        ];
    in
    ViSet.iter forEachSharedVar !sharedVars

  (*********************************************************************)
  (*** creating the main function **)
  (*********************************************************************)
  let createMain (f:file) = 
    let ftype = TFun (voidType, Some [], false, []) in
    addUsefulFunc mainFnName;
    let fd = findOrCreateFundec f mainFnName ftype in

    let userInitFn = findOrCreateFunc Args.file userInitialFnName ftype in
    let stmts = ref (Formatcil.cStmts "userInitFn();"
                       (fun n t -> makeTempVar fd ~name:n t) rekloc
                       [("userInitFn", Fv userInitFn);])
    in
    
    let hyperFn = findOrCreateFunc Args.file hyperFnName ftype in
    let snapshotFn = findOrCreateFunc Args.file snapshotFnName ftype in
    if (!unboundedSeq) then (
      let loopBody = (Formatcil.cStmts "hyperFn(); snapshotFn();"
                        (fun n t -> makeTempVar fd ~name:n t) rekloc
                        [ ("hyperFn", Fv hyperFn); ("snapshotFn", Fv snapshotFn); ])  
      in
      stmts := !stmts @ (mkWhile (integer 1) loopBody)
    ) else (
      for i = 1 to !hyperPeriods do
        stmts := !stmts @ Formatcil.cStmts "hyperFn(); snapshotFn();"
          (fun n t -> makeTempVar fd ~name:n t) rekloc
          [ ("hyperFn", Fv hyperFn); ("snapshotFn", Fv snapshotFn); ]
      done ;
    );
    let finalFn = findOrCreateFunc Args.file userFinalFnName ftype in
    stmts := !stmts @ Formatcil.cStmts "finalFn();"
      (fun n t -> makeTempVar fd ~name:n t) rekloc
      [ ("finalFn", Fv finalFn); ];
    fd.sbody <- mkBlock !stmts      
      
  (*********************************************************************)
  (** remove all globals that we don't like *)
  (*********************************************************************)
  let cleanupGlobals unit : unit = 
    Args.file.globals <- 
      List.filter 
      (fun g -> 
	match g with 
	    (* remove all prototypes for functions used to mark shared
	       reads and writes*)
	    GVarDecl (v, _) -> 
	      (not (Util.hasPrefix sharedReadPrefix v.vname)) && 
		(not (Util.hasPrefix sharedWritePrefix v.vname))
          (** remove all functions that are not useful *)
          | GFun(fd,_) -> isUsefulFunc(fd.svar.vname)
	  | _ -> true
      ) 
      Args.file.globals
      
  (*********************************************************************)
  (** register additional globals *)
  (*********************************************************************)   
  (** register all globals created so far with the file *)
  let registerGlobals unit : unit = 
    Sched.registerGlobals Args.file

  (*********************************************************************)
  (*** Entry point **)
  (*********************************************************************)      
  (** run everything *)
  let run unit : unit =
    (* create job ordering *)
    createJobOrdering ();
    (* create user initial and final functions *)
    createUserInitial Args.file;
    createUserFinal Args.file;
    (* create the functions for each job *)
    createJobFuncs Args.file;
    (* create check assumption functions *)
    createJobAssumps ();    
    (* schedule jobs *)
    createScheduleJobs ();
    (* create context switch function *)
    createCS Args.file;
    (* create function to initialize lamport clocks *)
    createInitLC ();
    (* create hyper-period function *)
    createHyperPeriod Args.file;
    (* create the variable snapshot function *)
    createSnapshot ();
    (* create the main function *)
    createMain Args.file;
    (* remove unused globals *)
    cleanupGlobals ();
    (* add newly created useful globals *)
    registerGlobals ();
    ()
end;;

(*********************************************************************)
(*** the Cil feature, where everything begins *)
(*********************************************************************)

(* command line option for first pass *)
let feature : featureDescr = { 
  fd_name = "reksc";              
  fd_enabled = ref false;
  fd_description = "REK for RTES using Sequential Consistency";
  fd_extraopt = 
    [
      "--rekhScRounds", Arg.Set_int numRounds, " number of sequentialization rounds";
      "--rekhScPrintNumJobs", Arg.Set printNumJobs, " print number of jobs and exit";
      "--rekhScUnbounded", Arg.Set unboundedSeq, " do unbounded sequentialization";
    ];
  fd_doit = (
    function (f: file) -> 
      print_string ("Analyzing " ^ f.fileName ^ "\n");

      (* compute task specific info *)
      initTaskList f ;
      
      (** print number of jobs and exit if necessary *)
      let jobNum = List.fold_left (fun res t -> res + t.tjobs) 0 !tasks in
      print_string ("number of jobs = " ^ (soi jobNum) ^ "\n");
      if (!printNumJobs) then exit 0;

      (** print number of rounds and check if they are enough *)
      print_string ("number of rounds = " ^ (soi !numRounds) ^ "\n");

      (* package arguments into a module *)
      let module A = struct 
	let file = f 
        let rounds = !numRounds (* List.fold_left (fun res t -> res + t.tjobs) 0 !tasks *)
	let hyperPeriodVar = findOrCreateGlobalVar f "__startrek_hyper_period" ucharType {init = None} builtinLoc
	let lockVar = findOrCreateGlobalVar f "__startrek_lock" boolType {init = Some (makeZeroInit boolType)} builtinLoc
	let jobEndVar = findOrCreateGlobalVar f "__startrek_job_end" ucharType {init = None} builtinLoc
	let jobVar = findOrCreateGlobalVar f "__startrek_job" ucharType {init = None} builtinLoc
	let taskVar = findOrCreateGlobalVar f "__startrek_task" ucharType {init = None} builtinLoc
	let roundVar = findOrCreateGlobalVar f "__startrek_round" ucharType {init = None} builtinLoc
        let jobNum = jobNum
	let clockVar = findOrCreateGlobalVar f "__startrek_clock" ucharType {init = None} builtinLoc
      end in
      let module M = 
	    RekSC
	      (A) 
	      (ArraySchedVarManager (A))
      in
      M.run ()
  );
  fd_post_check = false
} 
