open Cil
open Rtcbautil
open Rekargs

(** Varinfo map *)
module VM = ViMap

(*********************************************************************)
(** Signature of a manger of shared variables. *)
(*********************************************************************)
module type SHARED_VAR_MANAGER = sig
  (** opaque information about a variable *)
  type sv
  
  (** returns true if a given variable is managed by this module*)
  val mem : varinfo -> bool

  (** promote varinfo to a shared variable *)
  val makeSharedVar : varinfo -> unit

  (** iterate over all shared variables *)
  val iterShared : (varinfo -> sv -> unit) -> unit 

  val foldShared : (varinfo -> sv -> 'b -> 'b) -> 'b -> 'b

  (** get shared variable information for a given task *)
  val getSharedVar : varinfo  -> sv

  (** lval for a shared variable given a round expression *)
  val getSharedLval : sv -> int  -> lval
  val getSharedExpLval : sv -> exp -> lval

  (** lval for a symbolic constant that initialized a shared vairable *)
  val getSharedInitLval : sv -> int -> lval
  val getSharedInitExpLval : sv -> exp -> lval
    
  (** mkSharedReadStmt l s e returns a stmt that loads s[e] into l *)
  val mkSharedReadStmt : lval -> sv -> exp -> stmt

  (** mkSharedWriteStmt s e a retuns a stmt that writes a into s[e] *)
  val mkSharedWriteStmt : sv -> exp -> exp -> stmt

  (** register any global created as side-effects with with the
     current file. Must be called at the end. *)
  val registerGlobals : file -> unit
end

(*********************************************************************)
(** an implementation of SHARED_VAR_MANAGER *)
(*********************************************************************)
module ArraySharedVarManager (Args : ARGS) =
struct
  
  type sv = 
      {
	(** array for round values *)
	xs : varinfo;
	(** array for initial values *)
	sc : varinfo;
      }

  (** map from variables to shared info. The state. *)
  let sharedVars :  sv VM.t ref = ref VM.empty

  (** returns shared variable info for a variable *)
  let svi (v) = (VM.find v !sharedVars)

  (** returns true if a variable is managed *)
  let mem (v) = (VM.mem v !sharedVars)

  (** Creates a new shared variable from a varinfo *)
  let makeSharedVar (v:varinfo) : unit =

    let makeSharedVarName (v:varinfo) : string = 
      "_" ^ v.vname ^ "_" in

    let makeSharedInitName (v:varinfo) : string = 
      "_i_" ^ v.vname ^ "_" in
      
    let atype = 
      Formatcil.cType "%t:t [%e:R]"
	[
	  ("t", Ft v.vtype); 
	  ("R", Fe (integer Args.rounds))
	] in 
    
    let xs = makeGlobalVar (makeSharedVarName v) atype and
	sc = makeGlobalVar (makeSharedInitName v) atype; 
    in
    sharedVars := 
      VM.add v {xs = xs; sc = sc} !sharedVars

  (** Returns a shared varinfo for varinfo v*)
  let getSharedVar = svi

  (** Returns an Lval for accessing a shared variable v in a given round *)
  let getSharedLval (sv:sv) (r:int) : lval =  
    Formatcil.cLval "xs[%e:r]" [("xs", Fv sv.xs); ("r", Fe (integer r))]

  let getSharedExpLval (sv:sv) (r:exp) : lval = 
    Formatcil.cLval "xs[%e:r]" [("xs", Fv sv.xs); ("r", Fe r)]
    
  (** Returns an Lval for the symbolic constant representing an
      intial value of a shared variable in a given round *)
  let getSharedInitLval (sv:sv) (r:int) : lval =
    Formatcil.cLval "sc[%e:r]" [("sc", Fv sv.sc); ("r", Fe (integer r))]

  let getSharedInitExpLval (sv:sv) (r:exp) : lval =
    Formatcil.cLval "sc[%e:r]" [("sc", Fv sv.sc); ("r", Fe r)]

  (** Returns a statement that sets out to the value of shared
      variable sv in round re *)
  let mkSharedReadStmt (out:lval) (sv:sv) (re:exp) : stmt = 
    Formatcil.cStmt 
      "%l:out = xs[%e:re];"
      (fun n t -> assert false)
      builtinLoc
      [
	("out", Fl out);
	("xs", Fv sv.xs);
	("re", Fe re)
      ]
      
  (** Returns a statement that writes arg to the shared variable sv
      in round re *)
  let mkSharedWriteStmt (sv:sv) (re:exp) (arg:exp) : stmt = 
    Formatcil.cStmt
      "xs[%e:re] = %e:arg;"
      (fun n t -> assert false)
      builtinLoc
      [
	("xs", Fv sv.xs);
	("re", Fe re);
	("arg", Fe arg)
      ]
      
  (** iterate over shared variables*)
  let iterShared (f : varinfo -> sv -> unit) : unit =
    VM.iter f !sharedVars

  let foldShared (f : varinfo -> sv -> 'b -> 'b) (r:'b) : 'b = 
    VM.fold f !sharedVars r
      
  let registerGlobals f : unit = 
    (* register global variables *)
    let fn res g = 
      let out = g :: res in 
      match g with 
	| GVar (v, _, _) when mem v -> 
	  let sv = svi v in
	  GVar (sv.xs, {init = None}, builtinLoc) :: 
	    GVar (sv.sc, {init = None}, builtinLoc) :: out
	| _ -> out
    in
    f.globals <- List.fold_left fn [] (List.rev f.globals)
    (* iterShared  *)
    (*   (fun v sv -> *)
    (* 	 f.globals <-  *)
    (* 	   GVar(sv.xs, { init = None }, builtinLoc) ::  *)
    (* 	   GVar(sv.sc, { init = None }, builtinLoc) :: *)
    (* 	   f.globals) *)
end

(*********************************************************************)
(** Dummy Feature description for CIL compilation *)
(*********************************************************************)
let feature : featureDescr = 
{
  fd_name = "reksvm";
  fd_enabled = ref false;
  fd_description = "reksvm";
  fd_extraopt = [ ];
  fd_doit = (fun f -> ());
  fd_post_check = false;
}
