open Cil
open Rtcbautil

(*********************************************************************)
(* a module capturing inputs in terms of real-time tasks. the complete
   input is a list of RT tasks, and the length of the hyper-period to be
   analyzed. each RT task consists of a function name, period and a
   WCET. tasks are listed in order of increasing priority. *)
(*********************************************************************)

type rt_task = {
  rtname : string;
  rtwcet : int;
  rtperiod : int;
}

class rtInput =
object
  (* the list of RT tasks *)
  val mutable tasks : rt_task list = []

  (* add a task *)
  method addTask (t:rt_task) = tasks <- tasks@[t]

  (* get the tasks *)
  method getTasks = tasks
end

(* the global RT task input *)
let rtInp = new rtInput

(* parse real-time task descriptor func:wcet:period *)
let parseRtTaskDesc (arg:string) : unit =
  let sl = Str.split (Str.regexp ":") arg in
  let task : rt_task = {
    rtname = List.nth sl 0; 
    rtwcet = int_of_string(List.nth sl 1); 
    rtperiod = int_of_string(List.nth sl 2);
  } in
    rtInp#addTask task

(*********************************************************************)
(* input in terms of list of tasks and number of hyper-periods
   (currently assumed to be 1). each task has a function name, # of jobs,
   and preemptions bounds for each task of higher priority. tasks are
   listed in increasing priority order. *)
(*********************************************************************)

type pb_task = {
  pbname : string;
  pbperiod : int;
  pblatency : int;
  pbpbounds : int list;
}

class pbInput =
object
  (* the list of PB tasks *)
  val mutable tasks : pb_task list = []

  (* copy the object *)
  method copy = begin
    let res = new pbInput in
      res#setTasks tasks;
      res
  end

  (* get the tasks *)
  method getTasks = tasks

  (* set the tasks *)
  method setTasks t = tasks <- t

  (* add a task directly to the beginning of the list *)
  method addPbTask (t:pb_task) = begin
    tasks <- t::tasks
  end

  (* add a task given a task descriptor of the form func:period *)
  method addTask (tdesc:string) = begin    
    let sl = Str.split (Str.regexp ":") tdesc in
    let task : pb_task = {
      pbname = List.nth sl 0; 
      pbperiod = int_of_string (List.nth sl 1); 
      pblatency = 0;
      pbpbounds = [];
    } in
      tasks <- tasks@[task]
  end

  (* add preemption bounds for an existing task given a pb descriptor
     t:b where t is the id of the target task and b are the bounds for
     each task with higher priority *)
  method addPbound (pbdesc:string) = begin
    let sl = List.map int_of_string (Str.split (Str.regexp ":") pbdesc) in
    let ta = Array.of_list tasks in
    let oldtask = ta.(List.hd sl) in
    let newtask : pb_task = {
      pbname = oldtask.pbname;
      pbperiod = oldtask.pbperiod;
      pblatency = oldtask.pblatency;
      pbpbounds = List.tl sl;
    } in     
      ta.(List.hd sl) <- newtask;
      tasks <- Array.to_list ta
  end

  (* return number of jobs of a task *)
  method getNumJobs (t:pb_task) = begin
    intceil (getTimeBound()) t.pbperiod
  end

  (* return the preemption bound of i w.r.t. j, for i < j *)
  method getPB (i:int) (j:int) : int =
    let t = List.nth tasks i in
      List.nth t.pbpbounds (j - i - 1)

  (* the number of hyper-periods to analyze *)
  val mutable hpnum : int = 1

  (* get the hyper-period *)
  method getHpNum = hpnum

  (* set the hyper-period *)
  method setHpNum (hp:int) = hpnum <- hp  

  (* populate the input from a rtInput *)
  method populate (rti : rtInput) = begin
    tasks <- [];
    (* solve RMA equations to compute latencies *)
    let rtits = rti#getTasks in
    let taskNum = List.length rtits in
    let lat = Array.make taskNum 0 in
      (* function to compute new latency estimate *)
    let newLat (tid:int) : int =
      let rtit = List.nth rtits tid in
      let res = ref rtit.rtwcet in
        for i = tid + 1 to taskNum - 1 do (
          let rtitUp = List.nth rtits i in
            res := !res + (intceil lat.(tid) rtitUp.rtperiod) * rtitUp.rtwcet
        ) done;
        !res
    in
      (* compute each PB task *)
      for tid = (taskNum - 1) downto 0 do (
        let rtit = List.nth rtits tid in
          lat.(tid) <- rtit.rtwcet;
          try while (true) do (
            let newVal = newLat tid in
              if (newVal > lat.(tid)) then
                lat.(tid) <- newVal
              else
                raise Exit
          ) done;
          with Exit ->
            print_string ("lat[" ^ (string_of_int tid) ^ "] = " ^ 
                            (string_of_int lat.(tid)) ^ "\n");
            (* compute preemption bounds for the task *)
            let pb = ref [] in
              for i = tid + 1 to taskNum - 1 do (
                  (* compute upper preemption bound *)
                pb := !pb@[intceil lat.(tid) (List.nth rtits i).rtperiod];
                print_string ("PB[" ^ (string_of_int tid) ^ "][" ^ (string_of_int i) ^ "] = " ^ 
                                (string_of_int (intceil lat.(tid) (List.nth rtits i).rtperiod)) ^ "\n")
              ) done;
              (* create task *)
              let t:pb_task = {
                pbname = rtit.rtname;
                pbperiod = rtit.rtperiod;
                pblatency = lat.(tid);
                pbpbounds = !pb;
              } in 
              if (t.pblatency > t.pbperiod) then (
                print_string ("task " ^ (string_of_int tid) ^ " not schedulable !!\n");
                exit 1
              );
              tasks <- t::tasks
      ) done;    
  end
end

(* the global preemption bound task input *)
let pbInp = new pbInput

(*********************************************************************)
(* input in terms of list of tasks. each task has 
   - a function name, 
   - # of jobs
   - # of slices
   - and preemptions bounds (optionally) for each task of higher
   priority.   

   tasks are listed in increasing priority order. currently we assume
   1 hyper-period. *)
(*********************************************************************)

type sl_task = {
  slname : string;
  slperiod : int; 
  sljobs : int; (* inflated jobs due to linearization *)
  slslices : int;
  slpbounds : int list;
}

class slInput =
object
  (* the list of tasks *)
  val mutable tasks : sl_task list = []

  (* get the tasks *)
  method getTasks = tasks

  (* return number of jobs of a task *)
  method getNumJobs (t:sl_task) = begin
    intceil (getTimeBound()) t.slperiod
  end

  (* populate from a PB task descriptor -- this uses ofer's formula *)
  method populate (pbi:pbInput) = begin
    tasks <- [];
    (* create idle task *)
    let idle : pb_task = {
      pbname = "idle_task";
      pbperiod = getTimeBound();
      pblatency = 0;
      pbpbounds = List.map (fun (pbt:pb_task) -> pbInp#getNumJobs pbt) pbi#getTasks;
    }
    in
      (* make a copy of the input *)
    let newPbi = pbi#copy in
      (* add the idle task at the beginnning of the list *)
      newPbi#addPbTask idle;
      let pbits = newPbi#getTasks in
      let taskNum = List.length pbits in
        (* compute slices *)
      let slices = Array.make taskNum 0 in
        for i = (taskNum - 1) downto 0 do (
	  (* slice[i] = 1 + SUM_{j>i} (PB[i][j] - (S[j] - 1) * PB[j-1][j]) *)
          (* compute upper bound on slice -- use upper bounds in the
             positive parts and lower bounds in the negative parts *)
	  let slice = ref 1 in
            for j = i + 1 to taskNum - 1 do (
              slice := !slice + (newPbi#getPB i j);
	      (* should be fine. this line was extra *)
              slice := !slice - (slices.(j) - 1) * (newPbi#getPB (j-1) j)
            ) done;
	    slices.(i) <- !slice;
            print_string ("slices[" ^ (string_of_int i) ^ "] = " ^ 
			    (string_of_int !slice) ^ "\n")
        ) done;

        let jobs = Array.make taskNum 0 in
	  (* jobs[0] is unchanged and jobs[i] = jobs[i-1] * (slices[i-1] - 1) *)
	  for i = 0 to taskNum - 1 do (
            let pbit = List.nth pbits i in
	      jobs.(i) <- if (i = 0) then pbInp#getNumJobs pbit else (jobs.(i-1) * (slices.(i-1) - 1));
              print_string ("jobs[" ^ (string_of_int i) ^ "] = " ^ 
			      (string_of_int (jobs.(i))) ^ "\n");
              let t : sl_task = {
                slname = pbit.pbname;
                slperiod = pbit.pbperiod;
                sljobs = jobs.(i);
                slslices = slices.(i);
                slpbounds = pbit.pbpbounds;
              } in tasks <- tasks@[t]
	  ) done;
          (* finally remove the first idle task *)
          tasks <- List.tl tasks
  end

  (* populate from a PB task descriptor -- this uses periods, but only
     looks at next higher task *)
  (* WARNING: this is unsound *)
  method populate_PERIOD (pbi:pbInput) = begin
    tasks <- [];
    let pbits = pbi#getTasks in
    let taskNum = List.length pbits in
      (* compute slices *)
    let slices = Array.make taskNum 1 in
      for i = (taskNum - 1) downto 0 do (
	(* slice[i] = ceil (period[i] / period[i+1]) + 1 *)
        let slice = 
          if (i = taskNum - 1) then 1
          else (
            let p1 = (List.nth pbits i).pbperiod in 
            let p2 = (List.nth pbits (i + 1)).pbperiod in 
              (intceil p1 p2) + 1 
          )
        in
	  slices.(i) <- slice;
          print_string ("slices[" ^ (string_of_int i) ^ "] = " ^ 
			  (string_of_int slice) ^ "\n")
      ) done;

      let jobs = Array.make taskNum 0 in
	(* jobs[0] is unchanged and jobs[i] = jobs[i-1] * (slices[i-1] - 1) *)
	for i = 0 to taskNum - 1 do (
          let pbit = List.nth pbits i in
	    jobs.(i) <- if (i = 0) then pbi#getNumJobs pbit else (jobs.(i-1) * (slices.(i-1) - 1));
            print_string ("jobs[" ^ (string_of_int i) ^ "] = " ^ 
			    (string_of_int (jobs.(i))) ^ "\n");
            let t : sl_task = {
              slname = pbit.pbname;
              slperiod = pbit.pbperiod;
              sljobs = jobs.(i);
              slslices = slices.(i);
              slpbounds = pbit.pbpbounds;
            } in tasks <- tasks@[t]
	) done
  end

  (* return a list of CIL tasks (defined in rtcbautil as task) *)
  method cilTasks : task list = begin
    let createCilTask (l:task list) (t:sl_task) : task list =
      let t : task = {
        name = t.slname;
        tid = List.length l;
        tperiod = t.slperiod;
        tjobs = t.sljobs;
        tslices = t.slslices;
      } in l@[t]
    in
      List.fold_left createCilTask [] tasks
  end
end

(* the global slice task input *)
(* let slInp = new slInput *)

(** Feature description for CIL *)
let feature : featureDescr = 
{
  fd_name = "rtcbainput";
  fd_enabled = ref false;
  fd_description = "rtcbainput";
  fd_extraopt = [ ];
  fd_doit = (fun f -> ());
  fd_post_check = false;
}

