open Cil

let rekloc : location = {line = 1; file = "<startrek builtins>"; byte=0}

(*********************************************************************)
(* Helper methods *)
(*********************************************************************)
  
(** Creates a switch statement. mkSwitchStmt x s len fn returns a
   switch statement on x with len cases starting with s, the body of
   each case is given by fn. Break statements are added automatically.
   It takes an optional argument default for the default case.
**)
let mkSwitchStmt
    (x:exp) (s:int) (len:int) ?d:(default : stmt list = []) 
    (fn: int -> stmt list) : stmt =
  let rec mkBody s len fn res = 
    if len = 0 then res
    else
      begin
	(* the case currently being processed *)
	(* work bottom up *)
	let case = s + len - 1 in
	let caseStmts = fn case in
	let hd = List.hd caseStmts in
	  hd.labels <- [Case (integer case, builtinLoc)];
	  mkBody s (len - 1) fn 
	    (caseStmts @ ((mkStmt (Break builtinLoc)) :: res))
      end
  in 
  begin
    match default with
      | hd::tl -> hd.labels <- [Default builtinLoc]
      | [] -> ()
  end;
  let slist = (mkBody s len fn [])@default in
  mkStmt (Switch (x, mkBlock slist, slist, builtinLoc))


(*********************************************************************)
(* data structures and functions related to tasks *)
(*********************************************************************)

(* _Bool type *)
let boolType = TInt (IBool, []) 

(* char type *)
let ucharType = TInt (IUChar, [])

(* given integers x and y return ceiling of x/y *)
let intceil (x:int) (y:int) : int =
  let r = x / y in
    if x mod y > 0 then r + 1 else r

(* The time bound to analyze *)
let timeBound : int ref = ref 0

(* set the time bound *)
let setTimeBound (tb:int) = 
  timeBound := tb

(* get the time bound *)
let getTimeBound () = !timeBound

(* a task description type -- is a tuple consisting of the function
   name for the task, the task id, the number of jobs in a hyper-period,
   and the number of slices per job. the id is 0 for the lowest priority
   job and increases by 1 for each task of higher priority. *)
type task = {
  name : string;
  tid : int;
  tperiod : int;
  tjobs : int;
  tslices : int;
}

(* return number of jobs of a task *)
let numJobsForTask (t:task) = intceil (getTimeBound()) t.tperiod

(* the list of tasks *)
let tasks : task list ref = ref []

(* returns a task given an id *)
let taskByTid (tid:int) = List.nth !tasks tid
(* returns the total number of tasks *)
let numOfTasks unit : int  = List.length !tasks

let isTopTask t = (t.tid = pred (List.length !tasks))

let oget a = 
  match a with 
    | Some x -> x 
    | None -> invalid_arg "Expecting None, got Some"

let rec gcd a b =
  if      a = 0 then b
  else if b = 0 then a
  else if a > b then gcd b (a mod b)
  else gcd a (b mod a)

let lcm m n =  (abs m * n) / (gcd m n)

let intceil a b = (a / b) + if (a mod b = 0) then 0 else 1


(* parse a task descriptor and add the task to the global list of
   tasks. a task descriptor is of the form [func_name:period]. it is
   assumed that tasks are specified in increasing priority order. ids are
   assigned to tasks automatically, starting with 0 and increasing with
   each new task. *)
let parseTaskDesc (arg:string) : unit =
  let sl = Str.split (Str.regexp ":") arg in
  let _period = int_of_string(List.nth sl 1) in
  let task : task = {
    name = List.nth sl 0;
    tid = List.length !tasks;
    tperiod = _period;
    tjobs = intceil (getTimeBound()) _period;
    tslices = 0
  } in
    tasks := !tasks@[task]

(*********************************************************************)
(* data structures and functions related to priority bounds *)
(*********************************************************************)

(* ordering of (int,int). used for maps and sets with (int,int) as
   key, e.g., priority bounds. *)
module IntIntOrder =
struct
  type t = int * int
  let compare ((i1,j1):t) ((i2,j2):t) : int =
    if(i1 = i2) then j1 - j2
    else i1 - i2
end

(* map with (int,int) key *)
module IntIntMap = Map.Make(IntIntOrder)

(* map for priority bounds matrix *)
let pbounds = ref (IntIntMap.add (0,0) 0 IntIntMap.empty)

(* add a pbound b between task i and task j *)
let addPbound i j b : unit =
  print_string ("pbound[" ^ (string_of_int i) ^ 
                  "][" ^ (string_of_int j) ^ "] = " ^ (string_of_int b) ^ "\n");
  pbounds := IntIntMap.add (i,j) b !pbounds

(* get the pbound between task i and task j *)
let getPbound i j : int =
  IntIntMap.find (i,j) !pbounds

(* parse a priority bound descriptor [i:b1: ... : bn] where i = id of
   task and b1 ... bn are the bounds for each higher task. *)
let parsePboundDesc (arg:string) : unit =
  let sl = List.map int_of_string (Str.split (Str.regexp ":") arg) in
  let tid = List.nth sl 0 in
    for i = 1 to (List.length sl) - 1 do (
      addPbound tid (tid + i) (List.nth sl i)
    ) done

(*********************************************************************)
(* data structures for representing maps from global variables in the
   original program to global variables in the new program *)
(*********************************************************************)

(* ordering of (varinfo,int). this is used for maps and sets over with
   (varinfo,int) as key. *)
module ViIntOrder =
struct
  type t = varinfo * int
  let compare ((v1,t1):varinfo*int) ((v2,t2):varinfo*int) : int =
    if(v1.vid = v2.vid) then t1 - t2
    else if (v1.vid < v2.vid) then -1
    else 1
end

(* map with (varinfo,int) key *)
module ViIntMap = Map.Make(ViIntOrder)

(* ordering of varinfo based on vid. this is used for maps and sets
   over varinfo as key *)
module ViOrder =
struct
  type t = varinfo
  let compare (v1:varinfo) (v2:varinfo) : int =
    if(v1.vid = v2.vid) then 0
    else if (v1.vid < v2.vid) then -1
    else 1
end

(* map with varinfo key *)
module ViMap = Map.Make(ViOrder)

(*********************************************************************)
(* useful types and routines *)
(*********************************************************************)

(* type of assert function *)
let assertType = TFun (voidType,Some [("arg",boolType,[])],false,[])

(* list of arguments to asserts *)
let assertArgs : varinfo list ref = ref []

(* type of assume function *)
let assumeType = TFun (voidType,Some [("arg",boolType,[])],false,[])

(* type of nondet_int function *)
let ndiType = TFun (intType,Some [],false,[])

(* type of nondet_Bool function *)
let ndbType = TFun (boolType,Some [],false,[])

let  findFundec (f:file) (fn:string) : fundec option =
  try
    let g = 
      List.find 
	(function 
	     GFun (fd, _) -> fd.svar.vname = fn 
	       | _ -> false) 
	f.globals in
      match g with GFun (fd, _) -> Some fd
	| _ -> None 
  with Not_found -> None


(* find or create a fundec with given name *)
let findOrCreateFundec (f:file) (fn:string) (ft:typ) : fundec =
  match findFundec f fn with
      Some fd -> fd
    | None -> 
      let fd = Cil.emptyFunction fn  in
      Cil.setFunctionTypeMakeFormals fd ft;
      f.globals <- f.globals @ [GFun(fd,Cil.builtinLoc)];
      fd

(* make a copy of a statement *)
let copyStmt (st:stmt) : stmt = mkStmt (st.skind)

(* a visitor that replaces returns with a target statement *)
class returnReplacer (tinit:stmt) = object(self)
  inherit nopCilVisitor as super
    
  (* target to replace a return by *)
  val target : stmt = tinit

  (* visit a statement *)
  method vstmt (st:stmt) = begin
    match st.skind with
        Return (_,l) -> 
          st.skind <- target.skind;
          SkipChildren
      | _ -> DoChildren
  end
end

(* a visitor that removes asserts and records their arguments in
   global variables *)
class assertReplacer = object(self)
  inherit nopCilVisitor as super

  (* visit an instruction *)
  method vinst (ins:instr) = begin
    match ins with
        Call(None,Lval(Var vi,NoOffset),[e2],l) ->
          if(vi.vname = "assert") then
            let gvname = "assert_arg_" ^ 
	      (string_of_int (List.length !assertArgs)) in
            let gv = makeGlobalVar gvname boolType  in
            assertArgs := !assertArgs@[gv];
            ChangeTo [Set(var gv,e2,Cil.builtinLoc)]
          else
            DoChildren            
      | _ -> DoChildren
  end
end

(*********************************************************************)
(* stuff related to setting and getting gloabal variables in the C
   code being targeted *)
(*********************************************************************)

(* name of the setter function for globals in the target C code *)
let sharedWriteFnName = "__startrek_shared_write_"

(* name of the getter function for globals in the target C code *)
let sharedReadFnName = "__startrek_shared_read_"

let cpuLockFnName = "__startrek_cpu_lock"
let cpuUnlockFnName = "__startrek_cpu_unlock"

(* job counter variable *)
let jobCounterName = "__startrek_job_counter"

(*********************************************************************)
(* a tree *)
(*********************************************************************)

(* tree node is parameterized by a data node *)
class ['a] tree d_init p_init =
object(self)
  (* the id of this node. we assume that in a well-formed tree, the id
     of a node indicates the path from the root to the node *)
  val mutable id : string = ""

  (* return the id *)
  method get_id = id

  (* the data associated with the node *)
  val data = (d_init : 'a)

  (* the parent of the node *)
  val parent : 'a tree option = None

  (* the children *)
  val mutable children : 'a tree array = Array.of_list []

  (* return the children *)
  method get_children = children

  (* label the tree canonically with the root getting the argument
     label l, and the i-th child getting labeled l.i. calling this method
     with any argument results in a well-formed tree rooted at the
     node. *)
  method cannonicalLabel (l:string) = begin
    id <- l;
    for i = 0 to Array.length children - 1 do (
      children.(i)#cannonicalLabel (l ^ string_of_int i)
    ) done
  end

  (*******************************************************************)
  (** the following methods should only be invoked on well-formed
      trees, i.e., after calling canonicalLabel *)
  (*******************************************************************)

  (* return the id assuming the tree is well-formed -- essentially the
     last digit in its id *)
  method get_child_id : int = begin
    (int_of_char id.[(String.length id) - 1]) - (int_of_char '0')
  end

  (* return the next sibling -- None if this is the last sibling *)
  method next : 'a tree option = begin
    match parent with
        None -> None
      | Some(x) -> 
          let s = x#get_children in
          let i = self#get_child_id in
            if ((i + 1) < (Array.length s)) then
              Some(s.(i + 1))
            else
              None
  end

  (* return first child -- None if no children *)
  method first_child : 'a tree option = begin
    if (Array.length children = 0) then None
    else Some(children.(0))
  end

  (* return last child -- None if no children *)
  method last_child : 'a tree option = begin
    let n = Array.length children in
      if (n = 0) then None
      else Some(children.(n - 1))
  end

  (* return true if this node is a descendant of n. since the tree is
  well-formed, this holds iff the id of n is a substring of the id of
  this node. *)
  method descendant (n : 'a tree) : bool = begin
    let nid = n#get_id in
      nid = (String.sub self#get_id 0 (String.length nid))
  end
end

(*********************************************************************)
(* module to manage the job tree *)
(*********************************************************************)
module JobTree =
struct
  (* return the canonical label given a task and its job *)
  let rec label (tid) (jid) : string =
    if (tid = 0) then string_of_int jid
    else
      let task = taskByTid tid in
      let ptid = tid - 1 in
      let ptask = taskByTid ptid in
        (* number of jobs per job of the parent *)
      let pslices = ptask.tperiod / task.tperiod in        
      let pjid = intceil jid pslices in
        (label ptid pjid) ^ (string_of_int (jid mod pslices))      

  (* decompose a label into a tid and a jid *)
  let decompose (l:string) : int * int =
    let tid = String.length l - 1 in
    let jid = (int_of_char l.[tid]) - (int_of_char '0') in
      (tid,jid)
         
  (* return the label of the next sibling -- return "" if this is the
     last child of its parent *)
  let next (l:string) : string =
    let (tid,jid) = decompose l in
    let numSibs = 
      if (tid = 0) then (taskByTid 0).tjobs
      else
        let task = taskByTid tid in
        let ptid = tid - 1 in
        let ptask = taskByTid ptid in
          ptask.tperiod / task.tperiod 
    in        
      if (jid = (numSibs - 1)) then ""
      else label tid (jid + 1)

  (* return the label of the first child -- return "" if there are no
  children *)
  let child (l:string) : string = 
    let (tid,jid) = decompose l in
      if (tid = numOfTasks () - 1) then ""
      else l ^ "0"
        
  (* return the label of the last child -- return "" if there are no
  children *)
  let last_child (l:string) : string = 
    let (tid,jid) = decompose l in
      if (tid = numOfTasks () - 1) then ""
      else
        let task = taskByTid tid in
        let ctid = tid + 1 in
        let ctask = taskByTid ctid in
        let slices = task.tperiod / ctask.tperiod in
          l ^ (string_of_int (slices - 1))

  (* return true iff k is a descendant of n, i.e., n is a strict substring of k *)
  let descendant n k : bool =
    if (n = k) then false
    else
      n = (String.sub k 0 (String.length n))
end

module BalancedTree =
struct

  (** node type *)
  type n = int list

  (** tree type *)
  type t = 
      { 
	btRoot : n;
	btDegrees : int array
      }


  (** create a new tree with the list of out degrees *)
  let make (degrees : int list) : t =
    { btRoot = []; btDegrees = Array.of_list degrees; }

  (** height of the tree*)
  let height (tree : t) = succ (Array.length tree.btDegrees)
      
  (** root node of the tree *)
  let root (tree : t) : n  = tree.btRoot

  (** level of a node *)
  let level_of (node : n) : int = List.length node
    
  (** true if node is a root *)
  let isRoot (node : n) : bool = (List.length node = 0)
  (** true if node is a leaf in a given tree *)
  let isLeaf (tree : t) (node : n) : bool =  
    (level_of node = pred (height tree))
    
  (** returns parent of a node *)
  let parent_of (node : n) : n =
    match node with 
      | hd :: tl -> tl
      |  _ -> failwith "root has no parent"
	
  (** outgoing degree of nodes at a given level. e.g., (level_degree t
      x) is the number of children for any node of t at level x *)
  let level_degree (tree : t) (level : int) = tree.btDegrees.(level)

  (** unique string name of a node *)
  let string_of_node (node : n) : string = 
    let rec string_of_node_rec (node: n) : string = 
      match node with 
	| hd :: tl -> 
	  (string_of_node_rec (parent_of node))  ^ "_" ^  (string_of_int hd)
	| _ -> ""
    in
    ("[" ^ (string_of_node_rec node) ^ "]")

  (** returns sibling id of a node *)
  let sibling_id (node : n) : int =  if isRoot node then 0 else List.hd node

  (** true if a node is a last sibling in a tree *)
  let isLastSibling (tree: t) (node : n) : bool =
    (isRoot node) ||
      (sibling_id node == pred (level_degree tree (level_of (parent_of node))))
      
  (** true if a node has a next sibling *)
  let hasNext (tree : t) (node : n) : bool = not (isLastSibling tree node)

  (** true if a node has a prev sibling *)    
  let hasPrev (node : n) : bool = (sibling_id node) <> 0
    
    
  (** next sibling of a node *)
  let next (tree : t) (node : n) : n =
    match node with 
      | hd :: tl -> 
	if hd < pred (level_degree tree (level_of tl)) then
	  (succ hd) :: tl
	else
	  failwith "last sibling"
      | _ -> failwith "Root has no siblings"

  (** prev sibling of a node *)
  let prev (node : n) : n = 
    match node with
      | hd :: tl -> 
	if hd <> 0 then (pred hd) :: tl
	else failwith "first sibling"
      | _ -> failwith "Root has no siblings"
	
	
  (** first child of a node *)
  let first_child (tree : t) (node : n) : n =   
    if isLeaf tree node then failwith "Leaf" else  0 :: node

  (** last child of a node *)
  let last_child (tree : t) (node : n) : n = 
    if isLeaf tree node then failwith "Leaf" 
    else
      (pred (level_degree tree (level_of node))) :: node

	
  let rec fold_parents  (fn:'a->t->n->'a) (r:'a) (tree:t) (node:n) : 'a =
    let x = fn r tree node in
    if not (isRoot node) then fold_parents fn x tree (parent_of node) else x

  (** iterate over all the parents (including node n*)
  let iter_parents (fn : t -> n -> unit) (tree : t) (node : n) : unit =
    fold_parents (fun x t n -> fn t n) () tree node


  (** iterate over the tree in DFS order. fn is called in pre-order 
      fn can return false to avoid exploring current subtree 
  *)
  let rec iter_dfs (fn : t -> n -> bool) (tree : t) (node : n) : unit = 
    (* DFS the current node and all its next siblings *)
    let rec iter_dfs_siblings (fn : t -> n -> bool) (tree : t) (node : n) =
      iter_dfs fn tree node ; 
      if hasNext tree node then iter_dfs_siblings fn tree (next tree node) 

    in
    if fn tree node && not (isLeaf tree node) then
      iter_dfs_siblings fn tree (first_child tree node)

  let forall (fn : t -> n -> bool) (tree:t) (node:n) : bool = 
    try
      iter_dfs 
	(fun tx tn -> 
	  if not (fn tx tn) then raise Exit; true) tree node;
      true
    with 
      | Exit -> false

  (** folds a function over a tree in DFS order *)
  let rec 
      fold_dfs_siblings fn init tree node = 
    let res = fold_dfs fn init tree node in
    if hasNext tree node then fold_dfs_siblings fn res tree (next tree node)
    else res
  and
      fold_dfs 
      (fn : 'a -> t -> n -> 'a)  (init: 'a)  (tree  : t) (node : n) : 'a = 
    let res = fn init tree node  in
    if (isLeaf tree node) then res
    else fold_dfs_siblings fn res tree (first_child tree node)

  (** True if a given node is last in its generation (level) *)
  let rec isLastInGeneration (tree  : t) (node : n) : bool = 
    if isRoot node then true
    else if isLastSibling tree node then 
      isLastInGeneration tree (parent_of node)
    else false

  let rec firstInGeneration (tree : t) (level : int) = 
    if level = 0 then root tree
    else first_child tree (firstInGeneration tree (pred level))

  (** Returns next node in the same generation (level) as a give node *)
  let rec nextInGeneration (tree : t) (node : n) : n = 
    if hasNext tree node then next tree node
    else first_child tree (nextInGeneration tree (parent_of node))

  let rec stepInGeneration (tree : t) (root : n) (node : n) (step : int): n
     =
    if node = root then
      (if step = 0 then root else invalid_arg "Step out of bound")
    else
      let parent = parent_of node in
      let b = level_degree tree (level_of parent) in
      let sid = sibling_id node in
      let mv = sid + step in
      let x1 = mv mod b in
      let x2 = mv / b in
      let y1 = if x1 < 0 then x1 + b else x1 in
      let y2 = if x1 < 0 then x2 - 1 else x2 in
      y1 :: (stepInGeneration tree root parent y2)
	

  (** index of the node in the generation. Left most node is 0 *)
  let rec generationId (tree : t) (node : n) : int = 
    if isRoot node then 0 
    else
      let parent = parent_of node in
      let b = level_degree tree (level_of parent) in
      (sibling_id node) + b * (generationId tree parent)
       
      

  (** Folds fn over the generation of a given node, starting with
      that node *)
  let rec fold_left_generation (fn : 'a -> t -> n -> 'a) (init : 'a) (tree :t)
      (node : n) : 'a = 
    
    let res = fn init tree node in 
    if isLastInGeneration tree node then res
    else fold_left_generation fn res tree (nextInGeneration tree node)      

  (** The size of the substree rooted at n, including n *)
  let rec size (tree : t) (node : n) : int = 
    if isLeaf tree node then 1 
    else 
      let kids = level_degree tree (level_of node) in
      kids * (size tree (first_child tree node)) + 1


  let rec 
  (** pre-order DFS numbering of a node *)
      id (tree : t) (node : n) : int = 
    let r = (if isRoot node then 0
      else if hasPrev node then (max_id tree (prev node)) + 1
      else (id tree (parent_of node)) + 1)  in
    (* print_string ("id of " ^ (string_of_node node) ^ 
       " is " ^ (string_of_int r) ^ "\n");  *)
    r
  and
      (** The largest id in the subtree rooted at n*)
      max_id (tree : t) (node : n) : int = 
    let r = (id tree node) + (size tree node) - 1 in
    (* print_string ("max_id of " ^ (string_of_node node) ^ 
		     " is " ^ (string_of_int r) ^ "\n"); *)
    r
end

type range = { rmin:int; rmax:int }

let rng_add (x:range) y = { rmin = x.rmin + y; rmax = x.rmax + y }
let rng_leq (x:range) (y:range) : bool = x.rmax <= y.rmin
let rng_lt (x:range) (y:range) : bool = x.rmax < y.rmin
  
(*********************************************************************)
(* find or create varinfo for global variable with given name, type
   and initializer *)
(*********************************************************************)
let findOrCreateGlobalVar (f:file) (n:string) (t:typ) (ii:initinfo) (loc:location) : varinfo =
  let fn (x:varinfo option) (g:global) : varinfo option = 
    match x with
        None -> (
          match g with
              GVarDecl(vi,_) when (vi.vname = n) -> Some(vi)
            | GVar(vi,_,_) when (vi.vname = n) -> Some(vi)
            | _ -> None
        )
      | _ -> x
  in
  match foldGlobals f fn None with
      None -> 
        let vi = makeGlobalVar n t in
        f.globals <- (GVar (vi,ii,loc))::f.globals;
        vi
    | Some(vi) -> vi

(***********************************************************************)
(**** Variables for command line switches ****)

(** Feature description for CIL *)
let feature : featureDescr = 
{
  fd_name = "rtcbautil";
  fd_enabled = ref false;
  fd_description = "rtcbautil";
  fd_extraopt = [ ];
  fd_doit = (fun f -> ());
  fd_post_check = false;
}
